// bslmf_addcv.t.cpp                                                  -*-C++-*-
#include <bslmf_addcv.h>

#include <bslmf_issame.h>  // for testing only

#include <bsls_bsltestutil.h>

#include <stdio.h>   // 'printf'
#include <stdlib.h>  // 'atoi'

using namespace bsl;
using namespace BloombergLP;

//=============================================================================
//                                TEST PLAN
//-----------------------------------------------------------------------------
//                                Overview
//                                --------
// The component under test defines meta-functions, 'bsl::add_cv' and
// 'bsl::add_cv_t' that add a top-level 'const'-qualifier and a top-level
// 'volatile'-qualifier to a template parameter type.  Thus, we need to ensure
// that the values returned by the meta-function are correct for each possible
// category of types.
//
// ----------------------------------------------------------------------------
// PUBLIC TYPES
// [ 1] bsl::add_cv::type
// [ 1] bsl::add_cv_t
//
// ----------------------------------------------------------------------------
// [ 2] USAGE EXAMPLE

// ============================================================================
//                     STANDARD BSL ASSERT TEST FUNCTION
// ----------------------------------------------------------------------------

namespace {

int testStatus = 0;

void aSsErT(bool condition, const char *message, int line)
{
    if (condition) {
        printf("Error " __FILE__ "(%d): %s    (failed)\n", line, message);

        if (0 <= testStatus && testStatus <= 100) {
            ++testStatus;
        }
    }
}

}  // close unnamed namespace

// ============================================================================
//               STANDARD BSL TEST DRIVER MACRO ABBREVIATIONS
// ----------------------------------------------------------------------------

#define ASSERT       BSLS_BSLTESTUTIL_ASSERT
#define ASSERTV      BSLS_BSLTESTUTIL_ASSERTV

#define LOOP_ASSERT  BSLS_BSLTESTUTIL_LOOP_ASSERT
#define LOOP0_ASSERT BSLS_BSLTESTUTIL_LOOP0_ASSERT
#define LOOP1_ASSERT BSLS_BSLTESTUTIL_LOOP1_ASSERT
#define LOOP2_ASSERT BSLS_BSLTESTUTIL_LOOP2_ASSERT
#define LOOP3_ASSERT BSLS_BSLTESTUTIL_LOOP3_ASSERT
#define LOOP4_ASSERT BSLS_BSLTESTUTIL_LOOP4_ASSERT
#define LOOP5_ASSERT BSLS_BSLTESTUTIL_LOOP5_ASSERT
#define LOOP6_ASSERT BSLS_BSLTESTUTIL_LOOP6_ASSERT

#define Q            BSLS_BSLTESTUTIL_Q   // Quote identifier literally.
#define P            BSLS_BSLTESTUTIL_P   // Print identifier and value.
#define P_           BSLS_BSLTESTUTIL_P_  // P(X) without '\n'.
#define T_           BSLS_BSLTESTUTIL_T_  // Print a tab (w/o newline).
#define L_           BSLS_BSLTESTUTIL_L_  // current Line number

//=============================================================================
//                  GLOBAL TYPEDEFS/CONSTANTS FOR TESTING
//-----------------------------------------------------------------------------

namespace {

struct TestType {
   // This user-defined type is intended to be used during testing as an
   // argument for the template parameter 'TYPE' of 'bsl::add_cv'.
};

typedef void (TestType::*MethodPtrTestType) ();
    // This non-static function member type is intended to be used during
    // testing as an argument for the template parameter 'TYPE' of
    // 'bsl::add_cv'.

typedef int TestType::* PMD;
    // This class public data member pointer type is intended to be used during
    // testing as an argument as an argument for the template parameter 'TYPE'
    // of 'bsl::add_cv'.

}  // close unnamed namespace

//=============================================================================
//                              MAIN PROGRAM
//-----------------------------------------------------------------------------

int main(int argc, char *argv[])
{
    int                 test = argc > 1 ? atoi(argv[1]) : 0;
    bool             verbose = argc > 2;
    bool         veryVerbose = argc > 3;
    bool     veryVeryVerbose = argc > 4;
    bool veryVeryVeryVerbose = argc > 5;

    (void)veryVerbose;          // suppress warning
    (void)veryVeryVerbose;      // suppress warning
    (void)veryVeryVeryVerbose;  // suppress warning

    setbuf(stdout, NULL);       // Use unbuffered output

    printf("TEST " __FILE__ " CASE %d\n", test);

    switch (test) { case 0:
      case 2: {
        // --------------------------------------------------------------------
        // USAGE EXAMPLE
        //
        // Concerns:
        //: 1 The usage example provided in the component header file compiles,
        //:   links, and runs as shown.
        //
        // Plan:
        //: 1 Incorporate usage example from header into test driver, remove
        //:   leading comment characters, and replace 'assert' with 'ASSERT'.
        //:   (C-1)
        //
        // Testing:
        //   USAGE EXAMPLE
        // --------------------------------------------------------------------

        if (verbose) printf("\nUSAGE EXAMPLE\n"
                            "\n=============\n");

///Usage
///-----
// In this section we show intended use of this component.
//
///Example 1: Adding a 'const'-Qualifier and a 'volatile'-Qualifier to a Type
/// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
// Suppose that we want to add a 'const'-qualifier and a 'volatile'-qualifier
// to a particular type.
//
// First, we create two 'typedef's -- a 'const'-qualified and
// 'volatile'-qualified type ('MyCvType') and the same type without the
// cv-qualifiers ('MyType'):
//..
        typedef int                MyType;
        typedef const volatile int MyCvType;
//..
// Now, we add a 'const'-qualifier and a 'volatile'-qualifier to 'MyType' using
// 'bsl::add_cv' and verify that the resulting type is the same as 'MyCvType':
//..
        ASSERT(true ==
                   (bsl::is_same<bsl::add_cv<MyType>::type, MyCvType>::value));
//..
// Finally, if the current compiler supports alias templates C++11 feature, we
// add a 'const'-qualifier and a 'volatile'-qualifier to 'MyType' using
// 'bsl::add_cv_t' and verify that the resulting type is the same as
// 'MyCvType':
//..
#ifdef BSLS_COMPILERFEATURES_SUPPORT_ALIAS_TEMPLATES
        ASSERT(true == (bsl::is_same<bsl::add_cv_t<MyType>, MyCvType>::value));
#endif
//..

      } break;
      case 1: {
        // --------------------------------------------------------------------
        // 'bsl::add_cv::type'
        //   Ensure that the 'typedef' 'type' of 'bsl::add_cv' has the correct
        //   type for a variety of template parameter types.
        //
        //   Ensure that the 'bsl::add_cv_t' represents the correct type for a
        //   variety of template parameter types.
        //
        // Concerns:
        //: 1 'bsl::add_cv' adds a top-level 'const'-qualifier and a top-level
        //:   'volatile'-qualifier only to regular types (primitive, pointers,
        //:   and user-defined types).
        //:
        //: 2 'bsl::add_cv' does not add a 'const'-qualifier nor a
        //:   'volatile'-qualifier to reference types, function types, or types
        //:   that are already both 'const'-qualified and 'volatile'-qualified.
        //:
        //: 3 'bsl::add_cv_t' represents the return type of 'bsl::add_cv'
        //:   meta-function for a variety of template parameter types.
        //
        // Plan:
        //  1 Verify that 'bsl::add_cv::type' and 'bsl::add_cv_t' have the
        //    correct type for each concern. (C1-2)
        //
        //  2 Verify that 'bsl::add_cv_t' has the same type as the return
        //    type of 'bsl::add_cv' for a variety of template parameter types.
        //    (C-3)
        //
        // Testing:
        //   bsl::add_cv::type
        //   bsl::add_cv_t
        // --------------------------------------------------------------------

        if (verbose) printf("\n'bsl::add_cv::type'\n"
                            "\n===================\n");

        // C-1
        ASSERT((is_same<add_cv<int>::type, int const volatile>::value));
        ASSERT((is_same<add_cv<int const>::type, int const volatile>::value));
        ASSERT((is_same<add_cv<int volatile>::type,
                                                  int const volatile>::value));

        ASSERT((is_same<add_cv<int *>::type, int * const volatile>::value));
        ASSERT((is_same<add_cv<int const volatile *>::type,
                                 int const volatile * const volatile>::value));
        ASSERT((is_same<add_cv<TestType>::type,
                                             TestType const volatile>::value));

        // C-2
        ASSERT((is_same<add_cv<int &>::type, int &>::value));
        ASSERT((is_same<add_cv<TestType &>::type, TestType &>::value));
        ASSERT((is_same<add_cv<int (int)>::type, int (int)>::value));
        ASSERT((is_same<add_cv<int const volatile>::type,
                                                  int const volatile>::value));

#ifdef BSLS_COMPILERFEATURES_SUPPORT_ALIAS_TEMPLATES

        if (verbose) printf("\n'bsl::add_cv_t'\n"
                            "\n===============\n");

        // C-3
        ASSERT((is_same<add_cv  <int                 >::type,
                        add_cv_t<int                 >>::value));
        ASSERT((is_same<add_cv  <int const           >::type,
                        add_cv_t<int const           >>::value));
        ASSERT((is_same<add_cv  <int const volatile  >::type,
                        add_cv_t<int const volatile  >>::value));
        ASSERT((is_same<add_cv  <int *               >::type,
                        add_cv_t<int *               >>::value));
        ASSERT((is_same<add_cv  <int const volatile *>::type,
                        add_cv_t<int const volatile *>>::value));
        ASSERT((is_same<add_cv  <TestType            >::type,
                        add_cv_t<TestType            >>::value));

        ASSERT((is_same<add_cv  <int &             >::type,
                        add_cv_t<int &             >>::value));
        ASSERT((is_same<add_cv  <TestType &        >::type,
                        add_cv_t<TestType &        >>::value));
        ASSERT((is_same<add_cv  <int (int)         >::type,
                        add_cv_t<int (int)         >>::value));
        ASSERT((is_same<add_cv  <int const volatile>::type,
                        add_cv_t<int const volatile>>::value));

       ASSERT((is_same<add_const  <MethodPtrTestType               >::type,
                       add_const_t<MethodPtrTestType               >>::value));
       ASSERT((is_same<add_const  <MethodPtrTestType &             >::type,
                       add_const_t<MethodPtrTestType &             >>::value));
       ASSERT((is_same<add_const  <MethodPtrTestType *             >::type,
                       add_const_t<MethodPtrTestType *             >>::value));
       ASSERT((is_same<add_const  <MethodPtrTestType const         >::type,
                       add_const_t<MethodPtrTestType const         >>::value));
       ASSERT((is_same<add_const  <MethodPtrTestType volatile      >::type,
                       add_const_t<MethodPtrTestType volatile      >>::value));
       ASSERT((is_same<add_const  <MethodPtrTestType const volatile>::type,
                       add_const_t<MethodPtrTestType const volatile>>::value));

       ASSERT((is_same<add_const  <PMD               >::type,
                       add_const_t<PMD               >>::value));
       ASSERT((is_same<add_const  <PMD &             >::type,
                       add_const_t<PMD &             >>::value));
       ASSERT((is_same<add_const  <PMD *             >::type,
                       add_const_t<PMD *             >>::value));
       ASSERT((is_same<add_const  <PMD const         >::type,
                       add_const_t<PMD const         >>::value));
       ASSERT((is_same<add_const  <PMD volatile      >::type,
                       add_const_t<PMD volatile      >>::value));
       ASSERT((is_same<add_const  <PMD const volatile>::type,
                       add_const_t<PMD const volatile>>::value));

#endif  // BSLS_COMPILERFEATURES_SUPPORT_ALIAS_TEMPLATES
      } break;
      default: {
        fprintf(stderr, "WARNING: CASE `%d' NOT FOUND.\n", test);
        testStatus = -1;
      }
    }

    if (testStatus > 0) {
        fprintf(stderr, "Error, non-zero test status = %d.\n", testStatus);
    }

    return testStatus;
}

// ----------------------------------------------------------------------------
// Copyright 2013 Bloomberg Finance L.P.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// ----------------------------- END-OF-FILE ----------------------------------
