// bslstl_errc.t.cpp                                                  -*-C++-*-
#include <bslstl_errc.h>

#include <bsls_bsltestutil.h>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>

using namespace BloombergLP;
using namespace bsl;

// ============================================================================
//                             TEST PLAN
// ----------------------------------------------------------------------------
//
//                              Overview
//                              --------
// The component under test defines an enumeration defining longer and more
// descriptive symbolic names for 'errno' error value and a type alias
// designating the type of the enumeration.
//
// ----------------------------------------------------------------------------
//
// [ 2] USAGE EXAMPLE
// [ 1] bsl::ErrcEnum

// ============================================================================
//                     STANDARD BSL ASSERT TEST FUNCTION
// ----------------------------------------------------------------------------

namespace {

int testStatus = 0;

void aSsErT(bool condition, const char *message, int line)
{
    if (condition) {
        printf("Error " __FILE__ "(%d): %s    (failed)\n", line, message);

        if (0 <= testStatus && testStatus <= 100) {
            ++testStatus;
        }
    }
}

}  // close unnamed namespace

// ============================================================================
//               STANDARD BSL TEST DRIVER MACRO ABBREVIATIONS
// ----------------------------------------------------------------------------

#define ASSERT       BSLS_BSLTESTUTIL_ASSERT
#define ASSERTV      BSLS_BSLTESTUTIL_ASSERTV

#define LOOP_ASSERT  BSLS_BSLTESTUTIL_LOOP_ASSERT
#define LOOP0_ASSERT BSLS_BSLTESTUTIL_LOOP0_ASSERT
#define LOOP1_ASSERT BSLS_BSLTESTUTIL_LOOP1_ASSERT
#define LOOP2_ASSERT BSLS_BSLTESTUTIL_LOOP2_ASSERT
#define LOOP3_ASSERT BSLS_BSLTESTUTIL_LOOP3_ASSERT
#define LOOP4_ASSERT BSLS_BSLTESTUTIL_LOOP4_ASSERT
#define LOOP5_ASSERT BSLS_BSLTESTUTIL_LOOP5_ASSERT
#define LOOP6_ASSERT BSLS_BSLTESTUTIL_LOOP6_ASSERT

#define Q            BSLS_BSLTESTUTIL_Q   // Quote identifier literally.
#define P            BSLS_BSLTESTUTIL_P   // Print identifier and value.
#define P_           BSLS_BSLTESTUTIL_P_  // P(X) without '\n'.
#define T_           BSLS_BSLTESTUTIL_T_  // Print a tab (w/o newline).
#define L_           BSLS_BSLTESTUTIL_L_  // current Line number

//=============================================================================
//                      GLOBAL HELPER FUNCTIONS FOR TESTING
//-----------------------------------------------------------------------------

//=============================================================================
//                              MAIN PROGRAM
//-----------------------------------------------------------------------------
int main(int argc, char *argv[])
{
    int test                = argc > 1 ? atoi(argv[1]) : 0;
    int verbose             = argc > 2;
    int veryVerbose         = argc > 3; (void)veryVerbose;
    int veryVeryVerbose     = argc > 4; (void)veryVeryVerbose;
    int veryVeryVeryVerbose = argc > 5; (void)veryVeryVeryVerbose;

    printf("TEST " __FILE__ " CASE %d\n", test);

    switch (test) { case 0:  // Zero is always the leading case.
      case 2: {
        // --------------------------------------------------------------------
        // USAGE EXAMPLE
        //
        // Concerns:
        //: 1 The usage example provided in the component header file compiles,
        //:   links, and runs as shown.
        //
        // Plan:
        //: 1 Incorporate usage example from header into test driver, remove
        //:   leading comment characters, and replace 'assert' with 'ASSERT'.
        //:   (C-1)
        //
        // Testing:
        //   USAGE EXAMPLE
        // --------------------------------------------------------------------

        if (verbose) printf("\nUSAGE EXAMPLE"
                            "\n=============\n");

///Example 1: Descriptive System Error Value Names
///- - - - - - - - - - - - - - - - - - - - - - - -
// Suppose we would like to have more descriptive and non-macro names for
// 'errno' system error values.  We can use the 'errc' from the 'system_error'
// facility of the C++ standard for this.
//
// First, we do something that will set an error value.
//..
    strtod("1e2000", 0);
//..
// Finally, we check for that error using the descriptive name.
//..
    bsl::ErrcEnum expected = bsl::errc::result_out_of_range;
    ASSERT(static_cast<int>(expected) == errno);
//..
      } break;
      case 1: {
        // --------------------------------------------------------------------
        // TESTING 'errc'
        //
        // Concerns:
        //: 1 The 'bsl::ErrcEnum' type can be used to declare 'errc' objects.
        //: 2 'bsl::errc' can qualify 'errc' enumeration literals.
        //: 3 An 'errc' literal matches the 'POSIX' error code.
        //: 4 The 'errc' type is registered as an error condition.
        //
        // Plan:
        //: 1 Declare a 'bsl::ErrcEnum' object and initialize it with a
        //:   'bsl::errc' qualified value.  (C-1,2)
        //:
        //: 2 Compare the value of the object with the corresponding 'POSIX'
        //:   'errno' value.  (C-3)
        //:
        //: 3 Verify that 'is_error_condition' is true for 'bsl::ErrcEnum'.
        //:   (C-4)
        //
        // Testing:
        //   bsl::ErrcEnum
        // --------------------------------------------------------------------

        if (verbose) printf("\nTESTING 'errc'"
                            "\n==============\n");

        bsl::ErrcEnum value = bsl::errc::inappropriate_io_control_operation;
        ASSERT(ENOTTY == static_cast<int>(value));
        ASSERT(bsl::is_error_condition_enum<bsl::ErrcEnum>::value);
      } break;
      default: {
        fprintf(stderr, "WARNING: CASE `%d' NOT FOUND.\n", test);
        testStatus = -1;
      }
    }

    if (testStatus > 0) {
        fprintf(stderr, "Error, non-zero test status = %d.\n", testStatus);
    }

    return testStatus;
}

// ----------------------------------------------------------------------------
// Copyright 2019 Bloomberg Finance L.P.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// ----------------------------- END-OF-FILE ----------------------------------
