// bdlt_prolepticdateimputil.cpp                                      -*-C++-*-
#include <bdlt_prolepticdateimputil.h>

#include <bsls_ident.h>
BSLS_IDENT_RCSID(bdlt_prolepticdateimputil_cpp,"$Id$ $CSID$")

namespace BloombergLP {
namespace bdlt {

namespace {

enum {
    // months of the year used in the implementation

    k_JAN                   =     1,
    k_FEB                   =     2,
    k_DEC                   =    12,

    // other useful constants

    k_MIN_DAY               =     1,

    k_MIN_MONTH             = k_JAN,
    k_MAX_MONTH             = k_DEC,

    k_MIN_YEAR              =     1,
    k_MAX_YEAR              =  9999,

    k_MIN_DAY_OF_YEAR       =     1,

    k_DAYS_IN_NON_LEAP_YEAR =   365,
    k_DAYS_IN_LEAP_YEAR     =   366,

    k_DAYS_IN_4_YEARS       = k_DAYS_IN_NON_LEAP_YEAR * 4 + 1,  //   1,461
    k_DAYS_IN_100_YEARS     =  25 * k_DAYS_IN_4_YEARS     - 1,  //  36,524

    k_DAYS_IN_400_YEAR_ERA  =   4 * k_DAYS_IN_100_YEARS   + 1   // 146,097
};

// Note that, in each of the following arrays, the element at index position 0
// is always the value 0 (in order to facilitate asking questions involving all
// months up through the *previous* one).

static const int normDaysThroughMonth[] = {
    // Jan, Feb, Mar, Apr, May, Jun, Jul, Aug, Sep, Oct, Nov, Dec
    0,  31,  59,  90, 120, 151, 181, 212, 243, 273, 304, 334, 365
};

static const int leapDaysThroughMonth[] = {
    // Jan, Feb, Mar, Apr, May, Jun, Jul, Aug, Sep, Oct, Nov, Dec
    0,  31,  60,  91, 121, 152, 182, 213, 244, 274, 305, 335, 366
};

static const int normDaysPerMonth[] = {
    // Jan, Feb, Mar, Apr, May, Jun, Jul, Aug, Sep, Oct, Nov, Dec
    0,  31,  28,  31,  30,  31,  30,  31,  31,  30,  31,  30,  31
};

static const int leapDaysPerMonth[] = {
    // Jan, Feb, Mar, Apr, May, Jun, Jul, Aug, Sep, Oct, Nov, Dec
    0,  31,  29,  31,  30,  31,  30,  31,  31,  30,  31,  30,  31
};

static const unsigned char dayOfYear2Month[2][k_DAYS_IN_LEAP_YEAR] = {
    // used for 365-day years
    {
         1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,
         1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,

         2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,
         2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,          // 28 days

         3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,
         3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,

         4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,
         4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,

         5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,
         5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,

         6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,
         6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,

         7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,
         7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,

         8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,
         8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,

         9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,
         9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,

        10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
        10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,

        11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
        11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,

        12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12,
        12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12,

        255  // should never be accessed
    },

    // used for 366-day years
    {
         1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,
         1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,

         2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,
         2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,      // 29 days

         3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,
         3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,  3,

         4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,
         4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,

         5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,
         5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,

         6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,
         6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,  6,

         7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,
         7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,

         8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,
         8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,  8,

         9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,
         9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,  9,

        10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,
        10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10,

        11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,
        11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11, 11,

        12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12,
        12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12, 12
    }
};

// FORWARD DECLARATIONS OF STATIC HELPER FUNCTIONS

static inline
const int *getArrayDaysThroughMonth(int);

static inline
int numLeapYearsSoFar(int);

// STATIC HELPER FUNCTIONS

static inline
int calendarDaysThroughMonth(int year, int month)
    // Return the number of calendar days since the start of the specified
    // 'year' and including the days in the specified 'month'.  The behavior is
    // undefined unless 'k_MIN_YEAR <= year <= k_MAX_YEAR' and
    // '0 <= month <= k_MAX_MONTH'.
{
    BSLS_REVIEW(k_MIN_YEAR <= year);
    BSLS_REVIEW(              year  <= k_MAX_YEAR);
    BSLS_REVIEW(         0 <= month);
    BSLS_REVIEW(              month <= k_MAX_MONTH);

    return getArrayDaysThroughMonth(year)[month];
}

static inline
const int *getArrayDaysThroughMonth(int year)
    // Return the address of a static array that, for the specified 'year', can
    // be used to determine the number of days up to and including the month
    // indicated by an integer index in the range '[0 .. k_MAX_MONTH]', where
    // an index of 0 always results in the value 0.  The behavior is undefined
    // unless 'k_MIN_YEAR <= year <= k_MAX_YEAR'.
{
    BSLS_REVIEW(k_MIN_YEAR <= year);
    BSLS_REVIEW(              year <= k_MAX_YEAR);

    return bdlt::ProlepticDateImpUtil::isLeapYear(year)
           ? leapDaysThroughMonth
           : normDaysThroughMonth;
}

static inline
int numDaysInPreviousYears(int year)
    // Return the total number of days in all years, beginning with the year 1,
    // up to but not including the specified 'year'.  The behavior is undefined
    // unless 'k_MIN_YEAR <= year <= k_MAX_YEAR'.
{
    BSLS_REVIEW(k_MIN_YEAR <= year);
    BSLS_REVIEW(              year <= k_MAX_YEAR);

    const int y = year - 1;

    return y * k_DAYS_IN_NON_LEAP_YEAR + numLeapYearsSoFar(y);
}

static inline
int numLeapYearsSoFar(int year)
    // Return the number of leap years from year 1 to the specified 'year'.
    // The behavior is undefined unless '0 <= year <= k_MAX_YEAR'.
{
    BSLS_REVIEW(0 <= year);
    BSLS_REVIEW(     year <= k_MAX_YEAR);

    return year / 4 - year / 100 + year / 400;
}

}  // close unnamed namespace

                           // ---------------------------
                           // struct ProlepticDateImpUtil
                           // ---------------------------

// CLASS METHODS
int ProlepticDateImpUtil::lastDayOfMonth(int year, int month)
{
    BSLS_ASSERT(k_MIN_YEAR  <= year);
    BSLS_ASSERT(               year  <= k_MAX_YEAR);
    BSLS_ASSERT(k_MIN_MONTH <= month);
    BSLS_ASSERT(               month <= k_MAX_MONTH);

    return normDaysPerMonth[month] + (k_FEB == month && isLeapYear(year));
}

int ProlepticDateImpUtil::numLeapYears(int year1, int year2)
{
    BSLS_ASSERT(k_MIN_YEAR <= year1);
    BSLS_ASSERT(              year1 <= k_MAX_YEAR);
    BSLS_ASSERT(k_MIN_YEAR <= year2);
    BSLS_ASSERT(              year2 <= k_MAX_YEAR);
    BSLS_ASSERT(year1 <= year2);

    return numLeapYearsSoFar(year2) - numLeapYearsSoFar(year1 - 1);
}

                        // Is Valid Date

bool ProlepticDateImpUtil::isValidYearDay(int year, int dayOfYear)
{
    if (year < k_MIN_YEAR || year > k_MAX_YEAR) {
        return false;                                                 // RETURN
    }

    const int *daysThroughMonth = getArrayDaysThroughMonth(year);

    return k_MIN_DAY_OF_YEAR <= dayOfYear
        && dayOfYear         <= daysThroughMonth[k_DEC];
}

bool
ProlepticDateImpUtil::isValidYearMonthDayNoCache(int year, int month, int day)
{
    if (year < k_MIN_YEAR || month < k_MIN_MONTH || day < k_MIN_DAY
     || year > k_MAX_YEAR || month > k_MAX_MONTH) {

        return false;                                                 // RETURN
    }

    const int daysInMonth = isLeapYear(year)
                            ? leapDaysPerMonth[month]
                            : normDaysPerMonth[month];

    return day <= daysInMonth;
}

                        // To Serial Date (s)

int ProlepticDateImpUtil::ydToSerial(int year, int dayOfYear)
{
    BSLS_ASSERT(isValidYearDay(year, dayOfYear));

    return numDaysInPreviousYears(year) + dayOfYear;
}

int ProlepticDateImpUtil::ymdToSerial(int year, int month, int day)
{
    BSLS_ASSERT(isValidYearMonthDay(year, month, day));

    if (s_firstCachedYear <= year && year <= s_lastCachedYear) {
        return s_cachedSerialDate[year - s_firstCachedYear][month] + day;
                                                                      // RETURN
    }
    else {
        return ymdToSerialNoCache(year, month, day);                  // RETURN
    }
}

int ProlepticDateImpUtil::ymdToSerialNoCache(int year, int month, int day)
{
    BSLS_ASSERT(isValidYearMonthDay(year, month, day));

    const int y = year - 1;

    return y * k_DAYS_IN_NON_LEAP_YEAR     // standard days for each year
           + numLeapYearsSoFar(y)          // additional leap days
           + calendarDaysThroughMonth(year, month - 1)
                                           // additional days before this month
           + day;                          // additional days this month
}

                        // To Day-Of-Year Date (yd)

void ProlepticDateImpUtil::serialToYd(int *year, int *dayOfYear, int serialDay)
{
    BSLS_ASSERT(year);
    BSLS_ASSERT(dayOfYear);
    BSLS_ASSERT(isValidSerial(serialDay));

    // 'sm1': serial day minus 1

    const unsigned sm1 = static_cast<unsigned>(serialDay - 1);

    // 'era': 0-based 400-year "era"

    const unsigned era = sm1 / k_DAYS_IN_400_YEAR_ERA;

    // 'doe': 0-based day of the era

    const unsigned doe = sm1 - era * k_DAYS_IN_400_YEAR_ERA;   // [0 .. 146096]

    // 'yoe': 0-based year of the era

    const unsigned yoe =
     (doe - doe / 1460 + doe / 36524 - doe / 146096) / k_DAYS_IN_NON_LEAP_YEAR;
                                                               // [0 .. 399]

    // 'ym1': year minus 1

    const unsigned ym1 = era * 400 + yoe;

    // 'doy': 0-based day of the year

    const unsigned doy =
                sm1 - (ym1 * k_DAYS_IN_NON_LEAP_YEAR + numLeapYearsSoFar(ym1));

    *year      = static_cast<int>(ym1 + 1);

    *dayOfYear = static_cast<int>(doy + 1);
}

int ProlepticDateImpUtil::ymdToDayOfYear(int year, int month, int day)
{
    BSLS_ASSERT(isValidYearMonthDay(year, month, day));

    const int *daysThroughMonth = getArrayDaysThroughMonth(year);

    return daysThroughMonth[month - 1] + day;
}

                        // To Calendar Date (ymd)

int ProlepticDateImpUtil::serialToDay(int serialDay)
{
    BSLS_REVIEW(isValidSerial(serialDay));

    if (s_firstCachedSerialDate <= serialDay
                                && serialDay <= s_lastCachedSerialDate) {
        return s_cachedYearMonthDay[serialDay - s_firstCachedSerialDate].d_day;
                                                                      // RETURN
    }
    else {
        return serialToDayNoCache(serialDay);                         // RETURN
    }
}

int ProlepticDateImpUtil::serialToMonth(int serialDay)
{
    BSLS_REVIEW(isValidSerial(serialDay));

    if (s_firstCachedSerialDate <= serialDay
                                && serialDay <= s_lastCachedSerialDate) {
        return
            s_cachedYearMonthDay[serialDay - s_firstCachedSerialDate].d_month;
                                                                      // RETURN
    }
    else {
        return serialToMonthNoCache(serialDay);                       // RETURN
    }
}

int ProlepticDateImpUtil::serialToYear(int serialDay)
{
    BSLS_REVIEW(isValidSerial(serialDay));

    if (s_firstCachedSerialDate <= serialDay
                                && serialDay <= s_lastCachedSerialDate) {
        return
            s_cachedYearMonthDay[serialDay - s_firstCachedSerialDate].d_year;
                                                                      // RETURN
    }
    else {
        return serialToYearNoCache(serialDay);                        // RETURN
    }
}

void ProlepticDateImpUtil::serialToYmd(int *year,
                                       int *month,
                                       int *day,
                                       int  serialDay)
{
    BSLS_ASSERT(year);
    BSLS_ASSERT(month);
    BSLS_ASSERT(day);
    BSLS_REVIEW(isValidSerial(serialDay));

    if (s_firstCachedSerialDate <= serialDay
                                && serialDay <= s_lastCachedSerialDate) {
        const YearMonthDay *ymd =
                    s_cachedYearMonthDay + serialDay - s_firstCachedSerialDate;

        *year  = ymd->d_year;
        *month = ymd->d_month;
        *day   = ymd->d_day;
    }
    else {
        serialToYmdNoCache(year, month, day, serialDay);
    }
}

void ProlepticDateImpUtil::ydToMd(int *month,
                                  int *day,
                                  int  year,
                                  int  dayOfYear)
{
    BSLS_ASSERT(month);
    BSLS_ASSERT(day);
    BSLS_ASSERT(isValidYearDay(year, dayOfYear));

    const bool isLeapFlag = isLeapYear(year);

    *month = dayOfYear2Month[isLeapFlag][dayOfYear - 1];

    const int *daysThroughMonth = isLeapFlag
                                  ? leapDaysThroughMonth
                                  : normDaysThroughMonth;

    *day = dayOfYear - daysThroughMonth[*month - 1];
}

// ============================================================================
//                    MACHINE-GENERATED DATA GOES HERE
// ============================================================================

const int ProlepticDateImpUtil::s_firstCachedYear       = 1980;
const int ProlepticDateImpUtil::s_lastCachedYear        = 2040;
const int ProlepticDateImpUtil::s_firstCachedSerialDate = 722815;
const int ProlepticDateImpUtil::s_lastCachedSerialDate  = 745095;

const ProlepticDateImpUtil::YearMonthDay
    ProlepticDateImpUtil::s_cachedYearMonthDay[] = {
  { 1980,  1,  1 },  { 1980,  1,  2 },  { 1980,  1,  3 },  { 1980,  1,  4 },
  { 1980,  1,  5 },  { 1980,  1,  6 },  { 1980,  1,  7 },  { 1980,  1,  8 },
  { 1980,  1,  9 },  { 1980,  1, 10 },  { 1980,  1, 11 },  { 1980,  1, 12 },
  { 1980,  1, 13 },  { 1980,  1, 14 },  { 1980,  1, 15 },  { 1980,  1, 16 },
  { 1980,  1, 17 },  { 1980,  1, 18 },  { 1980,  1, 19 },  { 1980,  1, 20 },
  { 1980,  1, 21 },  { 1980,  1, 22 },  { 1980,  1, 23 },  { 1980,  1, 24 },
  { 1980,  1, 25 },  { 1980,  1, 26 },  { 1980,  1, 27 },  { 1980,  1, 28 },
  { 1980,  1, 29 },  { 1980,  1, 30 },  { 1980,  1, 31 },  { 1980,  2,  1 },
  { 1980,  2,  2 },  { 1980,  2,  3 },  { 1980,  2,  4 },  { 1980,  2,  5 },
  { 1980,  2,  6 },  { 1980,  2,  7 },  { 1980,  2,  8 },  { 1980,  2,  9 },
  { 1980,  2, 10 },  { 1980,  2, 11 },  { 1980,  2, 12 },  { 1980,  2, 13 },
  { 1980,  2, 14 },  { 1980,  2, 15 },  { 1980,  2, 16 },  { 1980,  2, 17 },
  { 1980,  2, 18 },  { 1980,  2, 19 },  { 1980,  2, 20 },  { 1980,  2, 21 },
  { 1980,  2, 22 },  { 1980,  2, 23 },  { 1980,  2, 24 },  { 1980,  2, 25 },
  { 1980,  2, 26 },  { 1980,  2, 27 },  { 1980,  2, 28 },  { 1980,  2, 29 },
  { 1980,  3,  1 },  { 1980,  3,  2 },  { 1980,  3,  3 },  { 1980,  3,  4 },
  { 1980,  3,  5 },  { 1980,  3,  6 },  { 1980,  3,  7 },  { 1980,  3,  8 },
  { 1980,  3,  9 },  { 1980,  3, 10 },  { 1980,  3, 11 },  { 1980,  3, 12 },
  { 1980,  3, 13 },  { 1980,  3, 14 },  { 1980,  3, 15 },  { 1980,  3, 16 },
  { 1980,  3, 17 },  { 1980,  3, 18 },  { 1980,  3, 19 },  { 1980,  3, 20 },
  { 1980,  3, 21 },  { 1980,  3, 22 },  { 1980,  3, 23 },  { 1980,  3, 24 },
  { 1980,  3, 25 },  { 1980,  3, 26 },  { 1980,  3, 27 },  { 1980,  3, 28 },
  { 1980,  3, 29 },  { 1980,  3, 30 },  { 1980,  3, 31 },  { 1980,  4,  1 },
  { 1980,  4,  2 },  { 1980,  4,  3 },  { 1980,  4,  4 },  { 1980,  4,  5 },
  { 1980,  4,  6 },  { 1980,  4,  7 },  { 1980,  4,  8 },  { 1980,  4,  9 },
  { 1980,  4, 10 },  { 1980,  4, 11 },  { 1980,  4, 12 },  { 1980,  4, 13 },
  { 1980,  4, 14 },  { 1980,  4, 15 },  { 1980,  4, 16 },  { 1980,  4, 17 },
  { 1980,  4, 18 },  { 1980,  4, 19 },  { 1980,  4, 20 },  { 1980,  4, 21 },
  { 1980,  4, 22 },  { 1980,  4, 23 },  { 1980,  4, 24 },  { 1980,  4, 25 },
  { 1980,  4, 26 },  { 1980,  4, 27 },  { 1980,  4, 28 },  { 1980,  4, 29 },
  { 1980,  4, 30 },  { 1980,  5,  1 },  { 1980,  5,  2 },  { 1980,  5,  3 },
  { 1980,  5,  4 },  { 1980,  5,  5 },  { 1980,  5,  6 },  { 1980,  5,  7 },
  { 1980,  5,  8 },  { 1980,  5,  9 },  { 1980,  5, 10 },  { 1980,  5, 11 },
  { 1980,  5, 12 },  { 1980,  5, 13 },  { 1980,  5, 14 },  { 1980,  5, 15 },
  { 1980,  5, 16 },  { 1980,  5, 17 },  { 1980,  5, 18 },  { 1980,  5, 19 },
  { 1980,  5, 20 },  { 1980,  5, 21 },  { 1980,  5, 22 },  { 1980,  5, 23 },
  { 1980,  5, 24 },  { 1980,  5, 25 },  { 1980,  5, 26 },  { 1980,  5, 27 },
  { 1980,  5, 28 },  { 1980,  5, 29 },  { 1980,  5, 30 },  { 1980,  5, 31 },
  { 1980,  6,  1 },  { 1980,  6,  2 },  { 1980,  6,  3 },  { 1980,  6,  4 },
  { 1980,  6,  5 },  { 1980,  6,  6 },  { 1980,  6,  7 },  { 1980,  6,  8 },
  { 1980,  6,  9 },  { 1980,  6, 10 },  { 1980,  6, 11 },  { 1980,  6, 12 },
  { 1980,  6, 13 },  { 1980,  6, 14 },  { 1980,  6, 15 },  { 1980,  6, 16 },
  { 1980,  6, 17 },  { 1980,  6, 18 },  { 1980,  6, 19 },  { 1980,  6, 20 },
  { 1980,  6, 21 },  { 1980,  6, 22 },  { 1980,  6, 23 },  { 1980,  6, 24 },
  { 1980,  6, 25 },  { 1980,  6, 26 },  { 1980,  6, 27 },  { 1980,  6, 28 },
  { 1980,  6, 29 },  { 1980,  6, 30 },  { 1980,  7,  1 },  { 1980,  7,  2 },
  { 1980,  7,  3 },  { 1980,  7,  4 },  { 1980,  7,  5 },  { 1980,  7,  6 },
  { 1980,  7,  7 },  { 1980,  7,  8 },  { 1980,  7,  9 },  { 1980,  7, 10 },
  { 1980,  7, 11 },  { 1980,  7, 12 },  { 1980,  7, 13 },  { 1980,  7, 14 },
  { 1980,  7, 15 },  { 1980,  7, 16 },  { 1980,  7, 17 },  { 1980,  7, 18 },
  { 1980,  7, 19 },  { 1980,  7, 20 },  { 1980,  7, 21 },  { 1980,  7, 22 },
  { 1980,  7, 23 },  { 1980,  7, 24 },  { 1980,  7, 25 },  { 1980,  7, 26 },
  { 1980,  7, 27 },  { 1980,  7, 28 },  { 1980,  7, 29 },  { 1980,  7, 30 },
  { 1980,  7, 31 },  { 1980,  8,  1 },  { 1980,  8,  2 },  { 1980,  8,  3 },
  { 1980,  8,  4 },  { 1980,  8,  5 },  { 1980,  8,  6 },  { 1980,  8,  7 },
  { 1980,  8,  8 },  { 1980,  8,  9 },  { 1980,  8, 10 },  { 1980,  8, 11 },
  { 1980,  8, 12 },  { 1980,  8, 13 },  { 1980,  8, 14 },  { 1980,  8, 15 },
  { 1980,  8, 16 },  { 1980,  8, 17 },  { 1980,  8, 18 },  { 1980,  8, 19 },
  { 1980,  8, 20 },  { 1980,  8, 21 },  { 1980,  8, 22 },  { 1980,  8, 23 },
  { 1980,  8, 24 },  { 1980,  8, 25 },  { 1980,  8, 26 },  { 1980,  8, 27 },
  { 1980,  8, 28 },  { 1980,  8, 29 },  { 1980,  8, 30 },  { 1980,  8, 31 },
  { 1980,  9,  1 },  { 1980,  9,  2 },  { 1980,  9,  3 },  { 1980,  9,  4 },
  { 1980,  9,  5 },  { 1980,  9,  6 },  { 1980,  9,  7 },  { 1980,  9,  8 },
  { 1980,  9,  9 },  { 1980,  9, 10 },  { 1980,  9, 11 },  { 1980,  9, 12 },
  { 1980,  9, 13 },  { 1980,  9, 14 },  { 1980,  9, 15 },  { 1980,  9, 16 },
  { 1980,  9, 17 },  { 1980,  9, 18 },  { 1980,  9, 19 },  { 1980,  9, 20 },
  { 1980,  9, 21 },  { 1980,  9, 22 },  { 1980,  9, 23 },  { 1980,  9, 24 },
  { 1980,  9, 25 },  { 1980,  9, 26 },  { 1980,  9, 27 },  { 1980,  9, 28 },
  { 1980,  9, 29 },  { 1980,  9, 30 },  { 1980, 10,  1 },  { 1980, 10,  2 },
  { 1980, 10,  3 },  { 1980, 10,  4 },  { 1980, 10,  5 },  { 1980, 10,  6 },
  { 1980, 10,  7 },  { 1980, 10,  8 },  { 1980, 10,  9 },  { 1980, 10, 10 },
  { 1980, 10, 11 },  { 1980, 10, 12 },  { 1980, 10, 13 },  { 1980, 10, 14 },
  { 1980, 10, 15 },  { 1980, 10, 16 },  { 1980, 10, 17 },  { 1980, 10, 18 },
  { 1980, 10, 19 },  { 1980, 10, 20 },  { 1980, 10, 21 },  { 1980, 10, 22 },
  { 1980, 10, 23 },  { 1980, 10, 24 },  { 1980, 10, 25 },  { 1980, 10, 26 },
  { 1980, 10, 27 },  { 1980, 10, 28 },  { 1980, 10, 29 },  { 1980, 10, 30 },
  { 1980, 10, 31 },  { 1980, 11,  1 },  { 1980, 11,  2 },  { 1980, 11,  3 },
  { 1980, 11,  4 },  { 1980, 11,  5 },  { 1980, 11,  6 },  { 1980, 11,  7 },
  { 1980, 11,  8 },  { 1980, 11,  9 },  { 1980, 11, 10 },  { 1980, 11, 11 },
  { 1980, 11, 12 },  { 1980, 11, 13 },  { 1980, 11, 14 },  { 1980, 11, 15 },
  { 1980, 11, 16 },  { 1980, 11, 17 },  { 1980, 11, 18 },  { 1980, 11, 19 },
  { 1980, 11, 20 },  { 1980, 11, 21 },  { 1980, 11, 22 },  { 1980, 11, 23 },
  { 1980, 11, 24 },  { 1980, 11, 25 },  { 1980, 11, 26 },  { 1980, 11, 27 },
  { 1980, 11, 28 },  { 1980, 11, 29 },  { 1980, 11, 30 },  { 1980, 12,  1 },
  { 1980, 12,  2 },  { 1980, 12,  3 },  { 1980, 12,  4 },  { 1980, 12,  5 },
  { 1980, 12,  6 },  { 1980, 12,  7 },  { 1980, 12,  8 },  { 1980, 12,  9 },
  { 1980, 12, 10 },  { 1980, 12, 11 },  { 1980, 12, 12 },  { 1980, 12, 13 },
  { 1980, 12, 14 },  { 1980, 12, 15 },  { 1980, 12, 16 },  { 1980, 12, 17 },
  { 1980, 12, 18 },  { 1980, 12, 19 },  { 1980, 12, 20 },  { 1980, 12, 21 },
  { 1980, 12, 22 },  { 1980, 12, 23 },  { 1980, 12, 24 },  { 1980, 12, 25 },
  { 1980, 12, 26 },  { 1980, 12, 27 },  { 1980, 12, 28 },  { 1980, 12, 29 },
  { 1980, 12, 30 },  { 1980, 12, 31 },  { 1981,  1,  1 },  { 1981,  1,  2 },
  { 1981,  1,  3 },  { 1981,  1,  4 },  { 1981,  1,  5 },  { 1981,  1,  6 },
  { 1981,  1,  7 },  { 1981,  1,  8 },  { 1981,  1,  9 },  { 1981,  1, 10 },
  { 1981,  1, 11 },  { 1981,  1, 12 },  { 1981,  1, 13 },  { 1981,  1, 14 },
  { 1981,  1, 15 },  { 1981,  1, 16 },  { 1981,  1, 17 },  { 1981,  1, 18 },
  { 1981,  1, 19 },  { 1981,  1, 20 },  { 1981,  1, 21 },  { 1981,  1, 22 },
  { 1981,  1, 23 },  { 1981,  1, 24 },  { 1981,  1, 25 },  { 1981,  1, 26 },
  { 1981,  1, 27 },  { 1981,  1, 28 },  { 1981,  1, 29 },  { 1981,  1, 30 },
  { 1981,  1, 31 },  { 1981,  2,  1 },  { 1981,  2,  2 },  { 1981,  2,  3 },
  { 1981,  2,  4 },  { 1981,  2,  5 },  { 1981,  2,  6 },  { 1981,  2,  7 },
  { 1981,  2,  8 },  { 1981,  2,  9 },  { 1981,  2, 10 },  { 1981,  2, 11 },
  { 1981,  2, 12 },  { 1981,  2, 13 },  { 1981,  2, 14 },  { 1981,  2, 15 },
  { 1981,  2, 16 },  { 1981,  2, 17 },  { 1981,  2, 18 },  { 1981,  2, 19 },
  { 1981,  2, 20 },  { 1981,  2, 21 },  { 1981,  2, 22 },  { 1981,  2, 23 },
  { 1981,  2, 24 },  { 1981,  2, 25 },  { 1981,  2, 26 },  { 1981,  2, 27 },
  { 1981,  2, 28 },  { 1981,  3,  1 },  { 1981,  3,  2 },  { 1981,  3,  3 },
  { 1981,  3,  4 },  { 1981,  3,  5 },  { 1981,  3,  6 },  { 1981,  3,  7 },
  { 1981,  3,  8 },  { 1981,  3,  9 },  { 1981,  3, 10 },  { 1981,  3, 11 },
  { 1981,  3, 12 },  { 1981,  3, 13 },  { 1981,  3, 14 },  { 1981,  3, 15 },
  { 1981,  3, 16 },  { 1981,  3, 17 },  { 1981,  3, 18 },  { 1981,  3, 19 },
  { 1981,  3, 20 },  { 1981,  3, 21 },  { 1981,  3, 22 },  { 1981,  3, 23 },
  { 1981,  3, 24 },  { 1981,  3, 25 },  { 1981,  3, 26 },  { 1981,  3, 27 },
  { 1981,  3, 28 },  { 1981,  3, 29 },  { 1981,  3, 30 },  { 1981,  3, 31 },
  { 1981,  4,  1 },  { 1981,  4,  2 },  { 1981,  4,  3 },  { 1981,  4,  4 },
  { 1981,  4,  5 },  { 1981,  4,  6 },  { 1981,  4,  7 },  { 1981,  4,  8 },
  { 1981,  4,  9 },  { 1981,  4, 10 },  { 1981,  4, 11 },  { 1981,  4, 12 },
  { 1981,  4, 13 },  { 1981,  4, 14 },  { 1981,  4, 15 },  { 1981,  4, 16 },
  { 1981,  4, 17 },  { 1981,  4, 18 },  { 1981,  4, 19 },  { 1981,  4, 20 },
  { 1981,  4, 21 },  { 1981,  4, 22 },  { 1981,  4, 23 },  { 1981,  4, 24 },
  { 1981,  4, 25 },  { 1981,  4, 26 },  { 1981,  4, 27 },  { 1981,  4, 28 },
  { 1981,  4, 29 },  { 1981,  4, 30 },  { 1981,  5,  1 },  { 1981,  5,  2 },
  { 1981,  5,  3 },  { 1981,  5,  4 },  { 1981,  5,  5 },  { 1981,  5,  6 },
  { 1981,  5,  7 },  { 1981,  5,  8 },  { 1981,  5,  9 },  { 1981,  5, 10 },
  { 1981,  5, 11 },  { 1981,  5, 12 },  { 1981,  5, 13 },  { 1981,  5, 14 },
  { 1981,  5, 15 },  { 1981,  5, 16 },  { 1981,  5, 17 },  { 1981,  5, 18 },
  { 1981,  5, 19 },  { 1981,  5, 20 },  { 1981,  5, 21 },  { 1981,  5, 22 },
  { 1981,  5, 23 },  { 1981,  5, 24 },  { 1981,  5, 25 },  { 1981,  5, 26 },
  { 1981,  5, 27 },  { 1981,  5, 28 },  { 1981,  5, 29 },  { 1981,  5, 30 },
  { 1981,  5, 31 },  { 1981,  6,  1 },  { 1981,  6,  2 },  { 1981,  6,  3 },
  { 1981,  6,  4 },  { 1981,  6,  5 },  { 1981,  6,  6 },  { 1981,  6,  7 },
  { 1981,  6,  8 },  { 1981,  6,  9 },  { 1981,  6, 10 },  { 1981,  6, 11 },
  { 1981,  6, 12 },  { 1981,  6, 13 },  { 1981,  6, 14 },  { 1981,  6, 15 },
  { 1981,  6, 16 },  { 1981,  6, 17 },  { 1981,  6, 18 },  { 1981,  6, 19 },
  { 1981,  6, 20 },  { 1981,  6, 21 },  { 1981,  6, 22 },  { 1981,  6, 23 },
  { 1981,  6, 24 },  { 1981,  6, 25 },  { 1981,  6, 26 },  { 1981,  6, 27 },
  { 1981,  6, 28 },  { 1981,  6, 29 },  { 1981,  6, 30 },  { 1981,  7,  1 },
  { 1981,  7,  2 },  { 1981,  7,  3 },  { 1981,  7,  4 },  { 1981,  7,  5 },
  { 1981,  7,  6 },  { 1981,  7,  7 },  { 1981,  7,  8 },  { 1981,  7,  9 },
  { 1981,  7, 10 },  { 1981,  7, 11 },  { 1981,  7, 12 },  { 1981,  7, 13 },
  { 1981,  7, 14 },  { 1981,  7, 15 },  { 1981,  7, 16 },  { 1981,  7, 17 },
  { 1981,  7, 18 },  { 1981,  7, 19 },  { 1981,  7, 20 },  { 1981,  7, 21 },
  { 1981,  7, 22 },  { 1981,  7, 23 },  { 1981,  7, 24 },  { 1981,  7, 25 },
  { 1981,  7, 26 },  { 1981,  7, 27 },  { 1981,  7, 28 },  { 1981,  7, 29 },
  { 1981,  7, 30 },  { 1981,  7, 31 },  { 1981,  8,  1 },  { 1981,  8,  2 },
  { 1981,  8,  3 },  { 1981,  8,  4 },  { 1981,  8,  5 },  { 1981,  8,  6 },
  { 1981,  8,  7 },  { 1981,  8,  8 },  { 1981,  8,  9 },  { 1981,  8, 10 },
  { 1981,  8, 11 },  { 1981,  8, 12 },  { 1981,  8, 13 },  { 1981,  8, 14 },
  { 1981,  8, 15 },  { 1981,  8, 16 },  { 1981,  8, 17 },  { 1981,  8, 18 },
  { 1981,  8, 19 },  { 1981,  8, 20 },  { 1981,  8, 21 },  { 1981,  8, 22 },
  { 1981,  8, 23 },  { 1981,  8, 24 },  { 1981,  8, 25 },  { 1981,  8, 26 },
  { 1981,  8, 27 },  { 1981,  8, 28 },  { 1981,  8, 29 },  { 1981,  8, 30 },
  { 1981,  8, 31 },  { 1981,  9,  1 },  { 1981,  9,  2 },  { 1981,  9,  3 },
  { 1981,  9,  4 },  { 1981,  9,  5 },  { 1981,  9,  6 },  { 1981,  9,  7 },
  { 1981,  9,  8 },  { 1981,  9,  9 },  { 1981,  9, 10 },  { 1981,  9, 11 },
  { 1981,  9, 12 },  { 1981,  9, 13 },  { 1981,  9, 14 },  { 1981,  9, 15 },
  { 1981,  9, 16 },  { 1981,  9, 17 },  { 1981,  9, 18 },  { 1981,  9, 19 },
  { 1981,  9, 20 },  { 1981,  9, 21 },  { 1981,  9, 22 },  { 1981,  9, 23 },
  { 1981,  9, 24 },  { 1981,  9, 25 },  { 1981,  9, 26 },  { 1981,  9, 27 },
  { 1981,  9, 28 },  { 1981,  9, 29 },  { 1981,  9, 30 },  { 1981, 10,  1 },
  { 1981, 10,  2 },  { 1981, 10,  3 },  { 1981, 10,  4 },  { 1981, 10,  5 },
  { 1981, 10,  6 },  { 1981, 10,  7 },  { 1981, 10,  8 },  { 1981, 10,  9 },
  { 1981, 10, 10 },  { 1981, 10, 11 },  { 1981, 10, 12 },  { 1981, 10, 13 },
  { 1981, 10, 14 },  { 1981, 10, 15 },  { 1981, 10, 16 },  { 1981, 10, 17 },
  { 1981, 10, 18 },  { 1981, 10, 19 },  { 1981, 10, 20 },  { 1981, 10, 21 },
  { 1981, 10, 22 },  { 1981, 10, 23 },  { 1981, 10, 24 },  { 1981, 10, 25 },
  { 1981, 10, 26 },  { 1981, 10, 27 },  { 1981, 10, 28 },  { 1981, 10, 29 },
  { 1981, 10, 30 },  { 1981, 10, 31 },  { 1981, 11,  1 },  { 1981, 11,  2 },
  { 1981, 11,  3 },  { 1981, 11,  4 },  { 1981, 11,  5 },  { 1981, 11,  6 },
  { 1981, 11,  7 },  { 1981, 11,  8 },  { 1981, 11,  9 },  { 1981, 11, 10 },
  { 1981, 11, 11 },  { 1981, 11, 12 },  { 1981, 11, 13 },  { 1981, 11, 14 },
  { 1981, 11, 15 },  { 1981, 11, 16 },  { 1981, 11, 17 },  { 1981, 11, 18 },
  { 1981, 11, 19 },  { 1981, 11, 20 },  { 1981, 11, 21 },  { 1981, 11, 22 },
  { 1981, 11, 23 },  { 1981, 11, 24 },  { 1981, 11, 25 },  { 1981, 11, 26 },
  { 1981, 11, 27 },  { 1981, 11, 28 },  { 1981, 11, 29 },  { 1981, 11, 30 },
  { 1981, 12,  1 },  { 1981, 12,  2 },  { 1981, 12,  3 },  { 1981, 12,  4 },
  { 1981, 12,  5 },  { 1981, 12,  6 },  { 1981, 12,  7 },  { 1981, 12,  8 },
  { 1981, 12,  9 },  { 1981, 12, 10 },  { 1981, 12, 11 },  { 1981, 12, 12 },
  { 1981, 12, 13 },  { 1981, 12, 14 },  { 1981, 12, 15 },  { 1981, 12, 16 },
  { 1981, 12, 17 },  { 1981, 12, 18 },  { 1981, 12, 19 },  { 1981, 12, 20 },
  { 1981, 12, 21 },  { 1981, 12, 22 },  { 1981, 12, 23 },  { 1981, 12, 24 },
  { 1981, 12, 25 },  { 1981, 12, 26 },  { 1981, 12, 27 },  { 1981, 12, 28 },
  { 1981, 12, 29 },  { 1981, 12, 30 },  { 1981, 12, 31 },  { 1982,  1,  1 },
  { 1982,  1,  2 },  { 1982,  1,  3 },  { 1982,  1,  4 },  { 1982,  1,  5 },
  { 1982,  1,  6 },  { 1982,  1,  7 },  { 1982,  1,  8 },  { 1982,  1,  9 },
  { 1982,  1, 10 },  { 1982,  1, 11 },  { 1982,  1, 12 },  { 1982,  1, 13 },
  { 1982,  1, 14 },  { 1982,  1, 15 },  { 1982,  1, 16 },  { 1982,  1, 17 },
  { 1982,  1, 18 },  { 1982,  1, 19 },  { 1982,  1, 20 },  { 1982,  1, 21 },
  { 1982,  1, 22 },  { 1982,  1, 23 },  { 1982,  1, 24 },  { 1982,  1, 25 },
  { 1982,  1, 26 },  { 1982,  1, 27 },  { 1982,  1, 28 },  { 1982,  1, 29 },
  { 1982,  1, 30 },  { 1982,  1, 31 },  { 1982,  2,  1 },  { 1982,  2,  2 },
  { 1982,  2,  3 },  { 1982,  2,  4 },  { 1982,  2,  5 },  { 1982,  2,  6 },
  { 1982,  2,  7 },  { 1982,  2,  8 },  { 1982,  2,  9 },  { 1982,  2, 10 },
  { 1982,  2, 11 },  { 1982,  2, 12 },  { 1982,  2, 13 },  { 1982,  2, 14 },
  { 1982,  2, 15 },  { 1982,  2, 16 },  { 1982,  2, 17 },  { 1982,  2, 18 },
  { 1982,  2, 19 },  { 1982,  2, 20 },  { 1982,  2, 21 },  { 1982,  2, 22 },
  { 1982,  2, 23 },  { 1982,  2, 24 },  { 1982,  2, 25 },  { 1982,  2, 26 },
  { 1982,  2, 27 },  { 1982,  2, 28 },  { 1982,  3,  1 },  { 1982,  3,  2 },
  { 1982,  3,  3 },  { 1982,  3,  4 },  { 1982,  3,  5 },  { 1982,  3,  6 },
  { 1982,  3,  7 },  { 1982,  3,  8 },  { 1982,  3,  9 },  { 1982,  3, 10 },
  { 1982,  3, 11 },  { 1982,  3, 12 },  { 1982,  3, 13 },  { 1982,  3, 14 },
  { 1982,  3, 15 },  { 1982,  3, 16 },  { 1982,  3, 17 },  { 1982,  3, 18 },
  { 1982,  3, 19 },  { 1982,  3, 20 },  { 1982,  3, 21 },  { 1982,  3, 22 },
  { 1982,  3, 23 },  { 1982,  3, 24 },  { 1982,  3, 25 },  { 1982,  3, 26 },
  { 1982,  3, 27 },  { 1982,  3, 28 },  { 1982,  3, 29 },  { 1982,  3, 30 },
  { 1982,  3, 31 },  { 1982,  4,  1 },  { 1982,  4,  2 },  { 1982,  4,  3 },
  { 1982,  4,  4 },  { 1982,  4,  5 },  { 1982,  4,  6 },  { 1982,  4,  7 },
  { 1982,  4,  8 },  { 1982,  4,  9 },  { 1982,  4, 10 },  { 1982,  4, 11 },
  { 1982,  4, 12 },  { 1982,  4, 13 },  { 1982,  4, 14 },  { 1982,  4, 15 },
  { 1982,  4, 16 },  { 1982,  4, 17 },  { 1982,  4, 18 },  { 1982,  4, 19 },
  { 1982,  4, 20 },  { 1982,  4, 21 },  { 1982,  4, 22 },  { 1982,  4, 23 },
  { 1982,  4, 24 },  { 1982,  4, 25 },  { 1982,  4, 26 },  { 1982,  4, 27 },
  { 1982,  4, 28 },  { 1982,  4, 29 },  { 1982,  4, 30 },  { 1982,  5,  1 },
  { 1982,  5,  2 },  { 1982,  5,  3 },  { 1982,  5,  4 },  { 1982,  5,  5 },
  { 1982,  5,  6 },  { 1982,  5,  7 },  { 1982,  5,  8 },  { 1982,  5,  9 },
  { 1982,  5, 10 },  { 1982,  5, 11 },  { 1982,  5, 12 },  { 1982,  5, 13 },
  { 1982,  5, 14 },  { 1982,  5, 15 },  { 1982,  5, 16 },  { 1982,  5, 17 },
  { 1982,  5, 18 },  { 1982,  5, 19 },  { 1982,  5, 20 },  { 1982,  5, 21 },
  { 1982,  5, 22 },  { 1982,  5, 23 },  { 1982,  5, 24 },  { 1982,  5, 25 },
  { 1982,  5, 26 },  { 1982,  5, 27 },  { 1982,  5, 28 },  { 1982,  5, 29 },
  { 1982,  5, 30 },  { 1982,  5, 31 },  { 1982,  6,  1 },  { 1982,  6,  2 },
  { 1982,  6,  3 },  { 1982,  6,  4 },  { 1982,  6,  5 },  { 1982,  6,  6 },
  { 1982,  6,  7 },  { 1982,  6,  8 },  { 1982,  6,  9 },  { 1982,  6, 10 },
  { 1982,  6, 11 },  { 1982,  6, 12 },  { 1982,  6, 13 },  { 1982,  6, 14 },
  { 1982,  6, 15 },  { 1982,  6, 16 },  { 1982,  6, 17 },  { 1982,  6, 18 },
  { 1982,  6, 19 },  { 1982,  6, 20 },  { 1982,  6, 21 },  { 1982,  6, 22 },
  { 1982,  6, 23 },  { 1982,  6, 24 },  { 1982,  6, 25 },  { 1982,  6, 26 },
  { 1982,  6, 27 },  { 1982,  6, 28 },  { 1982,  6, 29 },  { 1982,  6, 30 },
  { 1982,  7,  1 },  { 1982,  7,  2 },  { 1982,  7,  3 },  { 1982,  7,  4 },
  { 1982,  7,  5 },  { 1982,  7,  6 },  { 1982,  7,  7 },  { 1982,  7,  8 },
  { 1982,  7,  9 },  { 1982,  7, 10 },  { 1982,  7, 11 },  { 1982,  7, 12 },
  { 1982,  7, 13 },  { 1982,  7, 14 },  { 1982,  7, 15 },  { 1982,  7, 16 },
  { 1982,  7, 17 },  { 1982,  7, 18 },  { 1982,  7, 19 },  { 1982,  7, 20 },
  { 1982,  7, 21 },  { 1982,  7, 22 },  { 1982,  7, 23 },  { 1982,  7, 24 },
  { 1982,  7, 25 },  { 1982,  7, 26 },  { 1982,  7, 27 },  { 1982,  7, 28 },
  { 1982,  7, 29 },  { 1982,  7, 30 },  { 1982,  7, 31 },  { 1982,  8,  1 },
  { 1982,  8,  2 },  { 1982,  8,  3 },  { 1982,  8,  4 },  { 1982,  8,  5 },
  { 1982,  8,  6 },  { 1982,  8,  7 },  { 1982,  8,  8 },  { 1982,  8,  9 },
  { 1982,  8, 10 },  { 1982,  8, 11 },  { 1982,  8, 12 },  { 1982,  8, 13 },
  { 1982,  8, 14 },  { 1982,  8, 15 },  { 1982,  8, 16 },  { 1982,  8, 17 },
  { 1982,  8, 18 },  { 1982,  8, 19 },  { 1982,  8, 20 },  { 1982,  8, 21 },
  { 1982,  8, 22 },  { 1982,  8, 23 },  { 1982,  8, 24 },  { 1982,  8, 25 },
  { 1982,  8, 26 },  { 1982,  8, 27 },  { 1982,  8, 28 },  { 1982,  8, 29 },
  { 1982,  8, 30 },  { 1982,  8, 31 },  { 1982,  9,  1 },  { 1982,  9,  2 },
  { 1982,  9,  3 },  { 1982,  9,  4 },  { 1982,  9,  5 },  { 1982,  9,  6 },
  { 1982,  9,  7 },  { 1982,  9,  8 },  { 1982,  9,  9 },  { 1982,  9, 10 },
  { 1982,  9, 11 },  { 1982,  9, 12 },  { 1982,  9, 13 },  { 1982,  9, 14 },
  { 1982,  9, 15 },  { 1982,  9, 16 },  { 1982,  9, 17 },  { 1982,  9, 18 },
  { 1982,  9, 19 },  { 1982,  9, 20 },  { 1982,  9, 21 },  { 1982,  9, 22 },
  { 1982,  9, 23 },  { 1982,  9, 24 },  { 1982,  9, 25 },  { 1982,  9, 26 },
  { 1982,  9, 27 },  { 1982,  9, 28 },  { 1982,  9, 29 },  { 1982,  9, 30 },
  { 1982, 10,  1 },  { 1982, 10,  2 },  { 1982, 10,  3 },  { 1982, 10,  4 },
  { 1982, 10,  5 },  { 1982, 10,  6 },  { 1982, 10,  7 },  { 1982, 10,  8 },
  { 1982, 10,  9 },  { 1982, 10, 10 },  { 1982, 10, 11 },  { 1982, 10, 12 },
  { 1982, 10, 13 },  { 1982, 10, 14 },  { 1982, 10, 15 },  { 1982, 10, 16 },
  { 1982, 10, 17 },  { 1982, 10, 18 },  { 1982, 10, 19 },  { 1982, 10, 20 },
  { 1982, 10, 21 },  { 1982, 10, 22 },  { 1982, 10, 23 },  { 1982, 10, 24 },
  { 1982, 10, 25 },  { 1982, 10, 26 },  { 1982, 10, 27 },  { 1982, 10, 28 },
  { 1982, 10, 29 },  { 1982, 10, 30 },  { 1982, 10, 31 },  { 1982, 11,  1 },
  { 1982, 11,  2 },  { 1982, 11,  3 },  { 1982, 11,  4 },  { 1982, 11,  5 },
  { 1982, 11,  6 },  { 1982, 11,  7 },  { 1982, 11,  8 },  { 1982, 11,  9 },
  { 1982, 11, 10 },  { 1982, 11, 11 },  { 1982, 11, 12 },  { 1982, 11, 13 },
  { 1982, 11, 14 },  { 1982, 11, 15 },  { 1982, 11, 16 },  { 1982, 11, 17 },
  { 1982, 11, 18 },  { 1982, 11, 19 },  { 1982, 11, 20 },  { 1982, 11, 21 },
  { 1982, 11, 22 },  { 1982, 11, 23 },  { 1982, 11, 24 },  { 1982, 11, 25 },
  { 1982, 11, 26 },  { 1982, 11, 27 },  { 1982, 11, 28 },  { 1982, 11, 29 },
  { 1982, 11, 30 },  { 1982, 12,  1 },  { 1982, 12,  2 },  { 1982, 12,  3 },
  { 1982, 12,  4 },  { 1982, 12,  5 },  { 1982, 12,  6 },  { 1982, 12,  7 },
  { 1982, 12,  8 },  { 1982, 12,  9 },  { 1982, 12, 10 },  { 1982, 12, 11 },
  { 1982, 12, 12 },  { 1982, 12, 13 },  { 1982, 12, 14 },  { 1982, 12, 15 },
  { 1982, 12, 16 },  { 1982, 12, 17 },  { 1982, 12, 18 },  { 1982, 12, 19 },
  { 1982, 12, 20 },  { 1982, 12, 21 },  { 1982, 12, 22 },  { 1982, 12, 23 },
  { 1982, 12, 24 },  { 1982, 12, 25 },  { 1982, 12, 26 },  { 1982, 12, 27 },
  { 1982, 12, 28 },  { 1982, 12, 29 },  { 1982, 12, 30 },  { 1982, 12, 31 },
  { 1983,  1,  1 },  { 1983,  1,  2 },  { 1983,  1,  3 },  { 1983,  1,  4 },
  { 1983,  1,  5 },  { 1983,  1,  6 },  { 1983,  1,  7 },  { 1983,  1,  8 },
  { 1983,  1,  9 },  { 1983,  1, 10 },  { 1983,  1, 11 },  { 1983,  1, 12 },
  { 1983,  1, 13 },  { 1983,  1, 14 },  { 1983,  1, 15 },  { 1983,  1, 16 },
  { 1983,  1, 17 },  { 1983,  1, 18 },  { 1983,  1, 19 },  { 1983,  1, 20 },
  { 1983,  1, 21 },  { 1983,  1, 22 },  { 1983,  1, 23 },  { 1983,  1, 24 },
  { 1983,  1, 25 },  { 1983,  1, 26 },  { 1983,  1, 27 },  { 1983,  1, 28 },
  { 1983,  1, 29 },  { 1983,  1, 30 },  { 1983,  1, 31 },  { 1983,  2,  1 },
  { 1983,  2,  2 },  { 1983,  2,  3 },  { 1983,  2,  4 },  { 1983,  2,  5 },
  { 1983,  2,  6 },  { 1983,  2,  7 },  { 1983,  2,  8 },  { 1983,  2,  9 },
  { 1983,  2, 10 },  { 1983,  2, 11 },  { 1983,  2, 12 },  { 1983,  2, 13 },
  { 1983,  2, 14 },  { 1983,  2, 15 },  { 1983,  2, 16 },  { 1983,  2, 17 },
  { 1983,  2, 18 },  { 1983,  2, 19 },  { 1983,  2, 20 },  { 1983,  2, 21 },
  { 1983,  2, 22 },  { 1983,  2, 23 },  { 1983,  2, 24 },  { 1983,  2, 25 },
  { 1983,  2, 26 },  { 1983,  2, 27 },  { 1983,  2, 28 },  { 1983,  3,  1 },
  { 1983,  3,  2 },  { 1983,  3,  3 },  { 1983,  3,  4 },  { 1983,  3,  5 },
  { 1983,  3,  6 },  { 1983,  3,  7 },  { 1983,  3,  8 },  { 1983,  3,  9 },
  { 1983,  3, 10 },  { 1983,  3, 11 },  { 1983,  3, 12 },  { 1983,  3, 13 },
  { 1983,  3, 14 },  { 1983,  3, 15 },  { 1983,  3, 16 },  { 1983,  3, 17 },
  { 1983,  3, 18 },  { 1983,  3, 19 },  { 1983,  3, 20 },  { 1983,  3, 21 },
  { 1983,  3, 22 },  { 1983,  3, 23 },  { 1983,  3, 24 },  { 1983,  3, 25 },
  { 1983,  3, 26 },  { 1983,  3, 27 },  { 1983,  3, 28 },  { 1983,  3, 29 },
  { 1983,  3, 30 },  { 1983,  3, 31 },  { 1983,  4,  1 },  { 1983,  4,  2 },
  { 1983,  4,  3 },  { 1983,  4,  4 },  { 1983,  4,  5 },  { 1983,  4,  6 },
  { 1983,  4,  7 },  { 1983,  4,  8 },  { 1983,  4,  9 },  { 1983,  4, 10 },
  { 1983,  4, 11 },  { 1983,  4, 12 },  { 1983,  4, 13 },  { 1983,  4, 14 },
  { 1983,  4, 15 },  { 1983,  4, 16 },  { 1983,  4, 17 },  { 1983,  4, 18 },
  { 1983,  4, 19 },  { 1983,  4, 20 },  { 1983,  4, 21 },  { 1983,  4, 22 },
  { 1983,  4, 23 },  { 1983,  4, 24 },  { 1983,  4, 25 },  { 1983,  4, 26 },
  { 1983,  4, 27 },  { 1983,  4, 28 },  { 1983,  4, 29 },  { 1983,  4, 30 },
  { 1983,  5,  1 },  { 1983,  5,  2 },  { 1983,  5,  3 },  { 1983,  5,  4 },
  { 1983,  5,  5 },  { 1983,  5,  6 },  { 1983,  5,  7 },  { 1983,  5,  8 },
  { 1983,  5,  9 },  { 1983,  5, 10 },  { 1983,  5, 11 },  { 1983,  5, 12 },
  { 1983,  5, 13 },  { 1983,  5, 14 },  { 1983,  5, 15 },  { 1983,  5, 16 },
  { 1983,  5, 17 },  { 1983,  5, 18 },  { 1983,  5, 19 },  { 1983,  5, 20 },
  { 1983,  5, 21 },  { 1983,  5, 22 },  { 1983,  5, 23 },  { 1983,  5, 24 },
  { 1983,  5, 25 },  { 1983,  5, 26 },  { 1983,  5, 27 },  { 1983,  5, 28 },
  { 1983,  5, 29 },  { 1983,  5, 30 },  { 1983,  5, 31 },  { 1983,  6,  1 },
  { 1983,  6,  2 },  { 1983,  6,  3 },  { 1983,  6,  4 },  { 1983,  6,  5 },
  { 1983,  6,  6 },  { 1983,  6,  7 },  { 1983,  6,  8 },  { 1983,  6,  9 },
  { 1983,  6, 10 },  { 1983,  6, 11 },  { 1983,  6, 12 },  { 1983,  6, 13 },
  { 1983,  6, 14 },  { 1983,  6, 15 },  { 1983,  6, 16 },  { 1983,  6, 17 },
  { 1983,  6, 18 },  { 1983,  6, 19 },  { 1983,  6, 20 },  { 1983,  6, 21 },
  { 1983,  6, 22 },  { 1983,  6, 23 },  { 1983,  6, 24 },  { 1983,  6, 25 },
  { 1983,  6, 26 },  { 1983,  6, 27 },  { 1983,  6, 28 },  { 1983,  6, 29 },
  { 1983,  6, 30 },  { 1983,  7,  1 },  { 1983,  7,  2 },  { 1983,  7,  3 },
  { 1983,  7,  4 },  { 1983,  7,  5 },  { 1983,  7,  6 },  { 1983,  7,  7 },
  { 1983,  7,  8 },  { 1983,  7,  9 },  { 1983,  7, 10 },  { 1983,  7, 11 },
  { 1983,  7, 12 },  { 1983,  7, 13 },  { 1983,  7, 14 },  { 1983,  7, 15 },
  { 1983,  7, 16 },  { 1983,  7, 17 },  { 1983,  7, 18 },  { 1983,  7, 19 },
  { 1983,  7, 20 },  { 1983,  7, 21 },  { 1983,  7, 22 },  { 1983,  7, 23 },
  { 1983,  7, 24 },  { 1983,  7, 25 },  { 1983,  7, 26 },  { 1983,  7, 27 },
  { 1983,  7, 28 },  { 1983,  7, 29 },  { 1983,  7, 30 },  { 1983,  7, 31 },
  { 1983,  8,  1 },  { 1983,  8,  2 },  { 1983,  8,  3 },  { 1983,  8,  4 },
  { 1983,  8,  5 },  { 1983,  8,  6 },  { 1983,  8,  7 },  { 1983,  8,  8 },
  { 1983,  8,  9 },  { 1983,  8, 10 },  { 1983,  8, 11 },  { 1983,  8, 12 },
  { 1983,  8, 13 },  { 1983,  8, 14 },  { 1983,  8, 15 },  { 1983,  8, 16 },
  { 1983,  8, 17 },  { 1983,  8, 18 },  { 1983,  8, 19 },  { 1983,  8, 20 },
  { 1983,  8, 21 },  { 1983,  8, 22 },  { 1983,  8, 23 },  { 1983,  8, 24 },
  { 1983,  8, 25 },  { 1983,  8, 26 },  { 1983,  8, 27 },  { 1983,  8, 28 },
  { 1983,  8, 29 },  { 1983,  8, 30 },  { 1983,  8, 31 },  { 1983,  9,  1 },
  { 1983,  9,  2 },  { 1983,  9,  3 },  { 1983,  9,  4 },  { 1983,  9,  5 },
  { 1983,  9,  6 },  { 1983,  9,  7 },  { 1983,  9,  8 },  { 1983,  9,  9 },
  { 1983,  9, 10 },  { 1983,  9, 11 },  { 1983,  9, 12 },  { 1983,  9, 13 },
  { 1983,  9, 14 },  { 1983,  9, 15 },  { 1983,  9, 16 },  { 1983,  9, 17 },
  { 1983,  9, 18 },  { 1983,  9, 19 },  { 1983,  9, 20 },  { 1983,  9, 21 },
  { 1983,  9, 22 },  { 1983,  9, 23 },  { 1983,  9, 24 },  { 1983,  9, 25 },
  { 1983,  9, 26 },  { 1983,  9, 27 },  { 1983,  9, 28 },  { 1983,  9, 29 },
  { 1983,  9, 30 },  { 1983, 10,  1 },  { 1983, 10,  2 },  { 1983, 10,  3 },
  { 1983, 10,  4 },  { 1983, 10,  5 },  { 1983, 10,  6 },  { 1983, 10,  7 },
  { 1983, 10,  8 },  { 1983, 10,  9 },  { 1983, 10, 10 },  { 1983, 10, 11 },
  { 1983, 10, 12 },  { 1983, 10, 13 },  { 1983, 10, 14 },  { 1983, 10, 15 },
  { 1983, 10, 16 },  { 1983, 10, 17 },  { 1983, 10, 18 },  { 1983, 10, 19 },
  { 1983, 10, 20 },  { 1983, 10, 21 },  { 1983, 10, 22 },  { 1983, 10, 23 },
  { 1983, 10, 24 },  { 1983, 10, 25 },  { 1983, 10, 26 },  { 1983, 10, 27 },
  { 1983, 10, 28 },  { 1983, 10, 29 },  { 1983, 10, 30 },  { 1983, 10, 31 },
  { 1983, 11,  1 },  { 1983, 11,  2 },  { 1983, 11,  3 },  { 1983, 11,  4 },
  { 1983, 11,  5 },  { 1983, 11,  6 },  { 1983, 11,  7 },  { 1983, 11,  8 },
  { 1983, 11,  9 },  { 1983, 11, 10 },  { 1983, 11, 11 },  { 1983, 11, 12 },
  { 1983, 11, 13 },  { 1983, 11, 14 },  { 1983, 11, 15 },  { 1983, 11, 16 },
  { 1983, 11, 17 },  { 1983, 11, 18 },  { 1983, 11, 19 },  { 1983, 11, 20 },
  { 1983, 11, 21 },  { 1983, 11, 22 },  { 1983, 11, 23 },  { 1983, 11, 24 },
  { 1983, 11, 25 },  { 1983, 11, 26 },  { 1983, 11, 27 },  { 1983, 11, 28 },
  { 1983, 11, 29 },  { 1983, 11, 30 },  { 1983, 12,  1 },  { 1983, 12,  2 },
  { 1983, 12,  3 },  { 1983, 12,  4 },  { 1983, 12,  5 },  { 1983, 12,  6 },
  { 1983, 12,  7 },  { 1983, 12,  8 },  { 1983, 12,  9 },  { 1983, 12, 10 },
  { 1983, 12, 11 },  { 1983, 12, 12 },  { 1983, 12, 13 },  { 1983, 12, 14 },
  { 1983, 12, 15 },  { 1983, 12, 16 },  { 1983, 12, 17 },  { 1983, 12, 18 },
  { 1983, 12, 19 },  { 1983, 12, 20 },  { 1983, 12, 21 },  { 1983, 12, 22 },
  { 1983, 12, 23 },  { 1983, 12, 24 },  { 1983, 12, 25 },  { 1983, 12, 26 },
  { 1983, 12, 27 },  { 1983, 12, 28 },  { 1983, 12, 29 },  { 1983, 12, 30 },
  { 1983, 12, 31 },  { 1984,  1,  1 },  { 1984,  1,  2 },  { 1984,  1,  3 },
  { 1984,  1,  4 },  { 1984,  1,  5 },  { 1984,  1,  6 },  { 1984,  1,  7 },
  { 1984,  1,  8 },  { 1984,  1,  9 },  { 1984,  1, 10 },  { 1984,  1, 11 },
  { 1984,  1, 12 },  { 1984,  1, 13 },  { 1984,  1, 14 },  { 1984,  1, 15 },
  { 1984,  1, 16 },  { 1984,  1, 17 },  { 1984,  1, 18 },  { 1984,  1, 19 },
  { 1984,  1, 20 },  { 1984,  1, 21 },  { 1984,  1, 22 },  { 1984,  1, 23 },
  { 1984,  1, 24 },  { 1984,  1, 25 },  { 1984,  1, 26 },  { 1984,  1, 27 },
  { 1984,  1, 28 },  { 1984,  1, 29 },  { 1984,  1, 30 },  { 1984,  1, 31 },
  { 1984,  2,  1 },  { 1984,  2,  2 },  { 1984,  2,  3 },  { 1984,  2,  4 },
  { 1984,  2,  5 },  { 1984,  2,  6 },  { 1984,  2,  7 },  { 1984,  2,  8 },
  { 1984,  2,  9 },  { 1984,  2, 10 },  { 1984,  2, 11 },  { 1984,  2, 12 },
  { 1984,  2, 13 },  { 1984,  2, 14 },  { 1984,  2, 15 },  { 1984,  2, 16 },
  { 1984,  2, 17 },  { 1984,  2, 18 },  { 1984,  2, 19 },  { 1984,  2, 20 },
  { 1984,  2, 21 },  { 1984,  2, 22 },  { 1984,  2, 23 },  { 1984,  2, 24 },
  { 1984,  2, 25 },  { 1984,  2, 26 },  { 1984,  2, 27 },  { 1984,  2, 28 },
  { 1984,  2, 29 },  { 1984,  3,  1 },  { 1984,  3,  2 },  { 1984,  3,  3 },
  { 1984,  3,  4 },  { 1984,  3,  5 },  { 1984,  3,  6 },  { 1984,  3,  7 },
  { 1984,  3,  8 },  { 1984,  3,  9 },  { 1984,  3, 10 },  { 1984,  3, 11 },
  { 1984,  3, 12 },  { 1984,  3, 13 },  { 1984,  3, 14 },  { 1984,  3, 15 },
  { 1984,  3, 16 },  { 1984,  3, 17 },  { 1984,  3, 18 },  { 1984,  3, 19 },
  { 1984,  3, 20 },  { 1984,  3, 21 },  { 1984,  3, 22 },  { 1984,  3, 23 },
  { 1984,  3, 24 },  { 1984,  3, 25 },  { 1984,  3, 26 },  { 1984,  3, 27 },
  { 1984,  3, 28 },  { 1984,  3, 29 },  { 1984,  3, 30 },  { 1984,  3, 31 },
  { 1984,  4,  1 },  { 1984,  4,  2 },  { 1984,  4,  3 },  { 1984,  4,  4 },
  { 1984,  4,  5 },  { 1984,  4,  6 },  { 1984,  4,  7 },  { 1984,  4,  8 },
  { 1984,  4,  9 },  { 1984,  4, 10 },  { 1984,  4, 11 },  { 1984,  4, 12 },
  { 1984,  4, 13 },  { 1984,  4, 14 },  { 1984,  4, 15 },  { 1984,  4, 16 },
  { 1984,  4, 17 },  { 1984,  4, 18 },  { 1984,  4, 19 },  { 1984,  4, 20 },
  { 1984,  4, 21 },  { 1984,  4, 22 },  { 1984,  4, 23 },  { 1984,  4, 24 },
  { 1984,  4, 25 },  { 1984,  4, 26 },  { 1984,  4, 27 },  { 1984,  4, 28 },
  { 1984,  4, 29 },  { 1984,  4, 30 },  { 1984,  5,  1 },  { 1984,  5,  2 },
  { 1984,  5,  3 },  { 1984,  5,  4 },  { 1984,  5,  5 },  { 1984,  5,  6 },
  { 1984,  5,  7 },  { 1984,  5,  8 },  { 1984,  5,  9 },  { 1984,  5, 10 },
  { 1984,  5, 11 },  { 1984,  5, 12 },  { 1984,  5, 13 },  { 1984,  5, 14 },
  { 1984,  5, 15 },  { 1984,  5, 16 },  { 1984,  5, 17 },  { 1984,  5, 18 },
  { 1984,  5, 19 },  { 1984,  5, 20 },  { 1984,  5, 21 },  { 1984,  5, 22 },
  { 1984,  5, 23 },  { 1984,  5, 24 },  { 1984,  5, 25 },  { 1984,  5, 26 },
  { 1984,  5, 27 },  { 1984,  5, 28 },  { 1984,  5, 29 },  { 1984,  5, 30 },
  { 1984,  5, 31 },  { 1984,  6,  1 },  { 1984,  6,  2 },  { 1984,  6,  3 },
  { 1984,  6,  4 },  { 1984,  6,  5 },  { 1984,  6,  6 },  { 1984,  6,  7 },
  { 1984,  6,  8 },  { 1984,  6,  9 },  { 1984,  6, 10 },  { 1984,  6, 11 },
  { 1984,  6, 12 },  { 1984,  6, 13 },  { 1984,  6, 14 },  { 1984,  6, 15 },
  { 1984,  6, 16 },  { 1984,  6, 17 },  { 1984,  6, 18 },  { 1984,  6, 19 },
  { 1984,  6, 20 },  { 1984,  6, 21 },  { 1984,  6, 22 },  { 1984,  6, 23 },
  { 1984,  6, 24 },  { 1984,  6, 25 },  { 1984,  6, 26 },  { 1984,  6, 27 },
  { 1984,  6, 28 },  { 1984,  6, 29 },  { 1984,  6, 30 },  { 1984,  7,  1 },
  { 1984,  7,  2 },  { 1984,  7,  3 },  { 1984,  7,  4 },  { 1984,  7,  5 },
  { 1984,  7,  6 },  { 1984,  7,  7 },  { 1984,  7,  8 },  { 1984,  7,  9 },
  { 1984,  7, 10 },  { 1984,  7, 11 },  { 1984,  7, 12 },  { 1984,  7, 13 },
  { 1984,  7, 14 },  { 1984,  7, 15 },  { 1984,  7, 16 },  { 1984,  7, 17 },
  { 1984,  7, 18 },  { 1984,  7, 19 },  { 1984,  7, 20 },  { 1984,  7, 21 },
  { 1984,  7, 22 },  { 1984,  7, 23 },  { 1984,  7, 24 },  { 1984,  7, 25 },
  { 1984,  7, 26 },  { 1984,  7, 27 },  { 1984,  7, 28 },  { 1984,  7, 29 },
  { 1984,  7, 30 },  { 1984,  7, 31 },  { 1984,  8,  1 },  { 1984,  8,  2 },
  { 1984,  8,  3 },  { 1984,  8,  4 },  { 1984,  8,  5 },  { 1984,  8,  6 },
  { 1984,  8,  7 },  { 1984,  8,  8 },  { 1984,  8,  9 },  { 1984,  8, 10 },
  { 1984,  8, 11 },  { 1984,  8, 12 },  { 1984,  8, 13 },  { 1984,  8, 14 },
  { 1984,  8, 15 },  { 1984,  8, 16 },  { 1984,  8, 17 },  { 1984,  8, 18 },
  { 1984,  8, 19 },  { 1984,  8, 20 },  { 1984,  8, 21 },  { 1984,  8, 22 },
  { 1984,  8, 23 },  { 1984,  8, 24 },  { 1984,  8, 25 },  { 1984,  8, 26 },
  { 1984,  8, 27 },  { 1984,  8, 28 },  { 1984,  8, 29 },  { 1984,  8, 30 },
  { 1984,  8, 31 },  { 1984,  9,  1 },  { 1984,  9,  2 },  { 1984,  9,  3 },
  { 1984,  9,  4 },  { 1984,  9,  5 },  { 1984,  9,  6 },  { 1984,  9,  7 },
  { 1984,  9,  8 },  { 1984,  9,  9 },  { 1984,  9, 10 },  { 1984,  9, 11 },
  { 1984,  9, 12 },  { 1984,  9, 13 },  { 1984,  9, 14 },  { 1984,  9, 15 },
  { 1984,  9, 16 },  { 1984,  9, 17 },  { 1984,  9, 18 },  { 1984,  9, 19 },
  { 1984,  9, 20 },  { 1984,  9, 21 },  { 1984,  9, 22 },  { 1984,  9, 23 },
  { 1984,  9, 24 },  { 1984,  9, 25 },  { 1984,  9, 26 },  { 1984,  9, 27 },
  { 1984,  9, 28 },  { 1984,  9, 29 },  { 1984,  9, 30 },  { 1984, 10,  1 },
  { 1984, 10,  2 },  { 1984, 10,  3 },  { 1984, 10,  4 },  { 1984, 10,  5 },
  { 1984, 10,  6 },  { 1984, 10,  7 },  { 1984, 10,  8 },  { 1984, 10,  9 },
  { 1984, 10, 10 },  { 1984, 10, 11 },  { 1984, 10, 12 },  { 1984, 10, 13 },
  { 1984, 10, 14 },  { 1984, 10, 15 },  { 1984, 10, 16 },  { 1984, 10, 17 },
  { 1984, 10, 18 },  { 1984, 10, 19 },  { 1984, 10, 20 },  { 1984, 10, 21 },
  { 1984, 10, 22 },  { 1984, 10, 23 },  { 1984, 10, 24 },  { 1984, 10, 25 },
  { 1984, 10, 26 },  { 1984, 10, 27 },  { 1984, 10, 28 },  { 1984, 10, 29 },
  { 1984, 10, 30 },  { 1984, 10, 31 },  { 1984, 11,  1 },  { 1984, 11,  2 },
  { 1984, 11,  3 },  { 1984, 11,  4 },  { 1984, 11,  5 },  { 1984, 11,  6 },
  { 1984, 11,  7 },  { 1984, 11,  8 },  { 1984, 11,  9 },  { 1984, 11, 10 },
  { 1984, 11, 11 },  { 1984, 11, 12 },  { 1984, 11, 13 },  { 1984, 11, 14 },
  { 1984, 11, 15 },  { 1984, 11, 16 },  { 1984, 11, 17 },  { 1984, 11, 18 },
  { 1984, 11, 19 },  { 1984, 11, 20 },  { 1984, 11, 21 },  { 1984, 11, 22 },
  { 1984, 11, 23 },  { 1984, 11, 24 },  { 1984, 11, 25 },  { 1984, 11, 26 },
  { 1984, 11, 27 },  { 1984, 11, 28 },  { 1984, 11, 29 },  { 1984, 11, 30 },
  { 1984, 12,  1 },  { 1984, 12,  2 },  { 1984, 12,  3 },  { 1984, 12,  4 },
  { 1984, 12,  5 },  { 1984, 12,  6 },  { 1984, 12,  7 },  { 1984, 12,  8 },
  { 1984, 12,  9 },  { 1984, 12, 10 },  { 1984, 12, 11 },  { 1984, 12, 12 },
  { 1984, 12, 13 },  { 1984, 12, 14 },  { 1984, 12, 15 },  { 1984, 12, 16 },
  { 1984, 12, 17 },  { 1984, 12, 18 },  { 1984, 12, 19 },  { 1984, 12, 20 },
  { 1984, 12, 21 },  { 1984, 12, 22 },  { 1984, 12, 23 },  { 1984, 12, 24 },
  { 1984, 12, 25 },  { 1984, 12, 26 },  { 1984, 12, 27 },  { 1984, 12, 28 },
  { 1984, 12, 29 },  { 1984, 12, 30 },  { 1984, 12, 31 },  { 1985,  1,  1 },
  { 1985,  1,  2 },  { 1985,  1,  3 },  { 1985,  1,  4 },  { 1985,  1,  5 },
  { 1985,  1,  6 },  { 1985,  1,  7 },  { 1985,  1,  8 },  { 1985,  1,  9 },
  { 1985,  1, 10 },  { 1985,  1, 11 },  { 1985,  1, 12 },  { 1985,  1, 13 },
  { 1985,  1, 14 },  { 1985,  1, 15 },  { 1985,  1, 16 },  { 1985,  1, 17 },
  { 1985,  1, 18 },  { 1985,  1, 19 },  { 1985,  1, 20 },  { 1985,  1, 21 },
  { 1985,  1, 22 },  { 1985,  1, 23 },  { 1985,  1, 24 },  { 1985,  1, 25 },
  { 1985,  1, 26 },  { 1985,  1, 27 },  { 1985,  1, 28 },  { 1985,  1, 29 },
  { 1985,  1, 30 },  { 1985,  1, 31 },  { 1985,  2,  1 },  { 1985,  2,  2 },
  { 1985,  2,  3 },  { 1985,  2,  4 },  { 1985,  2,  5 },  { 1985,  2,  6 },
  { 1985,  2,  7 },  { 1985,  2,  8 },  { 1985,  2,  9 },  { 1985,  2, 10 },
  { 1985,  2, 11 },  { 1985,  2, 12 },  { 1985,  2, 13 },  { 1985,  2, 14 },
  { 1985,  2, 15 },  { 1985,  2, 16 },  { 1985,  2, 17 },  { 1985,  2, 18 },
  { 1985,  2, 19 },  { 1985,  2, 20 },  { 1985,  2, 21 },  { 1985,  2, 22 },
  { 1985,  2, 23 },  { 1985,  2, 24 },  { 1985,  2, 25 },  { 1985,  2, 26 },
  { 1985,  2, 27 },  { 1985,  2, 28 },  { 1985,  3,  1 },  { 1985,  3,  2 },
  { 1985,  3,  3 },  { 1985,  3,  4 },  { 1985,  3,  5 },  { 1985,  3,  6 },
  { 1985,  3,  7 },  { 1985,  3,  8 },  { 1985,  3,  9 },  { 1985,  3, 10 },
  { 1985,  3, 11 },  { 1985,  3, 12 },  { 1985,  3, 13 },  { 1985,  3, 14 },
  { 1985,  3, 15 },  { 1985,  3, 16 },  { 1985,  3, 17 },  { 1985,  3, 18 },
  { 1985,  3, 19 },  { 1985,  3, 20 },  { 1985,  3, 21 },  { 1985,  3, 22 },
  { 1985,  3, 23 },  { 1985,  3, 24 },  { 1985,  3, 25 },  { 1985,  3, 26 },
  { 1985,  3, 27 },  { 1985,  3, 28 },  { 1985,  3, 29 },  { 1985,  3, 30 },
  { 1985,  3, 31 },  { 1985,  4,  1 },  { 1985,  4,  2 },  { 1985,  4,  3 },
  { 1985,  4,  4 },  { 1985,  4,  5 },  { 1985,  4,  6 },  { 1985,  4,  7 },
  { 1985,  4,  8 },  { 1985,  4,  9 },  { 1985,  4, 10 },  { 1985,  4, 11 },
  { 1985,  4, 12 },  { 1985,  4, 13 },  { 1985,  4, 14 },  { 1985,  4, 15 },
  { 1985,  4, 16 },  { 1985,  4, 17 },  { 1985,  4, 18 },  { 1985,  4, 19 },
  { 1985,  4, 20 },  { 1985,  4, 21 },  { 1985,  4, 22 },  { 1985,  4, 23 },
  { 1985,  4, 24 },  { 1985,  4, 25 },  { 1985,  4, 26 },  { 1985,  4, 27 },
  { 1985,  4, 28 },  { 1985,  4, 29 },  { 1985,  4, 30 },  { 1985,  5,  1 },
  { 1985,  5,  2 },  { 1985,  5,  3 },  { 1985,  5,  4 },  { 1985,  5,  5 },
  { 1985,  5,  6 },  { 1985,  5,  7 },  { 1985,  5,  8 },  { 1985,  5,  9 },
  { 1985,  5, 10 },  { 1985,  5, 11 },  { 1985,  5, 12 },  { 1985,  5, 13 },
  { 1985,  5, 14 },  { 1985,  5, 15 },  { 1985,  5, 16 },  { 1985,  5, 17 },
  { 1985,  5, 18 },  { 1985,  5, 19 },  { 1985,  5, 20 },  { 1985,  5, 21 },
  { 1985,  5, 22 },  { 1985,  5, 23 },  { 1985,  5, 24 },  { 1985,  5, 25 },
  { 1985,  5, 26 },  { 1985,  5, 27 },  { 1985,  5, 28 },  { 1985,  5, 29 },
  { 1985,  5, 30 },  { 1985,  5, 31 },  { 1985,  6,  1 },  { 1985,  6,  2 },
  { 1985,  6,  3 },  { 1985,  6,  4 },  { 1985,  6,  5 },  { 1985,  6,  6 },
  { 1985,  6,  7 },  { 1985,  6,  8 },  { 1985,  6,  9 },  { 1985,  6, 10 },
  { 1985,  6, 11 },  { 1985,  6, 12 },  { 1985,  6, 13 },  { 1985,  6, 14 },
  { 1985,  6, 15 },  { 1985,  6, 16 },  { 1985,  6, 17 },  { 1985,  6, 18 },
  { 1985,  6, 19 },  { 1985,  6, 20 },  { 1985,  6, 21 },  { 1985,  6, 22 },
  { 1985,  6, 23 },  { 1985,  6, 24 },  { 1985,  6, 25 },  { 1985,  6, 26 },
  { 1985,  6, 27 },  { 1985,  6, 28 },  { 1985,  6, 29 },  { 1985,  6, 30 },
  { 1985,  7,  1 },  { 1985,  7,  2 },  { 1985,  7,  3 },  { 1985,  7,  4 },
  { 1985,  7,  5 },  { 1985,  7,  6 },  { 1985,  7,  7 },  { 1985,  7,  8 },
  { 1985,  7,  9 },  { 1985,  7, 10 },  { 1985,  7, 11 },  { 1985,  7, 12 },
  { 1985,  7, 13 },  { 1985,  7, 14 },  { 1985,  7, 15 },  { 1985,  7, 16 },
  { 1985,  7, 17 },  { 1985,  7, 18 },  { 1985,  7, 19 },  { 1985,  7, 20 },
  { 1985,  7, 21 },  { 1985,  7, 22 },  { 1985,  7, 23 },  { 1985,  7, 24 },
  { 1985,  7, 25 },  { 1985,  7, 26 },  { 1985,  7, 27 },  { 1985,  7, 28 },
  { 1985,  7, 29 },  { 1985,  7, 30 },  { 1985,  7, 31 },  { 1985,  8,  1 },
  { 1985,  8,  2 },  { 1985,  8,  3 },  { 1985,  8,  4 },  { 1985,  8,  5 },
  { 1985,  8,  6 },  { 1985,  8,  7 },  { 1985,  8,  8 },  { 1985,  8,  9 },
  { 1985,  8, 10 },  { 1985,  8, 11 },  { 1985,  8, 12 },  { 1985,  8, 13 },
  { 1985,  8, 14 },  { 1985,  8, 15 },  { 1985,  8, 16 },  { 1985,  8, 17 },
  { 1985,  8, 18 },  { 1985,  8, 19 },  { 1985,  8, 20 },  { 1985,  8, 21 },
  { 1985,  8, 22 },  { 1985,  8, 23 },  { 1985,  8, 24 },  { 1985,  8, 25 },
  { 1985,  8, 26 },  { 1985,  8, 27 },  { 1985,  8, 28 },  { 1985,  8, 29 },
  { 1985,  8, 30 },  { 1985,  8, 31 },  { 1985,  9,  1 },  { 1985,  9,  2 },
  { 1985,  9,  3 },  { 1985,  9,  4 },  { 1985,  9,  5 },  { 1985,  9,  6 },
  { 1985,  9,  7 },  { 1985,  9,  8 },  { 1985,  9,  9 },  { 1985,  9, 10 },
  { 1985,  9, 11 },  { 1985,  9, 12 },  { 1985,  9, 13 },  { 1985,  9, 14 },
  { 1985,  9, 15 },  { 1985,  9, 16 },  { 1985,  9, 17 },  { 1985,  9, 18 },
  { 1985,  9, 19 },  { 1985,  9, 20 },  { 1985,  9, 21 },  { 1985,  9, 22 },
  { 1985,  9, 23 },  { 1985,  9, 24 },  { 1985,  9, 25 },  { 1985,  9, 26 },
  { 1985,  9, 27 },  { 1985,  9, 28 },  { 1985,  9, 29 },  { 1985,  9, 30 },
  { 1985, 10,  1 },  { 1985, 10,  2 },  { 1985, 10,  3 },  { 1985, 10,  4 },
  { 1985, 10,  5 },  { 1985, 10,  6 },  { 1985, 10,  7 },  { 1985, 10,  8 },
  { 1985, 10,  9 },  { 1985, 10, 10 },  { 1985, 10, 11 },  { 1985, 10, 12 },
  { 1985, 10, 13 },  { 1985, 10, 14 },  { 1985, 10, 15 },  { 1985, 10, 16 },
  { 1985, 10, 17 },  { 1985, 10, 18 },  { 1985, 10, 19 },  { 1985, 10, 20 },
  { 1985, 10, 21 },  { 1985, 10, 22 },  { 1985, 10, 23 },  { 1985, 10, 24 },
  { 1985, 10, 25 },  { 1985, 10, 26 },  { 1985, 10, 27 },  { 1985, 10, 28 },
  { 1985, 10, 29 },  { 1985, 10, 30 },  { 1985, 10, 31 },  { 1985, 11,  1 },
  { 1985, 11,  2 },  { 1985, 11,  3 },  { 1985, 11,  4 },  { 1985, 11,  5 },
  { 1985, 11,  6 },  { 1985, 11,  7 },  { 1985, 11,  8 },  { 1985, 11,  9 },
  { 1985, 11, 10 },  { 1985, 11, 11 },  { 1985, 11, 12 },  { 1985, 11, 13 },
  { 1985, 11, 14 },  { 1985, 11, 15 },  { 1985, 11, 16 },  { 1985, 11, 17 },
  { 1985, 11, 18 },  { 1985, 11, 19 },  { 1985, 11, 20 },  { 1985, 11, 21 },
  { 1985, 11, 22 },  { 1985, 11, 23 },  { 1985, 11, 24 },  { 1985, 11, 25 },
  { 1985, 11, 26 },  { 1985, 11, 27 },  { 1985, 11, 28 },  { 1985, 11, 29 },
  { 1985, 11, 30 },  { 1985, 12,  1 },  { 1985, 12,  2 },  { 1985, 12,  3 },
  { 1985, 12,  4 },  { 1985, 12,  5 },  { 1985, 12,  6 },  { 1985, 12,  7 },
  { 1985, 12,  8 },  { 1985, 12,  9 },  { 1985, 12, 10 },  { 1985, 12, 11 },
  { 1985, 12, 12 },  { 1985, 12, 13 },  { 1985, 12, 14 },  { 1985, 12, 15 },
  { 1985, 12, 16 },  { 1985, 12, 17 },  { 1985, 12, 18 },  { 1985, 12, 19 },
  { 1985, 12, 20 },  { 1985, 12, 21 },  { 1985, 12, 22 },  { 1985, 12, 23 },
  { 1985, 12, 24 },  { 1985, 12, 25 },  { 1985, 12, 26 },  { 1985, 12, 27 },
  { 1985, 12, 28 },  { 1985, 12, 29 },  { 1985, 12, 30 },  { 1985, 12, 31 },
  { 1986,  1,  1 },  { 1986,  1,  2 },  { 1986,  1,  3 },  { 1986,  1,  4 },
  { 1986,  1,  5 },  { 1986,  1,  6 },  { 1986,  1,  7 },  { 1986,  1,  8 },
  { 1986,  1,  9 },  { 1986,  1, 10 },  { 1986,  1, 11 },  { 1986,  1, 12 },
  { 1986,  1, 13 },  { 1986,  1, 14 },  { 1986,  1, 15 },  { 1986,  1, 16 },
  { 1986,  1, 17 },  { 1986,  1, 18 },  { 1986,  1, 19 },  { 1986,  1, 20 },
  { 1986,  1, 21 },  { 1986,  1, 22 },  { 1986,  1, 23 },  { 1986,  1, 24 },
  { 1986,  1, 25 },  { 1986,  1, 26 },  { 1986,  1, 27 },  { 1986,  1, 28 },
  { 1986,  1, 29 },  { 1986,  1, 30 },  { 1986,  1, 31 },  { 1986,  2,  1 },
  { 1986,  2,  2 },  { 1986,  2,  3 },  { 1986,  2,  4 },  { 1986,  2,  5 },
  { 1986,  2,  6 },  { 1986,  2,  7 },  { 1986,  2,  8 },  { 1986,  2,  9 },
  { 1986,  2, 10 },  { 1986,  2, 11 },  { 1986,  2, 12 },  { 1986,  2, 13 },
  { 1986,  2, 14 },  { 1986,  2, 15 },  { 1986,  2, 16 },  { 1986,  2, 17 },
  { 1986,  2, 18 },  { 1986,  2, 19 },  { 1986,  2, 20 },  { 1986,  2, 21 },
  { 1986,  2, 22 },  { 1986,  2, 23 },  { 1986,  2, 24 },  { 1986,  2, 25 },
  { 1986,  2, 26 },  { 1986,  2, 27 },  { 1986,  2, 28 },  { 1986,  3,  1 },
  { 1986,  3,  2 },  { 1986,  3,  3 },  { 1986,  3,  4 },  { 1986,  3,  5 },
  { 1986,  3,  6 },  { 1986,  3,  7 },  { 1986,  3,  8 },  { 1986,  3,  9 },
  { 1986,  3, 10 },  { 1986,  3, 11 },  { 1986,  3, 12 },  { 1986,  3, 13 },
  { 1986,  3, 14 },  { 1986,  3, 15 },  { 1986,  3, 16 },  { 1986,  3, 17 },
  { 1986,  3, 18 },  { 1986,  3, 19 },  { 1986,  3, 20 },  { 1986,  3, 21 },
  { 1986,  3, 22 },  { 1986,  3, 23 },  { 1986,  3, 24 },  { 1986,  3, 25 },
  { 1986,  3, 26 },  { 1986,  3, 27 },  { 1986,  3, 28 },  { 1986,  3, 29 },
  { 1986,  3, 30 },  { 1986,  3, 31 },  { 1986,  4,  1 },  { 1986,  4,  2 },
  { 1986,  4,  3 },  { 1986,  4,  4 },  { 1986,  4,  5 },  { 1986,  4,  6 },
  { 1986,  4,  7 },  { 1986,  4,  8 },  { 1986,  4,  9 },  { 1986,  4, 10 },
  { 1986,  4, 11 },  { 1986,  4, 12 },  { 1986,  4, 13 },  { 1986,  4, 14 },
  { 1986,  4, 15 },  { 1986,  4, 16 },  { 1986,  4, 17 },  { 1986,  4, 18 },
  { 1986,  4, 19 },  { 1986,  4, 20 },  { 1986,  4, 21 },  { 1986,  4, 22 },
  { 1986,  4, 23 },  { 1986,  4, 24 },  { 1986,  4, 25 },  { 1986,  4, 26 },
  { 1986,  4, 27 },  { 1986,  4, 28 },  { 1986,  4, 29 },  { 1986,  4, 30 },
  { 1986,  5,  1 },  { 1986,  5,  2 },  { 1986,  5,  3 },  { 1986,  5,  4 },
  { 1986,  5,  5 },  { 1986,  5,  6 },  { 1986,  5,  7 },  { 1986,  5,  8 },
  { 1986,  5,  9 },  { 1986,  5, 10 },  { 1986,  5, 11 },  { 1986,  5, 12 },
  { 1986,  5, 13 },  { 1986,  5, 14 },  { 1986,  5, 15 },  { 1986,  5, 16 },
  { 1986,  5, 17 },  { 1986,  5, 18 },  { 1986,  5, 19 },  { 1986,  5, 20 },
  { 1986,  5, 21 },  { 1986,  5, 22 },  { 1986,  5, 23 },  { 1986,  5, 24 },
  { 1986,  5, 25 },  { 1986,  5, 26 },  { 1986,  5, 27 },  { 1986,  5, 28 },
  { 1986,  5, 29 },  { 1986,  5, 30 },  { 1986,  5, 31 },  { 1986,  6,  1 },
  { 1986,  6,  2 },  { 1986,  6,  3 },  { 1986,  6,  4 },  { 1986,  6,  5 },
  { 1986,  6,  6 },  { 1986,  6,  7 },  { 1986,  6,  8 },  { 1986,  6,  9 },
  { 1986,  6, 10 },  { 1986,  6, 11 },  { 1986,  6, 12 },  { 1986,  6, 13 },
  { 1986,  6, 14 },  { 1986,  6, 15 },  { 1986,  6, 16 },  { 1986,  6, 17 },
  { 1986,  6, 18 },  { 1986,  6, 19 },  { 1986,  6, 20 },  { 1986,  6, 21 },
  { 1986,  6, 22 },  { 1986,  6, 23 },  { 1986,  6, 24 },  { 1986,  6, 25 },
  { 1986,  6, 26 },  { 1986,  6, 27 },  { 1986,  6, 28 },  { 1986,  6, 29 },
  { 1986,  6, 30 },  { 1986,  7,  1 },  { 1986,  7,  2 },  { 1986,  7,  3 },
  { 1986,  7,  4 },  { 1986,  7,  5 },  { 1986,  7,  6 },  { 1986,  7,  7 },
  { 1986,  7,  8 },  { 1986,  7,  9 },  { 1986,  7, 10 },  { 1986,  7, 11 },
  { 1986,  7, 12 },  { 1986,  7, 13 },  { 1986,  7, 14 },  { 1986,  7, 15 },
  { 1986,  7, 16 },  { 1986,  7, 17 },  { 1986,  7, 18 },  { 1986,  7, 19 },
  { 1986,  7, 20 },  { 1986,  7, 21 },  { 1986,  7, 22 },  { 1986,  7, 23 },
  { 1986,  7, 24 },  { 1986,  7, 25 },  { 1986,  7, 26 },  { 1986,  7, 27 },
  { 1986,  7, 28 },  { 1986,  7, 29 },  { 1986,  7, 30 },  { 1986,  7, 31 },
  { 1986,  8,  1 },  { 1986,  8,  2 },  { 1986,  8,  3 },  { 1986,  8,  4 },
  { 1986,  8,  5 },  { 1986,  8,  6 },  { 1986,  8,  7 },  { 1986,  8,  8 },
  { 1986,  8,  9 },  { 1986,  8, 10 },  { 1986,  8, 11 },  { 1986,  8, 12 },
  { 1986,  8, 13 },  { 1986,  8, 14 },  { 1986,  8, 15 },  { 1986,  8, 16 },
  { 1986,  8, 17 },  { 1986,  8, 18 },  { 1986,  8, 19 },  { 1986,  8, 20 },
  { 1986,  8, 21 },  { 1986,  8, 22 },  { 1986,  8, 23 },  { 1986,  8, 24 },
  { 1986,  8, 25 },  { 1986,  8, 26 },  { 1986,  8, 27 },  { 1986,  8, 28 },
  { 1986,  8, 29 },  { 1986,  8, 30 },  { 1986,  8, 31 },  { 1986,  9,  1 },
  { 1986,  9,  2 },  { 1986,  9,  3 },  { 1986,  9,  4 },  { 1986,  9,  5 },
  { 1986,  9,  6 },  { 1986,  9,  7 },  { 1986,  9,  8 },  { 1986,  9,  9 },
  { 1986,  9, 10 },  { 1986,  9, 11 },  { 1986,  9, 12 },  { 1986,  9, 13 },
  { 1986,  9, 14 },  { 1986,  9, 15 },  { 1986,  9, 16 },  { 1986,  9, 17 },
  { 1986,  9, 18 },  { 1986,  9, 19 },  { 1986,  9, 20 },  { 1986,  9, 21 },
  { 1986,  9, 22 },  { 1986,  9, 23 },  { 1986,  9, 24 },  { 1986,  9, 25 },
  { 1986,  9, 26 },  { 1986,  9, 27 },  { 1986,  9, 28 },  { 1986,  9, 29 },
  { 1986,  9, 30 },  { 1986, 10,  1 },  { 1986, 10,  2 },  { 1986, 10,  3 },
  { 1986, 10,  4 },  { 1986, 10,  5 },  { 1986, 10,  6 },  { 1986, 10,  7 },
  { 1986, 10,  8 },  { 1986, 10,  9 },  { 1986, 10, 10 },  { 1986, 10, 11 },
  { 1986, 10, 12 },  { 1986, 10, 13 },  { 1986, 10, 14 },  { 1986, 10, 15 },
  { 1986, 10, 16 },  { 1986, 10, 17 },  { 1986, 10, 18 },  { 1986, 10, 19 },
  { 1986, 10, 20 },  { 1986, 10, 21 },  { 1986, 10, 22 },  { 1986, 10, 23 },
  { 1986, 10, 24 },  { 1986, 10, 25 },  { 1986, 10, 26 },  { 1986, 10, 27 },
  { 1986, 10, 28 },  { 1986, 10, 29 },  { 1986, 10, 30 },  { 1986, 10, 31 },
  { 1986, 11,  1 },  { 1986, 11,  2 },  { 1986, 11,  3 },  { 1986, 11,  4 },
  { 1986, 11,  5 },  { 1986, 11,  6 },  { 1986, 11,  7 },  { 1986, 11,  8 },
  { 1986, 11,  9 },  { 1986, 11, 10 },  { 1986, 11, 11 },  { 1986, 11, 12 },
  { 1986, 11, 13 },  { 1986, 11, 14 },  { 1986, 11, 15 },  { 1986, 11, 16 },
  { 1986, 11, 17 },  { 1986, 11, 18 },  { 1986, 11, 19 },  { 1986, 11, 20 },
  { 1986, 11, 21 },  { 1986, 11, 22 },  { 1986, 11, 23 },  { 1986, 11, 24 },
  { 1986, 11, 25 },  { 1986, 11, 26 },  { 1986, 11, 27 },  { 1986, 11, 28 },
  { 1986, 11, 29 },  { 1986, 11, 30 },  { 1986, 12,  1 },  { 1986, 12,  2 },
  { 1986, 12,  3 },  { 1986, 12,  4 },  { 1986, 12,  5 },  { 1986, 12,  6 },
  { 1986, 12,  7 },  { 1986, 12,  8 },  { 1986, 12,  9 },  { 1986, 12, 10 },
  { 1986, 12, 11 },  { 1986, 12, 12 },  { 1986, 12, 13 },  { 1986, 12, 14 },
  { 1986, 12, 15 },  { 1986, 12, 16 },  { 1986, 12, 17 },  { 1986, 12, 18 },
  { 1986, 12, 19 },  { 1986, 12, 20 },  { 1986, 12, 21 },  { 1986, 12, 22 },
  { 1986, 12, 23 },  { 1986, 12, 24 },  { 1986, 12, 25 },  { 1986, 12, 26 },
  { 1986, 12, 27 },  { 1986, 12, 28 },  { 1986, 12, 29 },  { 1986, 12, 30 },
  { 1986, 12, 31 },  { 1987,  1,  1 },  { 1987,  1,  2 },  { 1987,  1,  3 },
  { 1987,  1,  4 },  { 1987,  1,  5 },  { 1987,  1,  6 },  { 1987,  1,  7 },
  { 1987,  1,  8 },  { 1987,  1,  9 },  { 1987,  1, 10 },  { 1987,  1, 11 },
  { 1987,  1, 12 },  { 1987,  1, 13 },  { 1987,  1, 14 },  { 1987,  1, 15 },
  { 1987,  1, 16 },  { 1987,  1, 17 },  { 1987,  1, 18 },  { 1987,  1, 19 },
  { 1987,  1, 20 },  { 1987,  1, 21 },  { 1987,  1, 22 },  { 1987,  1, 23 },
  { 1987,  1, 24 },  { 1987,  1, 25 },  { 1987,  1, 26 },  { 1987,  1, 27 },
  { 1987,  1, 28 },  { 1987,  1, 29 },  { 1987,  1, 30 },  { 1987,  1, 31 },
  { 1987,  2,  1 },  { 1987,  2,  2 },  { 1987,  2,  3 },  { 1987,  2,  4 },
  { 1987,  2,  5 },  { 1987,  2,  6 },  { 1987,  2,  7 },  { 1987,  2,  8 },
  { 1987,  2,  9 },  { 1987,  2, 10 },  { 1987,  2, 11 },  { 1987,  2, 12 },
  { 1987,  2, 13 },  { 1987,  2, 14 },  { 1987,  2, 15 },  { 1987,  2, 16 },
  { 1987,  2, 17 },  { 1987,  2, 18 },  { 1987,  2, 19 },  { 1987,  2, 20 },
  { 1987,  2, 21 },  { 1987,  2, 22 },  { 1987,  2, 23 },  { 1987,  2, 24 },
  { 1987,  2, 25 },  { 1987,  2, 26 },  { 1987,  2, 27 },  { 1987,  2, 28 },
  { 1987,  3,  1 },  { 1987,  3,  2 },  { 1987,  3,  3 },  { 1987,  3,  4 },
  { 1987,  3,  5 },  { 1987,  3,  6 },  { 1987,  3,  7 },  { 1987,  3,  8 },
  { 1987,  3,  9 },  { 1987,  3, 10 },  { 1987,  3, 11 },  { 1987,  3, 12 },
  { 1987,  3, 13 },  { 1987,  3, 14 },  { 1987,  3, 15 },  { 1987,  3, 16 },
  { 1987,  3, 17 },  { 1987,  3, 18 },  { 1987,  3, 19 },  { 1987,  3, 20 },
  { 1987,  3, 21 },  { 1987,  3, 22 },  { 1987,  3, 23 },  { 1987,  3, 24 },
  { 1987,  3, 25 },  { 1987,  3, 26 },  { 1987,  3, 27 },  { 1987,  3, 28 },
  { 1987,  3, 29 },  { 1987,  3, 30 },  { 1987,  3, 31 },  { 1987,  4,  1 },
  { 1987,  4,  2 },  { 1987,  4,  3 },  { 1987,  4,  4 },  { 1987,  4,  5 },
  { 1987,  4,  6 },  { 1987,  4,  7 },  { 1987,  4,  8 },  { 1987,  4,  9 },
  { 1987,  4, 10 },  { 1987,  4, 11 },  { 1987,  4, 12 },  { 1987,  4, 13 },
  { 1987,  4, 14 },  { 1987,  4, 15 },  { 1987,  4, 16 },  { 1987,  4, 17 },
  { 1987,  4, 18 },  { 1987,  4, 19 },  { 1987,  4, 20 },  { 1987,  4, 21 },
  { 1987,  4, 22 },  { 1987,  4, 23 },  { 1987,  4, 24 },  { 1987,  4, 25 },
  { 1987,  4, 26 },  { 1987,  4, 27 },  { 1987,  4, 28 },  { 1987,  4, 29 },
  { 1987,  4, 30 },  { 1987,  5,  1 },  { 1987,  5,  2 },  { 1987,  5,  3 },
  { 1987,  5,  4 },  { 1987,  5,  5 },  { 1987,  5,  6 },  { 1987,  5,  7 },
  { 1987,  5,  8 },  { 1987,  5,  9 },  { 1987,  5, 10 },  { 1987,  5, 11 },
  { 1987,  5, 12 },  { 1987,  5, 13 },  { 1987,  5, 14 },  { 1987,  5, 15 },
  { 1987,  5, 16 },  { 1987,  5, 17 },  { 1987,  5, 18 },  { 1987,  5, 19 },
  { 1987,  5, 20 },  { 1987,  5, 21 },  { 1987,  5, 22 },  { 1987,  5, 23 },
  { 1987,  5, 24 },  { 1987,  5, 25 },  { 1987,  5, 26 },  { 1987,  5, 27 },
  { 1987,  5, 28 },  { 1987,  5, 29 },  { 1987,  5, 30 },  { 1987,  5, 31 },
  { 1987,  6,  1 },  { 1987,  6,  2 },  { 1987,  6,  3 },  { 1987,  6,  4 },
  { 1987,  6,  5 },  { 1987,  6,  6 },  { 1987,  6,  7 },  { 1987,  6,  8 },
  { 1987,  6,  9 },  { 1987,  6, 10 },  { 1987,  6, 11 },  { 1987,  6, 12 },
  { 1987,  6, 13 },  { 1987,  6, 14 },  { 1987,  6, 15 },  { 1987,  6, 16 },
  { 1987,  6, 17 },  { 1987,  6, 18 },  { 1987,  6, 19 },  { 1987,  6, 20 },
  { 1987,  6, 21 },  { 1987,  6, 22 },  { 1987,  6, 23 },  { 1987,  6, 24 },
  { 1987,  6, 25 },  { 1987,  6, 26 },  { 1987,  6, 27 },  { 1987,  6, 28 },
  { 1987,  6, 29 },  { 1987,  6, 30 },  { 1987,  7,  1 },  { 1987,  7,  2 },
  { 1987,  7,  3 },  { 1987,  7,  4 },  { 1987,  7,  5 },  { 1987,  7,  6 },
  { 1987,  7,  7 },  { 1987,  7,  8 },  { 1987,  7,  9 },  { 1987,  7, 10 },
  { 1987,  7, 11 },  { 1987,  7, 12 },  { 1987,  7, 13 },  { 1987,  7, 14 },
  { 1987,  7, 15 },  { 1987,  7, 16 },  { 1987,  7, 17 },  { 1987,  7, 18 },
  { 1987,  7, 19 },  { 1987,  7, 20 },  { 1987,  7, 21 },  { 1987,  7, 22 },
  { 1987,  7, 23 },  { 1987,  7, 24 },  { 1987,  7, 25 },  { 1987,  7, 26 },
  { 1987,  7, 27 },  { 1987,  7, 28 },  { 1987,  7, 29 },  { 1987,  7, 30 },
  { 1987,  7, 31 },  { 1987,  8,  1 },  { 1987,  8,  2 },  { 1987,  8,  3 },
  { 1987,  8,  4 },  { 1987,  8,  5 },  { 1987,  8,  6 },  { 1987,  8,  7 },
  { 1987,  8,  8 },  { 1987,  8,  9 },  { 1987,  8, 10 },  { 1987,  8, 11 },
  { 1987,  8, 12 },  { 1987,  8, 13 },  { 1987,  8, 14 },  { 1987,  8, 15 },
  { 1987,  8, 16 },  { 1987,  8, 17 },  { 1987,  8, 18 },  { 1987,  8, 19 },
  { 1987,  8, 20 },  { 1987,  8, 21 },  { 1987,  8, 22 },  { 1987,  8, 23 },
  { 1987,  8, 24 },  { 1987,  8, 25 },  { 1987,  8, 26 },  { 1987,  8, 27 },
  { 1987,  8, 28 },  { 1987,  8, 29 },  { 1987,  8, 30 },  { 1987,  8, 31 },
  { 1987,  9,  1 },  { 1987,  9,  2 },  { 1987,  9,  3 },  { 1987,  9,  4 },
  { 1987,  9,  5 },  { 1987,  9,  6 },  { 1987,  9,  7 },  { 1987,  9,  8 },
  { 1987,  9,  9 },  { 1987,  9, 10 },  { 1987,  9, 11 },  { 1987,  9, 12 },
  { 1987,  9, 13 },  { 1987,  9, 14 },  { 1987,  9, 15 },  { 1987,  9, 16 },
  { 1987,  9, 17 },  { 1987,  9, 18 },  { 1987,  9, 19 },  { 1987,  9, 20 },
  { 1987,  9, 21 },  { 1987,  9, 22 },  { 1987,  9, 23 },  { 1987,  9, 24 },
  { 1987,  9, 25 },  { 1987,  9, 26 },  { 1987,  9, 27 },  { 1987,  9, 28 },
  { 1987,  9, 29 },  { 1987,  9, 30 },  { 1987, 10,  1 },  { 1987, 10,  2 },
  { 1987, 10,  3 },  { 1987, 10,  4 },  { 1987, 10,  5 },  { 1987, 10,  6 },
  { 1987, 10,  7 },  { 1987, 10,  8 },  { 1987, 10,  9 },  { 1987, 10, 10 },
  { 1987, 10, 11 },  { 1987, 10, 12 },  { 1987, 10, 13 },  { 1987, 10, 14 },
  { 1987, 10, 15 },  { 1987, 10, 16 },  { 1987, 10, 17 },  { 1987, 10, 18 },
  { 1987, 10, 19 },  { 1987, 10, 20 },  { 1987, 10, 21 },  { 1987, 10, 22 },
  { 1987, 10, 23 },  { 1987, 10, 24 },  { 1987, 10, 25 },  { 1987, 10, 26 },
  { 1987, 10, 27 },  { 1987, 10, 28 },  { 1987, 10, 29 },  { 1987, 10, 30 },
  { 1987, 10, 31 },  { 1987, 11,  1 },  { 1987, 11,  2 },  { 1987, 11,  3 },
  { 1987, 11,  4 },  { 1987, 11,  5 },  { 1987, 11,  6 },  { 1987, 11,  7 },
  { 1987, 11,  8 },  { 1987, 11,  9 },  { 1987, 11, 10 },  { 1987, 11, 11 },
  { 1987, 11, 12 },  { 1987, 11, 13 },  { 1987, 11, 14 },  { 1987, 11, 15 },
  { 1987, 11, 16 },  { 1987, 11, 17 },  { 1987, 11, 18 },  { 1987, 11, 19 },
  { 1987, 11, 20 },  { 1987, 11, 21 },  { 1987, 11, 22 },  { 1987, 11, 23 },
  { 1987, 11, 24 },  { 1987, 11, 25 },  { 1987, 11, 26 },  { 1987, 11, 27 },
  { 1987, 11, 28 },  { 1987, 11, 29 },  { 1987, 11, 30 },  { 1987, 12,  1 },
  { 1987, 12,  2 },  { 1987, 12,  3 },  { 1987, 12,  4 },  { 1987, 12,  5 },
  { 1987, 12,  6 },  { 1987, 12,  7 },  { 1987, 12,  8 },  { 1987, 12,  9 },
  { 1987, 12, 10 },  { 1987, 12, 11 },  { 1987, 12, 12 },  { 1987, 12, 13 },
  { 1987, 12, 14 },  { 1987, 12, 15 },  { 1987, 12, 16 },  { 1987, 12, 17 },
  { 1987, 12, 18 },  { 1987, 12, 19 },  { 1987, 12, 20 },  { 1987, 12, 21 },
  { 1987, 12, 22 },  { 1987, 12, 23 },  { 1987, 12, 24 },  { 1987, 12, 25 },
  { 1987, 12, 26 },  { 1987, 12, 27 },  { 1987, 12, 28 },  { 1987, 12, 29 },
  { 1987, 12, 30 },  { 1987, 12, 31 },  { 1988,  1,  1 },  { 1988,  1,  2 },
  { 1988,  1,  3 },  { 1988,  1,  4 },  { 1988,  1,  5 },  { 1988,  1,  6 },
  { 1988,  1,  7 },  { 1988,  1,  8 },  { 1988,  1,  9 },  { 1988,  1, 10 },
  { 1988,  1, 11 },  { 1988,  1, 12 },  { 1988,  1, 13 },  { 1988,  1, 14 },
  { 1988,  1, 15 },  { 1988,  1, 16 },  { 1988,  1, 17 },  { 1988,  1, 18 },
  { 1988,  1, 19 },  { 1988,  1, 20 },  { 1988,  1, 21 },  { 1988,  1, 22 },
  { 1988,  1, 23 },  { 1988,  1, 24 },  { 1988,  1, 25 },  { 1988,  1, 26 },
  { 1988,  1, 27 },  { 1988,  1, 28 },  { 1988,  1, 29 },  { 1988,  1, 30 },
  { 1988,  1, 31 },  { 1988,  2,  1 },  { 1988,  2,  2 },  { 1988,  2,  3 },
  { 1988,  2,  4 },  { 1988,  2,  5 },  { 1988,  2,  6 },  { 1988,  2,  7 },
  { 1988,  2,  8 },  { 1988,  2,  9 },  { 1988,  2, 10 },  { 1988,  2, 11 },
  { 1988,  2, 12 },  { 1988,  2, 13 },  { 1988,  2, 14 },  { 1988,  2, 15 },
  { 1988,  2, 16 },  { 1988,  2, 17 },  { 1988,  2, 18 },  { 1988,  2, 19 },
  { 1988,  2, 20 },  { 1988,  2, 21 },  { 1988,  2, 22 },  { 1988,  2, 23 },
  { 1988,  2, 24 },  { 1988,  2, 25 },  { 1988,  2, 26 },  { 1988,  2, 27 },
  { 1988,  2, 28 },  { 1988,  2, 29 },  { 1988,  3,  1 },  { 1988,  3,  2 },
  { 1988,  3,  3 },  { 1988,  3,  4 },  { 1988,  3,  5 },  { 1988,  3,  6 },
  { 1988,  3,  7 },  { 1988,  3,  8 },  { 1988,  3,  9 },  { 1988,  3, 10 },
  { 1988,  3, 11 },  { 1988,  3, 12 },  { 1988,  3, 13 },  { 1988,  3, 14 },
  { 1988,  3, 15 },  { 1988,  3, 16 },  { 1988,  3, 17 },  { 1988,  3, 18 },
  { 1988,  3, 19 },  { 1988,  3, 20 },  { 1988,  3, 21 },  { 1988,  3, 22 },
  { 1988,  3, 23 },  { 1988,  3, 24 },  { 1988,  3, 25 },  { 1988,  3, 26 },
  { 1988,  3, 27 },  { 1988,  3, 28 },  { 1988,  3, 29 },  { 1988,  3, 30 },
  { 1988,  3, 31 },  { 1988,  4,  1 },  { 1988,  4,  2 },  { 1988,  4,  3 },
  { 1988,  4,  4 },  { 1988,  4,  5 },  { 1988,  4,  6 },  { 1988,  4,  7 },
  { 1988,  4,  8 },  { 1988,  4,  9 },  { 1988,  4, 10 },  { 1988,  4, 11 },
  { 1988,  4, 12 },  { 1988,  4, 13 },  { 1988,  4, 14 },  { 1988,  4, 15 },
  { 1988,  4, 16 },  { 1988,  4, 17 },  { 1988,  4, 18 },  { 1988,  4, 19 },
  { 1988,  4, 20 },  { 1988,  4, 21 },  { 1988,  4, 22 },  { 1988,  4, 23 },
  { 1988,  4, 24 },  { 1988,  4, 25 },  { 1988,  4, 26 },  { 1988,  4, 27 },
  { 1988,  4, 28 },  { 1988,  4, 29 },  { 1988,  4, 30 },  { 1988,  5,  1 },
  { 1988,  5,  2 },  { 1988,  5,  3 },  { 1988,  5,  4 },  { 1988,  5,  5 },
  { 1988,  5,  6 },  { 1988,  5,  7 },  { 1988,  5,  8 },  { 1988,  5,  9 },
  { 1988,  5, 10 },  { 1988,  5, 11 },  { 1988,  5, 12 },  { 1988,  5, 13 },
  { 1988,  5, 14 },  { 1988,  5, 15 },  { 1988,  5, 16 },  { 1988,  5, 17 },
  { 1988,  5, 18 },  { 1988,  5, 19 },  { 1988,  5, 20 },  { 1988,  5, 21 },
  { 1988,  5, 22 },  { 1988,  5, 23 },  { 1988,  5, 24 },  { 1988,  5, 25 },
  { 1988,  5, 26 },  { 1988,  5, 27 },  { 1988,  5, 28 },  { 1988,  5, 29 },
  { 1988,  5, 30 },  { 1988,  5, 31 },  { 1988,  6,  1 },  { 1988,  6,  2 },
  { 1988,  6,  3 },  { 1988,  6,  4 },  { 1988,  6,  5 },  { 1988,  6,  6 },
  { 1988,  6,  7 },  { 1988,  6,  8 },  { 1988,  6,  9 },  { 1988,  6, 10 },
  { 1988,  6, 11 },  { 1988,  6, 12 },  { 1988,  6, 13 },  { 1988,  6, 14 },
  { 1988,  6, 15 },  { 1988,  6, 16 },  { 1988,  6, 17 },  { 1988,  6, 18 },
  { 1988,  6, 19 },  { 1988,  6, 20 },  { 1988,  6, 21 },  { 1988,  6, 22 },
  { 1988,  6, 23 },  { 1988,  6, 24 },  { 1988,  6, 25 },  { 1988,  6, 26 },
  { 1988,  6, 27 },  { 1988,  6, 28 },  { 1988,  6, 29 },  { 1988,  6, 30 },
  { 1988,  7,  1 },  { 1988,  7,  2 },  { 1988,  7,  3 },  { 1988,  7,  4 },
  { 1988,  7,  5 },  { 1988,  7,  6 },  { 1988,  7,  7 },  { 1988,  7,  8 },
  { 1988,  7,  9 },  { 1988,  7, 10 },  { 1988,  7, 11 },  { 1988,  7, 12 },
  { 1988,  7, 13 },  { 1988,  7, 14 },  { 1988,  7, 15 },  { 1988,  7, 16 },
  { 1988,  7, 17 },  { 1988,  7, 18 },  { 1988,  7, 19 },  { 1988,  7, 20 },
  { 1988,  7, 21 },  { 1988,  7, 22 },  { 1988,  7, 23 },  { 1988,  7, 24 },
  { 1988,  7, 25 },  { 1988,  7, 26 },  { 1988,  7, 27 },  { 1988,  7, 28 },
  { 1988,  7, 29 },  { 1988,  7, 30 },  { 1988,  7, 31 },  { 1988,  8,  1 },
  { 1988,  8,  2 },  { 1988,  8,  3 },  { 1988,  8,  4 },  { 1988,  8,  5 },
  { 1988,  8,  6 },  { 1988,  8,  7 },  { 1988,  8,  8 },  { 1988,  8,  9 },
  { 1988,  8, 10 },  { 1988,  8, 11 },  { 1988,  8, 12 },  { 1988,  8, 13 },
  { 1988,  8, 14 },  { 1988,  8, 15 },  { 1988,  8, 16 },  { 1988,  8, 17 },
  { 1988,  8, 18 },  { 1988,  8, 19 },  { 1988,  8, 20 },  { 1988,  8, 21 },
  { 1988,  8, 22 },  { 1988,  8, 23 },  { 1988,  8, 24 },  { 1988,  8, 25 },
  { 1988,  8, 26 },  { 1988,  8, 27 },  { 1988,  8, 28 },  { 1988,  8, 29 },
  { 1988,  8, 30 },  { 1988,  8, 31 },  { 1988,  9,  1 },  { 1988,  9,  2 },
  { 1988,  9,  3 },  { 1988,  9,  4 },  { 1988,  9,  5 },  { 1988,  9,  6 },
  { 1988,  9,  7 },  { 1988,  9,  8 },  { 1988,  9,  9 },  { 1988,  9, 10 },
  { 1988,  9, 11 },  { 1988,  9, 12 },  { 1988,  9, 13 },  { 1988,  9, 14 },
  { 1988,  9, 15 },  { 1988,  9, 16 },  { 1988,  9, 17 },  { 1988,  9, 18 },
  { 1988,  9, 19 },  { 1988,  9, 20 },  { 1988,  9, 21 },  { 1988,  9, 22 },
  { 1988,  9, 23 },  { 1988,  9, 24 },  { 1988,  9, 25 },  { 1988,  9, 26 },
  { 1988,  9, 27 },  { 1988,  9, 28 },  { 1988,  9, 29 },  { 1988,  9, 30 },
  { 1988, 10,  1 },  { 1988, 10,  2 },  { 1988, 10,  3 },  { 1988, 10,  4 },
  { 1988, 10,  5 },  { 1988, 10,  6 },  { 1988, 10,  7 },  { 1988, 10,  8 },
  { 1988, 10,  9 },  { 1988, 10, 10 },  { 1988, 10, 11 },  { 1988, 10, 12 },
  { 1988, 10, 13 },  { 1988, 10, 14 },  { 1988, 10, 15 },  { 1988, 10, 16 },
  { 1988, 10, 17 },  { 1988, 10, 18 },  { 1988, 10, 19 },  { 1988, 10, 20 },
  { 1988, 10, 21 },  { 1988, 10, 22 },  { 1988, 10, 23 },  { 1988, 10, 24 },
  { 1988, 10, 25 },  { 1988, 10, 26 },  { 1988, 10, 27 },  { 1988, 10, 28 },
  { 1988, 10, 29 },  { 1988, 10, 30 },  { 1988, 10, 31 },  { 1988, 11,  1 },
  { 1988, 11,  2 },  { 1988, 11,  3 },  { 1988, 11,  4 },  { 1988, 11,  5 },
  { 1988, 11,  6 },  { 1988, 11,  7 },  { 1988, 11,  8 },  { 1988, 11,  9 },
  { 1988, 11, 10 },  { 1988, 11, 11 },  { 1988, 11, 12 },  { 1988, 11, 13 },
  { 1988, 11, 14 },  { 1988, 11, 15 },  { 1988, 11, 16 },  { 1988, 11, 17 },
  { 1988, 11, 18 },  { 1988, 11, 19 },  { 1988, 11, 20 },  { 1988, 11, 21 },
  { 1988, 11, 22 },  { 1988, 11, 23 },  { 1988, 11, 24 },  { 1988, 11, 25 },
  { 1988, 11, 26 },  { 1988, 11, 27 },  { 1988, 11, 28 },  { 1988, 11, 29 },
  { 1988, 11, 30 },  { 1988, 12,  1 },  { 1988, 12,  2 },  { 1988, 12,  3 },
  { 1988, 12,  4 },  { 1988, 12,  5 },  { 1988, 12,  6 },  { 1988, 12,  7 },
  { 1988, 12,  8 },  { 1988, 12,  9 },  { 1988, 12, 10 },  { 1988, 12, 11 },
  { 1988, 12, 12 },  { 1988, 12, 13 },  { 1988, 12, 14 },  { 1988, 12, 15 },
  { 1988, 12, 16 },  { 1988, 12, 17 },  { 1988, 12, 18 },  { 1988, 12, 19 },
  { 1988, 12, 20 },  { 1988, 12, 21 },  { 1988, 12, 22 },  { 1988, 12, 23 },
  { 1988, 12, 24 },  { 1988, 12, 25 },  { 1988, 12, 26 },  { 1988, 12, 27 },
  { 1988, 12, 28 },  { 1988, 12, 29 },  { 1988, 12, 30 },  { 1988, 12, 31 },
  { 1989,  1,  1 },  { 1989,  1,  2 },  { 1989,  1,  3 },  { 1989,  1,  4 },
  { 1989,  1,  5 },  { 1989,  1,  6 },  { 1989,  1,  7 },  { 1989,  1,  8 },
  { 1989,  1,  9 },  { 1989,  1, 10 },  { 1989,  1, 11 },  { 1989,  1, 12 },
  { 1989,  1, 13 },  { 1989,  1, 14 },  { 1989,  1, 15 },  { 1989,  1, 16 },
  { 1989,  1, 17 },  { 1989,  1, 18 },  { 1989,  1, 19 },  { 1989,  1, 20 },
  { 1989,  1, 21 },  { 1989,  1, 22 },  { 1989,  1, 23 },  { 1989,  1, 24 },
  { 1989,  1, 25 },  { 1989,  1, 26 },  { 1989,  1, 27 },  { 1989,  1, 28 },
  { 1989,  1, 29 },  { 1989,  1, 30 },  { 1989,  1, 31 },  { 1989,  2,  1 },
  { 1989,  2,  2 },  { 1989,  2,  3 },  { 1989,  2,  4 },  { 1989,  2,  5 },
  { 1989,  2,  6 },  { 1989,  2,  7 },  { 1989,  2,  8 },  { 1989,  2,  9 },
  { 1989,  2, 10 },  { 1989,  2, 11 },  { 1989,  2, 12 },  { 1989,  2, 13 },
  { 1989,  2, 14 },  { 1989,  2, 15 },  { 1989,  2, 16 },  { 1989,  2, 17 },
  { 1989,  2, 18 },  { 1989,  2, 19 },  { 1989,  2, 20 },  { 1989,  2, 21 },
  { 1989,  2, 22 },  { 1989,  2, 23 },  { 1989,  2, 24 },  { 1989,  2, 25 },
  { 1989,  2, 26 },  { 1989,  2, 27 },  { 1989,  2, 28 },  { 1989,  3,  1 },
  { 1989,  3,  2 },  { 1989,  3,  3 },  { 1989,  3,  4 },  { 1989,  3,  5 },
  { 1989,  3,  6 },  { 1989,  3,  7 },  { 1989,  3,  8 },  { 1989,  3,  9 },
  { 1989,  3, 10 },  { 1989,  3, 11 },  { 1989,  3, 12 },  { 1989,  3, 13 },
  { 1989,  3, 14 },  { 1989,  3, 15 },  { 1989,  3, 16 },  { 1989,  3, 17 },
  { 1989,  3, 18 },  { 1989,  3, 19 },  { 1989,  3, 20 },  { 1989,  3, 21 },
  { 1989,  3, 22 },  { 1989,  3, 23 },  { 1989,  3, 24 },  { 1989,  3, 25 },
  { 1989,  3, 26 },  { 1989,  3, 27 },  { 1989,  3, 28 },  { 1989,  3, 29 },
  { 1989,  3, 30 },  { 1989,  3, 31 },  { 1989,  4,  1 },  { 1989,  4,  2 },
  { 1989,  4,  3 },  { 1989,  4,  4 },  { 1989,  4,  5 },  { 1989,  4,  6 },
  { 1989,  4,  7 },  { 1989,  4,  8 },  { 1989,  4,  9 },  { 1989,  4, 10 },
  { 1989,  4, 11 },  { 1989,  4, 12 },  { 1989,  4, 13 },  { 1989,  4, 14 },
  { 1989,  4, 15 },  { 1989,  4, 16 },  { 1989,  4, 17 },  { 1989,  4, 18 },
  { 1989,  4, 19 },  { 1989,  4, 20 },  { 1989,  4, 21 },  { 1989,  4, 22 },
  { 1989,  4, 23 },  { 1989,  4, 24 },  { 1989,  4, 25 },  { 1989,  4, 26 },
  { 1989,  4, 27 },  { 1989,  4, 28 },  { 1989,  4, 29 },  { 1989,  4, 30 },
  { 1989,  5,  1 },  { 1989,  5,  2 },  { 1989,  5,  3 },  { 1989,  5,  4 },
  { 1989,  5,  5 },  { 1989,  5,  6 },  { 1989,  5,  7 },  { 1989,  5,  8 },
  { 1989,  5,  9 },  { 1989,  5, 10 },  { 1989,  5, 11 },  { 1989,  5, 12 },
  { 1989,  5, 13 },  { 1989,  5, 14 },  { 1989,  5, 15 },  { 1989,  5, 16 },
  { 1989,  5, 17 },  { 1989,  5, 18 },  { 1989,  5, 19 },  { 1989,  5, 20 },
  { 1989,  5, 21 },  { 1989,  5, 22 },  { 1989,  5, 23 },  { 1989,  5, 24 },
  { 1989,  5, 25 },  { 1989,  5, 26 },  { 1989,  5, 27 },  { 1989,  5, 28 },
  { 1989,  5, 29 },  { 1989,  5, 30 },  { 1989,  5, 31 },  { 1989,  6,  1 },
  { 1989,  6,  2 },  { 1989,  6,  3 },  { 1989,  6,  4 },  { 1989,  6,  5 },
  { 1989,  6,  6 },  { 1989,  6,  7 },  { 1989,  6,  8 },  { 1989,  6,  9 },
  { 1989,  6, 10 },  { 1989,  6, 11 },  { 1989,  6, 12 },  { 1989,  6, 13 },
  { 1989,  6, 14 },  { 1989,  6, 15 },  { 1989,  6, 16 },  { 1989,  6, 17 },
  { 1989,  6, 18 },  { 1989,  6, 19 },  { 1989,  6, 20 },  { 1989,  6, 21 },
  { 1989,  6, 22 },  { 1989,  6, 23 },  { 1989,  6, 24 },  { 1989,  6, 25 },
  { 1989,  6, 26 },  { 1989,  6, 27 },  { 1989,  6, 28 },  { 1989,  6, 29 },
  { 1989,  6, 30 },  { 1989,  7,  1 },  { 1989,  7,  2 },  { 1989,  7,  3 },
  { 1989,  7,  4 },  { 1989,  7,  5 },  { 1989,  7,  6 },  { 1989,  7,  7 },
  { 1989,  7,  8 },  { 1989,  7,  9 },  { 1989,  7, 10 },  { 1989,  7, 11 },
  { 1989,  7, 12 },  { 1989,  7, 13 },  { 1989,  7, 14 },  { 1989,  7, 15 },
  { 1989,  7, 16 },  { 1989,  7, 17 },  { 1989,  7, 18 },  { 1989,  7, 19 },
  { 1989,  7, 20 },  { 1989,  7, 21 },  { 1989,  7, 22 },  { 1989,  7, 23 },
  { 1989,  7, 24 },  { 1989,  7, 25 },  { 1989,  7, 26 },  { 1989,  7, 27 },
  { 1989,  7, 28 },  { 1989,  7, 29 },  { 1989,  7, 30 },  { 1989,  7, 31 },
  { 1989,  8,  1 },  { 1989,  8,  2 },  { 1989,  8,  3 },  { 1989,  8,  4 },
  { 1989,  8,  5 },  { 1989,  8,  6 },  { 1989,  8,  7 },  { 1989,  8,  8 },
  { 1989,  8,  9 },  { 1989,  8, 10 },  { 1989,  8, 11 },  { 1989,  8, 12 },
  { 1989,  8, 13 },  { 1989,  8, 14 },  { 1989,  8, 15 },  { 1989,  8, 16 },
  { 1989,  8, 17 },  { 1989,  8, 18 },  { 1989,  8, 19 },  { 1989,  8, 20 },
  { 1989,  8, 21 },  { 1989,  8, 22 },  { 1989,  8, 23 },  { 1989,  8, 24 },
  { 1989,  8, 25 },  { 1989,  8, 26 },  { 1989,  8, 27 },  { 1989,  8, 28 },
  { 1989,  8, 29 },  { 1989,  8, 30 },  { 1989,  8, 31 },  { 1989,  9,  1 },
  { 1989,  9,  2 },  { 1989,  9,  3 },  { 1989,  9,  4 },  { 1989,  9,  5 },
  { 1989,  9,  6 },  { 1989,  9,  7 },  { 1989,  9,  8 },  { 1989,  9,  9 },
  { 1989,  9, 10 },  { 1989,  9, 11 },  { 1989,  9, 12 },  { 1989,  9, 13 },
  { 1989,  9, 14 },  { 1989,  9, 15 },  { 1989,  9, 16 },  { 1989,  9, 17 },
  { 1989,  9, 18 },  { 1989,  9, 19 },  { 1989,  9, 20 },  { 1989,  9, 21 },
  { 1989,  9, 22 },  { 1989,  9, 23 },  { 1989,  9, 24 },  { 1989,  9, 25 },
  { 1989,  9, 26 },  { 1989,  9, 27 },  { 1989,  9, 28 },  { 1989,  9, 29 },
  { 1989,  9, 30 },  { 1989, 10,  1 },  { 1989, 10,  2 },  { 1989, 10,  3 },
  { 1989, 10,  4 },  { 1989, 10,  5 },  { 1989, 10,  6 },  { 1989, 10,  7 },
  { 1989, 10,  8 },  { 1989, 10,  9 },  { 1989, 10, 10 },  { 1989, 10, 11 },
  { 1989, 10, 12 },  { 1989, 10, 13 },  { 1989, 10, 14 },  { 1989, 10, 15 },
  { 1989, 10, 16 },  { 1989, 10, 17 },  { 1989, 10, 18 },  { 1989, 10, 19 },
  { 1989, 10, 20 },  { 1989, 10, 21 },  { 1989, 10, 22 },  { 1989, 10, 23 },
  { 1989, 10, 24 },  { 1989, 10, 25 },  { 1989, 10, 26 },  { 1989, 10, 27 },
  { 1989, 10, 28 },  { 1989, 10, 29 },  { 1989, 10, 30 },  { 1989, 10, 31 },
  { 1989, 11,  1 },  { 1989, 11,  2 },  { 1989, 11,  3 },  { 1989, 11,  4 },
  { 1989, 11,  5 },  { 1989, 11,  6 },  { 1989, 11,  7 },  { 1989, 11,  8 },
  { 1989, 11,  9 },  { 1989, 11, 10 },  { 1989, 11, 11 },  { 1989, 11, 12 },
  { 1989, 11, 13 },  { 1989, 11, 14 },  { 1989, 11, 15 },  { 1989, 11, 16 },
  { 1989, 11, 17 },  { 1989, 11, 18 },  { 1989, 11, 19 },  { 1989, 11, 20 },
  { 1989, 11, 21 },  { 1989, 11, 22 },  { 1989, 11, 23 },  { 1989, 11, 24 },
  { 1989, 11, 25 },  { 1989, 11, 26 },  { 1989, 11, 27 },  { 1989, 11, 28 },
  { 1989, 11, 29 },  { 1989, 11, 30 },  { 1989, 12,  1 },  { 1989, 12,  2 },
  { 1989, 12,  3 },  { 1989, 12,  4 },  { 1989, 12,  5 },  { 1989, 12,  6 },
  { 1989, 12,  7 },  { 1989, 12,  8 },  { 1989, 12,  9 },  { 1989, 12, 10 },
  { 1989, 12, 11 },  { 1989, 12, 12 },  { 1989, 12, 13 },  { 1989, 12, 14 },
  { 1989, 12, 15 },  { 1989, 12, 16 },  { 1989, 12, 17 },  { 1989, 12, 18 },
  { 1989, 12, 19 },  { 1989, 12, 20 },  { 1989, 12, 21 },  { 1989, 12, 22 },
  { 1989, 12, 23 },  { 1989, 12, 24 },  { 1989, 12, 25 },  { 1989, 12, 26 },
  { 1989, 12, 27 },  { 1989, 12, 28 },  { 1989, 12, 29 },  { 1989, 12, 30 },
  { 1989, 12, 31 },  { 1990,  1,  1 },  { 1990,  1,  2 },  { 1990,  1,  3 },
  { 1990,  1,  4 },  { 1990,  1,  5 },  { 1990,  1,  6 },  { 1990,  1,  7 },
  { 1990,  1,  8 },  { 1990,  1,  9 },  { 1990,  1, 10 },  { 1990,  1, 11 },
  { 1990,  1, 12 },  { 1990,  1, 13 },  { 1990,  1, 14 },  { 1990,  1, 15 },
  { 1990,  1, 16 },  { 1990,  1, 17 },  { 1990,  1, 18 },  { 1990,  1, 19 },
  { 1990,  1, 20 },  { 1990,  1, 21 },  { 1990,  1, 22 },  { 1990,  1, 23 },
  { 1990,  1, 24 },  { 1990,  1, 25 },  { 1990,  1, 26 },  { 1990,  1, 27 },
  { 1990,  1, 28 },  { 1990,  1, 29 },  { 1990,  1, 30 },  { 1990,  1, 31 },
  { 1990,  2,  1 },  { 1990,  2,  2 },  { 1990,  2,  3 },  { 1990,  2,  4 },
  { 1990,  2,  5 },  { 1990,  2,  6 },  { 1990,  2,  7 },  { 1990,  2,  8 },
  { 1990,  2,  9 },  { 1990,  2, 10 },  { 1990,  2, 11 },  { 1990,  2, 12 },
  { 1990,  2, 13 },  { 1990,  2, 14 },  { 1990,  2, 15 },  { 1990,  2, 16 },
  { 1990,  2, 17 },  { 1990,  2, 18 },  { 1990,  2, 19 },  { 1990,  2, 20 },
  { 1990,  2, 21 },  { 1990,  2, 22 },  { 1990,  2, 23 },  { 1990,  2, 24 },
  { 1990,  2, 25 },  { 1990,  2, 26 },  { 1990,  2, 27 },  { 1990,  2, 28 },
  { 1990,  3,  1 },  { 1990,  3,  2 },  { 1990,  3,  3 },  { 1990,  3,  4 },
  { 1990,  3,  5 },  { 1990,  3,  6 },  { 1990,  3,  7 },  { 1990,  3,  8 },
  { 1990,  3,  9 },  { 1990,  3, 10 },  { 1990,  3, 11 },  { 1990,  3, 12 },
  { 1990,  3, 13 },  { 1990,  3, 14 },  { 1990,  3, 15 },  { 1990,  3, 16 },
  { 1990,  3, 17 },  { 1990,  3, 18 },  { 1990,  3, 19 },  { 1990,  3, 20 },
  { 1990,  3, 21 },  { 1990,  3, 22 },  { 1990,  3, 23 },  { 1990,  3, 24 },
  { 1990,  3, 25 },  { 1990,  3, 26 },  { 1990,  3, 27 },  { 1990,  3, 28 },
  { 1990,  3, 29 },  { 1990,  3, 30 },  { 1990,  3, 31 },  { 1990,  4,  1 },
  { 1990,  4,  2 },  { 1990,  4,  3 },  { 1990,  4,  4 },  { 1990,  4,  5 },
  { 1990,  4,  6 },  { 1990,  4,  7 },  { 1990,  4,  8 },  { 1990,  4,  9 },
  { 1990,  4, 10 },  { 1990,  4, 11 },  { 1990,  4, 12 },  { 1990,  4, 13 },
  { 1990,  4, 14 },  { 1990,  4, 15 },  { 1990,  4, 16 },  { 1990,  4, 17 },
  { 1990,  4, 18 },  { 1990,  4, 19 },  { 1990,  4, 20 },  { 1990,  4, 21 },
  { 1990,  4, 22 },  { 1990,  4, 23 },  { 1990,  4, 24 },  { 1990,  4, 25 },
  { 1990,  4, 26 },  { 1990,  4, 27 },  { 1990,  4, 28 },  { 1990,  4, 29 },
  { 1990,  4, 30 },  { 1990,  5,  1 },  { 1990,  5,  2 },  { 1990,  5,  3 },
  { 1990,  5,  4 },  { 1990,  5,  5 },  { 1990,  5,  6 },  { 1990,  5,  7 },
  { 1990,  5,  8 },  { 1990,  5,  9 },  { 1990,  5, 10 },  { 1990,  5, 11 },
  { 1990,  5, 12 },  { 1990,  5, 13 },  { 1990,  5, 14 },  { 1990,  5, 15 },
  { 1990,  5, 16 },  { 1990,  5, 17 },  { 1990,  5, 18 },  { 1990,  5, 19 },
  { 1990,  5, 20 },  { 1990,  5, 21 },  { 1990,  5, 22 },  { 1990,  5, 23 },
  { 1990,  5, 24 },  { 1990,  5, 25 },  { 1990,  5, 26 },  { 1990,  5, 27 },
  { 1990,  5, 28 },  { 1990,  5, 29 },  { 1990,  5, 30 },  { 1990,  5, 31 },
  { 1990,  6,  1 },  { 1990,  6,  2 },  { 1990,  6,  3 },  { 1990,  6,  4 },
  { 1990,  6,  5 },  { 1990,  6,  6 },  { 1990,  6,  7 },  { 1990,  6,  8 },
  { 1990,  6,  9 },  { 1990,  6, 10 },  { 1990,  6, 11 },  { 1990,  6, 12 },
  { 1990,  6, 13 },  { 1990,  6, 14 },  { 1990,  6, 15 },  { 1990,  6, 16 },
  { 1990,  6, 17 },  { 1990,  6, 18 },  { 1990,  6, 19 },  { 1990,  6, 20 },
  { 1990,  6, 21 },  { 1990,  6, 22 },  { 1990,  6, 23 },  { 1990,  6, 24 },
  { 1990,  6, 25 },  { 1990,  6, 26 },  { 1990,  6, 27 },  { 1990,  6, 28 },
  { 1990,  6, 29 },  { 1990,  6, 30 },  { 1990,  7,  1 },  { 1990,  7,  2 },
  { 1990,  7,  3 },  { 1990,  7,  4 },  { 1990,  7,  5 },  { 1990,  7,  6 },
  { 1990,  7,  7 },  { 1990,  7,  8 },  { 1990,  7,  9 },  { 1990,  7, 10 },
  { 1990,  7, 11 },  { 1990,  7, 12 },  { 1990,  7, 13 },  { 1990,  7, 14 },
  { 1990,  7, 15 },  { 1990,  7, 16 },  { 1990,  7, 17 },  { 1990,  7, 18 },
  { 1990,  7, 19 },  { 1990,  7, 20 },  { 1990,  7, 21 },  { 1990,  7, 22 },
  { 1990,  7, 23 },  { 1990,  7, 24 },  { 1990,  7, 25 },  { 1990,  7, 26 },
  { 1990,  7, 27 },  { 1990,  7, 28 },  { 1990,  7, 29 },  { 1990,  7, 30 },
  { 1990,  7, 31 },  { 1990,  8,  1 },  { 1990,  8,  2 },  { 1990,  8,  3 },
  { 1990,  8,  4 },  { 1990,  8,  5 },  { 1990,  8,  6 },  { 1990,  8,  7 },
  { 1990,  8,  8 },  { 1990,  8,  9 },  { 1990,  8, 10 },  { 1990,  8, 11 },
  { 1990,  8, 12 },  { 1990,  8, 13 },  { 1990,  8, 14 },  { 1990,  8, 15 },
  { 1990,  8, 16 },  { 1990,  8, 17 },  { 1990,  8, 18 },  { 1990,  8, 19 },
  { 1990,  8, 20 },  { 1990,  8, 21 },  { 1990,  8, 22 },  { 1990,  8, 23 },
  { 1990,  8, 24 },  { 1990,  8, 25 },  { 1990,  8, 26 },  { 1990,  8, 27 },
  { 1990,  8, 28 },  { 1990,  8, 29 },  { 1990,  8, 30 },  { 1990,  8, 31 },
  { 1990,  9,  1 },  { 1990,  9,  2 },  { 1990,  9,  3 },  { 1990,  9,  4 },
  { 1990,  9,  5 },  { 1990,  9,  6 },  { 1990,  9,  7 },  { 1990,  9,  8 },
  { 1990,  9,  9 },  { 1990,  9, 10 },  { 1990,  9, 11 },  { 1990,  9, 12 },
  { 1990,  9, 13 },  { 1990,  9, 14 },  { 1990,  9, 15 },  { 1990,  9, 16 },
  { 1990,  9, 17 },  { 1990,  9, 18 },  { 1990,  9, 19 },  { 1990,  9, 20 },
  { 1990,  9, 21 },  { 1990,  9, 22 },  { 1990,  9, 23 },  { 1990,  9, 24 },
  { 1990,  9, 25 },  { 1990,  9, 26 },  { 1990,  9, 27 },  { 1990,  9, 28 },
  { 1990,  9, 29 },  { 1990,  9, 30 },  { 1990, 10,  1 },  { 1990, 10,  2 },
  { 1990, 10,  3 },  { 1990, 10,  4 },  { 1990, 10,  5 },  { 1990, 10,  6 },
  { 1990, 10,  7 },  { 1990, 10,  8 },  { 1990, 10,  9 },  { 1990, 10, 10 },
  { 1990, 10, 11 },  { 1990, 10, 12 },  { 1990, 10, 13 },  { 1990, 10, 14 },
  { 1990, 10, 15 },  { 1990, 10, 16 },  { 1990, 10, 17 },  { 1990, 10, 18 },
  { 1990, 10, 19 },  { 1990, 10, 20 },  { 1990, 10, 21 },  { 1990, 10, 22 },
  { 1990, 10, 23 },  { 1990, 10, 24 },  { 1990, 10, 25 },  { 1990, 10, 26 },
  { 1990, 10, 27 },  { 1990, 10, 28 },  { 1990, 10, 29 },  { 1990, 10, 30 },
  { 1990, 10, 31 },  { 1990, 11,  1 },  { 1990, 11,  2 },  { 1990, 11,  3 },
  { 1990, 11,  4 },  { 1990, 11,  5 },  { 1990, 11,  6 },  { 1990, 11,  7 },
  { 1990, 11,  8 },  { 1990, 11,  9 },  { 1990, 11, 10 },  { 1990, 11, 11 },
  { 1990, 11, 12 },  { 1990, 11, 13 },  { 1990, 11, 14 },  { 1990, 11, 15 },
  { 1990, 11, 16 },  { 1990, 11, 17 },  { 1990, 11, 18 },  { 1990, 11, 19 },
  { 1990, 11, 20 },  { 1990, 11, 21 },  { 1990, 11, 22 },  { 1990, 11, 23 },
  { 1990, 11, 24 },  { 1990, 11, 25 },  { 1990, 11, 26 },  { 1990, 11, 27 },
  { 1990, 11, 28 },  { 1990, 11, 29 },  { 1990, 11, 30 },  { 1990, 12,  1 },
  { 1990, 12,  2 },  { 1990, 12,  3 },  { 1990, 12,  4 },  { 1990, 12,  5 },
  { 1990, 12,  6 },  { 1990, 12,  7 },  { 1990, 12,  8 },  { 1990, 12,  9 },
  { 1990, 12, 10 },  { 1990, 12, 11 },  { 1990, 12, 12 },  { 1990, 12, 13 },
  { 1990, 12, 14 },  { 1990, 12, 15 },  { 1990, 12, 16 },  { 1990, 12, 17 },
  { 1990, 12, 18 },  { 1990, 12, 19 },  { 1990, 12, 20 },  { 1990, 12, 21 },
  { 1990, 12, 22 },  { 1990, 12, 23 },  { 1990, 12, 24 },  { 1990, 12, 25 },
  { 1990, 12, 26 },  { 1990, 12, 27 },  { 1990, 12, 28 },  { 1990, 12, 29 },
  { 1990, 12, 30 },  { 1990, 12, 31 },  { 1991,  1,  1 },  { 1991,  1,  2 },
  { 1991,  1,  3 },  { 1991,  1,  4 },  { 1991,  1,  5 },  { 1991,  1,  6 },
  { 1991,  1,  7 },  { 1991,  1,  8 },  { 1991,  1,  9 },  { 1991,  1, 10 },
  { 1991,  1, 11 },  { 1991,  1, 12 },  { 1991,  1, 13 },  { 1991,  1, 14 },
  { 1991,  1, 15 },  { 1991,  1, 16 },  { 1991,  1, 17 },  { 1991,  1, 18 },
  { 1991,  1, 19 },  { 1991,  1, 20 },  { 1991,  1, 21 },  { 1991,  1, 22 },
  { 1991,  1, 23 },  { 1991,  1, 24 },  { 1991,  1, 25 },  { 1991,  1, 26 },
  { 1991,  1, 27 },  { 1991,  1, 28 },  { 1991,  1, 29 },  { 1991,  1, 30 },
  { 1991,  1, 31 },  { 1991,  2,  1 },  { 1991,  2,  2 },  { 1991,  2,  3 },
  { 1991,  2,  4 },  { 1991,  2,  5 },  { 1991,  2,  6 },  { 1991,  2,  7 },
  { 1991,  2,  8 },  { 1991,  2,  9 },  { 1991,  2, 10 },  { 1991,  2, 11 },
  { 1991,  2, 12 },  { 1991,  2, 13 },  { 1991,  2, 14 },  { 1991,  2, 15 },
  { 1991,  2, 16 },  { 1991,  2, 17 },  { 1991,  2, 18 },  { 1991,  2, 19 },
  { 1991,  2, 20 },  { 1991,  2, 21 },  { 1991,  2, 22 },  { 1991,  2, 23 },
  { 1991,  2, 24 },  { 1991,  2, 25 },  { 1991,  2, 26 },  { 1991,  2, 27 },
  { 1991,  2, 28 },  { 1991,  3,  1 },  { 1991,  3,  2 },  { 1991,  3,  3 },
  { 1991,  3,  4 },  { 1991,  3,  5 },  { 1991,  3,  6 },  { 1991,  3,  7 },
  { 1991,  3,  8 },  { 1991,  3,  9 },  { 1991,  3, 10 },  { 1991,  3, 11 },
  { 1991,  3, 12 },  { 1991,  3, 13 },  { 1991,  3, 14 },  { 1991,  3, 15 },
  { 1991,  3, 16 },  { 1991,  3, 17 },  { 1991,  3, 18 },  { 1991,  3, 19 },
  { 1991,  3, 20 },  { 1991,  3, 21 },  { 1991,  3, 22 },  { 1991,  3, 23 },
  { 1991,  3, 24 },  { 1991,  3, 25 },  { 1991,  3, 26 },  { 1991,  3, 27 },
  { 1991,  3, 28 },  { 1991,  3, 29 },  { 1991,  3, 30 },  { 1991,  3, 31 },
  { 1991,  4,  1 },  { 1991,  4,  2 },  { 1991,  4,  3 },  { 1991,  4,  4 },
  { 1991,  4,  5 },  { 1991,  4,  6 },  { 1991,  4,  7 },  { 1991,  4,  8 },
  { 1991,  4,  9 },  { 1991,  4, 10 },  { 1991,  4, 11 },  { 1991,  4, 12 },
  { 1991,  4, 13 },  { 1991,  4, 14 },  { 1991,  4, 15 },  { 1991,  4, 16 },
  { 1991,  4, 17 },  { 1991,  4, 18 },  { 1991,  4, 19 },  { 1991,  4, 20 },
  { 1991,  4, 21 },  { 1991,  4, 22 },  { 1991,  4, 23 },  { 1991,  4, 24 },
  { 1991,  4, 25 },  { 1991,  4, 26 },  { 1991,  4, 27 },  { 1991,  4, 28 },
  { 1991,  4, 29 },  { 1991,  4, 30 },  { 1991,  5,  1 },  { 1991,  5,  2 },
  { 1991,  5,  3 },  { 1991,  5,  4 },  { 1991,  5,  5 },  { 1991,  5,  6 },
  { 1991,  5,  7 },  { 1991,  5,  8 },  { 1991,  5,  9 },  { 1991,  5, 10 },
  { 1991,  5, 11 },  { 1991,  5, 12 },  { 1991,  5, 13 },  { 1991,  5, 14 },
  { 1991,  5, 15 },  { 1991,  5, 16 },  { 1991,  5, 17 },  { 1991,  5, 18 },
  { 1991,  5, 19 },  { 1991,  5, 20 },  { 1991,  5, 21 },  { 1991,  5, 22 },
  { 1991,  5, 23 },  { 1991,  5, 24 },  { 1991,  5, 25 },  { 1991,  5, 26 },
  { 1991,  5, 27 },  { 1991,  5, 28 },  { 1991,  5, 29 },  { 1991,  5, 30 },
  { 1991,  5, 31 },  { 1991,  6,  1 },  { 1991,  6,  2 },  { 1991,  6,  3 },
  { 1991,  6,  4 },  { 1991,  6,  5 },  { 1991,  6,  6 },  { 1991,  6,  7 },
  { 1991,  6,  8 },  { 1991,  6,  9 },  { 1991,  6, 10 },  { 1991,  6, 11 },
  { 1991,  6, 12 },  { 1991,  6, 13 },  { 1991,  6, 14 },  { 1991,  6, 15 },
  { 1991,  6, 16 },  { 1991,  6, 17 },  { 1991,  6, 18 },  { 1991,  6, 19 },
  { 1991,  6, 20 },  { 1991,  6, 21 },  { 1991,  6, 22 },  { 1991,  6, 23 },
  { 1991,  6, 24 },  { 1991,  6, 25 },  { 1991,  6, 26 },  { 1991,  6, 27 },
  { 1991,  6, 28 },  { 1991,  6, 29 },  { 1991,  6, 30 },  { 1991,  7,  1 },
  { 1991,  7,  2 },  { 1991,  7,  3 },  { 1991,  7,  4 },  { 1991,  7,  5 },
  { 1991,  7,  6 },  { 1991,  7,  7 },  { 1991,  7,  8 },  { 1991,  7,  9 },
  { 1991,  7, 10 },  { 1991,  7, 11 },  { 1991,  7, 12 },  { 1991,  7, 13 },
  { 1991,  7, 14 },  { 1991,  7, 15 },  { 1991,  7, 16 },  { 1991,  7, 17 },
  { 1991,  7, 18 },  { 1991,  7, 19 },  { 1991,  7, 20 },  { 1991,  7, 21 },
  { 1991,  7, 22 },  { 1991,  7, 23 },  { 1991,  7, 24 },  { 1991,  7, 25 },
  { 1991,  7, 26 },  { 1991,  7, 27 },  { 1991,  7, 28 },  { 1991,  7, 29 },
  { 1991,  7, 30 },  { 1991,  7, 31 },  { 1991,  8,  1 },  { 1991,  8,  2 },
  { 1991,  8,  3 },  { 1991,  8,  4 },  { 1991,  8,  5 },  { 1991,  8,  6 },
  { 1991,  8,  7 },  { 1991,  8,  8 },  { 1991,  8,  9 },  { 1991,  8, 10 },
  { 1991,  8, 11 },  { 1991,  8, 12 },  { 1991,  8, 13 },  { 1991,  8, 14 },
  { 1991,  8, 15 },  { 1991,  8, 16 },  { 1991,  8, 17 },  { 1991,  8, 18 },
  { 1991,  8, 19 },  { 1991,  8, 20 },  { 1991,  8, 21 },  { 1991,  8, 22 },
  { 1991,  8, 23 },  { 1991,  8, 24 },  { 1991,  8, 25 },  { 1991,  8, 26 },
  { 1991,  8, 27 },  { 1991,  8, 28 },  { 1991,  8, 29 },  { 1991,  8, 30 },
  { 1991,  8, 31 },  { 1991,  9,  1 },  { 1991,  9,  2 },  { 1991,  9,  3 },
  { 1991,  9,  4 },  { 1991,  9,  5 },  { 1991,  9,  6 },  { 1991,  9,  7 },
  { 1991,  9,  8 },  { 1991,  9,  9 },  { 1991,  9, 10 },  { 1991,  9, 11 },
  { 1991,  9, 12 },  { 1991,  9, 13 },  { 1991,  9, 14 },  { 1991,  9, 15 },
  { 1991,  9, 16 },  { 1991,  9, 17 },  { 1991,  9, 18 },  { 1991,  9, 19 },
  { 1991,  9, 20 },  { 1991,  9, 21 },  { 1991,  9, 22 },  { 1991,  9, 23 },
  { 1991,  9, 24 },  { 1991,  9, 25 },  { 1991,  9, 26 },  { 1991,  9, 27 },
  { 1991,  9, 28 },  { 1991,  9, 29 },  { 1991,  9, 30 },  { 1991, 10,  1 },
  { 1991, 10,  2 },  { 1991, 10,  3 },  { 1991, 10,  4 },  { 1991, 10,  5 },
  { 1991, 10,  6 },  { 1991, 10,  7 },  { 1991, 10,  8 },  { 1991, 10,  9 },
  { 1991, 10, 10 },  { 1991, 10, 11 },  { 1991, 10, 12 },  { 1991, 10, 13 },
  { 1991, 10, 14 },  { 1991, 10, 15 },  { 1991, 10, 16 },  { 1991, 10, 17 },
  { 1991, 10, 18 },  { 1991, 10, 19 },  { 1991, 10, 20 },  { 1991, 10, 21 },
  { 1991, 10, 22 },  { 1991, 10, 23 },  { 1991, 10, 24 },  { 1991, 10, 25 },
  { 1991, 10, 26 },  { 1991, 10, 27 },  { 1991, 10, 28 },  { 1991, 10, 29 },
  { 1991, 10, 30 },  { 1991, 10, 31 },  { 1991, 11,  1 },  { 1991, 11,  2 },
  { 1991, 11,  3 },  { 1991, 11,  4 },  { 1991, 11,  5 },  { 1991, 11,  6 },
  { 1991, 11,  7 },  { 1991, 11,  8 },  { 1991, 11,  9 },  { 1991, 11, 10 },
  { 1991, 11, 11 },  { 1991, 11, 12 },  { 1991, 11, 13 },  { 1991, 11, 14 },
  { 1991, 11, 15 },  { 1991, 11, 16 },  { 1991, 11, 17 },  { 1991, 11, 18 },
  { 1991, 11, 19 },  { 1991, 11, 20 },  { 1991, 11, 21 },  { 1991, 11, 22 },
  { 1991, 11, 23 },  { 1991, 11, 24 },  { 1991, 11, 25 },  { 1991, 11, 26 },
  { 1991, 11, 27 },  { 1991, 11, 28 },  { 1991, 11, 29 },  { 1991, 11, 30 },
  { 1991, 12,  1 },  { 1991, 12,  2 },  { 1991, 12,  3 },  { 1991, 12,  4 },
  { 1991, 12,  5 },  { 1991, 12,  6 },  { 1991, 12,  7 },  { 1991, 12,  8 },
  { 1991, 12,  9 },  { 1991, 12, 10 },  { 1991, 12, 11 },  { 1991, 12, 12 },
  { 1991, 12, 13 },  { 1991, 12, 14 },  { 1991, 12, 15 },  { 1991, 12, 16 },
  { 1991, 12, 17 },  { 1991, 12, 18 },  { 1991, 12, 19 },  { 1991, 12, 20 },
  { 1991, 12, 21 },  { 1991, 12, 22 },  { 1991, 12, 23 },  { 1991, 12, 24 },
  { 1991, 12, 25 },  { 1991, 12, 26 },  { 1991, 12, 27 },  { 1991, 12, 28 },
  { 1991, 12, 29 },  { 1991, 12, 30 },  { 1991, 12, 31 },  { 1992,  1,  1 },
  { 1992,  1,  2 },  { 1992,  1,  3 },  { 1992,  1,  4 },  { 1992,  1,  5 },
  { 1992,  1,  6 },  { 1992,  1,  7 },  { 1992,  1,  8 },  { 1992,  1,  9 },
  { 1992,  1, 10 },  { 1992,  1, 11 },  { 1992,  1, 12 },  { 1992,  1, 13 },
  { 1992,  1, 14 },  { 1992,  1, 15 },  { 1992,  1, 16 },  { 1992,  1, 17 },
  { 1992,  1, 18 },  { 1992,  1, 19 },  { 1992,  1, 20 },  { 1992,  1, 21 },
  { 1992,  1, 22 },  { 1992,  1, 23 },  { 1992,  1, 24 },  { 1992,  1, 25 },
  { 1992,  1, 26 },  { 1992,  1, 27 },  { 1992,  1, 28 },  { 1992,  1, 29 },
  { 1992,  1, 30 },  { 1992,  1, 31 },  { 1992,  2,  1 },  { 1992,  2,  2 },
  { 1992,  2,  3 },  { 1992,  2,  4 },  { 1992,  2,  5 },  { 1992,  2,  6 },
  { 1992,  2,  7 },  { 1992,  2,  8 },  { 1992,  2,  9 },  { 1992,  2, 10 },
  { 1992,  2, 11 },  { 1992,  2, 12 },  { 1992,  2, 13 },  { 1992,  2, 14 },
  { 1992,  2, 15 },  { 1992,  2, 16 },  { 1992,  2, 17 },  { 1992,  2, 18 },
  { 1992,  2, 19 },  { 1992,  2, 20 },  { 1992,  2, 21 },  { 1992,  2, 22 },
  { 1992,  2, 23 },  { 1992,  2, 24 },  { 1992,  2, 25 },  { 1992,  2, 26 },
  { 1992,  2, 27 },  { 1992,  2, 28 },  { 1992,  2, 29 },  { 1992,  3,  1 },
  { 1992,  3,  2 },  { 1992,  3,  3 },  { 1992,  3,  4 },  { 1992,  3,  5 },
  { 1992,  3,  6 },  { 1992,  3,  7 },  { 1992,  3,  8 },  { 1992,  3,  9 },
  { 1992,  3, 10 },  { 1992,  3, 11 },  { 1992,  3, 12 },  { 1992,  3, 13 },
  { 1992,  3, 14 },  { 1992,  3, 15 },  { 1992,  3, 16 },  { 1992,  3, 17 },
  { 1992,  3, 18 },  { 1992,  3, 19 },  { 1992,  3, 20 },  { 1992,  3, 21 },
  { 1992,  3, 22 },  { 1992,  3, 23 },  { 1992,  3, 24 },  { 1992,  3, 25 },
  { 1992,  3, 26 },  { 1992,  3, 27 },  { 1992,  3, 28 },  { 1992,  3, 29 },
  { 1992,  3, 30 },  { 1992,  3, 31 },  { 1992,  4,  1 },  { 1992,  4,  2 },
  { 1992,  4,  3 },  { 1992,  4,  4 },  { 1992,  4,  5 },  { 1992,  4,  6 },
  { 1992,  4,  7 },  { 1992,  4,  8 },  { 1992,  4,  9 },  { 1992,  4, 10 },
  { 1992,  4, 11 },  { 1992,  4, 12 },  { 1992,  4, 13 },  { 1992,  4, 14 },
  { 1992,  4, 15 },  { 1992,  4, 16 },  { 1992,  4, 17 },  { 1992,  4, 18 },
  { 1992,  4, 19 },  { 1992,  4, 20 },  { 1992,  4, 21 },  { 1992,  4, 22 },
  { 1992,  4, 23 },  { 1992,  4, 24 },  { 1992,  4, 25 },  { 1992,  4, 26 },
  { 1992,  4, 27 },  { 1992,  4, 28 },  { 1992,  4, 29 },  { 1992,  4, 30 },
  { 1992,  5,  1 },  { 1992,  5,  2 },  { 1992,  5,  3 },  { 1992,  5,  4 },
  { 1992,  5,  5 },  { 1992,  5,  6 },  { 1992,  5,  7 },  { 1992,  5,  8 },
  { 1992,  5,  9 },  { 1992,  5, 10 },  { 1992,  5, 11 },  { 1992,  5, 12 },
  { 1992,  5, 13 },  { 1992,  5, 14 },  { 1992,  5, 15 },  { 1992,  5, 16 },
  { 1992,  5, 17 },  { 1992,  5, 18 },  { 1992,  5, 19 },  { 1992,  5, 20 },
  { 1992,  5, 21 },  { 1992,  5, 22 },  { 1992,  5, 23 },  { 1992,  5, 24 },
  { 1992,  5, 25 },  { 1992,  5, 26 },  { 1992,  5, 27 },  { 1992,  5, 28 },
  { 1992,  5, 29 },  { 1992,  5, 30 },  { 1992,  5, 31 },  { 1992,  6,  1 },
  { 1992,  6,  2 },  { 1992,  6,  3 },  { 1992,  6,  4 },  { 1992,  6,  5 },
  { 1992,  6,  6 },  { 1992,  6,  7 },  { 1992,  6,  8 },  { 1992,  6,  9 },
  { 1992,  6, 10 },  { 1992,  6, 11 },  { 1992,  6, 12 },  { 1992,  6, 13 },
  { 1992,  6, 14 },  { 1992,  6, 15 },  { 1992,  6, 16 },  { 1992,  6, 17 },
  { 1992,  6, 18 },  { 1992,  6, 19 },  { 1992,  6, 20 },  { 1992,  6, 21 },
  { 1992,  6, 22 },  { 1992,  6, 23 },  { 1992,  6, 24 },  { 1992,  6, 25 },
  { 1992,  6, 26 },  { 1992,  6, 27 },  { 1992,  6, 28 },  { 1992,  6, 29 },
  { 1992,  6, 30 },  { 1992,  7,  1 },  { 1992,  7,  2 },  { 1992,  7,  3 },
  { 1992,  7,  4 },  { 1992,  7,  5 },  { 1992,  7,  6 },  { 1992,  7,  7 },
  { 1992,  7,  8 },  { 1992,  7,  9 },  { 1992,  7, 10 },  { 1992,  7, 11 },
  { 1992,  7, 12 },  { 1992,  7, 13 },  { 1992,  7, 14 },  { 1992,  7, 15 },
  { 1992,  7, 16 },  { 1992,  7, 17 },  { 1992,  7, 18 },  { 1992,  7, 19 },
  { 1992,  7, 20 },  { 1992,  7, 21 },  { 1992,  7, 22 },  { 1992,  7, 23 },
  { 1992,  7, 24 },  { 1992,  7, 25 },  { 1992,  7, 26 },  { 1992,  7, 27 },
  { 1992,  7, 28 },  { 1992,  7, 29 },  { 1992,  7, 30 },  { 1992,  7, 31 },
  { 1992,  8,  1 },  { 1992,  8,  2 },  { 1992,  8,  3 },  { 1992,  8,  4 },
  { 1992,  8,  5 },  { 1992,  8,  6 },  { 1992,  8,  7 },  { 1992,  8,  8 },
  { 1992,  8,  9 },  { 1992,  8, 10 },  { 1992,  8, 11 },  { 1992,  8, 12 },
  { 1992,  8, 13 },  { 1992,  8, 14 },  { 1992,  8, 15 },  { 1992,  8, 16 },
  { 1992,  8, 17 },  { 1992,  8, 18 },  { 1992,  8, 19 },  { 1992,  8, 20 },
  { 1992,  8, 21 },  { 1992,  8, 22 },  { 1992,  8, 23 },  { 1992,  8, 24 },
  { 1992,  8, 25 },  { 1992,  8, 26 },  { 1992,  8, 27 },  { 1992,  8, 28 },
  { 1992,  8, 29 },  { 1992,  8, 30 },  { 1992,  8, 31 },  { 1992,  9,  1 },
  { 1992,  9,  2 },  { 1992,  9,  3 },  { 1992,  9,  4 },  { 1992,  9,  5 },
  { 1992,  9,  6 },  { 1992,  9,  7 },  { 1992,  9,  8 },  { 1992,  9,  9 },
  { 1992,  9, 10 },  { 1992,  9, 11 },  { 1992,  9, 12 },  { 1992,  9, 13 },
  { 1992,  9, 14 },  { 1992,  9, 15 },  { 1992,  9, 16 },  { 1992,  9, 17 },
  { 1992,  9, 18 },  { 1992,  9, 19 },  { 1992,  9, 20 },  { 1992,  9, 21 },
  { 1992,  9, 22 },  { 1992,  9, 23 },  { 1992,  9, 24 },  { 1992,  9, 25 },
  { 1992,  9, 26 },  { 1992,  9, 27 },  { 1992,  9, 28 },  { 1992,  9, 29 },
  { 1992,  9, 30 },  { 1992, 10,  1 },  { 1992, 10,  2 },  { 1992, 10,  3 },
  { 1992, 10,  4 },  { 1992, 10,  5 },  { 1992, 10,  6 },  { 1992, 10,  7 },
  { 1992, 10,  8 },  { 1992, 10,  9 },  { 1992, 10, 10 },  { 1992, 10, 11 },
  { 1992, 10, 12 },  { 1992, 10, 13 },  { 1992, 10, 14 },  { 1992, 10, 15 },
  { 1992, 10, 16 },  { 1992, 10, 17 },  { 1992, 10, 18 },  { 1992, 10, 19 },
  { 1992, 10, 20 },  { 1992, 10, 21 },  { 1992, 10, 22 },  { 1992, 10, 23 },
  { 1992, 10, 24 },  { 1992, 10, 25 },  { 1992, 10, 26 },  { 1992, 10, 27 },
  { 1992, 10, 28 },  { 1992, 10, 29 },  { 1992, 10, 30 },  { 1992, 10, 31 },
  { 1992, 11,  1 },  { 1992, 11,  2 },  { 1992, 11,  3 },  { 1992, 11,  4 },
  { 1992, 11,  5 },  { 1992, 11,  6 },  { 1992, 11,  7 },  { 1992, 11,  8 },
  { 1992, 11,  9 },  { 1992, 11, 10 },  { 1992, 11, 11 },  { 1992, 11, 12 },
  { 1992, 11, 13 },  { 1992, 11, 14 },  { 1992, 11, 15 },  { 1992, 11, 16 },
  { 1992, 11, 17 },  { 1992, 11, 18 },  { 1992, 11, 19 },  { 1992, 11, 20 },
  { 1992, 11, 21 },  { 1992, 11, 22 },  { 1992, 11, 23 },  { 1992, 11, 24 },
  { 1992, 11, 25 },  { 1992, 11, 26 },  { 1992, 11, 27 },  { 1992, 11, 28 },
  { 1992, 11, 29 },  { 1992, 11, 30 },  { 1992, 12,  1 },  { 1992, 12,  2 },
  { 1992, 12,  3 },  { 1992, 12,  4 },  { 1992, 12,  5 },  { 1992, 12,  6 },
  { 1992, 12,  7 },  { 1992, 12,  8 },  { 1992, 12,  9 },  { 1992, 12, 10 },
  { 1992, 12, 11 },  { 1992, 12, 12 },  { 1992, 12, 13 },  { 1992, 12, 14 },
  { 1992, 12, 15 },  { 1992, 12, 16 },  { 1992, 12, 17 },  { 1992, 12, 18 },
  { 1992, 12, 19 },  { 1992, 12, 20 },  { 1992, 12, 21 },  { 1992, 12, 22 },
  { 1992, 12, 23 },  { 1992, 12, 24 },  { 1992, 12, 25 },  { 1992, 12, 26 },
  { 1992, 12, 27 },  { 1992, 12, 28 },  { 1992, 12, 29 },  { 1992, 12, 30 },
  { 1992, 12, 31 },  { 1993,  1,  1 },  { 1993,  1,  2 },  { 1993,  1,  3 },
  { 1993,  1,  4 },  { 1993,  1,  5 },  { 1993,  1,  6 },  { 1993,  1,  7 },
  { 1993,  1,  8 },  { 1993,  1,  9 },  { 1993,  1, 10 },  { 1993,  1, 11 },
  { 1993,  1, 12 },  { 1993,  1, 13 },  { 1993,  1, 14 },  { 1993,  1, 15 },
  { 1993,  1, 16 },  { 1993,  1, 17 },  { 1993,  1, 18 },  { 1993,  1, 19 },
  { 1993,  1, 20 },  { 1993,  1, 21 },  { 1993,  1, 22 },  { 1993,  1, 23 },
  { 1993,  1, 24 },  { 1993,  1, 25 },  { 1993,  1, 26 },  { 1993,  1, 27 },
  { 1993,  1, 28 },  { 1993,  1, 29 },  { 1993,  1, 30 },  { 1993,  1, 31 },
  { 1993,  2,  1 },  { 1993,  2,  2 },  { 1993,  2,  3 },  { 1993,  2,  4 },
  { 1993,  2,  5 },  { 1993,  2,  6 },  { 1993,  2,  7 },  { 1993,  2,  8 },
  { 1993,  2,  9 },  { 1993,  2, 10 },  { 1993,  2, 11 },  { 1993,  2, 12 },
  { 1993,  2, 13 },  { 1993,  2, 14 },  { 1993,  2, 15 },  { 1993,  2, 16 },
  { 1993,  2, 17 },  { 1993,  2, 18 },  { 1993,  2, 19 },  { 1993,  2, 20 },
  { 1993,  2, 21 },  { 1993,  2, 22 },  { 1993,  2, 23 },  { 1993,  2, 24 },
  { 1993,  2, 25 },  { 1993,  2, 26 },  { 1993,  2, 27 },  { 1993,  2, 28 },
  { 1993,  3,  1 },  { 1993,  3,  2 },  { 1993,  3,  3 },  { 1993,  3,  4 },
  { 1993,  3,  5 },  { 1993,  3,  6 },  { 1993,  3,  7 },  { 1993,  3,  8 },
  { 1993,  3,  9 },  { 1993,  3, 10 },  { 1993,  3, 11 },  { 1993,  3, 12 },
  { 1993,  3, 13 },  { 1993,  3, 14 },  { 1993,  3, 15 },  { 1993,  3, 16 },
  { 1993,  3, 17 },  { 1993,  3, 18 },  { 1993,  3, 19 },  { 1993,  3, 20 },
  { 1993,  3, 21 },  { 1993,  3, 22 },  { 1993,  3, 23 },  { 1993,  3, 24 },
  { 1993,  3, 25 },  { 1993,  3, 26 },  { 1993,  3, 27 },  { 1993,  3, 28 },
  { 1993,  3, 29 },  { 1993,  3, 30 },  { 1993,  3, 31 },  { 1993,  4,  1 },
  { 1993,  4,  2 },  { 1993,  4,  3 },  { 1993,  4,  4 },  { 1993,  4,  5 },
  { 1993,  4,  6 },  { 1993,  4,  7 },  { 1993,  4,  8 },  { 1993,  4,  9 },
  { 1993,  4, 10 },  { 1993,  4, 11 },  { 1993,  4, 12 },  { 1993,  4, 13 },
  { 1993,  4, 14 },  { 1993,  4, 15 },  { 1993,  4, 16 },  { 1993,  4, 17 },
  { 1993,  4, 18 },  { 1993,  4, 19 },  { 1993,  4, 20 },  { 1993,  4, 21 },
  { 1993,  4, 22 },  { 1993,  4, 23 },  { 1993,  4, 24 },  { 1993,  4, 25 },
  { 1993,  4, 26 },  { 1993,  4, 27 },  { 1993,  4, 28 },  { 1993,  4, 29 },
  { 1993,  4, 30 },  { 1993,  5,  1 },  { 1993,  5,  2 },  { 1993,  5,  3 },
  { 1993,  5,  4 },  { 1993,  5,  5 },  { 1993,  5,  6 },  { 1993,  5,  7 },
  { 1993,  5,  8 },  { 1993,  5,  9 },  { 1993,  5, 10 },  { 1993,  5, 11 },
  { 1993,  5, 12 },  { 1993,  5, 13 },  { 1993,  5, 14 },  { 1993,  5, 15 },
  { 1993,  5, 16 },  { 1993,  5, 17 },  { 1993,  5, 18 },  { 1993,  5, 19 },
  { 1993,  5, 20 },  { 1993,  5, 21 },  { 1993,  5, 22 },  { 1993,  5, 23 },
  { 1993,  5, 24 },  { 1993,  5, 25 },  { 1993,  5, 26 },  { 1993,  5, 27 },
  { 1993,  5, 28 },  { 1993,  5, 29 },  { 1993,  5, 30 },  { 1993,  5, 31 },
  { 1993,  6,  1 },  { 1993,  6,  2 },  { 1993,  6,  3 },  { 1993,  6,  4 },
  { 1993,  6,  5 },  { 1993,  6,  6 },  { 1993,  6,  7 },  { 1993,  6,  8 },
  { 1993,  6,  9 },  { 1993,  6, 10 },  { 1993,  6, 11 },  { 1993,  6, 12 },
  { 1993,  6, 13 },  { 1993,  6, 14 },  { 1993,  6, 15 },  { 1993,  6, 16 },
  { 1993,  6, 17 },  { 1993,  6, 18 },  { 1993,  6, 19 },  { 1993,  6, 20 },
  { 1993,  6, 21 },  { 1993,  6, 22 },  { 1993,  6, 23 },  { 1993,  6, 24 },
  { 1993,  6, 25 },  { 1993,  6, 26 },  { 1993,  6, 27 },  { 1993,  6, 28 },
  { 1993,  6, 29 },  { 1993,  6, 30 },  { 1993,  7,  1 },  { 1993,  7,  2 },
  { 1993,  7,  3 },  { 1993,  7,  4 },  { 1993,  7,  5 },  { 1993,  7,  6 },
  { 1993,  7,  7 },  { 1993,  7,  8 },  { 1993,  7,  9 },  { 1993,  7, 10 },
  { 1993,  7, 11 },  { 1993,  7, 12 },  { 1993,  7, 13 },  { 1993,  7, 14 },
  { 1993,  7, 15 },  { 1993,  7, 16 },  { 1993,  7, 17 },  { 1993,  7, 18 },
  { 1993,  7, 19 },  { 1993,  7, 20 },  { 1993,  7, 21 },  { 1993,  7, 22 },
  { 1993,  7, 23 },  { 1993,  7, 24 },  { 1993,  7, 25 },  { 1993,  7, 26 },
  { 1993,  7, 27 },  { 1993,  7, 28 },  { 1993,  7, 29 },  { 1993,  7, 30 },
  { 1993,  7, 31 },  { 1993,  8,  1 },  { 1993,  8,  2 },  { 1993,  8,  3 },
  { 1993,  8,  4 },  { 1993,  8,  5 },  { 1993,  8,  6 },  { 1993,  8,  7 },
  { 1993,  8,  8 },  { 1993,  8,  9 },  { 1993,  8, 10 },  { 1993,  8, 11 },
  { 1993,  8, 12 },  { 1993,  8, 13 },  { 1993,  8, 14 },  { 1993,  8, 15 },
  { 1993,  8, 16 },  { 1993,  8, 17 },  { 1993,  8, 18 },  { 1993,  8, 19 },
  { 1993,  8, 20 },  { 1993,  8, 21 },  { 1993,  8, 22 },  { 1993,  8, 23 },
  { 1993,  8, 24 },  { 1993,  8, 25 },  { 1993,  8, 26 },  { 1993,  8, 27 },
  { 1993,  8, 28 },  { 1993,  8, 29 },  { 1993,  8, 30 },  { 1993,  8, 31 },
  { 1993,  9,  1 },  { 1993,  9,  2 },  { 1993,  9,  3 },  { 1993,  9,  4 },
  { 1993,  9,  5 },  { 1993,  9,  6 },  { 1993,  9,  7 },  { 1993,  9,  8 },
  { 1993,  9,  9 },  { 1993,  9, 10 },  { 1993,  9, 11 },  { 1993,  9, 12 },
  { 1993,  9, 13 },  { 1993,  9, 14 },  { 1993,  9, 15 },  { 1993,  9, 16 },
  { 1993,  9, 17 },  { 1993,  9, 18 },  { 1993,  9, 19 },  { 1993,  9, 20 },
  { 1993,  9, 21 },  { 1993,  9, 22 },  { 1993,  9, 23 },  { 1993,  9, 24 },
  { 1993,  9, 25 },  { 1993,  9, 26 },  { 1993,  9, 27 },  { 1993,  9, 28 },
  { 1993,  9, 29 },  { 1993,  9, 30 },  { 1993, 10,  1 },  { 1993, 10,  2 },
  { 1993, 10,  3 },  { 1993, 10,  4 },  { 1993, 10,  5 },  { 1993, 10,  6 },
  { 1993, 10,  7 },  { 1993, 10,  8 },  { 1993, 10,  9 },  { 1993, 10, 10 },
  { 1993, 10, 11 },  { 1993, 10, 12 },  { 1993, 10, 13 },  { 1993, 10, 14 },
  { 1993, 10, 15 },  { 1993, 10, 16 },  { 1993, 10, 17 },  { 1993, 10, 18 },
  { 1993, 10, 19 },  { 1993, 10, 20 },  { 1993, 10, 21 },  { 1993, 10, 22 },
  { 1993, 10, 23 },  { 1993, 10, 24 },  { 1993, 10, 25 },  { 1993, 10, 26 },
  { 1993, 10, 27 },  { 1993, 10, 28 },  { 1993, 10, 29 },  { 1993, 10, 30 },
  { 1993, 10, 31 },  { 1993, 11,  1 },  { 1993, 11,  2 },  { 1993, 11,  3 },
  { 1993, 11,  4 },  { 1993, 11,  5 },  { 1993, 11,  6 },  { 1993, 11,  7 },
  { 1993, 11,  8 },  { 1993, 11,  9 },  { 1993, 11, 10 },  { 1993, 11, 11 },
  { 1993, 11, 12 },  { 1993, 11, 13 },  { 1993, 11, 14 },  { 1993, 11, 15 },
  { 1993, 11, 16 },  { 1993, 11, 17 },  { 1993, 11, 18 },  { 1993, 11, 19 },
  { 1993, 11, 20 },  { 1993, 11, 21 },  { 1993, 11, 22 },  { 1993, 11, 23 },
  { 1993, 11, 24 },  { 1993, 11, 25 },  { 1993, 11, 26 },  { 1993, 11, 27 },
  { 1993, 11, 28 },  { 1993, 11, 29 },  { 1993, 11, 30 },  { 1993, 12,  1 },
  { 1993, 12,  2 },  { 1993, 12,  3 },  { 1993, 12,  4 },  { 1993, 12,  5 },
  { 1993, 12,  6 },  { 1993, 12,  7 },  { 1993, 12,  8 },  { 1993, 12,  9 },
  { 1993, 12, 10 },  { 1993, 12, 11 },  { 1993, 12, 12 },  { 1993, 12, 13 },
  { 1993, 12, 14 },  { 1993, 12, 15 },  { 1993, 12, 16 },  { 1993, 12, 17 },
  { 1993, 12, 18 },  { 1993, 12, 19 },  { 1993, 12, 20 },  { 1993, 12, 21 },
  { 1993, 12, 22 },  { 1993, 12, 23 },  { 1993, 12, 24 },  { 1993, 12, 25 },
  { 1993, 12, 26 },  { 1993, 12, 27 },  { 1993, 12, 28 },  { 1993, 12, 29 },
  { 1993, 12, 30 },  { 1993, 12, 31 },  { 1994,  1,  1 },  { 1994,  1,  2 },
  { 1994,  1,  3 },  { 1994,  1,  4 },  { 1994,  1,  5 },  { 1994,  1,  6 },
  { 1994,  1,  7 },  { 1994,  1,  8 },  { 1994,  1,  9 },  { 1994,  1, 10 },
  { 1994,  1, 11 },  { 1994,  1, 12 },  { 1994,  1, 13 },  { 1994,  1, 14 },
  { 1994,  1, 15 },  { 1994,  1, 16 },  { 1994,  1, 17 },  { 1994,  1, 18 },
  { 1994,  1, 19 },  { 1994,  1, 20 },  { 1994,  1, 21 },  { 1994,  1, 22 },
  { 1994,  1, 23 },  { 1994,  1, 24 },  { 1994,  1, 25 },  { 1994,  1, 26 },
  { 1994,  1, 27 },  { 1994,  1, 28 },  { 1994,  1, 29 },  { 1994,  1, 30 },
  { 1994,  1, 31 },  { 1994,  2,  1 },  { 1994,  2,  2 },  { 1994,  2,  3 },
  { 1994,  2,  4 },  { 1994,  2,  5 },  { 1994,  2,  6 },  { 1994,  2,  7 },
  { 1994,  2,  8 },  { 1994,  2,  9 },  { 1994,  2, 10 },  { 1994,  2, 11 },
  { 1994,  2, 12 },  { 1994,  2, 13 },  { 1994,  2, 14 },  { 1994,  2, 15 },
  { 1994,  2, 16 },  { 1994,  2, 17 },  { 1994,  2, 18 },  { 1994,  2, 19 },
  { 1994,  2, 20 },  { 1994,  2, 21 },  { 1994,  2, 22 },  { 1994,  2, 23 },
  { 1994,  2, 24 },  { 1994,  2, 25 },  { 1994,  2, 26 },  { 1994,  2, 27 },
  { 1994,  2, 28 },  { 1994,  3,  1 },  { 1994,  3,  2 },  { 1994,  3,  3 },
  { 1994,  3,  4 },  { 1994,  3,  5 },  { 1994,  3,  6 },  { 1994,  3,  7 },
  { 1994,  3,  8 },  { 1994,  3,  9 },  { 1994,  3, 10 },  { 1994,  3, 11 },
  { 1994,  3, 12 },  { 1994,  3, 13 },  { 1994,  3, 14 },  { 1994,  3, 15 },
  { 1994,  3, 16 },  { 1994,  3, 17 },  { 1994,  3, 18 },  { 1994,  3, 19 },
  { 1994,  3, 20 },  { 1994,  3, 21 },  { 1994,  3, 22 },  { 1994,  3, 23 },
  { 1994,  3, 24 },  { 1994,  3, 25 },  { 1994,  3, 26 },  { 1994,  3, 27 },
  { 1994,  3, 28 },  { 1994,  3, 29 },  { 1994,  3, 30 },  { 1994,  3, 31 },
  { 1994,  4,  1 },  { 1994,  4,  2 },  { 1994,  4,  3 },  { 1994,  4,  4 },
  { 1994,  4,  5 },  { 1994,  4,  6 },  { 1994,  4,  7 },  { 1994,  4,  8 },
  { 1994,  4,  9 },  { 1994,  4, 10 },  { 1994,  4, 11 },  { 1994,  4, 12 },
  { 1994,  4, 13 },  { 1994,  4, 14 },  { 1994,  4, 15 },  { 1994,  4, 16 },
  { 1994,  4, 17 },  { 1994,  4, 18 },  { 1994,  4, 19 },  { 1994,  4, 20 },
  { 1994,  4, 21 },  { 1994,  4, 22 },  { 1994,  4, 23 },  { 1994,  4, 24 },
  { 1994,  4, 25 },  { 1994,  4, 26 },  { 1994,  4, 27 },  { 1994,  4, 28 },
  { 1994,  4, 29 },  { 1994,  4, 30 },  { 1994,  5,  1 },  { 1994,  5,  2 },
  { 1994,  5,  3 },  { 1994,  5,  4 },  { 1994,  5,  5 },  { 1994,  5,  6 },
  { 1994,  5,  7 },  { 1994,  5,  8 },  { 1994,  5,  9 },  { 1994,  5, 10 },
  { 1994,  5, 11 },  { 1994,  5, 12 },  { 1994,  5, 13 },  { 1994,  5, 14 },
  { 1994,  5, 15 },  { 1994,  5, 16 },  { 1994,  5, 17 },  { 1994,  5, 18 },
  { 1994,  5, 19 },  { 1994,  5, 20 },  { 1994,  5, 21 },  { 1994,  5, 22 },
  { 1994,  5, 23 },  { 1994,  5, 24 },  { 1994,  5, 25 },  { 1994,  5, 26 },
  { 1994,  5, 27 },  { 1994,  5, 28 },  { 1994,  5, 29 },  { 1994,  5, 30 },
  { 1994,  5, 31 },  { 1994,  6,  1 },  { 1994,  6,  2 },  { 1994,  6,  3 },
  { 1994,  6,  4 },  { 1994,  6,  5 },  { 1994,  6,  6 },  { 1994,  6,  7 },
  { 1994,  6,  8 },  { 1994,  6,  9 },  { 1994,  6, 10 },  { 1994,  6, 11 },
  { 1994,  6, 12 },  { 1994,  6, 13 },  { 1994,  6, 14 },  { 1994,  6, 15 },
  { 1994,  6, 16 },  { 1994,  6, 17 },  { 1994,  6, 18 },  { 1994,  6, 19 },
  { 1994,  6, 20 },  { 1994,  6, 21 },  { 1994,  6, 22 },  { 1994,  6, 23 },
  { 1994,  6, 24 },  { 1994,  6, 25 },  { 1994,  6, 26 },  { 1994,  6, 27 },
  { 1994,  6, 28 },  { 1994,  6, 29 },  { 1994,  6, 30 },  { 1994,  7,  1 },
  { 1994,  7,  2 },  { 1994,  7,  3 },  { 1994,  7,  4 },  { 1994,  7,  5 },
  { 1994,  7,  6 },  { 1994,  7,  7 },  { 1994,  7,  8 },  { 1994,  7,  9 },
  { 1994,  7, 10 },  { 1994,  7, 11 },  { 1994,  7, 12 },  { 1994,  7, 13 },
  { 1994,  7, 14 },  { 1994,  7, 15 },  { 1994,  7, 16 },  { 1994,  7, 17 },
  { 1994,  7, 18 },  { 1994,  7, 19 },  { 1994,  7, 20 },  { 1994,  7, 21 },
  { 1994,  7, 22 },  { 1994,  7, 23 },  { 1994,  7, 24 },  { 1994,  7, 25 },
  { 1994,  7, 26 },  { 1994,  7, 27 },  { 1994,  7, 28 },  { 1994,  7, 29 },
  { 1994,  7, 30 },  { 1994,  7, 31 },  { 1994,  8,  1 },  { 1994,  8,  2 },
  { 1994,  8,  3 },  { 1994,  8,  4 },  { 1994,  8,  5 },  { 1994,  8,  6 },
  { 1994,  8,  7 },  { 1994,  8,  8 },  { 1994,  8,  9 },  { 1994,  8, 10 },
  { 1994,  8, 11 },  { 1994,  8, 12 },  { 1994,  8, 13 },  { 1994,  8, 14 },
  { 1994,  8, 15 },  { 1994,  8, 16 },  { 1994,  8, 17 },  { 1994,  8, 18 },
  { 1994,  8, 19 },  { 1994,  8, 20 },  { 1994,  8, 21 },  { 1994,  8, 22 },
  { 1994,  8, 23 },  { 1994,  8, 24 },  { 1994,  8, 25 },  { 1994,  8, 26 },
  { 1994,  8, 27 },  { 1994,  8, 28 },  { 1994,  8, 29 },  { 1994,  8, 30 },
  { 1994,  8, 31 },  { 1994,  9,  1 },  { 1994,  9,  2 },  { 1994,  9,  3 },
  { 1994,  9,  4 },  { 1994,  9,  5 },  { 1994,  9,  6 },  { 1994,  9,  7 },
  { 1994,  9,  8 },  { 1994,  9,  9 },  { 1994,  9, 10 },  { 1994,  9, 11 },
  { 1994,  9, 12 },  { 1994,  9, 13 },  { 1994,  9, 14 },  { 1994,  9, 15 },
  { 1994,  9, 16 },  { 1994,  9, 17 },  { 1994,  9, 18 },  { 1994,  9, 19 },
  { 1994,  9, 20 },  { 1994,  9, 21 },  { 1994,  9, 22 },  { 1994,  9, 23 },
  { 1994,  9, 24 },  { 1994,  9, 25 },  { 1994,  9, 26 },  { 1994,  9, 27 },
  { 1994,  9, 28 },  { 1994,  9, 29 },  { 1994,  9, 30 },  { 1994, 10,  1 },
  { 1994, 10,  2 },  { 1994, 10,  3 },  { 1994, 10,  4 },  { 1994, 10,  5 },
  { 1994, 10,  6 },  { 1994, 10,  7 },  { 1994, 10,  8 },  { 1994, 10,  9 },
  { 1994, 10, 10 },  { 1994, 10, 11 },  { 1994, 10, 12 },  { 1994, 10, 13 },
  { 1994, 10, 14 },  { 1994, 10, 15 },  { 1994, 10, 16 },  { 1994, 10, 17 },
  { 1994, 10, 18 },  { 1994, 10, 19 },  { 1994, 10, 20 },  { 1994, 10, 21 },
  { 1994, 10, 22 },  { 1994, 10, 23 },  { 1994, 10, 24 },  { 1994, 10, 25 },
  { 1994, 10, 26 },  { 1994, 10, 27 },  { 1994, 10, 28 },  { 1994, 10, 29 },
  { 1994, 10, 30 },  { 1994, 10, 31 },  { 1994, 11,  1 },  { 1994, 11,  2 },
  { 1994, 11,  3 },  { 1994, 11,  4 },  { 1994, 11,  5 },  { 1994, 11,  6 },
  { 1994, 11,  7 },  { 1994, 11,  8 },  { 1994, 11,  9 },  { 1994, 11, 10 },
  { 1994, 11, 11 },  { 1994, 11, 12 },  { 1994, 11, 13 },  { 1994, 11, 14 },
  { 1994, 11, 15 },  { 1994, 11, 16 },  { 1994, 11, 17 },  { 1994, 11, 18 },
  { 1994, 11, 19 },  { 1994, 11, 20 },  { 1994, 11, 21 },  { 1994, 11, 22 },
  { 1994, 11, 23 },  { 1994, 11, 24 },  { 1994, 11, 25 },  { 1994, 11, 26 },
  { 1994, 11, 27 },  { 1994, 11, 28 },  { 1994, 11, 29 },  { 1994, 11, 30 },
  { 1994, 12,  1 },  { 1994, 12,  2 },  { 1994, 12,  3 },  { 1994, 12,  4 },
  { 1994, 12,  5 },  { 1994, 12,  6 },  { 1994, 12,  7 },  { 1994, 12,  8 },
  { 1994, 12,  9 },  { 1994, 12, 10 },  { 1994, 12, 11 },  { 1994, 12, 12 },
  { 1994, 12, 13 },  { 1994, 12, 14 },  { 1994, 12, 15 },  { 1994, 12, 16 },
  { 1994, 12, 17 },  { 1994, 12, 18 },  { 1994, 12, 19 },  { 1994, 12, 20 },
  { 1994, 12, 21 },  { 1994, 12, 22 },  { 1994, 12, 23 },  { 1994, 12, 24 },
  { 1994, 12, 25 },  { 1994, 12, 26 },  { 1994, 12, 27 },  { 1994, 12, 28 },
  { 1994, 12, 29 },  { 1994, 12, 30 },  { 1994, 12, 31 },  { 1995,  1,  1 },
  { 1995,  1,  2 },  { 1995,  1,  3 },  { 1995,  1,  4 },  { 1995,  1,  5 },
  { 1995,  1,  6 },  { 1995,  1,  7 },  { 1995,  1,  8 },  { 1995,  1,  9 },
  { 1995,  1, 10 },  { 1995,  1, 11 },  { 1995,  1, 12 },  { 1995,  1, 13 },
  { 1995,  1, 14 },  { 1995,  1, 15 },  { 1995,  1, 16 },  { 1995,  1, 17 },
  { 1995,  1, 18 },  { 1995,  1, 19 },  { 1995,  1, 20 },  { 1995,  1, 21 },
  { 1995,  1, 22 },  { 1995,  1, 23 },  { 1995,  1, 24 },  { 1995,  1, 25 },
  { 1995,  1, 26 },  { 1995,  1, 27 },  { 1995,  1, 28 },  { 1995,  1, 29 },
  { 1995,  1, 30 },  { 1995,  1, 31 },  { 1995,  2,  1 },  { 1995,  2,  2 },
  { 1995,  2,  3 },  { 1995,  2,  4 },  { 1995,  2,  5 },  { 1995,  2,  6 },
  { 1995,  2,  7 },  { 1995,  2,  8 },  { 1995,  2,  9 },  { 1995,  2, 10 },
  { 1995,  2, 11 },  { 1995,  2, 12 },  { 1995,  2, 13 },  { 1995,  2, 14 },
  { 1995,  2, 15 },  { 1995,  2, 16 },  { 1995,  2, 17 },  { 1995,  2, 18 },
  { 1995,  2, 19 },  { 1995,  2, 20 },  { 1995,  2, 21 },  { 1995,  2, 22 },
  { 1995,  2, 23 },  { 1995,  2, 24 },  { 1995,  2, 25 },  { 1995,  2, 26 },
  { 1995,  2, 27 },  { 1995,  2, 28 },  { 1995,  3,  1 },  { 1995,  3,  2 },
  { 1995,  3,  3 },  { 1995,  3,  4 },  { 1995,  3,  5 },  { 1995,  3,  6 },
  { 1995,  3,  7 },  { 1995,  3,  8 },  { 1995,  3,  9 },  { 1995,  3, 10 },
  { 1995,  3, 11 },  { 1995,  3, 12 },  { 1995,  3, 13 },  { 1995,  3, 14 },
  { 1995,  3, 15 },  { 1995,  3, 16 },  { 1995,  3, 17 },  { 1995,  3, 18 },
  { 1995,  3, 19 },  { 1995,  3, 20 },  { 1995,  3, 21 },  { 1995,  3, 22 },
  { 1995,  3, 23 },  { 1995,  3, 24 },  { 1995,  3, 25 },  { 1995,  3, 26 },
  { 1995,  3, 27 },  { 1995,  3, 28 },  { 1995,  3, 29 },  { 1995,  3, 30 },
  { 1995,  3, 31 },  { 1995,  4,  1 },  { 1995,  4,  2 },  { 1995,  4,  3 },
  { 1995,  4,  4 },  { 1995,  4,  5 },  { 1995,  4,  6 },  { 1995,  4,  7 },
  { 1995,  4,  8 },  { 1995,  4,  9 },  { 1995,  4, 10 },  { 1995,  4, 11 },
  { 1995,  4, 12 },  { 1995,  4, 13 },  { 1995,  4, 14 },  { 1995,  4, 15 },
  { 1995,  4, 16 },  { 1995,  4, 17 },  { 1995,  4, 18 },  { 1995,  4, 19 },
  { 1995,  4, 20 },  { 1995,  4, 21 },  { 1995,  4, 22 },  { 1995,  4, 23 },
  { 1995,  4, 24 },  { 1995,  4, 25 },  { 1995,  4, 26 },  { 1995,  4, 27 },
  { 1995,  4, 28 },  { 1995,  4, 29 },  { 1995,  4, 30 },  { 1995,  5,  1 },
  { 1995,  5,  2 },  { 1995,  5,  3 },  { 1995,  5,  4 },  { 1995,  5,  5 },
  { 1995,  5,  6 },  { 1995,  5,  7 },  { 1995,  5,  8 },  { 1995,  5,  9 },
  { 1995,  5, 10 },  { 1995,  5, 11 },  { 1995,  5, 12 },  { 1995,  5, 13 },
  { 1995,  5, 14 },  { 1995,  5, 15 },  { 1995,  5, 16 },  { 1995,  5, 17 },
  { 1995,  5, 18 },  { 1995,  5, 19 },  { 1995,  5, 20 },  { 1995,  5, 21 },
  { 1995,  5, 22 },  { 1995,  5, 23 },  { 1995,  5, 24 },  { 1995,  5, 25 },
  { 1995,  5, 26 },  { 1995,  5, 27 },  { 1995,  5, 28 },  { 1995,  5, 29 },
  { 1995,  5, 30 },  { 1995,  5, 31 },  { 1995,  6,  1 },  { 1995,  6,  2 },
  { 1995,  6,  3 },  { 1995,  6,  4 },  { 1995,  6,  5 },  { 1995,  6,  6 },
  { 1995,  6,  7 },  { 1995,  6,  8 },  { 1995,  6,  9 },  { 1995,  6, 10 },
  { 1995,  6, 11 },  { 1995,  6, 12 },  { 1995,  6, 13 },  { 1995,  6, 14 },
  { 1995,  6, 15 },  { 1995,  6, 16 },  { 1995,  6, 17 },  { 1995,  6, 18 },
  { 1995,  6, 19 },  { 1995,  6, 20 },  { 1995,  6, 21 },  { 1995,  6, 22 },
  { 1995,  6, 23 },  { 1995,  6, 24 },  { 1995,  6, 25 },  { 1995,  6, 26 },
  { 1995,  6, 27 },  { 1995,  6, 28 },  { 1995,  6, 29 },  { 1995,  6, 30 },
  { 1995,  7,  1 },  { 1995,  7,  2 },  { 1995,  7,  3 },  { 1995,  7,  4 },
  { 1995,  7,  5 },  { 1995,  7,  6 },  { 1995,  7,  7 },  { 1995,  7,  8 },
  { 1995,  7,  9 },  { 1995,  7, 10 },  { 1995,  7, 11 },  { 1995,  7, 12 },
  { 1995,  7, 13 },  { 1995,  7, 14 },  { 1995,  7, 15 },  { 1995,  7, 16 },
  { 1995,  7, 17 },  { 1995,  7, 18 },  { 1995,  7, 19 },  { 1995,  7, 20 },
  { 1995,  7, 21 },  { 1995,  7, 22 },  { 1995,  7, 23 },  { 1995,  7, 24 },
  { 1995,  7, 25 },  { 1995,  7, 26 },  { 1995,  7, 27 },  { 1995,  7, 28 },
  { 1995,  7, 29 },  { 1995,  7, 30 },  { 1995,  7, 31 },  { 1995,  8,  1 },
  { 1995,  8,  2 },  { 1995,  8,  3 },  { 1995,  8,  4 },  { 1995,  8,  5 },
  { 1995,  8,  6 },  { 1995,  8,  7 },  { 1995,  8,  8 },  { 1995,  8,  9 },
  { 1995,  8, 10 },  { 1995,  8, 11 },  { 1995,  8, 12 },  { 1995,  8, 13 },
  { 1995,  8, 14 },  { 1995,  8, 15 },  { 1995,  8, 16 },  { 1995,  8, 17 },
  { 1995,  8, 18 },  { 1995,  8, 19 },  { 1995,  8, 20 },  { 1995,  8, 21 },
  { 1995,  8, 22 },  { 1995,  8, 23 },  { 1995,  8, 24 },  { 1995,  8, 25 },
  { 1995,  8, 26 },  { 1995,  8, 27 },  { 1995,  8, 28 },  { 1995,  8, 29 },
  { 1995,  8, 30 },  { 1995,  8, 31 },  { 1995,  9,  1 },  { 1995,  9,  2 },
  { 1995,  9,  3 },  { 1995,  9,  4 },  { 1995,  9,  5 },  { 1995,  9,  6 },
  { 1995,  9,  7 },  { 1995,  9,  8 },  { 1995,  9,  9 },  { 1995,  9, 10 },
  { 1995,  9, 11 },  { 1995,  9, 12 },  { 1995,  9, 13 },  { 1995,  9, 14 },
  { 1995,  9, 15 },  { 1995,  9, 16 },  { 1995,  9, 17 },  { 1995,  9, 18 },
  { 1995,  9, 19 },  { 1995,  9, 20 },  { 1995,  9, 21 },  { 1995,  9, 22 },
  { 1995,  9, 23 },  { 1995,  9, 24 },  { 1995,  9, 25 },  { 1995,  9, 26 },
  { 1995,  9, 27 },  { 1995,  9, 28 },  { 1995,  9, 29 },  { 1995,  9, 30 },
  { 1995, 10,  1 },  { 1995, 10,  2 },  { 1995, 10,  3 },  { 1995, 10,  4 },
  { 1995, 10,  5 },  { 1995, 10,  6 },  { 1995, 10,  7 },  { 1995, 10,  8 },
  { 1995, 10,  9 },  { 1995, 10, 10 },  { 1995, 10, 11 },  { 1995, 10, 12 },
  { 1995, 10, 13 },  { 1995, 10, 14 },  { 1995, 10, 15 },  { 1995, 10, 16 },
  { 1995, 10, 17 },  { 1995, 10, 18 },  { 1995, 10, 19 },  { 1995, 10, 20 },
  { 1995, 10, 21 },  { 1995, 10, 22 },  { 1995, 10, 23 },  { 1995, 10, 24 },
  { 1995, 10, 25 },  { 1995, 10, 26 },  { 1995, 10, 27 },  { 1995, 10, 28 },
  { 1995, 10, 29 },  { 1995, 10, 30 },  { 1995, 10, 31 },  { 1995, 11,  1 },
  { 1995, 11,  2 },  { 1995, 11,  3 },  { 1995, 11,  4 },  { 1995, 11,  5 },
  { 1995, 11,  6 },  { 1995, 11,  7 },  { 1995, 11,  8 },  { 1995, 11,  9 },
  { 1995, 11, 10 },  { 1995, 11, 11 },  { 1995, 11, 12 },  { 1995, 11, 13 },
  { 1995, 11, 14 },  { 1995, 11, 15 },  { 1995, 11, 16 },  { 1995, 11, 17 },
  { 1995, 11, 18 },  { 1995, 11, 19 },  { 1995, 11, 20 },  { 1995, 11, 21 },
  { 1995, 11, 22 },  { 1995, 11, 23 },  { 1995, 11, 24 },  { 1995, 11, 25 },
  { 1995, 11, 26 },  { 1995, 11, 27 },  { 1995, 11, 28 },  { 1995, 11, 29 },
  { 1995, 11, 30 },  { 1995, 12,  1 },  { 1995, 12,  2 },  { 1995, 12,  3 },
  { 1995, 12,  4 },  { 1995, 12,  5 },  { 1995, 12,  6 },  { 1995, 12,  7 },
  { 1995, 12,  8 },  { 1995, 12,  9 },  { 1995, 12, 10 },  { 1995, 12, 11 },
  { 1995, 12, 12 },  { 1995, 12, 13 },  { 1995, 12, 14 },  { 1995, 12, 15 },
  { 1995, 12, 16 },  { 1995, 12, 17 },  { 1995, 12, 18 },  { 1995, 12, 19 },
  { 1995, 12, 20 },  { 1995, 12, 21 },  { 1995, 12, 22 },  { 1995, 12, 23 },
  { 1995, 12, 24 },  { 1995, 12, 25 },  { 1995, 12, 26 },  { 1995, 12, 27 },
  { 1995, 12, 28 },  { 1995, 12, 29 },  { 1995, 12, 30 },  { 1995, 12, 31 },
  { 1996,  1,  1 },  { 1996,  1,  2 },  { 1996,  1,  3 },  { 1996,  1,  4 },
  { 1996,  1,  5 },  { 1996,  1,  6 },  { 1996,  1,  7 },  { 1996,  1,  8 },
  { 1996,  1,  9 },  { 1996,  1, 10 },  { 1996,  1, 11 },  { 1996,  1, 12 },
  { 1996,  1, 13 },  { 1996,  1, 14 },  { 1996,  1, 15 },  { 1996,  1, 16 },
  { 1996,  1, 17 },  { 1996,  1, 18 },  { 1996,  1, 19 },  { 1996,  1, 20 },
  { 1996,  1, 21 },  { 1996,  1, 22 },  { 1996,  1, 23 },  { 1996,  1, 24 },
  { 1996,  1, 25 },  { 1996,  1, 26 },  { 1996,  1, 27 },  { 1996,  1, 28 },
  { 1996,  1, 29 },  { 1996,  1, 30 },  { 1996,  1, 31 },  { 1996,  2,  1 },
  { 1996,  2,  2 },  { 1996,  2,  3 },  { 1996,  2,  4 },  { 1996,  2,  5 },
  { 1996,  2,  6 },  { 1996,  2,  7 },  { 1996,  2,  8 },  { 1996,  2,  9 },
  { 1996,  2, 10 },  { 1996,  2, 11 },  { 1996,  2, 12 },  { 1996,  2, 13 },
  { 1996,  2, 14 },  { 1996,  2, 15 },  { 1996,  2, 16 },  { 1996,  2, 17 },
  { 1996,  2, 18 },  { 1996,  2, 19 },  { 1996,  2, 20 },  { 1996,  2, 21 },
  { 1996,  2, 22 },  { 1996,  2, 23 },  { 1996,  2, 24 },  { 1996,  2, 25 },
  { 1996,  2, 26 },  { 1996,  2, 27 },  { 1996,  2, 28 },  { 1996,  2, 29 },
  { 1996,  3,  1 },  { 1996,  3,  2 },  { 1996,  3,  3 },  { 1996,  3,  4 },
  { 1996,  3,  5 },  { 1996,  3,  6 },  { 1996,  3,  7 },  { 1996,  3,  8 },
  { 1996,  3,  9 },  { 1996,  3, 10 },  { 1996,  3, 11 },  { 1996,  3, 12 },
  { 1996,  3, 13 },  { 1996,  3, 14 },  { 1996,  3, 15 },  { 1996,  3, 16 },
  { 1996,  3, 17 },  { 1996,  3, 18 },  { 1996,  3, 19 },  { 1996,  3, 20 },
  { 1996,  3, 21 },  { 1996,  3, 22 },  { 1996,  3, 23 },  { 1996,  3, 24 },
  { 1996,  3, 25 },  { 1996,  3, 26 },  { 1996,  3, 27 },  { 1996,  3, 28 },
  { 1996,  3, 29 },  { 1996,  3, 30 },  { 1996,  3, 31 },  { 1996,  4,  1 },
  { 1996,  4,  2 },  { 1996,  4,  3 },  { 1996,  4,  4 },  { 1996,  4,  5 },
  { 1996,  4,  6 },  { 1996,  4,  7 },  { 1996,  4,  8 },  { 1996,  4,  9 },
  { 1996,  4, 10 },  { 1996,  4, 11 },  { 1996,  4, 12 },  { 1996,  4, 13 },
  { 1996,  4, 14 },  { 1996,  4, 15 },  { 1996,  4, 16 },  { 1996,  4, 17 },
  { 1996,  4, 18 },  { 1996,  4, 19 },  { 1996,  4, 20 },  { 1996,  4, 21 },
  { 1996,  4, 22 },  { 1996,  4, 23 },  { 1996,  4, 24 },  { 1996,  4, 25 },
  { 1996,  4, 26 },  { 1996,  4, 27 },  { 1996,  4, 28 },  { 1996,  4, 29 },
  { 1996,  4, 30 },  { 1996,  5,  1 },  { 1996,  5,  2 },  { 1996,  5,  3 },
  { 1996,  5,  4 },  { 1996,  5,  5 },  { 1996,  5,  6 },  { 1996,  5,  7 },
  { 1996,  5,  8 },  { 1996,  5,  9 },  { 1996,  5, 10 },  { 1996,  5, 11 },
  { 1996,  5, 12 },  { 1996,  5, 13 },  { 1996,  5, 14 },  { 1996,  5, 15 },
  { 1996,  5, 16 },  { 1996,  5, 17 },  { 1996,  5, 18 },  { 1996,  5, 19 },
  { 1996,  5, 20 },  { 1996,  5, 21 },  { 1996,  5, 22 },  { 1996,  5, 23 },
  { 1996,  5, 24 },  { 1996,  5, 25 },  { 1996,  5, 26 },  { 1996,  5, 27 },
  { 1996,  5, 28 },  { 1996,  5, 29 },  { 1996,  5, 30 },  { 1996,  5, 31 },
  { 1996,  6,  1 },  { 1996,  6,  2 },  { 1996,  6,  3 },  { 1996,  6,  4 },
  { 1996,  6,  5 },  { 1996,  6,  6 },  { 1996,  6,  7 },  { 1996,  6,  8 },
  { 1996,  6,  9 },  { 1996,  6, 10 },  { 1996,  6, 11 },  { 1996,  6, 12 },
  { 1996,  6, 13 },  { 1996,  6, 14 },  { 1996,  6, 15 },  { 1996,  6, 16 },
  { 1996,  6, 17 },  { 1996,  6, 18 },  { 1996,  6, 19 },  { 1996,  6, 20 },
  { 1996,  6, 21 },  { 1996,  6, 22 },  { 1996,  6, 23 },  { 1996,  6, 24 },
  { 1996,  6, 25 },  { 1996,  6, 26 },  { 1996,  6, 27 },  { 1996,  6, 28 },
  { 1996,  6, 29 },  { 1996,  6, 30 },  { 1996,  7,  1 },  { 1996,  7,  2 },
  { 1996,  7,  3 },  { 1996,  7,  4 },  { 1996,  7,  5 },  { 1996,  7,  6 },
  { 1996,  7,  7 },  { 1996,  7,  8 },  { 1996,  7,  9 },  { 1996,  7, 10 },
  { 1996,  7, 11 },  { 1996,  7, 12 },  { 1996,  7, 13 },  { 1996,  7, 14 },
  { 1996,  7, 15 },  { 1996,  7, 16 },  { 1996,  7, 17 },  { 1996,  7, 18 },
  { 1996,  7, 19 },  { 1996,  7, 20 },  { 1996,  7, 21 },  { 1996,  7, 22 },
  { 1996,  7, 23 },  { 1996,  7, 24 },  { 1996,  7, 25 },  { 1996,  7, 26 },
  { 1996,  7, 27 },  { 1996,  7, 28 },  { 1996,  7, 29 },  { 1996,  7, 30 },
  { 1996,  7, 31 },  { 1996,  8,  1 },  { 1996,  8,  2 },  { 1996,  8,  3 },
  { 1996,  8,  4 },  { 1996,  8,  5 },  { 1996,  8,  6 },  { 1996,  8,  7 },
  { 1996,  8,  8 },  { 1996,  8,  9 },  { 1996,  8, 10 },  { 1996,  8, 11 },
  { 1996,  8, 12 },  { 1996,  8, 13 },  { 1996,  8, 14 },  { 1996,  8, 15 },
  { 1996,  8, 16 },  { 1996,  8, 17 },  { 1996,  8, 18 },  { 1996,  8, 19 },
  { 1996,  8, 20 },  { 1996,  8, 21 },  { 1996,  8, 22 },  { 1996,  8, 23 },
  { 1996,  8, 24 },  { 1996,  8, 25 },  { 1996,  8, 26 },  { 1996,  8, 27 },
  { 1996,  8, 28 },  { 1996,  8, 29 },  { 1996,  8, 30 },  { 1996,  8, 31 },
  { 1996,  9,  1 },  { 1996,  9,  2 },  { 1996,  9,  3 },  { 1996,  9,  4 },
  { 1996,  9,  5 },  { 1996,  9,  6 },  { 1996,  9,  7 },  { 1996,  9,  8 },
  { 1996,  9,  9 },  { 1996,  9, 10 },  { 1996,  9, 11 },  { 1996,  9, 12 },
  { 1996,  9, 13 },  { 1996,  9, 14 },  { 1996,  9, 15 },  { 1996,  9, 16 },
  { 1996,  9, 17 },  { 1996,  9, 18 },  { 1996,  9, 19 },  { 1996,  9, 20 },
  { 1996,  9, 21 },  { 1996,  9, 22 },  { 1996,  9, 23 },  { 1996,  9, 24 },
  { 1996,  9, 25 },  { 1996,  9, 26 },  { 1996,  9, 27 },  { 1996,  9, 28 },
  { 1996,  9, 29 },  { 1996,  9, 30 },  { 1996, 10,  1 },  { 1996, 10,  2 },
  { 1996, 10,  3 },  { 1996, 10,  4 },  { 1996, 10,  5 },  { 1996, 10,  6 },
  { 1996, 10,  7 },  { 1996, 10,  8 },  { 1996, 10,  9 },  { 1996, 10, 10 },
  { 1996, 10, 11 },  { 1996, 10, 12 },  { 1996, 10, 13 },  { 1996, 10, 14 },
  { 1996, 10, 15 },  { 1996, 10, 16 },  { 1996, 10, 17 },  { 1996, 10, 18 },
  { 1996, 10, 19 },  { 1996, 10, 20 },  { 1996, 10, 21 },  { 1996, 10, 22 },
  { 1996, 10, 23 },  { 1996, 10, 24 },  { 1996, 10, 25 },  { 1996, 10, 26 },
  { 1996, 10, 27 },  { 1996, 10, 28 },  { 1996, 10, 29 },  { 1996, 10, 30 },
  { 1996, 10, 31 },  { 1996, 11,  1 },  { 1996, 11,  2 },  { 1996, 11,  3 },
  { 1996, 11,  4 },  { 1996, 11,  5 },  { 1996, 11,  6 },  { 1996, 11,  7 },
  { 1996, 11,  8 },  { 1996, 11,  9 },  { 1996, 11, 10 },  { 1996, 11, 11 },
  { 1996, 11, 12 },  { 1996, 11, 13 },  { 1996, 11, 14 },  { 1996, 11, 15 },
  { 1996, 11, 16 },  { 1996, 11, 17 },  { 1996, 11, 18 },  { 1996, 11, 19 },
  { 1996, 11, 20 },  { 1996, 11, 21 },  { 1996, 11, 22 },  { 1996, 11, 23 },
  { 1996, 11, 24 },  { 1996, 11, 25 },  { 1996, 11, 26 },  { 1996, 11, 27 },
  { 1996, 11, 28 },  { 1996, 11, 29 },  { 1996, 11, 30 },  { 1996, 12,  1 },
  { 1996, 12,  2 },  { 1996, 12,  3 },  { 1996, 12,  4 },  { 1996, 12,  5 },
  { 1996, 12,  6 },  { 1996, 12,  7 },  { 1996, 12,  8 },  { 1996, 12,  9 },
  { 1996, 12, 10 },  { 1996, 12, 11 },  { 1996, 12, 12 },  { 1996, 12, 13 },
  { 1996, 12, 14 },  { 1996, 12, 15 },  { 1996, 12, 16 },  { 1996, 12, 17 },
  { 1996, 12, 18 },  { 1996, 12, 19 },  { 1996, 12, 20 },  { 1996, 12, 21 },
  { 1996, 12, 22 },  { 1996, 12, 23 },  { 1996, 12, 24 },  { 1996, 12, 25 },
  { 1996, 12, 26 },  { 1996, 12, 27 },  { 1996, 12, 28 },  { 1996, 12, 29 },
  { 1996, 12, 30 },  { 1996, 12, 31 },  { 1997,  1,  1 },  { 1997,  1,  2 },
  { 1997,  1,  3 },  { 1997,  1,  4 },  { 1997,  1,  5 },  { 1997,  1,  6 },
  { 1997,  1,  7 },  { 1997,  1,  8 },  { 1997,  1,  9 },  { 1997,  1, 10 },
  { 1997,  1, 11 },  { 1997,  1, 12 },  { 1997,  1, 13 },  { 1997,  1, 14 },
  { 1997,  1, 15 },  { 1997,  1, 16 },  { 1997,  1, 17 },  { 1997,  1, 18 },
  { 1997,  1, 19 },  { 1997,  1, 20 },  { 1997,  1, 21 },  { 1997,  1, 22 },
  { 1997,  1, 23 },  { 1997,  1, 24 },  { 1997,  1, 25 },  { 1997,  1, 26 },
  { 1997,  1, 27 },  { 1997,  1, 28 },  { 1997,  1, 29 },  { 1997,  1, 30 },
  { 1997,  1, 31 },  { 1997,  2,  1 },  { 1997,  2,  2 },  { 1997,  2,  3 },
  { 1997,  2,  4 },  { 1997,  2,  5 },  { 1997,  2,  6 },  { 1997,  2,  7 },
  { 1997,  2,  8 },  { 1997,  2,  9 },  { 1997,  2, 10 },  { 1997,  2, 11 },
  { 1997,  2, 12 },  { 1997,  2, 13 },  { 1997,  2, 14 },  { 1997,  2, 15 },
  { 1997,  2, 16 },  { 1997,  2, 17 },  { 1997,  2, 18 },  { 1997,  2, 19 },
  { 1997,  2, 20 },  { 1997,  2, 21 },  { 1997,  2, 22 },  { 1997,  2, 23 },
  { 1997,  2, 24 },  { 1997,  2, 25 },  { 1997,  2, 26 },  { 1997,  2, 27 },
  { 1997,  2, 28 },  { 1997,  3,  1 },  { 1997,  3,  2 },  { 1997,  3,  3 },
  { 1997,  3,  4 },  { 1997,  3,  5 },  { 1997,  3,  6 },  { 1997,  3,  7 },
  { 1997,  3,  8 },  { 1997,  3,  9 },  { 1997,  3, 10 },  { 1997,  3, 11 },
  { 1997,  3, 12 },  { 1997,  3, 13 },  { 1997,  3, 14 },  { 1997,  3, 15 },
  { 1997,  3, 16 },  { 1997,  3, 17 },  { 1997,  3, 18 },  { 1997,  3, 19 },
  { 1997,  3, 20 },  { 1997,  3, 21 },  { 1997,  3, 22 },  { 1997,  3, 23 },
  { 1997,  3, 24 },  { 1997,  3, 25 },  { 1997,  3, 26 },  { 1997,  3, 27 },
  { 1997,  3, 28 },  { 1997,  3, 29 },  { 1997,  3, 30 },  { 1997,  3, 31 },
  { 1997,  4,  1 },  { 1997,  4,  2 },  { 1997,  4,  3 },  { 1997,  4,  4 },
  { 1997,  4,  5 },  { 1997,  4,  6 },  { 1997,  4,  7 },  { 1997,  4,  8 },
  { 1997,  4,  9 },  { 1997,  4, 10 },  { 1997,  4, 11 },  { 1997,  4, 12 },
  { 1997,  4, 13 },  { 1997,  4, 14 },  { 1997,  4, 15 },  { 1997,  4, 16 },
  { 1997,  4, 17 },  { 1997,  4, 18 },  { 1997,  4, 19 },  { 1997,  4, 20 },
  { 1997,  4, 21 },  { 1997,  4, 22 },  { 1997,  4, 23 },  { 1997,  4, 24 },
  { 1997,  4, 25 },  { 1997,  4, 26 },  { 1997,  4, 27 },  { 1997,  4, 28 },
  { 1997,  4, 29 },  { 1997,  4, 30 },  { 1997,  5,  1 },  { 1997,  5,  2 },
  { 1997,  5,  3 },  { 1997,  5,  4 },  { 1997,  5,  5 },  { 1997,  5,  6 },
  { 1997,  5,  7 },  { 1997,  5,  8 },  { 1997,  5,  9 },  { 1997,  5, 10 },
  { 1997,  5, 11 },  { 1997,  5, 12 },  { 1997,  5, 13 },  { 1997,  5, 14 },
  { 1997,  5, 15 },  { 1997,  5, 16 },  { 1997,  5, 17 },  { 1997,  5, 18 },
  { 1997,  5, 19 },  { 1997,  5, 20 },  { 1997,  5, 21 },  { 1997,  5, 22 },
  { 1997,  5, 23 },  { 1997,  5, 24 },  { 1997,  5, 25 },  { 1997,  5, 26 },
  { 1997,  5, 27 },  { 1997,  5, 28 },  { 1997,  5, 29 },  { 1997,  5, 30 },
  { 1997,  5, 31 },  { 1997,  6,  1 },  { 1997,  6,  2 },  { 1997,  6,  3 },
  { 1997,  6,  4 },  { 1997,  6,  5 },  { 1997,  6,  6 },  { 1997,  6,  7 },
  { 1997,  6,  8 },  { 1997,  6,  9 },  { 1997,  6, 10 },  { 1997,  6, 11 },
  { 1997,  6, 12 },  { 1997,  6, 13 },  { 1997,  6, 14 },  { 1997,  6, 15 },
  { 1997,  6, 16 },  { 1997,  6, 17 },  { 1997,  6, 18 },  { 1997,  6, 19 },
  { 1997,  6, 20 },  { 1997,  6, 21 },  { 1997,  6, 22 },  { 1997,  6, 23 },
  { 1997,  6, 24 },  { 1997,  6, 25 },  { 1997,  6, 26 },  { 1997,  6, 27 },
  { 1997,  6, 28 },  { 1997,  6, 29 },  { 1997,  6, 30 },  { 1997,  7,  1 },
  { 1997,  7,  2 },  { 1997,  7,  3 },  { 1997,  7,  4 },  { 1997,  7,  5 },
  { 1997,  7,  6 },  { 1997,  7,  7 },  { 1997,  7,  8 },  { 1997,  7,  9 },
  { 1997,  7, 10 },  { 1997,  7, 11 },  { 1997,  7, 12 },  { 1997,  7, 13 },
  { 1997,  7, 14 },  { 1997,  7, 15 },  { 1997,  7, 16 },  { 1997,  7, 17 },
  { 1997,  7, 18 },  { 1997,  7, 19 },  { 1997,  7, 20 },  { 1997,  7, 21 },
  { 1997,  7, 22 },  { 1997,  7, 23 },  { 1997,  7, 24 },  { 1997,  7, 25 },
  { 1997,  7, 26 },  { 1997,  7, 27 },  { 1997,  7, 28 },  { 1997,  7, 29 },
  { 1997,  7, 30 },  { 1997,  7, 31 },  { 1997,  8,  1 },  { 1997,  8,  2 },
  { 1997,  8,  3 },  { 1997,  8,  4 },  { 1997,  8,  5 },  { 1997,  8,  6 },
  { 1997,  8,  7 },  { 1997,  8,  8 },  { 1997,  8,  9 },  { 1997,  8, 10 },
  { 1997,  8, 11 },  { 1997,  8, 12 },  { 1997,  8, 13 },  { 1997,  8, 14 },
  { 1997,  8, 15 },  { 1997,  8, 16 },  { 1997,  8, 17 },  { 1997,  8, 18 },
  { 1997,  8, 19 },  { 1997,  8, 20 },  { 1997,  8, 21 },  { 1997,  8, 22 },
  { 1997,  8, 23 },  { 1997,  8, 24 },  { 1997,  8, 25 },  { 1997,  8, 26 },
  { 1997,  8, 27 },  { 1997,  8, 28 },  { 1997,  8, 29 },  { 1997,  8, 30 },
  { 1997,  8, 31 },  { 1997,  9,  1 },  { 1997,  9,  2 },  { 1997,  9,  3 },
  { 1997,  9,  4 },  { 1997,  9,  5 },  { 1997,  9,  6 },  { 1997,  9,  7 },
  { 1997,  9,  8 },  { 1997,  9,  9 },  { 1997,  9, 10 },  { 1997,  9, 11 },
  { 1997,  9, 12 },  { 1997,  9, 13 },  { 1997,  9, 14 },  { 1997,  9, 15 },
  { 1997,  9, 16 },  { 1997,  9, 17 },  { 1997,  9, 18 },  { 1997,  9, 19 },
  { 1997,  9, 20 },  { 1997,  9, 21 },  { 1997,  9, 22 },  { 1997,  9, 23 },
  { 1997,  9, 24 },  { 1997,  9, 25 },  { 1997,  9, 26 },  { 1997,  9, 27 },
  { 1997,  9, 28 },  { 1997,  9, 29 },  { 1997,  9, 30 },  { 1997, 10,  1 },
  { 1997, 10,  2 },  { 1997, 10,  3 },  { 1997, 10,  4 },  { 1997, 10,  5 },
  { 1997, 10,  6 },  { 1997, 10,  7 },  { 1997, 10,  8 },  { 1997, 10,  9 },
  { 1997, 10, 10 },  { 1997, 10, 11 },  { 1997, 10, 12 },  { 1997, 10, 13 },
  { 1997, 10, 14 },  { 1997, 10, 15 },  { 1997, 10, 16 },  { 1997, 10, 17 },
  { 1997, 10, 18 },  { 1997, 10, 19 },  { 1997, 10, 20 },  { 1997, 10, 21 },
  { 1997, 10, 22 },  { 1997, 10, 23 },  { 1997, 10, 24 },  { 1997, 10, 25 },
  { 1997, 10, 26 },  { 1997, 10, 27 },  { 1997, 10, 28 },  { 1997, 10, 29 },
  { 1997, 10, 30 },  { 1997, 10, 31 },  { 1997, 11,  1 },  { 1997, 11,  2 },
  { 1997, 11,  3 },  { 1997, 11,  4 },  { 1997, 11,  5 },  { 1997, 11,  6 },
  { 1997, 11,  7 },  { 1997, 11,  8 },  { 1997, 11,  9 },  { 1997, 11, 10 },
  { 1997, 11, 11 },  { 1997, 11, 12 },  { 1997, 11, 13 },  { 1997, 11, 14 },
  { 1997, 11, 15 },  { 1997, 11, 16 },  { 1997, 11, 17 },  { 1997, 11, 18 },
  { 1997, 11, 19 },  { 1997, 11, 20 },  { 1997, 11, 21 },  { 1997, 11, 22 },
  { 1997, 11, 23 },  { 1997, 11, 24 },  { 1997, 11, 25 },  { 1997, 11, 26 },
  { 1997, 11, 27 },  { 1997, 11, 28 },  { 1997, 11, 29 },  { 1997, 11, 30 },
  { 1997, 12,  1 },  { 1997, 12,  2 },  { 1997, 12,  3 },  { 1997, 12,  4 },
  { 1997, 12,  5 },  { 1997, 12,  6 },  { 1997, 12,  7 },  { 1997, 12,  8 },
  { 1997, 12,  9 },  { 1997, 12, 10 },  { 1997, 12, 11 },  { 1997, 12, 12 },
  { 1997, 12, 13 },  { 1997, 12, 14 },  { 1997, 12, 15 },  { 1997, 12, 16 },
  { 1997, 12, 17 },  { 1997, 12, 18 },  { 1997, 12, 19 },  { 1997, 12, 20 },
  { 1997, 12, 21 },  { 1997, 12, 22 },  { 1997, 12, 23 },  { 1997, 12, 24 },
  { 1997, 12, 25 },  { 1997, 12, 26 },  { 1997, 12, 27 },  { 1997, 12, 28 },
  { 1997, 12, 29 },  { 1997, 12, 30 },  { 1997, 12, 31 },  { 1998,  1,  1 },
  { 1998,  1,  2 },  { 1998,  1,  3 },  { 1998,  1,  4 },  { 1998,  1,  5 },
  { 1998,  1,  6 },  { 1998,  1,  7 },  { 1998,  1,  8 },  { 1998,  1,  9 },
  { 1998,  1, 10 },  { 1998,  1, 11 },  { 1998,  1, 12 },  { 1998,  1, 13 },
  { 1998,  1, 14 },  { 1998,  1, 15 },  { 1998,  1, 16 },  { 1998,  1, 17 },
  { 1998,  1, 18 },  { 1998,  1, 19 },  { 1998,  1, 20 },  { 1998,  1, 21 },
  { 1998,  1, 22 },  { 1998,  1, 23 },  { 1998,  1, 24 },  { 1998,  1, 25 },
  { 1998,  1, 26 },  { 1998,  1, 27 },  { 1998,  1, 28 },  { 1998,  1, 29 },
  { 1998,  1, 30 },  { 1998,  1, 31 },  { 1998,  2,  1 },  { 1998,  2,  2 },
  { 1998,  2,  3 },  { 1998,  2,  4 },  { 1998,  2,  5 },  { 1998,  2,  6 },
  { 1998,  2,  7 },  { 1998,  2,  8 },  { 1998,  2,  9 },  { 1998,  2, 10 },
  { 1998,  2, 11 },  { 1998,  2, 12 },  { 1998,  2, 13 },  { 1998,  2, 14 },
  { 1998,  2, 15 },  { 1998,  2, 16 },  { 1998,  2, 17 },  { 1998,  2, 18 },
  { 1998,  2, 19 },  { 1998,  2, 20 },  { 1998,  2, 21 },  { 1998,  2, 22 },
  { 1998,  2, 23 },  { 1998,  2, 24 },  { 1998,  2, 25 },  { 1998,  2, 26 },
  { 1998,  2, 27 },  { 1998,  2, 28 },  { 1998,  3,  1 },  { 1998,  3,  2 },
  { 1998,  3,  3 },  { 1998,  3,  4 },  { 1998,  3,  5 },  { 1998,  3,  6 },
  { 1998,  3,  7 },  { 1998,  3,  8 },  { 1998,  3,  9 },  { 1998,  3, 10 },
  { 1998,  3, 11 },  { 1998,  3, 12 },  { 1998,  3, 13 },  { 1998,  3, 14 },
  { 1998,  3, 15 },  { 1998,  3, 16 },  { 1998,  3, 17 },  { 1998,  3, 18 },
  { 1998,  3, 19 },  { 1998,  3, 20 },  { 1998,  3, 21 },  { 1998,  3, 22 },
  { 1998,  3, 23 },  { 1998,  3, 24 },  { 1998,  3, 25 },  { 1998,  3, 26 },
  { 1998,  3, 27 },  { 1998,  3, 28 },  { 1998,  3, 29 },  { 1998,  3, 30 },
  { 1998,  3, 31 },  { 1998,  4,  1 },  { 1998,  4,  2 },  { 1998,  4,  3 },
  { 1998,  4,  4 },  { 1998,  4,  5 },  { 1998,  4,  6 },  { 1998,  4,  7 },
  { 1998,  4,  8 },  { 1998,  4,  9 },  { 1998,  4, 10 },  { 1998,  4, 11 },
  { 1998,  4, 12 },  { 1998,  4, 13 },  { 1998,  4, 14 },  { 1998,  4, 15 },
  { 1998,  4, 16 },  { 1998,  4, 17 },  { 1998,  4, 18 },  { 1998,  4, 19 },
  { 1998,  4, 20 },  { 1998,  4, 21 },  { 1998,  4, 22 },  { 1998,  4, 23 },
  { 1998,  4, 24 },  { 1998,  4, 25 },  { 1998,  4, 26 },  { 1998,  4, 27 },
  { 1998,  4, 28 },  { 1998,  4, 29 },  { 1998,  4, 30 },  { 1998,  5,  1 },
  { 1998,  5,  2 },  { 1998,  5,  3 },  { 1998,  5,  4 },  { 1998,  5,  5 },
  { 1998,  5,  6 },  { 1998,  5,  7 },  { 1998,  5,  8 },  { 1998,  5,  9 },
  { 1998,  5, 10 },  { 1998,  5, 11 },  { 1998,  5, 12 },  { 1998,  5, 13 },
  { 1998,  5, 14 },  { 1998,  5, 15 },  { 1998,  5, 16 },  { 1998,  5, 17 },
  { 1998,  5, 18 },  { 1998,  5, 19 },  { 1998,  5, 20 },  { 1998,  5, 21 },
  { 1998,  5, 22 },  { 1998,  5, 23 },  { 1998,  5, 24 },  { 1998,  5, 25 },
  { 1998,  5, 26 },  { 1998,  5, 27 },  { 1998,  5, 28 },  { 1998,  5, 29 },
  { 1998,  5, 30 },  { 1998,  5, 31 },  { 1998,  6,  1 },  { 1998,  6,  2 },
  { 1998,  6,  3 },  { 1998,  6,  4 },  { 1998,  6,  5 },  { 1998,  6,  6 },
  { 1998,  6,  7 },  { 1998,  6,  8 },  { 1998,  6,  9 },  { 1998,  6, 10 },
  { 1998,  6, 11 },  { 1998,  6, 12 },  { 1998,  6, 13 },  { 1998,  6, 14 },
  { 1998,  6, 15 },  { 1998,  6, 16 },  { 1998,  6, 17 },  { 1998,  6, 18 },
  { 1998,  6, 19 },  { 1998,  6, 20 },  { 1998,  6, 21 },  { 1998,  6, 22 },
  { 1998,  6, 23 },  { 1998,  6, 24 },  { 1998,  6, 25 },  { 1998,  6, 26 },
  { 1998,  6, 27 },  { 1998,  6, 28 },  { 1998,  6, 29 },  { 1998,  6, 30 },
  { 1998,  7,  1 },  { 1998,  7,  2 },  { 1998,  7,  3 },  { 1998,  7,  4 },
  { 1998,  7,  5 },  { 1998,  7,  6 },  { 1998,  7,  7 },  { 1998,  7,  8 },
  { 1998,  7,  9 },  { 1998,  7, 10 },  { 1998,  7, 11 },  { 1998,  7, 12 },
  { 1998,  7, 13 },  { 1998,  7, 14 },  { 1998,  7, 15 },  { 1998,  7, 16 },
  { 1998,  7, 17 },  { 1998,  7, 18 },  { 1998,  7, 19 },  { 1998,  7, 20 },
  { 1998,  7, 21 },  { 1998,  7, 22 },  { 1998,  7, 23 },  { 1998,  7, 24 },
  { 1998,  7, 25 },  { 1998,  7, 26 },  { 1998,  7, 27 },  { 1998,  7, 28 },
  { 1998,  7, 29 },  { 1998,  7, 30 },  { 1998,  7, 31 },  { 1998,  8,  1 },
  { 1998,  8,  2 },  { 1998,  8,  3 },  { 1998,  8,  4 },  { 1998,  8,  5 },
  { 1998,  8,  6 },  { 1998,  8,  7 },  { 1998,  8,  8 },  { 1998,  8,  9 },
  { 1998,  8, 10 },  { 1998,  8, 11 },  { 1998,  8, 12 },  { 1998,  8, 13 },
  { 1998,  8, 14 },  { 1998,  8, 15 },  { 1998,  8, 16 },  { 1998,  8, 17 },
  { 1998,  8, 18 },  { 1998,  8, 19 },  { 1998,  8, 20 },  { 1998,  8, 21 },
  { 1998,  8, 22 },  { 1998,  8, 23 },  { 1998,  8, 24 },  { 1998,  8, 25 },
  { 1998,  8, 26 },  { 1998,  8, 27 },  { 1998,  8, 28 },  { 1998,  8, 29 },
  { 1998,  8, 30 },  { 1998,  8, 31 },  { 1998,  9,  1 },  { 1998,  9,  2 },
  { 1998,  9,  3 },  { 1998,  9,  4 },  { 1998,  9,  5 },  { 1998,  9,  6 },
  { 1998,  9,  7 },  { 1998,  9,  8 },  { 1998,  9,  9 },  { 1998,  9, 10 },
  { 1998,  9, 11 },  { 1998,  9, 12 },  { 1998,  9, 13 },  { 1998,  9, 14 },
  { 1998,  9, 15 },  { 1998,  9, 16 },  { 1998,  9, 17 },  { 1998,  9, 18 },
  { 1998,  9, 19 },  { 1998,  9, 20 },  { 1998,  9, 21 },  { 1998,  9, 22 },
  { 1998,  9, 23 },  { 1998,  9, 24 },  { 1998,  9, 25 },  { 1998,  9, 26 },
  { 1998,  9, 27 },  { 1998,  9, 28 },  { 1998,  9, 29 },  { 1998,  9, 30 },
  { 1998, 10,  1 },  { 1998, 10,  2 },  { 1998, 10,  3 },  { 1998, 10,  4 },
  { 1998, 10,  5 },  { 1998, 10,  6 },  { 1998, 10,  7 },  { 1998, 10,  8 },
  { 1998, 10,  9 },  { 1998, 10, 10 },  { 1998, 10, 11 },  { 1998, 10, 12 },
  { 1998, 10, 13 },  { 1998, 10, 14 },  { 1998, 10, 15 },  { 1998, 10, 16 },
  { 1998, 10, 17 },  { 1998, 10, 18 },  { 1998, 10, 19 },  { 1998, 10, 20 },
  { 1998, 10, 21 },  { 1998, 10, 22 },  { 1998, 10, 23 },  { 1998, 10, 24 },
  { 1998, 10, 25 },  { 1998, 10, 26 },  { 1998, 10, 27 },  { 1998, 10, 28 },
  { 1998, 10, 29 },  { 1998, 10, 30 },  { 1998, 10, 31 },  { 1998, 11,  1 },
  { 1998, 11,  2 },  { 1998, 11,  3 },  { 1998, 11,  4 },  { 1998, 11,  5 },
  { 1998, 11,  6 },  { 1998, 11,  7 },  { 1998, 11,  8 },  { 1998, 11,  9 },
  { 1998, 11, 10 },  { 1998, 11, 11 },  { 1998, 11, 12 },  { 1998, 11, 13 },
  { 1998, 11, 14 },  { 1998, 11, 15 },  { 1998, 11, 16 },  { 1998, 11, 17 },
  { 1998, 11, 18 },  { 1998, 11, 19 },  { 1998, 11, 20 },  { 1998, 11, 21 },
  { 1998, 11, 22 },  { 1998, 11, 23 },  { 1998, 11, 24 },  { 1998, 11, 25 },
  { 1998, 11, 26 },  { 1998, 11, 27 },  { 1998, 11, 28 },  { 1998, 11, 29 },
  { 1998, 11, 30 },  { 1998, 12,  1 },  { 1998, 12,  2 },  { 1998, 12,  3 },
  { 1998, 12,  4 },  { 1998, 12,  5 },  { 1998, 12,  6 },  { 1998, 12,  7 },
  { 1998, 12,  8 },  { 1998, 12,  9 },  { 1998, 12, 10 },  { 1998, 12, 11 },
  { 1998, 12, 12 },  { 1998, 12, 13 },  { 1998, 12, 14 },  { 1998, 12, 15 },
  { 1998, 12, 16 },  { 1998, 12, 17 },  { 1998, 12, 18 },  { 1998, 12, 19 },
  { 1998, 12, 20 },  { 1998, 12, 21 },  { 1998, 12, 22 },  { 1998, 12, 23 },
  { 1998, 12, 24 },  { 1998, 12, 25 },  { 1998, 12, 26 },  { 1998, 12, 27 },
  { 1998, 12, 28 },  { 1998, 12, 29 },  { 1998, 12, 30 },  { 1998, 12, 31 },
  { 1999,  1,  1 },  { 1999,  1,  2 },  { 1999,  1,  3 },  { 1999,  1,  4 },
  { 1999,  1,  5 },  { 1999,  1,  6 },  { 1999,  1,  7 },  { 1999,  1,  8 },
  { 1999,  1,  9 },  { 1999,  1, 10 },  { 1999,  1, 11 },  { 1999,  1, 12 },
  { 1999,  1, 13 },  { 1999,  1, 14 },  { 1999,  1, 15 },  { 1999,  1, 16 },
  { 1999,  1, 17 },  { 1999,  1, 18 },  { 1999,  1, 19 },  { 1999,  1, 20 },
  { 1999,  1, 21 },  { 1999,  1, 22 },  { 1999,  1, 23 },  { 1999,  1, 24 },
  { 1999,  1, 25 },  { 1999,  1, 26 },  { 1999,  1, 27 },  { 1999,  1, 28 },
  { 1999,  1, 29 },  { 1999,  1, 30 },  { 1999,  1, 31 },  { 1999,  2,  1 },
  { 1999,  2,  2 },  { 1999,  2,  3 },  { 1999,  2,  4 },  { 1999,  2,  5 },
  { 1999,  2,  6 },  { 1999,  2,  7 },  { 1999,  2,  8 },  { 1999,  2,  9 },
  { 1999,  2, 10 },  { 1999,  2, 11 },  { 1999,  2, 12 },  { 1999,  2, 13 },
  { 1999,  2, 14 },  { 1999,  2, 15 },  { 1999,  2, 16 },  { 1999,  2, 17 },
  { 1999,  2, 18 },  { 1999,  2, 19 },  { 1999,  2, 20 },  { 1999,  2, 21 },
  { 1999,  2, 22 },  { 1999,  2, 23 },  { 1999,  2, 24 },  { 1999,  2, 25 },
  { 1999,  2, 26 },  { 1999,  2, 27 },  { 1999,  2, 28 },  { 1999,  3,  1 },
  { 1999,  3,  2 },  { 1999,  3,  3 },  { 1999,  3,  4 },  { 1999,  3,  5 },
  { 1999,  3,  6 },  { 1999,  3,  7 },  { 1999,  3,  8 },  { 1999,  3,  9 },
  { 1999,  3, 10 },  { 1999,  3, 11 },  { 1999,  3, 12 },  { 1999,  3, 13 },
  { 1999,  3, 14 },  { 1999,  3, 15 },  { 1999,  3, 16 },  { 1999,  3, 17 },
  { 1999,  3, 18 },  { 1999,  3, 19 },  { 1999,  3, 20 },  { 1999,  3, 21 },
  { 1999,  3, 22 },  { 1999,  3, 23 },  { 1999,  3, 24 },  { 1999,  3, 25 },
  { 1999,  3, 26 },  { 1999,  3, 27 },  { 1999,  3, 28 },  { 1999,  3, 29 },
  { 1999,  3, 30 },  { 1999,  3, 31 },  { 1999,  4,  1 },  { 1999,  4,  2 },
  { 1999,  4,  3 },  { 1999,  4,  4 },  { 1999,  4,  5 },  { 1999,  4,  6 },
  { 1999,  4,  7 },  { 1999,  4,  8 },  { 1999,  4,  9 },  { 1999,  4, 10 },
  { 1999,  4, 11 },  { 1999,  4, 12 },  { 1999,  4, 13 },  { 1999,  4, 14 },
  { 1999,  4, 15 },  { 1999,  4, 16 },  { 1999,  4, 17 },  { 1999,  4, 18 },
  { 1999,  4, 19 },  { 1999,  4, 20 },  { 1999,  4, 21 },  { 1999,  4, 22 },
  { 1999,  4, 23 },  { 1999,  4, 24 },  { 1999,  4, 25 },  { 1999,  4, 26 },
  { 1999,  4, 27 },  { 1999,  4, 28 },  { 1999,  4, 29 },  { 1999,  4, 30 },
  { 1999,  5,  1 },  { 1999,  5,  2 },  { 1999,  5,  3 },  { 1999,  5,  4 },
  { 1999,  5,  5 },  { 1999,  5,  6 },  { 1999,  5,  7 },  { 1999,  5,  8 },
  { 1999,  5,  9 },  { 1999,  5, 10 },  { 1999,  5, 11 },  { 1999,  5, 12 },
  { 1999,  5, 13 },  { 1999,  5, 14 },  { 1999,  5, 15 },  { 1999,  5, 16 },
  { 1999,  5, 17 },  { 1999,  5, 18 },  { 1999,  5, 19 },  { 1999,  5, 20 },
  { 1999,  5, 21 },  { 1999,  5, 22 },  { 1999,  5, 23 },  { 1999,  5, 24 },
  { 1999,  5, 25 },  { 1999,  5, 26 },  { 1999,  5, 27 },  { 1999,  5, 28 },
  { 1999,  5, 29 },  { 1999,  5, 30 },  { 1999,  5, 31 },  { 1999,  6,  1 },
  { 1999,  6,  2 },  { 1999,  6,  3 },  { 1999,  6,  4 },  { 1999,  6,  5 },
  { 1999,  6,  6 },  { 1999,  6,  7 },  { 1999,  6,  8 },  { 1999,  6,  9 },
  { 1999,  6, 10 },  { 1999,  6, 11 },  { 1999,  6, 12 },  { 1999,  6, 13 },
  { 1999,  6, 14 },  { 1999,  6, 15 },  { 1999,  6, 16 },  { 1999,  6, 17 },
  { 1999,  6, 18 },  { 1999,  6, 19 },  { 1999,  6, 20 },  { 1999,  6, 21 },
  { 1999,  6, 22 },  { 1999,  6, 23 },  { 1999,  6, 24 },  { 1999,  6, 25 },
  { 1999,  6, 26 },  { 1999,  6, 27 },  { 1999,  6, 28 },  { 1999,  6, 29 },
  { 1999,  6, 30 },  { 1999,  7,  1 },  { 1999,  7,  2 },  { 1999,  7,  3 },
  { 1999,  7,  4 },  { 1999,  7,  5 },  { 1999,  7,  6 },  { 1999,  7,  7 },
  { 1999,  7,  8 },  { 1999,  7,  9 },  { 1999,  7, 10 },  { 1999,  7, 11 },
  { 1999,  7, 12 },  { 1999,  7, 13 },  { 1999,  7, 14 },  { 1999,  7, 15 },
  { 1999,  7, 16 },  { 1999,  7, 17 },  { 1999,  7, 18 },  { 1999,  7, 19 },
  { 1999,  7, 20 },  { 1999,  7, 21 },  { 1999,  7, 22 },  { 1999,  7, 23 },
  { 1999,  7, 24 },  { 1999,  7, 25 },  { 1999,  7, 26 },  { 1999,  7, 27 },
  { 1999,  7, 28 },  { 1999,  7, 29 },  { 1999,  7, 30 },  { 1999,  7, 31 },
  { 1999,  8,  1 },  { 1999,  8,  2 },  { 1999,  8,  3 },  { 1999,  8,  4 },
  { 1999,  8,  5 },  { 1999,  8,  6 },  { 1999,  8,  7 },  { 1999,  8,  8 },
  { 1999,  8,  9 },  { 1999,  8, 10 },  { 1999,  8, 11 },  { 1999,  8, 12 },
  { 1999,  8, 13 },  { 1999,  8, 14 },  { 1999,  8, 15 },  { 1999,  8, 16 },
  { 1999,  8, 17 },  { 1999,  8, 18 },  { 1999,  8, 19 },  { 1999,  8, 20 },
  { 1999,  8, 21 },  { 1999,  8, 22 },  { 1999,  8, 23 },  { 1999,  8, 24 },
  { 1999,  8, 25 },  { 1999,  8, 26 },  { 1999,  8, 27 },  { 1999,  8, 28 },
  { 1999,  8, 29 },  { 1999,  8, 30 },  { 1999,  8, 31 },  { 1999,  9,  1 },
  { 1999,  9,  2 },  { 1999,  9,  3 },  { 1999,  9,  4 },  { 1999,  9,  5 },
  { 1999,  9,  6 },  { 1999,  9,  7 },  { 1999,  9,  8 },  { 1999,  9,  9 },
  { 1999,  9, 10 },  { 1999,  9, 11 },  { 1999,  9, 12 },  { 1999,  9, 13 },
  { 1999,  9, 14 },  { 1999,  9, 15 },  { 1999,  9, 16 },  { 1999,  9, 17 },
  { 1999,  9, 18 },  { 1999,  9, 19 },  { 1999,  9, 20 },  { 1999,  9, 21 },
  { 1999,  9, 22 },  { 1999,  9, 23 },  { 1999,  9, 24 },  { 1999,  9, 25 },
  { 1999,  9, 26 },  { 1999,  9, 27 },  { 1999,  9, 28 },  { 1999,  9, 29 },
  { 1999,  9, 30 },  { 1999, 10,  1 },  { 1999, 10,  2 },  { 1999, 10,  3 },
  { 1999, 10,  4 },  { 1999, 10,  5 },  { 1999, 10,  6 },  { 1999, 10,  7 },
  { 1999, 10,  8 },  { 1999, 10,  9 },  { 1999, 10, 10 },  { 1999, 10, 11 },
  { 1999, 10, 12 },  { 1999, 10, 13 },  { 1999, 10, 14 },  { 1999, 10, 15 },
  { 1999, 10, 16 },  { 1999, 10, 17 },  { 1999, 10, 18 },  { 1999, 10, 19 },
  { 1999, 10, 20 },  { 1999, 10, 21 },  { 1999, 10, 22 },  { 1999, 10, 23 },
  { 1999, 10, 24 },  { 1999, 10, 25 },  { 1999, 10, 26 },  { 1999, 10, 27 },
  { 1999, 10, 28 },  { 1999, 10, 29 },  { 1999, 10, 30 },  { 1999, 10, 31 },
  { 1999, 11,  1 },  { 1999, 11,  2 },  { 1999, 11,  3 },  { 1999, 11,  4 },
  { 1999, 11,  5 },  { 1999, 11,  6 },  { 1999, 11,  7 },  { 1999, 11,  8 },
  { 1999, 11,  9 },  { 1999, 11, 10 },  { 1999, 11, 11 },  { 1999, 11, 12 },
  { 1999, 11, 13 },  { 1999, 11, 14 },  { 1999, 11, 15 },  { 1999, 11, 16 },
  { 1999, 11, 17 },  { 1999, 11, 18 },  { 1999, 11, 19 },  { 1999, 11, 20 },
  { 1999, 11, 21 },  { 1999, 11, 22 },  { 1999, 11, 23 },  { 1999, 11, 24 },
  { 1999, 11, 25 },  { 1999, 11, 26 },  { 1999, 11, 27 },  { 1999, 11, 28 },
  { 1999, 11, 29 },  { 1999, 11, 30 },  { 1999, 12,  1 },  { 1999, 12,  2 },
  { 1999, 12,  3 },  { 1999, 12,  4 },  { 1999, 12,  5 },  { 1999, 12,  6 },
  { 1999, 12,  7 },  { 1999, 12,  8 },  { 1999, 12,  9 },  { 1999, 12, 10 },
  { 1999, 12, 11 },  { 1999, 12, 12 },  { 1999, 12, 13 },  { 1999, 12, 14 },
  { 1999, 12, 15 },  { 1999, 12, 16 },  { 1999, 12, 17 },  { 1999, 12, 18 },
  { 1999, 12, 19 },  { 1999, 12, 20 },  { 1999, 12, 21 },  { 1999, 12, 22 },
  { 1999, 12, 23 },  { 1999, 12, 24 },  { 1999, 12, 25 },  { 1999, 12, 26 },
  { 1999, 12, 27 },  { 1999, 12, 28 },  { 1999, 12, 29 },  { 1999, 12, 30 },
  { 1999, 12, 31 },  { 2000,  1,  1 },  { 2000,  1,  2 },  { 2000,  1,  3 },
  { 2000,  1,  4 },  { 2000,  1,  5 },  { 2000,  1,  6 },  { 2000,  1,  7 },
  { 2000,  1,  8 },  { 2000,  1,  9 },  { 2000,  1, 10 },  { 2000,  1, 11 },
  { 2000,  1, 12 },  { 2000,  1, 13 },  { 2000,  1, 14 },  { 2000,  1, 15 },
  { 2000,  1, 16 },  { 2000,  1, 17 },  { 2000,  1, 18 },  { 2000,  1, 19 },
  { 2000,  1, 20 },  { 2000,  1, 21 },  { 2000,  1, 22 },  { 2000,  1, 23 },
  { 2000,  1, 24 },  { 2000,  1, 25 },  { 2000,  1, 26 },  { 2000,  1, 27 },
  { 2000,  1, 28 },  { 2000,  1, 29 },  { 2000,  1, 30 },  { 2000,  1, 31 },
  { 2000,  2,  1 },  { 2000,  2,  2 },  { 2000,  2,  3 },  { 2000,  2,  4 },
  { 2000,  2,  5 },  { 2000,  2,  6 },  { 2000,  2,  7 },  { 2000,  2,  8 },
  { 2000,  2,  9 },  { 2000,  2, 10 },  { 2000,  2, 11 },  { 2000,  2, 12 },
  { 2000,  2, 13 },  { 2000,  2, 14 },  { 2000,  2, 15 },  { 2000,  2, 16 },
  { 2000,  2, 17 },  { 2000,  2, 18 },  { 2000,  2, 19 },  { 2000,  2, 20 },
  { 2000,  2, 21 },  { 2000,  2, 22 },  { 2000,  2, 23 },  { 2000,  2, 24 },
  { 2000,  2, 25 },  { 2000,  2, 26 },  { 2000,  2, 27 },  { 2000,  2, 28 },
  { 2000,  2, 29 },  { 2000,  3,  1 },  { 2000,  3,  2 },  { 2000,  3,  3 },
  { 2000,  3,  4 },  { 2000,  3,  5 },  { 2000,  3,  6 },  { 2000,  3,  7 },
  { 2000,  3,  8 },  { 2000,  3,  9 },  { 2000,  3, 10 },  { 2000,  3, 11 },
  { 2000,  3, 12 },  { 2000,  3, 13 },  { 2000,  3, 14 },  { 2000,  3, 15 },
  { 2000,  3, 16 },  { 2000,  3, 17 },  { 2000,  3, 18 },  { 2000,  3, 19 },
  { 2000,  3, 20 },  { 2000,  3, 21 },  { 2000,  3, 22 },  { 2000,  3, 23 },
  { 2000,  3, 24 },  { 2000,  3, 25 },  { 2000,  3, 26 },  { 2000,  3, 27 },
  { 2000,  3, 28 },  { 2000,  3, 29 },  { 2000,  3, 30 },  { 2000,  3, 31 },
  { 2000,  4,  1 },  { 2000,  4,  2 },  { 2000,  4,  3 },  { 2000,  4,  4 },
  { 2000,  4,  5 },  { 2000,  4,  6 },  { 2000,  4,  7 },  { 2000,  4,  8 },
  { 2000,  4,  9 },  { 2000,  4, 10 },  { 2000,  4, 11 },  { 2000,  4, 12 },
  { 2000,  4, 13 },  { 2000,  4, 14 },  { 2000,  4, 15 },  { 2000,  4, 16 },
  { 2000,  4, 17 },  { 2000,  4, 18 },  { 2000,  4, 19 },  { 2000,  4, 20 },
  { 2000,  4, 21 },  { 2000,  4, 22 },  { 2000,  4, 23 },  { 2000,  4, 24 },
  { 2000,  4, 25 },  { 2000,  4, 26 },  { 2000,  4, 27 },  { 2000,  4, 28 },
  { 2000,  4, 29 },  { 2000,  4, 30 },  { 2000,  5,  1 },  { 2000,  5,  2 },
  { 2000,  5,  3 },  { 2000,  5,  4 },  { 2000,  5,  5 },  { 2000,  5,  6 },
  { 2000,  5,  7 },  { 2000,  5,  8 },  { 2000,  5,  9 },  { 2000,  5, 10 },
  { 2000,  5, 11 },  { 2000,  5, 12 },  { 2000,  5, 13 },  { 2000,  5, 14 },
  { 2000,  5, 15 },  { 2000,  5, 16 },  { 2000,  5, 17 },  { 2000,  5, 18 },
  { 2000,  5, 19 },  { 2000,  5, 20 },  { 2000,  5, 21 },  { 2000,  5, 22 },
  { 2000,  5, 23 },  { 2000,  5, 24 },  { 2000,  5, 25 },  { 2000,  5, 26 },
  { 2000,  5, 27 },  { 2000,  5, 28 },  { 2000,  5, 29 },  { 2000,  5, 30 },
  { 2000,  5, 31 },  { 2000,  6,  1 },  { 2000,  6,  2 },  { 2000,  6,  3 },
  { 2000,  6,  4 },  { 2000,  6,  5 },  { 2000,  6,  6 },  { 2000,  6,  7 },
  { 2000,  6,  8 },  { 2000,  6,  9 },  { 2000,  6, 10 },  { 2000,  6, 11 },
  { 2000,  6, 12 },  { 2000,  6, 13 },  { 2000,  6, 14 },  { 2000,  6, 15 },
  { 2000,  6, 16 },  { 2000,  6, 17 },  { 2000,  6, 18 },  { 2000,  6, 19 },
  { 2000,  6, 20 },  { 2000,  6, 21 },  { 2000,  6, 22 },  { 2000,  6, 23 },
  { 2000,  6, 24 },  { 2000,  6, 25 },  { 2000,  6, 26 },  { 2000,  6, 27 },
  { 2000,  6, 28 },  { 2000,  6, 29 },  { 2000,  6, 30 },  { 2000,  7,  1 },
  { 2000,  7,  2 },  { 2000,  7,  3 },  { 2000,  7,  4 },  { 2000,  7,  5 },
  { 2000,  7,  6 },  { 2000,  7,  7 },  { 2000,  7,  8 },  { 2000,  7,  9 },
  { 2000,  7, 10 },  { 2000,  7, 11 },  { 2000,  7, 12 },  { 2000,  7, 13 },
  { 2000,  7, 14 },  { 2000,  7, 15 },  { 2000,  7, 16 },  { 2000,  7, 17 },
  { 2000,  7, 18 },  { 2000,  7, 19 },  { 2000,  7, 20 },  { 2000,  7, 21 },
  { 2000,  7, 22 },  { 2000,  7, 23 },  { 2000,  7, 24 },  { 2000,  7, 25 },
  { 2000,  7, 26 },  { 2000,  7, 27 },  { 2000,  7, 28 },  { 2000,  7, 29 },
  { 2000,  7, 30 },  { 2000,  7, 31 },  { 2000,  8,  1 },  { 2000,  8,  2 },
  { 2000,  8,  3 },  { 2000,  8,  4 },  { 2000,  8,  5 },  { 2000,  8,  6 },
  { 2000,  8,  7 },  { 2000,  8,  8 },  { 2000,  8,  9 },  { 2000,  8, 10 },
  { 2000,  8, 11 },  { 2000,  8, 12 },  { 2000,  8, 13 },  { 2000,  8, 14 },
  { 2000,  8, 15 },  { 2000,  8, 16 },  { 2000,  8, 17 },  { 2000,  8, 18 },
  { 2000,  8, 19 },  { 2000,  8, 20 },  { 2000,  8, 21 },  { 2000,  8, 22 },
  { 2000,  8, 23 },  { 2000,  8, 24 },  { 2000,  8, 25 },  { 2000,  8, 26 },
  { 2000,  8, 27 },  { 2000,  8, 28 },  { 2000,  8, 29 },  { 2000,  8, 30 },
  { 2000,  8, 31 },  { 2000,  9,  1 },  { 2000,  9,  2 },  { 2000,  9,  3 },
  { 2000,  9,  4 },  { 2000,  9,  5 },  { 2000,  9,  6 },  { 2000,  9,  7 },
  { 2000,  9,  8 },  { 2000,  9,  9 },  { 2000,  9, 10 },  { 2000,  9, 11 },
  { 2000,  9, 12 },  { 2000,  9, 13 },  { 2000,  9, 14 },  { 2000,  9, 15 },
  { 2000,  9, 16 },  { 2000,  9, 17 },  { 2000,  9, 18 },  { 2000,  9, 19 },
  { 2000,  9, 20 },  { 2000,  9, 21 },  { 2000,  9, 22 },  { 2000,  9, 23 },
  { 2000,  9, 24 },  { 2000,  9, 25 },  { 2000,  9, 26 },  { 2000,  9, 27 },
  { 2000,  9, 28 },  { 2000,  9, 29 },  { 2000,  9, 30 },  { 2000, 10,  1 },
  { 2000, 10,  2 },  { 2000, 10,  3 },  { 2000, 10,  4 },  { 2000, 10,  5 },
  { 2000, 10,  6 },  { 2000, 10,  7 },  { 2000, 10,  8 },  { 2000, 10,  9 },
  { 2000, 10, 10 },  { 2000, 10, 11 },  { 2000, 10, 12 },  { 2000, 10, 13 },
  { 2000, 10, 14 },  { 2000, 10, 15 },  { 2000, 10, 16 },  { 2000, 10, 17 },
  { 2000, 10, 18 },  { 2000, 10, 19 },  { 2000, 10, 20 },  { 2000, 10, 21 },
  { 2000, 10, 22 },  { 2000, 10, 23 },  { 2000, 10, 24 },  { 2000, 10, 25 },
  { 2000, 10, 26 },  { 2000, 10, 27 },  { 2000, 10, 28 },  { 2000, 10, 29 },
  { 2000, 10, 30 },  { 2000, 10, 31 },  { 2000, 11,  1 },  { 2000, 11,  2 },
  { 2000, 11,  3 },  { 2000, 11,  4 },  { 2000, 11,  5 },  { 2000, 11,  6 },
  { 2000, 11,  7 },  { 2000, 11,  8 },  { 2000, 11,  9 },  { 2000, 11, 10 },
  { 2000, 11, 11 },  { 2000, 11, 12 },  { 2000, 11, 13 },  { 2000, 11, 14 },
  { 2000, 11, 15 },  { 2000, 11, 16 },  { 2000, 11, 17 },  { 2000, 11, 18 },
  { 2000, 11, 19 },  { 2000, 11, 20 },  { 2000, 11, 21 },  { 2000, 11, 22 },
  { 2000, 11, 23 },  { 2000, 11, 24 },  { 2000, 11, 25 },  { 2000, 11, 26 },
  { 2000, 11, 27 },  { 2000, 11, 28 },  { 2000, 11, 29 },  { 2000, 11, 30 },
  { 2000, 12,  1 },  { 2000, 12,  2 },  { 2000, 12,  3 },  { 2000, 12,  4 },
  { 2000, 12,  5 },  { 2000, 12,  6 },  { 2000, 12,  7 },  { 2000, 12,  8 },
  { 2000, 12,  9 },  { 2000, 12, 10 },  { 2000, 12, 11 },  { 2000, 12, 12 },
  { 2000, 12, 13 },  { 2000, 12, 14 },  { 2000, 12, 15 },  { 2000, 12, 16 },
  { 2000, 12, 17 },  { 2000, 12, 18 },  { 2000, 12, 19 },  { 2000, 12, 20 },
  { 2000, 12, 21 },  { 2000, 12, 22 },  { 2000, 12, 23 },  { 2000, 12, 24 },
  { 2000, 12, 25 },  { 2000, 12, 26 },  { 2000, 12, 27 },  { 2000, 12, 28 },
  { 2000, 12, 29 },  { 2000, 12, 30 },  { 2000, 12, 31 },  { 2001,  1,  1 },
  { 2001,  1,  2 },  { 2001,  1,  3 },  { 2001,  1,  4 },  { 2001,  1,  5 },
  { 2001,  1,  6 },  { 2001,  1,  7 },  { 2001,  1,  8 },  { 2001,  1,  9 },
  { 2001,  1, 10 },  { 2001,  1, 11 },  { 2001,  1, 12 },  { 2001,  1, 13 },
  { 2001,  1, 14 },  { 2001,  1, 15 },  { 2001,  1, 16 },  { 2001,  1, 17 },
  { 2001,  1, 18 },  { 2001,  1, 19 },  { 2001,  1, 20 },  { 2001,  1, 21 },
  { 2001,  1, 22 },  { 2001,  1, 23 },  { 2001,  1, 24 },  { 2001,  1, 25 },
  { 2001,  1, 26 },  { 2001,  1, 27 },  { 2001,  1, 28 },  { 2001,  1, 29 },
  { 2001,  1, 30 },  { 2001,  1, 31 },  { 2001,  2,  1 },  { 2001,  2,  2 },
  { 2001,  2,  3 },  { 2001,  2,  4 },  { 2001,  2,  5 },  { 2001,  2,  6 },
  { 2001,  2,  7 },  { 2001,  2,  8 },  { 2001,  2,  9 },  { 2001,  2, 10 },
  { 2001,  2, 11 },  { 2001,  2, 12 },  { 2001,  2, 13 },  { 2001,  2, 14 },
  { 2001,  2, 15 },  { 2001,  2, 16 },  { 2001,  2, 17 },  { 2001,  2, 18 },
  { 2001,  2, 19 },  { 2001,  2, 20 },  { 2001,  2, 21 },  { 2001,  2, 22 },
  { 2001,  2, 23 },  { 2001,  2, 24 },  { 2001,  2, 25 },  { 2001,  2, 26 },
  { 2001,  2, 27 },  { 2001,  2, 28 },  { 2001,  3,  1 },  { 2001,  3,  2 },
  { 2001,  3,  3 },  { 2001,  3,  4 },  { 2001,  3,  5 },  { 2001,  3,  6 },
  { 2001,  3,  7 },  { 2001,  3,  8 },  { 2001,  3,  9 },  { 2001,  3, 10 },
  { 2001,  3, 11 },  { 2001,  3, 12 },  { 2001,  3, 13 },  { 2001,  3, 14 },
  { 2001,  3, 15 },  { 2001,  3, 16 },  { 2001,  3, 17 },  { 2001,  3, 18 },
  { 2001,  3, 19 },  { 2001,  3, 20 },  { 2001,  3, 21 },  { 2001,  3, 22 },
  { 2001,  3, 23 },  { 2001,  3, 24 },  { 2001,  3, 25 },  { 2001,  3, 26 },
  { 2001,  3, 27 },  { 2001,  3, 28 },  { 2001,  3, 29 },  { 2001,  3, 30 },
  { 2001,  3, 31 },  { 2001,  4,  1 },  { 2001,  4,  2 },  { 2001,  4,  3 },
  { 2001,  4,  4 },  { 2001,  4,  5 },  { 2001,  4,  6 },  { 2001,  4,  7 },
  { 2001,  4,  8 },  { 2001,  4,  9 },  { 2001,  4, 10 },  { 2001,  4, 11 },
  { 2001,  4, 12 },  { 2001,  4, 13 },  { 2001,  4, 14 },  { 2001,  4, 15 },
  { 2001,  4, 16 },  { 2001,  4, 17 },  { 2001,  4, 18 },  { 2001,  4, 19 },
  { 2001,  4, 20 },  { 2001,  4, 21 },  { 2001,  4, 22 },  { 2001,  4, 23 },
  { 2001,  4, 24 },  { 2001,  4, 25 },  { 2001,  4, 26 },  { 2001,  4, 27 },
  { 2001,  4, 28 },  { 2001,  4, 29 },  { 2001,  4, 30 },  { 2001,  5,  1 },
  { 2001,  5,  2 },  { 2001,  5,  3 },  { 2001,  5,  4 },  { 2001,  5,  5 },
  { 2001,  5,  6 },  { 2001,  5,  7 },  { 2001,  5,  8 },  { 2001,  5,  9 },
  { 2001,  5, 10 },  { 2001,  5, 11 },  { 2001,  5, 12 },  { 2001,  5, 13 },
  { 2001,  5, 14 },  { 2001,  5, 15 },  { 2001,  5, 16 },  { 2001,  5, 17 },
  { 2001,  5, 18 },  { 2001,  5, 19 },  { 2001,  5, 20 },  { 2001,  5, 21 },
  { 2001,  5, 22 },  { 2001,  5, 23 },  { 2001,  5, 24 },  { 2001,  5, 25 },
  { 2001,  5, 26 },  { 2001,  5, 27 },  { 2001,  5, 28 },  { 2001,  5, 29 },
  { 2001,  5, 30 },  { 2001,  5, 31 },  { 2001,  6,  1 },  { 2001,  6,  2 },
  { 2001,  6,  3 },  { 2001,  6,  4 },  { 2001,  6,  5 },  { 2001,  6,  6 },
  { 2001,  6,  7 },  { 2001,  6,  8 },  { 2001,  6,  9 },  { 2001,  6, 10 },
  { 2001,  6, 11 },  { 2001,  6, 12 },  { 2001,  6, 13 },  { 2001,  6, 14 },
  { 2001,  6, 15 },  { 2001,  6, 16 },  { 2001,  6, 17 },  { 2001,  6, 18 },
  { 2001,  6, 19 },  { 2001,  6, 20 },  { 2001,  6, 21 },  { 2001,  6, 22 },
  { 2001,  6, 23 },  { 2001,  6, 24 },  { 2001,  6, 25 },  { 2001,  6, 26 },
  { 2001,  6, 27 },  { 2001,  6, 28 },  { 2001,  6, 29 },  { 2001,  6, 30 },
  { 2001,  7,  1 },  { 2001,  7,  2 },  { 2001,  7,  3 },  { 2001,  7,  4 },
  { 2001,  7,  5 },  { 2001,  7,  6 },  { 2001,  7,  7 },  { 2001,  7,  8 },
  { 2001,  7,  9 },  { 2001,  7, 10 },  { 2001,  7, 11 },  { 2001,  7, 12 },
  { 2001,  7, 13 },  { 2001,  7, 14 },  { 2001,  7, 15 },  { 2001,  7, 16 },
  { 2001,  7, 17 },  { 2001,  7, 18 },  { 2001,  7, 19 },  { 2001,  7, 20 },
  { 2001,  7, 21 },  { 2001,  7, 22 },  { 2001,  7, 23 },  { 2001,  7, 24 },
  { 2001,  7, 25 },  { 2001,  7, 26 },  { 2001,  7, 27 },  { 2001,  7, 28 },
  { 2001,  7, 29 },  { 2001,  7, 30 },  { 2001,  7, 31 },  { 2001,  8,  1 },
  { 2001,  8,  2 },  { 2001,  8,  3 },  { 2001,  8,  4 },  { 2001,  8,  5 },
  { 2001,  8,  6 },  { 2001,  8,  7 },  { 2001,  8,  8 },  { 2001,  8,  9 },
  { 2001,  8, 10 },  { 2001,  8, 11 },  { 2001,  8, 12 },  { 2001,  8, 13 },
  { 2001,  8, 14 },  { 2001,  8, 15 },  { 2001,  8, 16 },  { 2001,  8, 17 },
  { 2001,  8, 18 },  { 2001,  8, 19 },  { 2001,  8, 20 },  { 2001,  8, 21 },
  { 2001,  8, 22 },  { 2001,  8, 23 },  { 2001,  8, 24 },  { 2001,  8, 25 },
  { 2001,  8, 26 },  { 2001,  8, 27 },  { 2001,  8, 28 },  { 2001,  8, 29 },
  { 2001,  8, 30 },  { 2001,  8, 31 },  { 2001,  9,  1 },  { 2001,  9,  2 },
  { 2001,  9,  3 },  { 2001,  9,  4 },  { 2001,  9,  5 },  { 2001,  9,  6 },
  { 2001,  9,  7 },  { 2001,  9,  8 },  { 2001,  9,  9 },  { 2001,  9, 10 },
  { 2001,  9, 11 },  { 2001,  9, 12 },  { 2001,  9, 13 },  { 2001,  9, 14 },
  { 2001,  9, 15 },  { 2001,  9, 16 },  { 2001,  9, 17 },  { 2001,  9, 18 },
  { 2001,  9, 19 },  { 2001,  9, 20 },  { 2001,  9, 21 },  { 2001,  9, 22 },
  { 2001,  9, 23 },  { 2001,  9, 24 },  { 2001,  9, 25 },  { 2001,  9, 26 },
  { 2001,  9, 27 },  { 2001,  9, 28 },  { 2001,  9, 29 },  { 2001,  9, 30 },
  { 2001, 10,  1 },  { 2001, 10,  2 },  { 2001, 10,  3 },  { 2001, 10,  4 },
  { 2001, 10,  5 },  { 2001, 10,  6 },  { 2001, 10,  7 },  { 2001, 10,  8 },
  { 2001, 10,  9 },  { 2001, 10, 10 },  { 2001, 10, 11 },  { 2001, 10, 12 },
  { 2001, 10, 13 },  { 2001, 10, 14 },  { 2001, 10, 15 },  { 2001, 10, 16 },
  { 2001, 10, 17 },  { 2001, 10, 18 },  { 2001, 10, 19 },  { 2001, 10, 20 },
  { 2001, 10, 21 },  { 2001, 10, 22 },  { 2001, 10, 23 },  { 2001, 10, 24 },
  { 2001, 10, 25 },  { 2001, 10, 26 },  { 2001, 10, 27 },  { 2001, 10, 28 },
  { 2001, 10, 29 },  { 2001, 10, 30 },  { 2001, 10, 31 },  { 2001, 11,  1 },
  { 2001, 11,  2 },  { 2001, 11,  3 },  { 2001, 11,  4 },  { 2001, 11,  5 },
  { 2001, 11,  6 },  { 2001, 11,  7 },  { 2001, 11,  8 },  { 2001, 11,  9 },
  { 2001, 11, 10 },  { 2001, 11, 11 },  { 2001, 11, 12 },  { 2001, 11, 13 },
  { 2001, 11, 14 },  { 2001, 11, 15 },  { 2001, 11, 16 },  { 2001, 11, 17 },
  { 2001, 11, 18 },  { 2001, 11, 19 },  { 2001, 11, 20 },  { 2001, 11, 21 },
  { 2001, 11, 22 },  { 2001, 11, 23 },  { 2001, 11, 24 },  { 2001, 11, 25 },
  { 2001, 11, 26 },  { 2001, 11, 27 },  { 2001, 11, 28 },  { 2001, 11, 29 },
  { 2001, 11, 30 },  { 2001, 12,  1 },  { 2001, 12,  2 },  { 2001, 12,  3 },
  { 2001, 12,  4 },  { 2001, 12,  5 },  { 2001, 12,  6 },  { 2001, 12,  7 },
  { 2001, 12,  8 },  { 2001, 12,  9 },  { 2001, 12, 10 },  { 2001, 12, 11 },
  { 2001, 12, 12 },  { 2001, 12, 13 },  { 2001, 12, 14 },  { 2001, 12, 15 },
  { 2001, 12, 16 },  { 2001, 12, 17 },  { 2001, 12, 18 },  { 2001, 12, 19 },
  { 2001, 12, 20 },  { 2001, 12, 21 },  { 2001, 12, 22 },  { 2001, 12, 23 },
  { 2001, 12, 24 },  { 2001, 12, 25 },  { 2001, 12, 26 },  { 2001, 12, 27 },
  { 2001, 12, 28 },  { 2001, 12, 29 },  { 2001, 12, 30 },  { 2001, 12, 31 },
  { 2002,  1,  1 },  { 2002,  1,  2 },  { 2002,  1,  3 },  { 2002,  1,  4 },
  { 2002,  1,  5 },  { 2002,  1,  6 },  { 2002,  1,  7 },  { 2002,  1,  8 },
  { 2002,  1,  9 },  { 2002,  1, 10 },  { 2002,  1, 11 },  { 2002,  1, 12 },
  { 2002,  1, 13 },  { 2002,  1, 14 },  { 2002,  1, 15 },  { 2002,  1, 16 },
  { 2002,  1, 17 },  { 2002,  1, 18 },  { 2002,  1, 19 },  { 2002,  1, 20 },
  { 2002,  1, 21 },  { 2002,  1, 22 },  { 2002,  1, 23 },  { 2002,  1, 24 },
  { 2002,  1, 25 },  { 2002,  1, 26 },  { 2002,  1, 27 },  { 2002,  1, 28 },
  { 2002,  1, 29 },  { 2002,  1, 30 },  { 2002,  1, 31 },  { 2002,  2,  1 },
  { 2002,  2,  2 },  { 2002,  2,  3 },  { 2002,  2,  4 },  { 2002,  2,  5 },
  { 2002,  2,  6 },  { 2002,  2,  7 },  { 2002,  2,  8 },  { 2002,  2,  9 },
  { 2002,  2, 10 },  { 2002,  2, 11 },  { 2002,  2, 12 },  { 2002,  2, 13 },
  { 2002,  2, 14 },  { 2002,  2, 15 },  { 2002,  2, 16 },  { 2002,  2, 17 },
  { 2002,  2, 18 },  { 2002,  2, 19 },  { 2002,  2, 20 },  { 2002,  2, 21 },
  { 2002,  2, 22 },  { 2002,  2, 23 },  { 2002,  2, 24 },  { 2002,  2, 25 },
  { 2002,  2, 26 },  { 2002,  2, 27 },  { 2002,  2, 28 },  { 2002,  3,  1 },
  { 2002,  3,  2 },  { 2002,  3,  3 },  { 2002,  3,  4 },  { 2002,  3,  5 },
  { 2002,  3,  6 },  { 2002,  3,  7 },  { 2002,  3,  8 },  { 2002,  3,  9 },
  { 2002,  3, 10 },  { 2002,  3, 11 },  { 2002,  3, 12 },  { 2002,  3, 13 },
  { 2002,  3, 14 },  { 2002,  3, 15 },  { 2002,  3, 16 },  { 2002,  3, 17 },
  { 2002,  3, 18 },  { 2002,  3, 19 },  { 2002,  3, 20 },  { 2002,  3, 21 },
  { 2002,  3, 22 },  { 2002,  3, 23 },  { 2002,  3, 24 },  { 2002,  3, 25 },
  { 2002,  3, 26 },  { 2002,  3, 27 },  { 2002,  3, 28 },  { 2002,  3, 29 },
  { 2002,  3, 30 },  { 2002,  3, 31 },  { 2002,  4,  1 },  { 2002,  4,  2 },
  { 2002,  4,  3 },  { 2002,  4,  4 },  { 2002,  4,  5 },  { 2002,  4,  6 },
  { 2002,  4,  7 },  { 2002,  4,  8 },  { 2002,  4,  9 },  { 2002,  4, 10 },
  { 2002,  4, 11 },  { 2002,  4, 12 },  { 2002,  4, 13 },  { 2002,  4, 14 },
  { 2002,  4, 15 },  { 2002,  4, 16 },  { 2002,  4, 17 },  { 2002,  4, 18 },
  { 2002,  4, 19 },  { 2002,  4, 20 },  { 2002,  4, 21 },  { 2002,  4, 22 },
  { 2002,  4, 23 },  { 2002,  4, 24 },  { 2002,  4, 25 },  { 2002,  4, 26 },
  { 2002,  4, 27 },  { 2002,  4, 28 },  { 2002,  4, 29 },  { 2002,  4, 30 },
  { 2002,  5,  1 },  { 2002,  5,  2 },  { 2002,  5,  3 },  { 2002,  5,  4 },
  { 2002,  5,  5 },  { 2002,  5,  6 },  { 2002,  5,  7 },  { 2002,  5,  8 },
  { 2002,  5,  9 },  { 2002,  5, 10 },  { 2002,  5, 11 },  { 2002,  5, 12 },
  { 2002,  5, 13 },  { 2002,  5, 14 },  { 2002,  5, 15 },  { 2002,  5, 16 },
  { 2002,  5, 17 },  { 2002,  5, 18 },  { 2002,  5, 19 },  { 2002,  5, 20 },
  { 2002,  5, 21 },  { 2002,  5, 22 },  { 2002,  5, 23 },  { 2002,  5, 24 },
  { 2002,  5, 25 },  { 2002,  5, 26 },  { 2002,  5, 27 },  { 2002,  5, 28 },
  { 2002,  5, 29 },  { 2002,  5, 30 },  { 2002,  5, 31 },  { 2002,  6,  1 },
  { 2002,  6,  2 },  { 2002,  6,  3 },  { 2002,  6,  4 },  { 2002,  6,  5 },
  { 2002,  6,  6 },  { 2002,  6,  7 },  { 2002,  6,  8 },  { 2002,  6,  9 },
  { 2002,  6, 10 },  { 2002,  6, 11 },  { 2002,  6, 12 },  { 2002,  6, 13 },
  { 2002,  6, 14 },  { 2002,  6, 15 },  { 2002,  6, 16 },  { 2002,  6, 17 },
  { 2002,  6, 18 },  { 2002,  6, 19 },  { 2002,  6, 20 },  { 2002,  6, 21 },
  { 2002,  6, 22 },  { 2002,  6, 23 },  { 2002,  6, 24 },  { 2002,  6, 25 },
  { 2002,  6, 26 },  { 2002,  6, 27 },  { 2002,  6, 28 },  { 2002,  6, 29 },
  { 2002,  6, 30 },  { 2002,  7,  1 },  { 2002,  7,  2 },  { 2002,  7,  3 },
  { 2002,  7,  4 },  { 2002,  7,  5 },  { 2002,  7,  6 },  { 2002,  7,  7 },
  { 2002,  7,  8 },  { 2002,  7,  9 },  { 2002,  7, 10 },  { 2002,  7, 11 },
  { 2002,  7, 12 },  { 2002,  7, 13 },  { 2002,  7, 14 },  { 2002,  7, 15 },
  { 2002,  7, 16 },  { 2002,  7, 17 },  { 2002,  7, 18 },  { 2002,  7, 19 },
  { 2002,  7, 20 },  { 2002,  7, 21 },  { 2002,  7, 22 },  { 2002,  7, 23 },
  { 2002,  7, 24 },  { 2002,  7, 25 },  { 2002,  7, 26 },  { 2002,  7, 27 },
  { 2002,  7, 28 },  { 2002,  7, 29 },  { 2002,  7, 30 },  { 2002,  7, 31 },
  { 2002,  8,  1 },  { 2002,  8,  2 },  { 2002,  8,  3 },  { 2002,  8,  4 },
  { 2002,  8,  5 },  { 2002,  8,  6 },  { 2002,  8,  7 },  { 2002,  8,  8 },
  { 2002,  8,  9 },  { 2002,  8, 10 },  { 2002,  8, 11 },  { 2002,  8, 12 },
  { 2002,  8, 13 },  { 2002,  8, 14 },  { 2002,  8, 15 },  { 2002,  8, 16 },
  { 2002,  8, 17 },  { 2002,  8, 18 },  { 2002,  8, 19 },  { 2002,  8, 20 },
  { 2002,  8, 21 },  { 2002,  8, 22 },  { 2002,  8, 23 },  { 2002,  8, 24 },
  { 2002,  8, 25 },  { 2002,  8, 26 },  { 2002,  8, 27 },  { 2002,  8, 28 },
  { 2002,  8, 29 },  { 2002,  8, 30 },  { 2002,  8, 31 },  { 2002,  9,  1 },
  { 2002,  9,  2 },  { 2002,  9,  3 },  { 2002,  9,  4 },  { 2002,  9,  5 },
  { 2002,  9,  6 },  { 2002,  9,  7 },  { 2002,  9,  8 },  { 2002,  9,  9 },
  { 2002,  9, 10 },  { 2002,  9, 11 },  { 2002,  9, 12 },  { 2002,  9, 13 },
  { 2002,  9, 14 },  { 2002,  9, 15 },  { 2002,  9, 16 },  { 2002,  9, 17 },
  { 2002,  9, 18 },  { 2002,  9, 19 },  { 2002,  9, 20 },  { 2002,  9, 21 },
  { 2002,  9, 22 },  { 2002,  9, 23 },  { 2002,  9, 24 },  { 2002,  9, 25 },
  { 2002,  9, 26 },  { 2002,  9, 27 },  { 2002,  9, 28 },  { 2002,  9, 29 },
  { 2002,  9, 30 },  { 2002, 10,  1 },  { 2002, 10,  2 },  { 2002, 10,  3 },
  { 2002, 10,  4 },  { 2002, 10,  5 },  { 2002, 10,  6 },  { 2002, 10,  7 },
  { 2002, 10,  8 },  { 2002, 10,  9 },  { 2002, 10, 10 },  { 2002, 10, 11 },
  { 2002, 10, 12 },  { 2002, 10, 13 },  { 2002, 10, 14 },  { 2002, 10, 15 },
  { 2002, 10, 16 },  { 2002, 10, 17 },  { 2002, 10, 18 },  { 2002, 10, 19 },
  { 2002, 10, 20 },  { 2002, 10, 21 },  { 2002, 10, 22 },  { 2002, 10, 23 },
  { 2002, 10, 24 },  { 2002, 10, 25 },  { 2002, 10, 26 },  { 2002, 10, 27 },
  { 2002, 10, 28 },  { 2002, 10, 29 },  { 2002, 10, 30 },  { 2002, 10, 31 },
  { 2002, 11,  1 },  { 2002, 11,  2 },  { 2002, 11,  3 },  { 2002, 11,  4 },
  { 2002, 11,  5 },  { 2002, 11,  6 },  { 2002, 11,  7 },  { 2002, 11,  8 },
  { 2002, 11,  9 },  { 2002, 11, 10 },  { 2002, 11, 11 },  { 2002, 11, 12 },
  { 2002, 11, 13 },  { 2002, 11, 14 },  { 2002, 11, 15 },  { 2002, 11, 16 },
  { 2002, 11, 17 },  { 2002, 11, 18 },  { 2002, 11, 19 },  { 2002, 11, 20 },
  { 2002, 11, 21 },  { 2002, 11, 22 },  { 2002, 11, 23 },  { 2002, 11, 24 },
  { 2002, 11, 25 },  { 2002, 11, 26 },  { 2002, 11, 27 },  { 2002, 11, 28 },
  { 2002, 11, 29 },  { 2002, 11, 30 },  { 2002, 12,  1 },  { 2002, 12,  2 },
  { 2002, 12,  3 },  { 2002, 12,  4 },  { 2002, 12,  5 },  { 2002, 12,  6 },
  { 2002, 12,  7 },  { 2002, 12,  8 },  { 2002, 12,  9 },  { 2002, 12, 10 },
  { 2002, 12, 11 },  { 2002, 12, 12 },  { 2002, 12, 13 },  { 2002, 12, 14 },
  { 2002, 12, 15 },  { 2002, 12, 16 },  { 2002, 12, 17 },  { 2002, 12, 18 },
  { 2002, 12, 19 },  { 2002, 12, 20 },  { 2002, 12, 21 },  { 2002, 12, 22 },
  { 2002, 12, 23 },  { 2002, 12, 24 },  { 2002, 12, 25 },  { 2002, 12, 26 },
  { 2002, 12, 27 },  { 2002, 12, 28 },  { 2002, 12, 29 },  { 2002, 12, 30 },
  { 2002, 12, 31 },  { 2003,  1,  1 },  { 2003,  1,  2 },  { 2003,  1,  3 },
  { 2003,  1,  4 },  { 2003,  1,  5 },  { 2003,  1,  6 },  { 2003,  1,  7 },
  { 2003,  1,  8 },  { 2003,  1,  9 },  { 2003,  1, 10 },  { 2003,  1, 11 },
  { 2003,  1, 12 },  { 2003,  1, 13 },  { 2003,  1, 14 },  { 2003,  1, 15 },
  { 2003,  1, 16 },  { 2003,  1, 17 },  { 2003,  1, 18 },  { 2003,  1, 19 },
  { 2003,  1, 20 },  { 2003,  1, 21 },  { 2003,  1, 22 },  { 2003,  1, 23 },
  { 2003,  1, 24 },  { 2003,  1, 25 },  { 2003,  1, 26 },  { 2003,  1, 27 },
  { 2003,  1, 28 },  { 2003,  1, 29 },  { 2003,  1, 30 },  { 2003,  1, 31 },
  { 2003,  2,  1 },  { 2003,  2,  2 },  { 2003,  2,  3 },  { 2003,  2,  4 },
  { 2003,  2,  5 },  { 2003,  2,  6 },  { 2003,  2,  7 },  { 2003,  2,  8 },
  { 2003,  2,  9 },  { 2003,  2, 10 },  { 2003,  2, 11 },  { 2003,  2, 12 },
  { 2003,  2, 13 },  { 2003,  2, 14 },  { 2003,  2, 15 },  { 2003,  2, 16 },
  { 2003,  2, 17 },  { 2003,  2, 18 },  { 2003,  2, 19 },  { 2003,  2, 20 },
  { 2003,  2, 21 },  { 2003,  2, 22 },  { 2003,  2, 23 },  { 2003,  2, 24 },
  { 2003,  2, 25 },  { 2003,  2, 26 },  { 2003,  2, 27 },  { 2003,  2, 28 },
  { 2003,  3,  1 },  { 2003,  3,  2 },  { 2003,  3,  3 },  { 2003,  3,  4 },
  { 2003,  3,  5 },  { 2003,  3,  6 },  { 2003,  3,  7 },  { 2003,  3,  8 },
  { 2003,  3,  9 },  { 2003,  3, 10 },  { 2003,  3, 11 },  { 2003,  3, 12 },
  { 2003,  3, 13 },  { 2003,  3, 14 },  { 2003,  3, 15 },  { 2003,  3, 16 },
  { 2003,  3, 17 },  { 2003,  3, 18 },  { 2003,  3, 19 },  { 2003,  3, 20 },
  { 2003,  3, 21 },  { 2003,  3, 22 },  { 2003,  3, 23 },  { 2003,  3, 24 },
  { 2003,  3, 25 },  { 2003,  3, 26 },  { 2003,  3, 27 },  { 2003,  3, 28 },
  { 2003,  3, 29 },  { 2003,  3, 30 },  { 2003,  3, 31 },  { 2003,  4,  1 },
  { 2003,  4,  2 },  { 2003,  4,  3 },  { 2003,  4,  4 },  { 2003,  4,  5 },
  { 2003,  4,  6 },  { 2003,  4,  7 },  { 2003,  4,  8 },  { 2003,  4,  9 },
  { 2003,  4, 10 },  { 2003,  4, 11 },  { 2003,  4, 12 },  { 2003,  4, 13 },
  { 2003,  4, 14 },  { 2003,  4, 15 },  { 2003,  4, 16 },  { 2003,  4, 17 },
  { 2003,  4, 18 },  { 2003,  4, 19 },  { 2003,  4, 20 },  { 2003,  4, 21 },
  { 2003,  4, 22 },  { 2003,  4, 23 },  { 2003,  4, 24 },  { 2003,  4, 25 },
  { 2003,  4, 26 },  { 2003,  4, 27 },  { 2003,  4, 28 },  { 2003,  4, 29 },
  { 2003,  4, 30 },  { 2003,  5,  1 },  { 2003,  5,  2 },  { 2003,  5,  3 },
  { 2003,  5,  4 },  { 2003,  5,  5 },  { 2003,  5,  6 },  { 2003,  5,  7 },
  { 2003,  5,  8 },  { 2003,  5,  9 },  { 2003,  5, 10 },  { 2003,  5, 11 },
  { 2003,  5, 12 },  { 2003,  5, 13 },  { 2003,  5, 14 },  { 2003,  5, 15 },
  { 2003,  5, 16 },  { 2003,  5, 17 },  { 2003,  5, 18 },  { 2003,  5, 19 },
  { 2003,  5, 20 },  { 2003,  5, 21 },  { 2003,  5, 22 },  { 2003,  5, 23 },
  { 2003,  5, 24 },  { 2003,  5, 25 },  { 2003,  5, 26 },  { 2003,  5, 27 },
  { 2003,  5, 28 },  { 2003,  5, 29 },  { 2003,  5, 30 },  { 2003,  5, 31 },
  { 2003,  6,  1 },  { 2003,  6,  2 },  { 2003,  6,  3 },  { 2003,  6,  4 },
  { 2003,  6,  5 },  { 2003,  6,  6 },  { 2003,  6,  7 },  { 2003,  6,  8 },
  { 2003,  6,  9 },  { 2003,  6, 10 },  { 2003,  6, 11 },  { 2003,  6, 12 },
  { 2003,  6, 13 },  { 2003,  6, 14 },  { 2003,  6, 15 },  { 2003,  6, 16 },
  { 2003,  6, 17 },  { 2003,  6, 18 },  { 2003,  6, 19 },  { 2003,  6, 20 },
  { 2003,  6, 21 },  { 2003,  6, 22 },  { 2003,  6, 23 },  { 2003,  6, 24 },
  { 2003,  6, 25 },  { 2003,  6, 26 },  { 2003,  6, 27 },  { 2003,  6, 28 },
  { 2003,  6, 29 },  { 2003,  6, 30 },  { 2003,  7,  1 },  { 2003,  7,  2 },
  { 2003,  7,  3 },  { 2003,  7,  4 },  { 2003,  7,  5 },  { 2003,  7,  6 },
  { 2003,  7,  7 },  { 2003,  7,  8 },  { 2003,  7,  9 },  { 2003,  7, 10 },
  { 2003,  7, 11 },  { 2003,  7, 12 },  { 2003,  7, 13 },  { 2003,  7, 14 },
  { 2003,  7, 15 },  { 2003,  7, 16 },  { 2003,  7, 17 },  { 2003,  7, 18 },
  { 2003,  7, 19 },  { 2003,  7, 20 },  { 2003,  7, 21 },  { 2003,  7, 22 },
  { 2003,  7, 23 },  { 2003,  7, 24 },  { 2003,  7, 25 },  { 2003,  7, 26 },
  { 2003,  7, 27 },  { 2003,  7, 28 },  { 2003,  7, 29 },  { 2003,  7, 30 },
  { 2003,  7, 31 },  { 2003,  8,  1 },  { 2003,  8,  2 },  { 2003,  8,  3 },
  { 2003,  8,  4 },  { 2003,  8,  5 },  { 2003,  8,  6 },  { 2003,  8,  7 },
  { 2003,  8,  8 },  { 2003,  8,  9 },  { 2003,  8, 10 },  { 2003,  8, 11 },
  { 2003,  8, 12 },  { 2003,  8, 13 },  { 2003,  8, 14 },  { 2003,  8, 15 },
  { 2003,  8, 16 },  { 2003,  8, 17 },  { 2003,  8, 18 },  { 2003,  8, 19 },
  { 2003,  8, 20 },  { 2003,  8, 21 },  { 2003,  8, 22 },  { 2003,  8, 23 },
  { 2003,  8, 24 },  { 2003,  8, 25 },  { 2003,  8, 26 },  { 2003,  8, 27 },
  { 2003,  8, 28 },  { 2003,  8, 29 },  { 2003,  8, 30 },  { 2003,  8, 31 },
  { 2003,  9,  1 },  { 2003,  9,  2 },  { 2003,  9,  3 },  { 2003,  9,  4 },
  { 2003,  9,  5 },  { 2003,  9,  6 },  { 2003,  9,  7 },  { 2003,  9,  8 },
  { 2003,  9,  9 },  { 2003,  9, 10 },  { 2003,  9, 11 },  { 2003,  9, 12 },
  { 2003,  9, 13 },  { 2003,  9, 14 },  { 2003,  9, 15 },  { 2003,  9, 16 },
  { 2003,  9, 17 },  { 2003,  9, 18 },  { 2003,  9, 19 },  { 2003,  9, 20 },
  { 2003,  9, 21 },  { 2003,  9, 22 },  { 2003,  9, 23 },  { 2003,  9, 24 },
  { 2003,  9, 25 },  { 2003,  9, 26 },  { 2003,  9, 27 },  { 2003,  9, 28 },
  { 2003,  9, 29 },  { 2003,  9, 30 },  { 2003, 10,  1 },  { 2003, 10,  2 },
  { 2003, 10,  3 },  { 2003, 10,  4 },  { 2003, 10,  5 },  { 2003, 10,  6 },
  { 2003, 10,  7 },  { 2003, 10,  8 },  { 2003, 10,  9 },  { 2003, 10, 10 },
  { 2003, 10, 11 },  { 2003, 10, 12 },  { 2003, 10, 13 },  { 2003, 10, 14 },
  { 2003, 10, 15 },  { 2003, 10, 16 },  { 2003, 10, 17 },  { 2003, 10, 18 },
  { 2003, 10, 19 },  { 2003, 10, 20 },  { 2003, 10, 21 },  { 2003, 10, 22 },
  { 2003, 10, 23 },  { 2003, 10, 24 },  { 2003, 10, 25 },  { 2003, 10, 26 },
  { 2003, 10, 27 },  { 2003, 10, 28 },  { 2003, 10, 29 },  { 2003, 10, 30 },
  { 2003, 10, 31 },  { 2003, 11,  1 },  { 2003, 11,  2 },  { 2003, 11,  3 },
  { 2003, 11,  4 },  { 2003, 11,  5 },  { 2003, 11,  6 },  { 2003, 11,  7 },
  { 2003, 11,  8 },  { 2003, 11,  9 },  { 2003, 11, 10 },  { 2003, 11, 11 },
  { 2003, 11, 12 },  { 2003, 11, 13 },  { 2003, 11, 14 },  { 2003, 11, 15 },
  { 2003, 11, 16 },  { 2003, 11, 17 },  { 2003, 11, 18 },  { 2003, 11, 19 },
  { 2003, 11, 20 },  { 2003, 11, 21 },  { 2003, 11, 22 },  { 2003, 11, 23 },
  { 2003, 11, 24 },  { 2003, 11, 25 },  { 2003, 11, 26 },  { 2003, 11, 27 },
  { 2003, 11, 28 },  { 2003, 11, 29 },  { 2003, 11, 30 },  { 2003, 12,  1 },
  { 2003, 12,  2 },  { 2003, 12,  3 },  { 2003, 12,  4 },  { 2003, 12,  5 },
  { 2003, 12,  6 },  { 2003, 12,  7 },  { 2003, 12,  8 },  { 2003, 12,  9 },
  { 2003, 12, 10 },  { 2003, 12, 11 },  { 2003, 12, 12 },  { 2003, 12, 13 },
  { 2003, 12, 14 },  { 2003, 12, 15 },  { 2003, 12, 16 },  { 2003, 12, 17 },
  { 2003, 12, 18 },  { 2003, 12, 19 },  { 2003, 12, 20 },  { 2003, 12, 21 },
  { 2003, 12, 22 },  { 2003, 12, 23 },  { 2003, 12, 24 },  { 2003, 12, 25 },
  { 2003, 12, 26 },  { 2003, 12, 27 },  { 2003, 12, 28 },  { 2003, 12, 29 },
  { 2003, 12, 30 },  { 2003, 12, 31 },  { 2004,  1,  1 },  { 2004,  1,  2 },
  { 2004,  1,  3 },  { 2004,  1,  4 },  { 2004,  1,  5 },  { 2004,  1,  6 },
  { 2004,  1,  7 },  { 2004,  1,  8 },  { 2004,  1,  9 },  { 2004,  1, 10 },
  { 2004,  1, 11 },  { 2004,  1, 12 },  { 2004,  1, 13 },  { 2004,  1, 14 },
  { 2004,  1, 15 },  { 2004,  1, 16 },  { 2004,  1, 17 },  { 2004,  1, 18 },
  { 2004,  1, 19 },  { 2004,  1, 20 },  { 2004,  1, 21 },  { 2004,  1, 22 },
  { 2004,  1, 23 },  { 2004,  1, 24 },  { 2004,  1, 25 },  { 2004,  1, 26 },
  { 2004,  1, 27 },  { 2004,  1, 28 },  { 2004,  1, 29 },  { 2004,  1, 30 },
  { 2004,  1, 31 },  { 2004,  2,  1 },  { 2004,  2,  2 },  { 2004,  2,  3 },
  { 2004,  2,  4 },  { 2004,  2,  5 },  { 2004,  2,  6 },  { 2004,  2,  7 },
  { 2004,  2,  8 },  { 2004,  2,  9 },  { 2004,  2, 10 },  { 2004,  2, 11 },
  { 2004,  2, 12 },  { 2004,  2, 13 },  { 2004,  2, 14 },  { 2004,  2, 15 },
  { 2004,  2, 16 },  { 2004,  2, 17 },  { 2004,  2, 18 },  { 2004,  2, 19 },
  { 2004,  2, 20 },  { 2004,  2, 21 },  { 2004,  2, 22 },  { 2004,  2, 23 },
  { 2004,  2, 24 },  { 2004,  2, 25 },  { 2004,  2, 26 },  { 2004,  2, 27 },
  { 2004,  2, 28 },  { 2004,  2, 29 },  { 2004,  3,  1 },  { 2004,  3,  2 },
  { 2004,  3,  3 },  { 2004,  3,  4 },  { 2004,  3,  5 },  { 2004,  3,  6 },
  { 2004,  3,  7 },  { 2004,  3,  8 },  { 2004,  3,  9 },  { 2004,  3, 10 },
  { 2004,  3, 11 },  { 2004,  3, 12 },  { 2004,  3, 13 },  { 2004,  3, 14 },
  { 2004,  3, 15 },  { 2004,  3, 16 },  { 2004,  3, 17 },  { 2004,  3, 18 },
  { 2004,  3, 19 },  { 2004,  3, 20 },  { 2004,  3, 21 },  { 2004,  3, 22 },
  { 2004,  3, 23 },  { 2004,  3, 24 },  { 2004,  3, 25 },  { 2004,  3, 26 },
  { 2004,  3, 27 },  { 2004,  3, 28 },  { 2004,  3, 29 },  { 2004,  3, 30 },
  { 2004,  3, 31 },  { 2004,  4,  1 },  { 2004,  4,  2 },  { 2004,  4,  3 },
  { 2004,  4,  4 },  { 2004,  4,  5 },  { 2004,  4,  6 },  { 2004,  4,  7 },
  { 2004,  4,  8 },  { 2004,  4,  9 },  { 2004,  4, 10 },  { 2004,  4, 11 },
  { 2004,  4, 12 },  { 2004,  4, 13 },  { 2004,  4, 14 },  { 2004,  4, 15 },
  { 2004,  4, 16 },  { 2004,  4, 17 },  { 2004,  4, 18 },  { 2004,  4, 19 },
  { 2004,  4, 20 },  { 2004,  4, 21 },  { 2004,  4, 22 },  { 2004,  4, 23 },
  { 2004,  4, 24 },  { 2004,  4, 25 },  { 2004,  4, 26 },  { 2004,  4, 27 },
  { 2004,  4, 28 },  { 2004,  4, 29 },  { 2004,  4, 30 },  { 2004,  5,  1 },
  { 2004,  5,  2 },  { 2004,  5,  3 },  { 2004,  5,  4 },  { 2004,  5,  5 },
  { 2004,  5,  6 },  { 2004,  5,  7 },  { 2004,  5,  8 },  { 2004,  5,  9 },
  { 2004,  5, 10 },  { 2004,  5, 11 },  { 2004,  5, 12 },  { 2004,  5, 13 },
  { 2004,  5, 14 },  { 2004,  5, 15 },  { 2004,  5, 16 },  { 2004,  5, 17 },
  { 2004,  5, 18 },  { 2004,  5, 19 },  { 2004,  5, 20 },  { 2004,  5, 21 },
  { 2004,  5, 22 },  { 2004,  5, 23 },  { 2004,  5, 24 },  { 2004,  5, 25 },
  { 2004,  5, 26 },  { 2004,  5, 27 },  { 2004,  5, 28 },  { 2004,  5, 29 },
  { 2004,  5, 30 },  { 2004,  5, 31 },  { 2004,  6,  1 },  { 2004,  6,  2 },
  { 2004,  6,  3 },  { 2004,  6,  4 },  { 2004,  6,  5 },  { 2004,  6,  6 },
  { 2004,  6,  7 },  { 2004,  6,  8 },  { 2004,  6,  9 },  { 2004,  6, 10 },
  { 2004,  6, 11 },  { 2004,  6, 12 },  { 2004,  6, 13 },  { 2004,  6, 14 },
  { 2004,  6, 15 },  { 2004,  6, 16 },  { 2004,  6, 17 },  { 2004,  6, 18 },
  { 2004,  6, 19 },  { 2004,  6, 20 },  { 2004,  6, 21 },  { 2004,  6, 22 },
  { 2004,  6, 23 },  { 2004,  6, 24 },  { 2004,  6, 25 },  { 2004,  6, 26 },
  { 2004,  6, 27 },  { 2004,  6, 28 },  { 2004,  6, 29 },  { 2004,  6, 30 },
  { 2004,  7,  1 },  { 2004,  7,  2 },  { 2004,  7,  3 },  { 2004,  7,  4 },
  { 2004,  7,  5 },  { 2004,  7,  6 },  { 2004,  7,  7 },  { 2004,  7,  8 },
  { 2004,  7,  9 },  { 2004,  7, 10 },  { 2004,  7, 11 },  { 2004,  7, 12 },
  { 2004,  7, 13 },  { 2004,  7, 14 },  { 2004,  7, 15 },  { 2004,  7, 16 },
  { 2004,  7, 17 },  { 2004,  7, 18 },  { 2004,  7, 19 },  { 2004,  7, 20 },
  { 2004,  7, 21 },  { 2004,  7, 22 },  { 2004,  7, 23 },  { 2004,  7, 24 },
  { 2004,  7, 25 },  { 2004,  7, 26 },  { 2004,  7, 27 },  { 2004,  7, 28 },
  { 2004,  7, 29 },  { 2004,  7, 30 },  { 2004,  7, 31 },  { 2004,  8,  1 },
  { 2004,  8,  2 },  { 2004,  8,  3 },  { 2004,  8,  4 },  { 2004,  8,  5 },
  { 2004,  8,  6 },  { 2004,  8,  7 },  { 2004,  8,  8 },  { 2004,  8,  9 },
  { 2004,  8, 10 },  { 2004,  8, 11 },  { 2004,  8, 12 },  { 2004,  8, 13 },
  { 2004,  8, 14 },  { 2004,  8, 15 },  { 2004,  8, 16 },  { 2004,  8, 17 },
  { 2004,  8, 18 },  { 2004,  8, 19 },  { 2004,  8, 20 },  { 2004,  8, 21 },
  { 2004,  8, 22 },  { 2004,  8, 23 },  { 2004,  8, 24 },  { 2004,  8, 25 },
  { 2004,  8, 26 },  { 2004,  8, 27 },  { 2004,  8, 28 },  { 2004,  8, 29 },
  { 2004,  8, 30 },  { 2004,  8, 31 },  { 2004,  9,  1 },  { 2004,  9,  2 },
  { 2004,  9,  3 },  { 2004,  9,  4 },  { 2004,  9,  5 },  { 2004,  9,  6 },
  { 2004,  9,  7 },  { 2004,  9,  8 },  { 2004,  9,  9 },  { 2004,  9, 10 },
  { 2004,  9, 11 },  { 2004,  9, 12 },  { 2004,  9, 13 },  { 2004,  9, 14 },
  { 2004,  9, 15 },  { 2004,  9, 16 },  { 2004,  9, 17 },  { 2004,  9, 18 },
  { 2004,  9, 19 },  { 2004,  9, 20 },  { 2004,  9, 21 },  { 2004,  9, 22 },
  { 2004,  9, 23 },  { 2004,  9, 24 },  { 2004,  9, 25 },  { 2004,  9, 26 },
  { 2004,  9, 27 },  { 2004,  9, 28 },  { 2004,  9, 29 },  { 2004,  9, 30 },
  { 2004, 10,  1 },  { 2004, 10,  2 },  { 2004, 10,  3 },  { 2004, 10,  4 },
  { 2004, 10,  5 },  { 2004, 10,  6 },  { 2004, 10,  7 },  { 2004, 10,  8 },
  { 2004, 10,  9 },  { 2004, 10, 10 },  { 2004, 10, 11 },  { 2004, 10, 12 },
  { 2004, 10, 13 },  { 2004, 10, 14 },  { 2004, 10, 15 },  { 2004, 10, 16 },
  { 2004, 10, 17 },  { 2004, 10, 18 },  { 2004, 10, 19 },  { 2004, 10, 20 },
  { 2004, 10, 21 },  { 2004, 10, 22 },  { 2004, 10, 23 },  { 2004, 10, 24 },
  { 2004, 10, 25 },  { 2004, 10, 26 },  { 2004, 10, 27 },  { 2004, 10, 28 },
  { 2004, 10, 29 },  { 2004, 10, 30 },  { 2004, 10, 31 },  { 2004, 11,  1 },
  { 2004, 11,  2 },  { 2004, 11,  3 },  { 2004, 11,  4 },  { 2004, 11,  5 },
  { 2004, 11,  6 },  { 2004, 11,  7 },  { 2004, 11,  8 },  { 2004, 11,  9 },
  { 2004, 11, 10 },  { 2004, 11, 11 },  { 2004, 11, 12 },  { 2004, 11, 13 },
  { 2004, 11, 14 },  { 2004, 11, 15 },  { 2004, 11, 16 },  { 2004, 11, 17 },
  { 2004, 11, 18 },  { 2004, 11, 19 },  { 2004, 11, 20 },  { 2004, 11, 21 },
  { 2004, 11, 22 },  { 2004, 11, 23 },  { 2004, 11, 24 },  { 2004, 11, 25 },
  { 2004, 11, 26 },  { 2004, 11, 27 },  { 2004, 11, 28 },  { 2004, 11, 29 },
  { 2004, 11, 30 },  { 2004, 12,  1 },  { 2004, 12,  2 },  { 2004, 12,  3 },
  { 2004, 12,  4 },  { 2004, 12,  5 },  { 2004, 12,  6 },  { 2004, 12,  7 },
  { 2004, 12,  8 },  { 2004, 12,  9 },  { 2004, 12, 10 },  { 2004, 12, 11 },
  { 2004, 12, 12 },  { 2004, 12, 13 },  { 2004, 12, 14 },  { 2004, 12, 15 },
  { 2004, 12, 16 },  { 2004, 12, 17 },  { 2004, 12, 18 },  { 2004, 12, 19 },
  { 2004, 12, 20 },  { 2004, 12, 21 },  { 2004, 12, 22 },  { 2004, 12, 23 },
  { 2004, 12, 24 },  { 2004, 12, 25 },  { 2004, 12, 26 },  { 2004, 12, 27 },
  { 2004, 12, 28 },  { 2004, 12, 29 },  { 2004, 12, 30 },  { 2004, 12, 31 },
  { 2005,  1,  1 },  { 2005,  1,  2 },  { 2005,  1,  3 },  { 2005,  1,  4 },
  { 2005,  1,  5 },  { 2005,  1,  6 },  { 2005,  1,  7 },  { 2005,  1,  8 },
  { 2005,  1,  9 },  { 2005,  1, 10 },  { 2005,  1, 11 },  { 2005,  1, 12 },
  { 2005,  1, 13 },  { 2005,  1, 14 },  { 2005,  1, 15 },  { 2005,  1, 16 },
  { 2005,  1, 17 },  { 2005,  1, 18 },  { 2005,  1, 19 },  { 2005,  1, 20 },
  { 2005,  1, 21 },  { 2005,  1, 22 },  { 2005,  1, 23 },  { 2005,  1, 24 },
  { 2005,  1, 25 },  { 2005,  1, 26 },  { 2005,  1, 27 },  { 2005,  1, 28 },
  { 2005,  1, 29 },  { 2005,  1, 30 },  { 2005,  1, 31 },  { 2005,  2,  1 },
  { 2005,  2,  2 },  { 2005,  2,  3 },  { 2005,  2,  4 },  { 2005,  2,  5 },
  { 2005,  2,  6 },  { 2005,  2,  7 },  { 2005,  2,  8 },  { 2005,  2,  9 },
  { 2005,  2, 10 },  { 2005,  2, 11 },  { 2005,  2, 12 },  { 2005,  2, 13 },
  { 2005,  2, 14 },  { 2005,  2, 15 },  { 2005,  2, 16 },  { 2005,  2, 17 },
  { 2005,  2, 18 },  { 2005,  2, 19 },  { 2005,  2, 20 },  { 2005,  2, 21 },
  { 2005,  2, 22 },  { 2005,  2, 23 },  { 2005,  2, 24 },  { 2005,  2, 25 },
  { 2005,  2, 26 },  { 2005,  2, 27 },  { 2005,  2, 28 },  { 2005,  3,  1 },
  { 2005,  3,  2 },  { 2005,  3,  3 },  { 2005,  3,  4 },  { 2005,  3,  5 },
  { 2005,  3,  6 },  { 2005,  3,  7 },  { 2005,  3,  8 },  { 2005,  3,  9 },
  { 2005,  3, 10 },  { 2005,  3, 11 },  { 2005,  3, 12 },  { 2005,  3, 13 },
  { 2005,  3, 14 },  { 2005,  3, 15 },  { 2005,  3, 16 },  { 2005,  3, 17 },
  { 2005,  3, 18 },  { 2005,  3, 19 },  { 2005,  3, 20 },  { 2005,  3, 21 },
  { 2005,  3, 22 },  { 2005,  3, 23 },  { 2005,  3, 24 },  { 2005,  3, 25 },
  { 2005,  3, 26 },  { 2005,  3, 27 },  { 2005,  3, 28 },  { 2005,  3, 29 },
  { 2005,  3, 30 },  { 2005,  3, 31 },  { 2005,  4,  1 },  { 2005,  4,  2 },
  { 2005,  4,  3 },  { 2005,  4,  4 },  { 2005,  4,  5 },  { 2005,  4,  6 },
  { 2005,  4,  7 },  { 2005,  4,  8 },  { 2005,  4,  9 },  { 2005,  4, 10 },
  { 2005,  4, 11 },  { 2005,  4, 12 },  { 2005,  4, 13 },  { 2005,  4, 14 },
  { 2005,  4, 15 },  { 2005,  4, 16 },  { 2005,  4, 17 },  { 2005,  4, 18 },
  { 2005,  4, 19 },  { 2005,  4, 20 },  { 2005,  4, 21 },  { 2005,  4, 22 },
  { 2005,  4, 23 },  { 2005,  4, 24 },  { 2005,  4, 25 },  { 2005,  4, 26 },
  { 2005,  4, 27 },  { 2005,  4, 28 },  { 2005,  4, 29 },  { 2005,  4, 30 },
  { 2005,  5,  1 },  { 2005,  5,  2 },  { 2005,  5,  3 },  { 2005,  5,  4 },
  { 2005,  5,  5 },  { 2005,  5,  6 },  { 2005,  5,  7 },  { 2005,  5,  8 },
  { 2005,  5,  9 },  { 2005,  5, 10 },  { 2005,  5, 11 },  { 2005,  5, 12 },
  { 2005,  5, 13 },  { 2005,  5, 14 },  { 2005,  5, 15 },  { 2005,  5, 16 },
  { 2005,  5, 17 },  { 2005,  5, 18 },  { 2005,  5, 19 },  { 2005,  5, 20 },
  { 2005,  5, 21 },  { 2005,  5, 22 },  { 2005,  5, 23 },  { 2005,  5, 24 },
  { 2005,  5, 25 },  { 2005,  5, 26 },  { 2005,  5, 27 },  { 2005,  5, 28 },
  { 2005,  5, 29 },  { 2005,  5, 30 },  { 2005,  5, 31 },  { 2005,  6,  1 },
  { 2005,  6,  2 },  { 2005,  6,  3 },  { 2005,  6,  4 },  { 2005,  6,  5 },
  { 2005,  6,  6 },  { 2005,  6,  7 },  { 2005,  6,  8 },  { 2005,  6,  9 },
  { 2005,  6, 10 },  { 2005,  6, 11 },  { 2005,  6, 12 },  { 2005,  6, 13 },
  { 2005,  6, 14 },  { 2005,  6, 15 },  { 2005,  6, 16 },  { 2005,  6, 17 },
  { 2005,  6, 18 },  { 2005,  6, 19 },  { 2005,  6, 20 },  { 2005,  6, 21 },
  { 2005,  6, 22 },  { 2005,  6, 23 },  { 2005,  6, 24 },  { 2005,  6, 25 },
  { 2005,  6, 26 },  { 2005,  6, 27 },  { 2005,  6, 28 },  { 2005,  6, 29 },
  { 2005,  6, 30 },  { 2005,  7,  1 },  { 2005,  7,  2 },  { 2005,  7,  3 },
  { 2005,  7,  4 },  { 2005,  7,  5 },  { 2005,  7,  6 },  { 2005,  7,  7 },
  { 2005,  7,  8 },  { 2005,  7,  9 },  { 2005,  7, 10 },  { 2005,  7, 11 },
  { 2005,  7, 12 },  { 2005,  7, 13 },  { 2005,  7, 14 },  { 2005,  7, 15 },
  { 2005,  7, 16 },  { 2005,  7, 17 },  { 2005,  7, 18 },  { 2005,  7, 19 },
  { 2005,  7, 20 },  { 2005,  7, 21 },  { 2005,  7, 22 },  { 2005,  7, 23 },
  { 2005,  7, 24 },  { 2005,  7, 25 },  { 2005,  7, 26 },  { 2005,  7, 27 },
  { 2005,  7, 28 },  { 2005,  7, 29 },  { 2005,  7, 30 },  { 2005,  7, 31 },
  { 2005,  8,  1 },  { 2005,  8,  2 },  { 2005,  8,  3 },  { 2005,  8,  4 },
  { 2005,  8,  5 },  { 2005,  8,  6 },  { 2005,  8,  7 },  { 2005,  8,  8 },
  { 2005,  8,  9 },  { 2005,  8, 10 },  { 2005,  8, 11 },  { 2005,  8, 12 },
  { 2005,  8, 13 },  { 2005,  8, 14 },  { 2005,  8, 15 },  { 2005,  8, 16 },
  { 2005,  8, 17 },  { 2005,  8, 18 },  { 2005,  8, 19 },  { 2005,  8, 20 },
  { 2005,  8, 21 },  { 2005,  8, 22 },  { 2005,  8, 23 },  { 2005,  8, 24 },
  { 2005,  8, 25 },  { 2005,  8, 26 },  { 2005,  8, 27 },  { 2005,  8, 28 },
  { 2005,  8, 29 },  { 2005,  8, 30 },  { 2005,  8, 31 },  { 2005,  9,  1 },
  { 2005,  9,  2 },  { 2005,  9,  3 },  { 2005,  9,  4 },  { 2005,  9,  5 },
  { 2005,  9,  6 },  { 2005,  9,  7 },  { 2005,  9,  8 },  { 2005,  9,  9 },
  { 2005,  9, 10 },  { 2005,  9, 11 },  { 2005,  9, 12 },  { 2005,  9, 13 },
  { 2005,  9, 14 },  { 2005,  9, 15 },  { 2005,  9, 16 },  { 2005,  9, 17 },
  { 2005,  9, 18 },  { 2005,  9, 19 },  { 2005,  9, 20 },  { 2005,  9, 21 },
  { 2005,  9, 22 },  { 2005,  9, 23 },  { 2005,  9, 24 },  { 2005,  9, 25 },
  { 2005,  9, 26 },  { 2005,  9, 27 },  { 2005,  9, 28 },  { 2005,  9, 29 },
  { 2005,  9, 30 },  { 2005, 10,  1 },  { 2005, 10,  2 },  { 2005, 10,  3 },
  { 2005, 10,  4 },  { 2005, 10,  5 },  { 2005, 10,  6 },  { 2005, 10,  7 },
  { 2005, 10,  8 },  { 2005, 10,  9 },  { 2005, 10, 10 },  { 2005, 10, 11 },
  { 2005, 10, 12 },  { 2005, 10, 13 },  { 2005, 10, 14 },  { 2005, 10, 15 },
  { 2005, 10, 16 },  { 2005, 10, 17 },  { 2005, 10, 18 },  { 2005, 10, 19 },
  { 2005, 10, 20 },  { 2005, 10, 21 },  { 2005, 10, 22 },  { 2005, 10, 23 },
  { 2005, 10, 24 },  { 2005, 10, 25 },  { 2005, 10, 26 },  { 2005, 10, 27 },
  { 2005, 10, 28 },  { 2005, 10, 29 },  { 2005, 10, 30 },  { 2005, 10, 31 },
  { 2005, 11,  1 },  { 2005, 11,  2 },  { 2005, 11,  3 },  { 2005, 11,  4 },
  { 2005, 11,  5 },  { 2005, 11,  6 },  { 2005, 11,  7 },  { 2005, 11,  8 },
  { 2005, 11,  9 },  { 2005, 11, 10 },  { 2005, 11, 11 },  { 2005, 11, 12 },
  { 2005, 11, 13 },  { 2005, 11, 14 },  { 2005, 11, 15 },  { 2005, 11, 16 },
  { 2005, 11, 17 },  { 2005, 11, 18 },  { 2005, 11, 19 },  { 2005, 11, 20 },
  { 2005, 11, 21 },  { 2005, 11, 22 },  { 2005, 11, 23 },  { 2005, 11, 24 },
  { 2005, 11, 25 },  { 2005, 11, 26 },  { 2005, 11, 27 },  { 2005, 11, 28 },
  { 2005, 11, 29 },  { 2005, 11, 30 },  { 2005, 12,  1 },  { 2005, 12,  2 },
  { 2005, 12,  3 },  { 2005, 12,  4 },  { 2005, 12,  5 },  { 2005, 12,  6 },
  { 2005, 12,  7 },  { 2005, 12,  8 },  { 2005, 12,  9 },  { 2005, 12, 10 },
  { 2005, 12, 11 },  { 2005, 12, 12 },  { 2005, 12, 13 },  { 2005, 12, 14 },
  { 2005, 12, 15 },  { 2005, 12, 16 },  { 2005, 12, 17 },  { 2005, 12, 18 },
  { 2005, 12, 19 },  { 2005, 12, 20 },  { 2005, 12, 21 },  { 2005, 12, 22 },
  { 2005, 12, 23 },  { 2005, 12, 24 },  { 2005, 12, 25 },  { 2005, 12, 26 },
  { 2005, 12, 27 },  { 2005, 12, 28 },  { 2005, 12, 29 },  { 2005, 12, 30 },
  { 2005, 12, 31 },  { 2006,  1,  1 },  { 2006,  1,  2 },  { 2006,  1,  3 },
  { 2006,  1,  4 },  { 2006,  1,  5 },  { 2006,  1,  6 },  { 2006,  1,  7 },
  { 2006,  1,  8 },  { 2006,  1,  9 },  { 2006,  1, 10 },  { 2006,  1, 11 },
  { 2006,  1, 12 },  { 2006,  1, 13 },  { 2006,  1, 14 },  { 2006,  1, 15 },
  { 2006,  1, 16 },  { 2006,  1, 17 },  { 2006,  1, 18 },  { 2006,  1, 19 },
  { 2006,  1, 20 },  { 2006,  1, 21 },  { 2006,  1, 22 },  { 2006,  1, 23 },
  { 2006,  1, 24 },  { 2006,  1, 25 },  { 2006,  1, 26 },  { 2006,  1, 27 },
  { 2006,  1, 28 },  { 2006,  1, 29 },  { 2006,  1, 30 },  { 2006,  1, 31 },
  { 2006,  2,  1 },  { 2006,  2,  2 },  { 2006,  2,  3 },  { 2006,  2,  4 },
  { 2006,  2,  5 },  { 2006,  2,  6 },  { 2006,  2,  7 },  { 2006,  2,  8 },
  { 2006,  2,  9 },  { 2006,  2, 10 },  { 2006,  2, 11 },  { 2006,  2, 12 },
  { 2006,  2, 13 },  { 2006,  2, 14 },  { 2006,  2, 15 },  { 2006,  2, 16 },
  { 2006,  2, 17 },  { 2006,  2, 18 },  { 2006,  2, 19 },  { 2006,  2, 20 },
  { 2006,  2, 21 },  { 2006,  2, 22 },  { 2006,  2, 23 },  { 2006,  2, 24 },
  { 2006,  2, 25 },  { 2006,  2, 26 },  { 2006,  2, 27 },  { 2006,  2, 28 },
  { 2006,  3,  1 },  { 2006,  3,  2 },  { 2006,  3,  3 },  { 2006,  3,  4 },
  { 2006,  3,  5 },  { 2006,  3,  6 },  { 2006,  3,  7 },  { 2006,  3,  8 },
  { 2006,  3,  9 },  { 2006,  3, 10 },  { 2006,  3, 11 },  { 2006,  3, 12 },
  { 2006,  3, 13 },  { 2006,  3, 14 },  { 2006,  3, 15 },  { 2006,  3, 16 },
  { 2006,  3, 17 },  { 2006,  3, 18 },  { 2006,  3, 19 },  { 2006,  3, 20 },
  { 2006,  3, 21 },  { 2006,  3, 22 },  { 2006,  3, 23 },  { 2006,  3, 24 },
  { 2006,  3, 25 },  { 2006,  3, 26 },  { 2006,  3, 27 },  { 2006,  3, 28 },
  { 2006,  3, 29 },  { 2006,  3, 30 },  { 2006,  3, 31 },  { 2006,  4,  1 },
  { 2006,  4,  2 },  { 2006,  4,  3 },  { 2006,  4,  4 },  { 2006,  4,  5 },
  { 2006,  4,  6 },  { 2006,  4,  7 },  { 2006,  4,  8 },  { 2006,  4,  9 },
  { 2006,  4, 10 },  { 2006,  4, 11 },  { 2006,  4, 12 },  { 2006,  4, 13 },
  { 2006,  4, 14 },  { 2006,  4, 15 },  { 2006,  4, 16 },  { 2006,  4, 17 },
  { 2006,  4, 18 },  { 2006,  4, 19 },  { 2006,  4, 20 },  { 2006,  4, 21 },
  { 2006,  4, 22 },  { 2006,  4, 23 },  { 2006,  4, 24 },  { 2006,  4, 25 },
  { 2006,  4, 26 },  { 2006,  4, 27 },  { 2006,  4, 28 },  { 2006,  4, 29 },
  { 2006,  4, 30 },  { 2006,  5,  1 },  { 2006,  5,  2 },  { 2006,  5,  3 },
  { 2006,  5,  4 },  { 2006,  5,  5 },  { 2006,  5,  6 },  { 2006,  5,  7 },
  { 2006,  5,  8 },  { 2006,  5,  9 },  { 2006,  5, 10 },  { 2006,  5, 11 },
  { 2006,  5, 12 },  { 2006,  5, 13 },  { 2006,  5, 14 },  { 2006,  5, 15 },
  { 2006,  5, 16 },  { 2006,  5, 17 },  { 2006,  5, 18 },  { 2006,  5, 19 },
  { 2006,  5, 20 },  { 2006,  5, 21 },  { 2006,  5, 22 },  { 2006,  5, 23 },
  { 2006,  5, 24 },  { 2006,  5, 25 },  { 2006,  5, 26 },  { 2006,  5, 27 },
  { 2006,  5, 28 },  { 2006,  5, 29 },  { 2006,  5, 30 },  { 2006,  5, 31 },
  { 2006,  6,  1 },  { 2006,  6,  2 },  { 2006,  6,  3 },  { 2006,  6,  4 },
  { 2006,  6,  5 },  { 2006,  6,  6 },  { 2006,  6,  7 },  { 2006,  6,  8 },
  { 2006,  6,  9 },  { 2006,  6, 10 },  { 2006,  6, 11 },  { 2006,  6, 12 },
  { 2006,  6, 13 },  { 2006,  6, 14 },  { 2006,  6, 15 },  { 2006,  6, 16 },
  { 2006,  6, 17 },  { 2006,  6, 18 },  { 2006,  6, 19 },  { 2006,  6, 20 },
  { 2006,  6, 21 },  { 2006,  6, 22 },  { 2006,  6, 23 },  { 2006,  6, 24 },
  { 2006,  6, 25 },  { 2006,  6, 26 },  { 2006,  6, 27 },  { 2006,  6, 28 },
  { 2006,  6, 29 },  { 2006,  6, 30 },  { 2006,  7,  1 },  { 2006,  7,  2 },
  { 2006,  7,  3 },  { 2006,  7,  4 },  { 2006,  7,  5 },  { 2006,  7,  6 },
  { 2006,  7,  7 },  { 2006,  7,  8 },  { 2006,  7,  9 },  { 2006,  7, 10 },
  { 2006,  7, 11 },  { 2006,  7, 12 },  { 2006,  7, 13 },  { 2006,  7, 14 },
  { 2006,  7, 15 },  { 2006,  7, 16 },  { 2006,  7, 17 },  { 2006,  7, 18 },
  { 2006,  7, 19 },  { 2006,  7, 20 },  { 2006,  7, 21 },  { 2006,  7, 22 },
  { 2006,  7, 23 },  { 2006,  7, 24 },  { 2006,  7, 25 },  { 2006,  7, 26 },
  { 2006,  7, 27 },  { 2006,  7, 28 },  { 2006,  7, 29 },  { 2006,  7, 30 },
  { 2006,  7, 31 },  { 2006,  8,  1 },  { 2006,  8,  2 },  { 2006,  8,  3 },
  { 2006,  8,  4 },  { 2006,  8,  5 },  { 2006,  8,  6 },  { 2006,  8,  7 },
  { 2006,  8,  8 },  { 2006,  8,  9 },  { 2006,  8, 10 },  { 2006,  8, 11 },
  { 2006,  8, 12 },  { 2006,  8, 13 },  { 2006,  8, 14 },  { 2006,  8, 15 },
  { 2006,  8, 16 },  { 2006,  8, 17 },  { 2006,  8, 18 },  { 2006,  8, 19 },
  { 2006,  8, 20 },  { 2006,  8, 21 },  { 2006,  8, 22 },  { 2006,  8, 23 },
  { 2006,  8, 24 },  { 2006,  8, 25 },  { 2006,  8, 26 },  { 2006,  8, 27 },
  { 2006,  8, 28 },  { 2006,  8, 29 },  { 2006,  8, 30 },  { 2006,  8, 31 },
  { 2006,  9,  1 },  { 2006,  9,  2 },  { 2006,  9,  3 },  { 2006,  9,  4 },
  { 2006,  9,  5 },  { 2006,  9,  6 },  { 2006,  9,  7 },  { 2006,  9,  8 },
  { 2006,  9,  9 },  { 2006,  9, 10 },  { 2006,  9, 11 },  { 2006,  9, 12 },
  { 2006,  9, 13 },  { 2006,  9, 14 },  { 2006,  9, 15 },  { 2006,  9, 16 },
  { 2006,  9, 17 },  { 2006,  9, 18 },  { 2006,  9, 19 },  { 2006,  9, 20 },
  { 2006,  9, 21 },  { 2006,  9, 22 },  { 2006,  9, 23 },  { 2006,  9, 24 },
  { 2006,  9, 25 },  { 2006,  9, 26 },  { 2006,  9, 27 },  { 2006,  9, 28 },
  { 2006,  9, 29 },  { 2006,  9, 30 },  { 2006, 10,  1 },  { 2006, 10,  2 },
  { 2006, 10,  3 },  { 2006, 10,  4 },  { 2006, 10,  5 },  { 2006, 10,  6 },
  { 2006, 10,  7 },  { 2006, 10,  8 },  { 2006, 10,  9 },  { 2006, 10, 10 },
  { 2006, 10, 11 },  { 2006, 10, 12 },  { 2006, 10, 13 },  { 2006, 10, 14 },
  { 2006, 10, 15 },  { 2006, 10, 16 },  { 2006, 10, 17 },  { 2006, 10, 18 },
  { 2006, 10, 19 },  { 2006, 10, 20 },  { 2006, 10, 21 },  { 2006, 10, 22 },
  { 2006, 10, 23 },  { 2006, 10, 24 },  { 2006, 10, 25 },  { 2006, 10, 26 },
  { 2006, 10, 27 },  { 2006, 10, 28 },  { 2006, 10, 29 },  { 2006, 10, 30 },
  { 2006, 10, 31 },  { 2006, 11,  1 },  { 2006, 11,  2 },  { 2006, 11,  3 },
  { 2006, 11,  4 },  { 2006, 11,  5 },  { 2006, 11,  6 },  { 2006, 11,  7 },
  { 2006, 11,  8 },  { 2006, 11,  9 },  { 2006, 11, 10 },  { 2006, 11, 11 },
  { 2006, 11, 12 },  { 2006, 11, 13 },  { 2006, 11, 14 },  { 2006, 11, 15 },
  { 2006, 11, 16 },  { 2006, 11, 17 },  { 2006, 11, 18 },  { 2006, 11, 19 },
  { 2006, 11, 20 },  { 2006, 11, 21 },  { 2006, 11, 22 },  { 2006, 11, 23 },
  { 2006, 11, 24 },  { 2006, 11, 25 },  { 2006, 11, 26 },  { 2006, 11, 27 },
  { 2006, 11, 28 },  { 2006, 11, 29 },  { 2006, 11, 30 },  { 2006, 12,  1 },
  { 2006, 12,  2 },  { 2006, 12,  3 },  { 2006, 12,  4 },  { 2006, 12,  5 },
  { 2006, 12,  6 },  { 2006, 12,  7 },  { 2006, 12,  8 },  { 2006, 12,  9 },
  { 2006, 12, 10 },  { 2006, 12, 11 },  { 2006, 12, 12 },  { 2006, 12, 13 },
  { 2006, 12, 14 },  { 2006, 12, 15 },  { 2006, 12, 16 },  { 2006, 12, 17 },
  { 2006, 12, 18 },  { 2006, 12, 19 },  { 2006, 12, 20 },  { 2006, 12, 21 },
  { 2006, 12, 22 },  { 2006, 12, 23 },  { 2006, 12, 24 },  { 2006, 12, 25 },
  { 2006, 12, 26 },  { 2006, 12, 27 },  { 2006, 12, 28 },  { 2006, 12, 29 },
  { 2006, 12, 30 },  { 2006, 12, 31 },  { 2007,  1,  1 },  { 2007,  1,  2 },
  { 2007,  1,  3 },  { 2007,  1,  4 },  { 2007,  1,  5 },  { 2007,  1,  6 },
  { 2007,  1,  7 },  { 2007,  1,  8 },  { 2007,  1,  9 },  { 2007,  1, 10 },
  { 2007,  1, 11 },  { 2007,  1, 12 },  { 2007,  1, 13 },  { 2007,  1, 14 },
  { 2007,  1, 15 },  { 2007,  1, 16 },  { 2007,  1, 17 },  { 2007,  1, 18 },
  { 2007,  1, 19 },  { 2007,  1, 20 },  { 2007,  1, 21 },  { 2007,  1, 22 },
  { 2007,  1, 23 },  { 2007,  1, 24 },  { 2007,  1, 25 },  { 2007,  1, 26 },
  { 2007,  1, 27 },  { 2007,  1, 28 },  { 2007,  1, 29 },  { 2007,  1, 30 },
  { 2007,  1, 31 },  { 2007,  2,  1 },  { 2007,  2,  2 },  { 2007,  2,  3 },
  { 2007,  2,  4 },  { 2007,  2,  5 },  { 2007,  2,  6 },  { 2007,  2,  7 },
  { 2007,  2,  8 },  { 2007,  2,  9 },  { 2007,  2, 10 },  { 2007,  2, 11 },
  { 2007,  2, 12 },  { 2007,  2, 13 },  { 2007,  2, 14 },  { 2007,  2, 15 },
  { 2007,  2, 16 },  { 2007,  2, 17 },  { 2007,  2, 18 },  { 2007,  2, 19 },
  { 2007,  2, 20 },  { 2007,  2, 21 },  { 2007,  2, 22 },  { 2007,  2, 23 },
  { 2007,  2, 24 },  { 2007,  2, 25 },  { 2007,  2, 26 },  { 2007,  2, 27 },
  { 2007,  2, 28 },  { 2007,  3,  1 },  { 2007,  3,  2 },  { 2007,  3,  3 },
  { 2007,  3,  4 },  { 2007,  3,  5 },  { 2007,  3,  6 },  { 2007,  3,  7 },
  { 2007,  3,  8 },  { 2007,  3,  9 },  { 2007,  3, 10 },  { 2007,  3, 11 },
  { 2007,  3, 12 },  { 2007,  3, 13 },  { 2007,  3, 14 },  { 2007,  3, 15 },
  { 2007,  3, 16 },  { 2007,  3, 17 },  { 2007,  3, 18 },  { 2007,  3, 19 },
  { 2007,  3, 20 },  { 2007,  3, 21 },  { 2007,  3, 22 },  { 2007,  3, 23 },
  { 2007,  3, 24 },  { 2007,  3, 25 },  { 2007,  3, 26 },  { 2007,  3, 27 },
  { 2007,  3, 28 },  { 2007,  3, 29 },  { 2007,  3, 30 },  { 2007,  3, 31 },
  { 2007,  4,  1 },  { 2007,  4,  2 },  { 2007,  4,  3 },  { 2007,  4,  4 },
  { 2007,  4,  5 },  { 2007,  4,  6 },  { 2007,  4,  7 },  { 2007,  4,  8 },
  { 2007,  4,  9 },  { 2007,  4, 10 },  { 2007,  4, 11 },  { 2007,  4, 12 },
  { 2007,  4, 13 },  { 2007,  4, 14 },  { 2007,  4, 15 },  { 2007,  4, 16 },
  { 2007,  4, 17 },  { 2007,  4, 18 },  { 2007,  4, 19 },  { 2007,  4, 20 },
  { 2007,  4, 21 },  { 2007,  4, 22 },  { 2007,  4, 23 },  { 2007,  4, 24 },
  { 2007,  4, 25 },  { 2007,  4, 26 },  { 2007,  4, 27 },  { 2007,  4, 28 },
  { 2007,  4, 29 },  { 2007,  4, 30 },  { 2007,  5,  1 },  { 2007,  5,  2 },
  { 2007,  5,  3 },  { 2007,  5,  4 },  { 2007,  5,  5 },  { 2007,  5,  6 },
  { 2007,  5,  7 },  { 2007,  5,  8 },  { 2007,  5,  9 },  { 2007,  5, 10 },
  { 2007,  5, 11 },  { 2007,  5, 12 },  { 2007,  5, 13 },  { 2007,  5, 14 },
  { 2007,  5, 15 },  { 2007,  5, 16 },  { 2007,  5, 17 },  { 2007,  5, 18 },
  { 2007,  5, 19 },  { 2007,  5, 20 },  { 2007,  5, 21 },  { 2007,  5, 22 },
  { 2007,  5, 23 },  { 2007,  5, 24 },  { 2007,  5, 25 },  { 2007,  5, 26 },
  { 2007,  5, 27 },  { 2007,  5, 28 },  { 2007,  5, 29 },  { 2007,  5, 30 },
  { 2007,  5, 31 },  { 2007,  6,  1 },  { 2007,  6,  2 },  { 2007,  6,  3 },
  { 2007,  6,  4 },  { 2007,  6,  5 },  { 2007,  6,  6 },  { 2007,  6,  7 },
  { 2007,  6,  8 },  { 2007,  6,  9 },  { 2007,  6, 10 },  { 2007,  6, 11 },
  { 2007,  6, 12 },  { 2007,  6, 13 },  { 2007,  6, 14 },  { 2007,  6, 15 },
  { 2007,  6, 16 },  { 2007,  6, 17 },  { 2007,  6, 18 },  { 2007,  6, 19 },
  { 2007,  6, 20 },  { 2007,  6, 21 },  { 2007,  6, 22 },  { 2007,  6, 23 },
  { 2007,  6, 24 },  { 2007,  6, 25 },  { 2007,  6, 26 },  { 2007,  6, 27 },
  { 2007,  6, 28 },  { 2007,  6, 29 },  { 2007,  6, 30 },  { 2007,  7,  1 },
  { 2007,  7,  2 },  { 2007,  7,  3 },  { 2007,  7,  4 },  { 2007,  7,  5 },
  { 2007,  7,  6 },  { 2007,  7,  7 },  { 2007,  7,  8 },  { 2007,  7,  9 },
  { 2007,  7, 10 },  { 2007,  7, 11 },  { 2007,  7, 12 },  { 2007,  7, 13 },
  { 2007,  7, 14 },  { 2007,  7, 15 },  { 2007,  7, 16 },  { 2007,  7, 17 },
  { 2007,  7, 18 },  { 2007,  7, 19 },  { 2007,  7, 20 },  { 2007,  7, 21 },
  { 2007,  7, 22 },  { 2007,  7, 23 },  { 2007,  7, 24 },  { 2007,  7, 25 },
  { 2007,  7, 26 },  { 2007,  7, 27 },  { 2007,  7, 28 },  { 2007,  7, 29 },
  { 2007,  7, 30 },  { 2007,  7, 31 },  { 2007,  8,  1 },  { 2007,  8,  2 },
  { 2007,  8,  3 },  { 2007,  8,  4 },  { 2007,  8,  5 },  { 2007,  8,  6 },
  { 2007,  8,  7 },  { 2007,  8,  8 },  { 2007,  8,  9 },  { 2007,  8, 10 },
  { 2007,  8, 11 },  { 2007,  8, 12 },  { 2007,  8, 13 },  { 2007,  8, 14 },
  { 2007,  8, 15 },  { 2007,  8, 16 },  { 2007,  8, 17 },  { 2007,  8, 18 },
  { 2007,  8, 19 },  { 2007,  8, 20 },  { 2007,  8, 21 },  { 2007,  8, 22 },
  { 2007,  8, 23 },  { 2007,  8, 24 },  { 2007,  8, 25 },  { 2007,  8, 26 },
  { 2007,  8, 27 },  { 2007,  8, 28 },  { 2007,  8, 29 },  { 2007,  8, 30 },
  { 2007,  8, 31 },  { 2007,  9,  1 },  { 2007,  9,  2 },  { 2007,  9,  3 },
  { 2007,  9,  4 },  { 2007,  9,  5 },  { 2007,  9,  6 },  { 2007,  9,  7 },
  { 2007,  9,  8 },  { 2007,  9,  9 },  { 2007,  9, 10 },  { 2007,  9, 11 },
  { 2007,  9, 12 },  { 2007,  9, 13 },  { 2007,  9, 14 },  { 2007,  9, 15 },
  { 2007,  9, 16 },  { 2007,  9, 17 },  { 2007,  9, 18 },  { 2007,  9, 19 },
  { 2007,  9, 20 },  { 2007,  9, 21 },  { 2007,  9, 22 },  { 2007,  9, 23 },
  { 2007,  9, 24 },  { 2007,  9, 25 },  { 2007,  9, 26 },  { 2007,  9, 27 },
  { 2007,  9, 28 },  { 2007,  9, 29 },  { 2007,  9, 30 },  { 2007, 10,  1 },
  { 2007, 10,  2 },  { 2007, 10,  3 },  { 2007, 10,  4 },  { 2007, 10,  5 },
  { 2007, 10,  6 },  { 2007, 10,  7 },  { 2007, 10,  8 },  { 2007, 10,  9 },
  { 2007, 10, 10 },  { 2007, 10, 11 },  { 2007, 10, 12 },  { 2007, 10, 13 },
  { 2007, 10, 14 },  { 2007, 10, 15 },  { 2007, 10, 16 },  { 2007, 10, 17 },
  { 2007, 10, 18 },  { 2007, 10, 19 },  { 2007, 10, 20 },  { 2007, 10, 21 },
  { 2007, 10, 22 },  { 2007, 10, 23 },  { 2007, 10, 24 },  { 2007, 10, 25 },
  { 2007, 10, 26 },  { 2007, 10, 27 },  { 2007, 10, 28 },  { 2007, 10, 29 },
  { 2007, 10, 30 },  { 2007, 10, 31 },  { 2007, 11,  1 },  { 2007, 11,  2 },
  { 2007, 11,  3 },  { 2007, 11,  4 },  { 2007, 11,  5 },  { 2007, 11,  6 },
  { 2007, 11,  7 },  { 2007, 11,  8 },  { 2007, 11,  9 },  { 2007, 11, 10 },
  { 2007, 11, 11 },  { 2007, 11, 12 },  { 2007, 11, 13 },  { 2007, 11, 14 },
  { 2007, 11, 15 },  { 2007, 11, 16 },  { 2007, 11, 17 },  { 2007, 11, 18 },
  { 2007, 11, 19 },  { 2007, 11, 20 },  { 2007, 11, 21 },  { 2007, 11, 22 },
  { 2007, 11, 23 },  { 2007, 11, 24 },  { 2007, 11, 25 },  { 2007, 11, 26 },
  { 2007, 11, 27 },  { 2007, 11, 28 },  { 2007, 11, 29 },  { 2007, 11, 30 },
  { 2007, 12,  1 },  { 2007, 12,  2 },  { 2007, 12,  3 },  { 2007, 12,  4 },
  { 2007, 12,  5 },  { 2007, 12,  6 },  { 2007, 12,  7 },  { 2007, 12,  8 },
  { 2007, 12,  9 },  { 2007, 12, 10 },  { 2007, 12, 11 },  { 2007, 12, 12 },
  { 2007, 12, 13 },  { 2007, 12, 14 },  { 2007, 12, 15 },  { 2007, 12, 16 },
  { 2007, 12, 17 },  { 2007, 12, 18 },  { 2007, 12, 19 },  { 2007, 12, 20 },
  { 2007, 12, 21 },  { 2007, 12, 22 },  { 2007, 12, 23 },  { 2007, 12, 24 },
  { 2007, 12, 25 },  { 2007, 12, 26 },  { 2007, 12, 27 },  { 2007, 12, 28 },
  { 2007, 12, 29 },  { 2007, 12, 30 },  { 2007, 12, 31 },  { 2008,  1,  1 },
  { 2008,  1,  2 },  { 2008,  1,  3 },  { 2008,  1,  4 },  { 2008,  1,  5 },
  { 2008,  1,  6 },  { 2008,  1,  7 },  { 2008,  1,  8 },  { 2008,  1,  9 },
  { 2008,  1, 10 },  { 2008,  1, 11 },  { 2008,  1, 12 },  { 2008,  1, 13 },
  { 2008,  1, 14 },  { 2008,  1, 15 },  { 2008,  1, 16 },  { 2008,  1, 17 },
  { 2008,  1, 18 },  { 2008,  1, 19 },  { 2008,  1, 20 },  { 2008,  1, 21 },
  { 2008,  1, 22 },  { 2008,  1, 23 },  { 2008,  1, 24 },  { 2008,  1, 25 },
  { 2008,  1, 26 },  { 2008,  1, 27 },  { 2008,  1, 28 },  { 2008,  1, 29 },
  { 2008,  1, 30 },  { 2008,  1, 31 },  { 2008,  2,  1 },  { 2008,  2,  2 },
  { 2008,  2,  3 },  { 2008,  2,  4 },  { 2008,  2,  5 },  { 2008,  2,  6 },
  { 2008,  2,  7 },  { 2008,  2,  8 },  { 2008,  2,  9 },  { 2008,  2, 10 },
  { 2008,  2, 11 },  { 2008,  2, 12 },  { 2008,  2, 13 },  { 2008,  2, 14 },
  { 2008,  2, 15 },  { 2008,  2, 16 },  { 2008,  2, 17 },  { 2008,  2, 18 },
  { 2008,  2, 19 },  { 2008,  2, 20 },  { 2008,  2, 21 },  { 2008,  2, 22 },
  { 2008,  2, 23 },  { 2008,  2, 24 },  { 2008,  2, 25 },  { 2008,  2, 26 },
  { 2008,  2, 27 },  { 2008,  2, 28 },  { 2008,  2, 29 },  { 2008,  3,  1 },
  { 2008,  3,  2 },  { 2008,  3,  3 },  { 2008,  3,  4 },  { 2008,  3,  5 },
  { 2008,  3,  6 },  { 2008,  3,  7 },  { 2008,  3,  8 },  { 2008,  3,  9 },
  { 2008,  3, 10 },  { 2008,  3, 11 },  { 2008,  3, 12 },  { 2008,  3, 13 },
  { 2008,  3, 14 },  { 2008,  3, 15 },  { 2008,  3, 16 },  { 2008,  3, 17 },
  { 2008,  3, 18 },  { 2008,  3, 19 },  { 2008,  3, 20 },  { 2008,  3, 21 },
  { 2008,  3, 22 },  { 2008,  3, 23 },  { 2008,  3, 24 },  { 2008,  3, 25 },
  { 2008,  3, 26 },  { 2008,  3, 27 },  { 2008,  3, 28 },  { 2008,  3, 29 },
  { 2008,  3, 30 },  { 2008,  3, 31 },  { 2008,  4,  1 },  { 2008,  4,  2 },
  { 2008,  4,  3 },  { 2008,  4,  4 },  { 2008,  4,  5 },  { 2008,  4,  6 },
  { 2008,  4,  7 },  { 2008,  4,  8 },  { 2008,  4,  9 },  { 2008,  4, 10 },
  { 2008,  4, 11 },  { 2008,  4, 12 },  { 2008,  4, 13 },  { 2008,  4, 14 },
  { 2008,  4, 15 },  { 2008,  4, 16 },  { 2008,  4, 17 },  { 2008,  4, 18 },
  { 2008,  4, 19 },  { 2008,  4, 20 },  { 2008,  4, 21 },  { 2008,  4, 22 },
  { 2008,  4, 23 },  { 2008,  4, 24 },  { 2008,  4, 25 },  { 2008,  4, 26 },
  { 2008,  4, 27 },  { 2008,  4, 28 },  { 2008,  4, 29 },  { 2008,  4, 30 },
  { 2008,  5,  1 },  { 2008,  5,  2 },  { 2008,  5,  3 },  { 2008,  5,  4 },
  { 2008,  5,  5 },  { 2008,  5,  6 },  { 2008,  5,  7 },  { 2008,  5,  8 },
  { 2008,  5,  9 },  { 2008,  5, 10 },  { 2008,  5, 11 },  { 2008,  5, 12 },
  { 2008,  5, 13 },  { 2008,  5, 14 },  { 2008,  5, 15 },  { 2008,  5, 16 },
  { 2008,  5, 17 },  { 2008,  5, 18 },  { 2008,  5, 19 },  { 2008,  5, 20 },
  { 2008,  5, 21 },  { 2008,  5, 22 },  { 2008,  5, 23 },  { 2008,  5, 24 },
  { 2008,  5, 25 },  { 2008,  5, 26 },  { 2008,  5, 27 },  { 2008,  5, 28 },
  { 2008,  5, 29 },  { 2008,  5, 30 },  { 2008,  5, 31 },  { 2008,  6,  1 },
  { 2008,  6,  2 },  { 2008,  6,  3 },  { 2008,  6,  4 },  { 2008,  6,  5 },
  { 2008,  6,  6 },  { 2008,  6,  7 },  { 2008,  6,  8 },  { 2008,  6,  9 },
  { 2008,  6, 10 },  { 2008,  6, 11 },  { 2008,  6, 12 },  { 2008,  6, 13 },
  { 2008,  6, 14 },  { 2008,  6, 15 },  { 2008,  6, 16 },  { 2008,  6, 17 },
  { 2008,  6, 18 },  { 2008,  6, 19 },  { 2008,  6, 20 },  { 2008,  6, 21 },
  { 2008,  6, 22 },  { 2008,  6, 23 },  { 2008,  6, 24 },  { 2008,  6, 25 },
  { 2008,  6, 26 },  { 2008,  6, 27 },  { 2008,  6, 28 },  { 2008,  6, 29 },
  { 2008,  6, 30 },  { 2008,  7,  1 },  { 2008,  7,  2 },  { 2008,  7,  3 },
  { 2008,  7,  4 },  { 2008,  7,  5 },  { 2008,  7,  6 },  { 2008,  7,  7 },
  { 2008,  7,  8 },  { 2008,  7,  9 },  { 2008,  7, 10 },  { 2008,  7, 11 },
  { 2008,  7, 12 },  { 2008,  7, 13 },  { 2008,  7, 14 },  { 2008,  7, 15 },
  { 2008,  7, 16 },  { 2008,  7, 17 },  { 2008,  7, 18 },  { 2008,  7, 19 },
  { 2008,  7, 20 },  { 2008,  7, 21 },  { 2008,  7, 22 },  { 2008,  7, 23 },
  { 2008,  7, 24 },  { 2008,  7, 25 },  { 2008,  7, 26 },  { 2008,  7, 27 },
  { 2008,  7, 28 },  { 2008,  7, 29 },  { 2008,  7, 30 },  { 2008,  7, 31 },
  { 2008,  8,  1 },  { 2008,  8,  2 },  { 2008,  8,  3 },  { 2008,  8,  4 },
  { 2008,  8,  5 },  { 2008,  8,  6 },  { 2008,  8,  7 },  { 2008,  8,  8 },
  { 2008,  8,  9 },  { 2008,  8, 10 },  { 2008,  8, 11 },  { 2008,  8, 12 },
  { 2008,  8, 13 },  { 2008,  8, 14 },  { 2008,  8, 15 },  { 2008,  8, 16 },
  { 2008,  8, 17 },  { 2008,  8, 18 },  { 2008,  8, 19 },  { 2008,  8, 20 },
  { 2008,  8, 21 },  { 2008,  8, 22 },  { 2008,  8, 23 },  { 2008,  8, 24 },
  { 2008,  8, 25 },  { 2008,  8, 26 },  { 2008,  8, 27 },  { 2008,  8, 28 },
  { 2008,  8, 29 },  { 2008,  8, 30 },  { 2008,  8, 31 },  { 2008,  9,  1 },
  { 2008,  9,  2 },  { 2008,  9,  3 },  { 2008,  9,  4 },  { 2008,  9,  5 },
  { 2008,  9,  6 },  { 2008,  9,  7 },  { 2008,  9,  8 },  { 2008,  9,  9 },
  { 2008,  9, 10 },  { 2008,  9, 11 },  { 2008,  9, 12 },  { 2008,  9, 13 },
  { 2008,  9, 14 },  { 2008,  9, 15 },  { 2008,  9, 16 },  { 2008,  9, 17 },
  { 2008,  9, 18 },  { 2008,  9, 19 },  { 2008,  9, 20 },  { 2008,  9, 21 },
  { 2008,  9, 22 },  { 2008,  9, 23 },  { 2008,  9, 24 },  { 2008,  9, 25 },
  { 2008,  9, 26 },  { 2008,  9, 27 },  { 2008,  9, 28 },  { 2008,  9, 29 },
  { 2008,  9, 30 },  { 2008, 10,  1 },  { 2008, 10,  2 },  { 2008, 10,  3 },
  { 2008, 10,  4 },  { 2008, 10,  5 },  { 2008, 10,  6 },  { 2008, 10,  7 },
  { 2008, 10,  8 },  { 2008, 10,  9 },  { 2008, 10, 10 },  { 2008, 10, 11 },
  { 2008, 10, 12 },  { 2008, 10, 13 },  { 2008, 10, 14 },  { 2008, 10, 15 },
  { 2008, 10, 16 },  { 2008, 10, 17 },  { 2008, 10, 18 },  { 2008, 10, 19 },
  { 2008, 10, 20 },  { 2008, 10, 21 },  { 2008, 10, 22 },  { 2008, 10, 23 },
  { 2008, 10, 24 },  { 2008, 10, 25 },  { 2008, 10, 26 },  { 2008, 10, 27 },
  { 2008, 10, 28 },  { 2008, 10, 29 },  { 2008, 10, 30 },  { 2008, 10, 31 },
  { 2008, 11,  1 },  { 2008, 11,  2 },  { 2008, 11,  3 },  { 2008, 11,  4 },
  { 2008, 11,  5 },  { 2008, 11,  6 },  { 2008, 11,  7 },  { 2008, 11,  8 },
  { 2008, 11,  9 },  { 2008, 11, 10 },  { 2008, 11, 11 },  { 2008, 11, 12 },
  { 2008, 11, 13 },  { 2008, 11, 14 },  { 2008, 11, 15 },  { 2008, 11, 16 },
  { 2008, 11, 17 },  { 2008, 11, 18 },  { 2008, 11, 19 },  { 2008, 11, 20 },
  { 2008, 11, 21 },  { 2008, 11, 22 },  { 2008, 11, 23 },  { 2008, 11, 24 },
  { 2008, 11, 25 },  { 2008, 11, 26 },  { 2008, 11, 27 },  { 2008, 11, 28 },
  { 2008, 11, 29 },  { 2008, 11, 30 },  { 2008, 12,  1 },  { 2008, 12,  2 },
  { 2008, 12,  3 },  { 2008, 12,  4 },  { 2008, 12,  5 },  { 2008, 12,  6 },
  { 2008, 12,  7 },  { 2008, 12,  8 },  { 2008, 12,  9 },  { 2008, 12, 10 },
  { 2008, 12, 11 },  { 2008, 12, 12 },  { 2008, 12, 13 },  { 2008, 12, 14 },
  { 2008, 12, 15 },  { 2008, 12, 16 },  { 2008, 12, 17 },  { 2008, 12, 18 },
  { 2008, 12, 19 },  { 2008, 12, 20 },  { 2008, 12, 21 },  { 2008, 12, 22 },
  { 2008, 12, 23 },  { 2008, 12, 24 },  { 2008, 12, 25 },  { 2008, 12, 26 },
  { 2008, 12, 27 },  { 2008, 12, 28 },  { 2008, 12, 29 },  { 2008, 12, 30 },
  { 2008, 12, 31 },  { 2009,  1,  1 },  { 2009,  1,  2 },  { 2009,  1,  3 },
  { 2009,  1,  4 },  { 2009,  1,  5 },  { 2009,  1,  6 },  { 2009,  1,  7 },
  { 2009,  1,  8 },  { 2009,  1,  9 },  { 2009,  1, 10 },  { 2009,  1, 11 },
  { 2009,  1, 12 },  { 2009,  1, 13 },  { 2009,  1, 14 },  { 2009,  1, 15 },
  { 2009,  1, 16 },  { 2009,  1, 17 },  { 2009,  1, 18 },  { 2009,  1, 19 },
  { 2009,  1, 20 },  { 2009,  1, 21 },  { 2009,  1, 22 },  { 2009,  1, 23 },
  { 2009,  1, 24 },  { 2009,  1, 25 },  { 2009,  1, 26 },  { 2009,  1, 27 },
  { 2009,  1, 28 },  { 2009,  1, 29 },  { 2009,  1, 30 },  { 2009,  1, 31 },
  { 2009,  2,  1 },  { 2009,  2,  2 },  { 2009,  2,  3 },  { 2009,  2,  4 },
  { 2009,  2,  5 },  { 2009,  2,  6 },  { 2009,  2,  7 },  { 2009,  2,  8 },
  { 2009,  2,  9 },  { 2009,  2, 10 },  { 2009,  2, 11 },  { 2009,  2, 12 },
  { 2009,  2, 13 },  { 2009,  2, 14 },  { 2009,  2, 15 },  { 2009,  2, 16 },
  { 2009,  2, 17 },  { 2009,  2, 18 },  { 2009,  2, 19 },  { 2009,  2, 20 },
  { 2009,  2, 21 },  { 2009,  2, 22 },  { 2009,  2, 23 },  { 2009,  2, 24 },
  { 2009,  2, 25 },  { 2009,  2, 26 },  { 2009,  2, 27 },  { 2009,  2, 28 },
  { 2009,  3,  1 },  { 2009,  3,  2 },  { 2009,  3,  3 },  { 2009,  3,  4 },
  { 2009,  3,  5 },  { 2009,  3,  6 },  { 2009,  3,  7 },  { 2009,  3,  8 },
  { 2009,  3,  9 },  { 2009,  3, 10 },  { 2009,  3, 11 },  { 2009,  3, 12 },
  { 2009,  3, 13 },  { 2009,  3, 14 },  { 2009,  3, 15 },  { 2009,  3, 16 },
  { 2009,  3, 17 },  { 2009,  3, 18 },  { 2009,  3, 19 },  { 2009,  3, 20 },
  { 2009,  3, 21 },  { 2009,  3, 22 },  { 2009,  3, 23 },  { 2009,  3, 24 },
  { 2009,  3, 25 },  { 2009,  3, 26 },  { 2009,  3, 27 },  { 2009,  3, 28 },
  { 2009,  3, 29 },  { 2009,  3, 30 },  { 2009,  3, 31 },  { 2009,  4,  1 },
  { 2009,  4,  2 },  { 2009,  4,  3 },  { 2009,  4,  4 },  { 2009,  4,  5 },
  { 2009,  4,  6 },  { 2009,  4,  7 },  { 2009,  4,  8 },  { 2009,  4,  9 },
  { 2009,  4, 10 },  { 2009,  4, 11 },  { 2009,  4, 12 },  { 2009,  4, 13 },
  { 2009,  4, 14 },  { 2009,  4, 15 },  { 2009,  4, 16 },  { 2009,  4, 17 },
  { 2009,  4, 18 },  { 2009,  4, 19 },  { 2009,  4, 20 },  { 2009,  4, 21 },
  { 2009,  4, 22 },  { 2009,  4, 23 },  { 2009,  4, 24 },  { 2009,  4, 25 },
  { 2009,  4, 26 },  { 2009,  4, 27 },  { 2009,  4, 28 },  { 2009,  4, 29 },
  { 2009,  4, 30 },  { 2009,  5,  1 },  { 2009,  5,  2 },  { 2009,  5,  3 },
  { 2009,  5,  4 },  { 2009,  5,  5 },  { 2009,  5,  6 },  { 2009,  5,  7 },
  { 2009,  5,  8 },  { 2009,  5,  9 },  { 2009,  5, 10 },  { 2009,  5, 11 },
  { 2009,  5, 12 },  { 2009,  5, 13 },  { 2009,  5, 14 },  { 2009,  5, 15 },
  { 2009,  5, 16 },  { 2009,  5, 17 },  { 2009,  5, 18 },  { 2009,  5, 19 },
  { 2009,  5, 20 },  { 2009,  5, 21 },  { 2009,  5, 22 },  { 2009,  5, 23 },
  { 2009,  5, 24 },  { 2009,  5, 25 },  { 2009,  5, 26 },  { 2009,  5, 27 },
  { 2009,  5, 28 },  { 2009,  5, 29 },  { 2009,  5, 30 },  { 2009,  5, 31 },
  { 2009,  6,  1 },  { 2009,  6,  2 },  { 2009,  6,  3 },  { 2009,  6,  4 },
  { 2009,  6,  5 },  { 2009,  6,  6 },  { 2009,  6,  7 },  { 2009,  6,  8 },
  { 2009,  6,  9 },  { 2009,  6, 10 },  { 2009,  6, 11 },  { 2009,  6, 12 },
  { 2009,  6, 13 },  { 2009,  6, 14 },  { 2009,  6, 15 },  { 2009,  6, 16 },
  { 2009,  6, 17 },  { 2009,  6, 18 },  { 2009,  6, 19 },  { 2009,  6, 20 },
  { 2009,  6, 21 },  { 2009,  6, 22 },  { 2009,  6, 23 },  { 2009,  6, 24 },
  { 2009,  6, 25 },  { 2009,  6, 26 },  { 2009,  6, 27 },  { 2009,  6, 28 },
  { 2009,  6, 29 },  { 2009,  6, 30 },  { 2009,  7,  1 },  { 2009,  7,  2 },
  { 2009,  7,  3 },  { 2009,  7,  4 },  { 2009,  7,  5 },  { 2009,  7,  6 },
  { 2009,  7,  7 },  { 2009,  7,  8 },  { 2009,  7,  9 },  { 2009,  7, 10 },
  { 2009,  7, 11 },  { 2009,  7, 12 },  { 2009,  7, 13 },  { 2009,  7, 14 },
  { 2009,  7, 15 },  { 2009,  7, 16 },  { 2009,  7, 17 },  { 2009,  7, 18 },
  { 2009,  7, 19 },  { 2009,  7, 20 },  { 2009,  7, 21 },  { 2009,  7, 22 },
  { 2009,  7, 23 },  { 2009,  7, 24 },  { 2009,  7, 25 },  { 2009,  7, 26 },
  { 2009,  7, 27 },  { 2009,  7, 28 },  { 2009,  7, 29 },  { 2009,  7, 30 },
  { 2009,  7, 31 },  { 2009,  8,  1 },  { 2009,  8,  2 },  { 2009,  8,  3 },
  { 2009,  8,  4 },  { 2009,  8,  5 },  { 2009,  8,  6 },  { 2009,  8,  7 },
  { 2009,  8,  8 },  { 2009,  8,  9 },  { 2009,  8, 10 },  { 2009,  8, 11 },
  { 2009,  8, 12 },  { 2009,  8, 13 },  { 2009,  8, 14 },  { 2009,  8, 15 },
  { 2009,  8, 16 },  { 2009,  8, 17 },  { 2009,  8, 18 },  { 2009,  8, 19 },
  { 2009,  8, 20 },  { 2009,  8, 21 },  { 2009,  8, 22 },  { 2009,  8, 23 },
  { 2009,  8, 24 },  { 2009,  8, 25 },  { 2009,  8, 26 },  { 2009,  8, 27 },
  { 2009,  8, 28 },  { 2009,  8, 29 },  { 2009,  8, 30 },  { 2009,  8, 31 },
  { 2009,  9,  1 },  { 2009,  9,  2 },  { 2009,  9,  3 },  { 2009,  9,  4 },
  { 2009,  9,  5 },  { 2009,  9,  6 },  { 2009,  9,  7 },  { 2009,  9,  8 },
  { 2009,  9,  9 },  { 2009,  9, 10 },  { 2009,  9, 11 },  { 2009,  9, 12 },
  { 2009,  9, 13 },  { 2009,  9, 14 },  { 2009,  9, 15 },  { 2009,  9, 16 },
  { 2009,  9, 17 },  { 2009,  9, 18 },  { 2009,  9, 19 },  { 2009,  9, 20 },
  { 2009,  9, 21 },  { 2009,  9, 22 },  { 2009,  9, 23 },  { 2009,  9, 24 },
  { 2009,  9, 25 },  { 2009,  9, 26 },  { 2009,  9, 27 },  { 2009,  9, 28 },
  { 2009,  9, 29 },  { 2009,  9, 30 },  { 2009, 10,  1 },  { 2009, 10,  2 },
  { 2009, 10,  3 },  { 2009, 10,  4 },  { 2009, 10,  5 },  { 2009, 10,  6 },
  { 2009, 10,  7 },  { 2009, 10,  8 },  { 2009, 10,  9 },  { 2009, 10, 10 },
  { 2009, 10, 11 },  { 2009, 10, 12 },  { 2009, 10, 13 },  { 2009, 10, 14 },
  { 2009, 10, 15 },  { 2009, 10, 16 },  { 2009, 10, 17 },  { 2009, 10, 18 },
  { 2009, 10, 19 },  { 2009, 10, 20 },  { 2009, 10, 21 },  { 2009, 10, 22 },
  { 2009, 10, 23 },  { 2009, 10, 24 },  { 2009, 10, 25 },  { 2009, 10, 26 },
  { 2009, 10, 27 },  { 2009, 10, 28 },  { 2009, 10, 29 },  { 2009, 10, 30 },
  { 2009, 10, 31 },  { 2009, 11,  1 },  { 2009, 11,  2 },  { 2009, 11,  3 },
  { 2009, 11,  4 },  { 2009, 11,  5 },  { 2009, 11,  6 },  { 2009, 11,  7 },
  { 2009, 11,  8 },  { 2009, 11,  9 },  { 2009, 11, 10 },  { 2009, 11, 11 },
  { 2009, 11, 12 },  { 2009, 11, 13 },  { 2009, 11, 14 },  { 2009, 11, 15 },
  { 2009, 11, 16 },  { 2009, 11, 17 },  { 2009, 11, 18 },  { 2009, 11, 19 },
  { 2009, 11, 20 },  { 2009, 11, 21 },  { 2009, 11, 22 },  { 2009, 11, 23 },
  { 2009, 11, 24 },  { 2009, 11, 25 },  { 2009, 11, 26 },  { 2009, 11, 27 },
  { 2009, 11, 28 },  { 2009, 11, 29 },  { 2009, 11, 30 },  { 2009, 12,  1 },
  { 2009, 12,  2 },  { 2009, 12,  3 },  { 2009, 12,  4 },  { 2009, 12,  5 },
  { 2009, 12,  6 },  { 2009, 12,  7 },  { 2009, 12,  8 },  { 2009, 12,  9 },
  { 2009, 12, 10 },  { 2009, 12, 11 },  { 2009, 12, 12 },  { 2009, 12, 13 },
  { 2009, 12, 14 },  { 2009, 12, 15 },  { 2009, 12, 16 },  { 2009, 12, 17 },
  { 2009, 12, 18 },  { 2009, 12, 19 },  { 2009, 12, 20 },  { 2009, 12, 21 },
  { 2009, 12, 22 },  { 2009, 12, 23 },  { 2009, 12, 24 },  { 2009, 12, 25 },
  { 2009, 12, 26 },  { 2009, 12, 27 },  { 2009, 12, 28 },  { 2009, 12, 29 },
  { 2009, 12, 30 },  { 2009, 12, 31 },  { 2010,  1,  1 },  { 2010,  1,  2 },
  { 2010,  1,  3 },  { 2010,  1,  4 },  { 2010,  1,  5 },  { 2010,  1,  6 },
  { 2010,  1,  7 },  { 2010,  1,  8 },  { 2010,  1,  9 },  { 2010,  1, 10 },
  { 2010,  1, 11 },  { 2010,  1, 12 },  { 2010,  1, 13 },  { 2010,  1, 14 },
  { 2010,  1, 15 },  { 2010,  1, 16 },  { 2010,  1, 17 },  { 2010,  1, 18 },
  { 2010,  1, 19 },  { 2010,  1, 20 },  { 2010,  1, 21 },  { 2010,  1, 22 },
  { 2010,  1, 23 },  { 2010,  1, 24 },  { 2010,  1, 25 },  { 2010,  1, 26 },
  { 2010,  1, 27 },  { 2010,  1, 28 },  { 2010,  1, 29 },  { 2010,  1, 30 },
  { 2010,  1, 31 },  { 2010,  2,  1 },  { 2010,  2,  2 },  { 2010,  2,  3 },
  { 2010,  2,  4 },  { 2010,  2,  5 },  { 2010,  2,  6 },  { 2010,  2,  7 },
  { 2010,  2,  8 },  { 2010,  2,  9 },  { 2010,  2, 10 },  { 2010,  2, 11 },
  { 2010,  2, 12 },  { 2010,  2, 13 },  { 2010,  2, 14 },  { 2010,  2, 15 },
  { 2010,  2, 16 },  { 2010,  2, 17 },  { 2010,  2, 18 },  { 2010,  2, 19 },
  { 2010,  2, 20 },  { 2010,  2, 21 },  { 2010,  2, 22 },  { 2010,  2, 23 },
  { 2010,  2, 24 },  { 2010,  2, 25 },  { 2010,  2, 26 },  { 2010,  2, 27 },
  { 2010,  2, 28 },  { 2010,  3,  1 },  { 2010,  3,  2 },  { 2010,  3,  3 },
  { 2010,  3,  4 },  { 2010,  3,  5 },  { 2010,  3,  6 },  { 2010,  3,  7 },
  { 2010,  3,  8 },  { 2010,  3,  9 },  { 2010,  3, 10 },  { 2010,  3, 11 },
  { 2010,  3, 12 },  { 2010,  3, 13 },  { 2010,  3, 14 },  { 2010,  3, 15 },
  { 2010,  3, 16 },  { 2010,  3, 17 },  { 2010,  3, 18 },  { 2010,  3, 19 },
  { 2010,  3, 20 },  { 2010,  3, 21 },  { 2010,  3, 22 },  { 2010,  3, 23 },
  { 2010,  3, 24 },  { 2010,  3, 25 },  { 2010,  3, 26 },  { 2010,  3, 27 },
  { 2010,  3, 28 },  { 2010,  3, 29 },  { 2010,  3, 30 },  { 2010,  3, 31 },
  { 2010,  4,  1 },  { 2010,  4,  2 },  { 2010,  4,  3 },  { 2010,  4,  4 },
  { 2010,  4,  5 },  { 2010,  4,  6 },  { 2010,  4,  7 },  { 2010,  4,  8 },
  { 2010,  4,  9 },  { 2010,  4, 10 },  { 2010,  4, 11 },  { 2010,  4, 12 },
  { 2010,  4, 13 },  { 2010,  4, 14 },  { 2010,  4, 15 },  { 2010,  4, 16 },
  { 2010,  4, 17 },  { 2010,  4, 18 },  { 2010,  4, 19 },  { 2010,  4, 20 },
  { 2010,  4, 21 },  { 2010,  4, 22 },  { 2010,  4, 23 },  { 2010,  4, 24 },
  { 2010,  4, 25 },  { 2010,  4, 26 },  { 2010,  4, 27 },  { 2010,  4, 28 },
  { 2010,  4, 29 },  { 2010,  4, 30 },  { 2010,  5,  1 },  { 2010,  5,  2 },
  { 2010,  5,  3 },  { 2010,  5,  4 },  { 2010,  5,  5 },  { 2010,  5,  6 },
  { 2010,  5,  7 },  { 2010,  5,  8 },  { 2010,  5,  9 },  { 2010,  5, 10 },
  { 2010,  5, 11 },  { 2010,  5, 12 },  { 2010,  5, 13 },  { 2010,  5, 14 },
  { 2010,  5, 15 },  { 2010,  5, 16 },  { 2010,  5, 17 },  { 2010,  5, 18 },
  { 2010,  5, 19 },  { 2010,  5, 20 },  { 2010,  5, 21 },  { 2010,  5, 22 },
  { 2010,  5, 23 },  { 2010,  5, 24 },  { 2010,  5, 25 },  { 2010,  5, 26 },
  { 2010,  5, 27 },  { 2010,  5, 28 },  { 2010,  5, 29 },  { 2010,  5, 30 },
  { 2010,  5, 31 },  { 2010,  6,  1 },  { 2010,  6,  2 },  { 2010,  6,  3 },
  { 2010,  6,  4 },  { 2010,  6,  5 },  { 2010,  6,  6 },  { 2010,  6,  7 },
  { 2010,  6,  8 },  { 2010,  6,  9 },  { 2010,  6, 10 },  { 2010,  6, 11 },
  { 2010,  6, 12 },  { 2010,  6, 13 },  { 2010,  6, 14 },  { 2010,  6, 15 },
  { 2010,  6, 16 },  { 2010,  6, 17 },  { 2010,  6, 18 },  { 2010,  6, 19 },
  { 2010,  6, 20 },  { 2010,  6, 21 },  { 2010,  6, 22 },  { 2010,  6, 23 },
  { 2010,  6, 24 },  { 2010,  6, 25 },  { 2010,  6, 26 },  { 2010,  6, 27 },
  { 2010,  6, 28 },  { 2010,  6, 29 },  { 2010,  6, 30 },  { 2010,  7,  1 },
  { 2010,  7,  2 },  { 2010,  7,  3 },  { 2010,  7,  4 },  { 2010,  7,  5 },
  { 2010,  7,  6 },  { 2010,  7,  7 },  { 2010,  7,  8 },  { 2010,  7,  9 },
  { 2010,  7, 10 },  { 2010,  7, 11 },  { 2010,  7, 12 },  { 2010,  7, 13 },
  { 2010,  7, 14 },  { 2010,  7, 15 },  { 2010,  7, 16 },  { 2010,  7, 17 },
  { 2010,  7, 18 },  { 2010,  7, 19 },  { 2010,  7, 20 },  { 2010,  7, 21 },
  { 2010,  7, 22 },  { 2010,  7, 23 },  { 2010,  7, 24 },  { 2010,  7, 25 },
  { 2010,  7, 26 },  { 2010,  7, 27 },  { 2010,  7, 28 },  { 2010,  7, 29 },
  { 2010,  7, 30 },  { 2010,  7, 31 },  { 2010,  8,  1 },  { 2010,  8,  2 },
  { 2010,  8,  3 },  { 2010,  8,  4 },  { 2010,  8,  5 },  { 2010,  8,  6 },
  { 2010,  8,  7 },  { 2010,  8,  8 },  { 2010,  8,  9 },  { 2010,  8, 10 },
  { 2010,  8, 11 },  { 2010,  8, 12 },  { 2010,  8, 13 },  { 2010,  8, 14 },
  { 2010,  8, 15 },  { 2010,  8, 16 },  { 2010,  8, 17 },  { 2010,  8, 18 },
  { 2010,  8, 19 },  { 2010,  8, 20 },  { 2010,  8, 21 },  { 2010,  8, 22 },
  { 2010,  8, 23 },  { 2010,  8, 24 },  { 2010,  8, 25 },  { 2010,  8, 26 },
  { 2010,  8, 27 },  { 2010,  8, 28 },  { 2010,  8, 29 },  { 2010,  8, 30 },
  { 2010,  8, 31 },  { 2010,  9,  1 },  { 2010,  9,  2 },  { 2010,  9,  3 },
  { 2010,  9,  4 },  { 2010,  9,  5 },  { 2010,  9,  6 },  { 2010,  9,  7 },
  { 2010,  9,  8 },  { 2010,  9,  9 },  { 2010,  9, 10 },  { 2010,  9, 11 },
  { 2010,  9, 12 },  { 2010,  9, 13 },  { 2010,  9, 14 },  { 2010,  9, 15 },
  { 2010,  9, 16 },  { 2010,  9, 17 },  { 2010,  9, 18 },  { 2010,  9, 19 },
  { 2010,  9, 20 },  { 2010,  9, 21 },  { 2010,  9, 22 },  { 2010,  9, 23 },
  { 2010,  9, 24 },  { 2010,  9, 25 },  { 2010,  9, 26 },  { 2010,  9, 27 },
  { 2010,  9, 28 },  { 2010,  9, 29 },  { 2010,  9, 30 },  { 2010, 10,  1 },
  { 2010, 10,  2 },  { 2010, 10,  3 },  { 2010, 10,  4 },  { 2010, 10,  5 },
  { 2010, 10,  6 },  { 2010, 10,  7 },  { 2010, 10,  8 },  { 2010, 10,  9 },
  { 2010, 10, 10 },  { 2010, 10, 11 },  { 2010, 10, 12 },  { 2010, 10, 13 },
  { 2010, 10, 14 },  { 2010, 10, 15 },  { 2010, 10, 16 },  { 2010, 10, 17 },
  { 2010, 10, 18 },  { 2010, 10, 19 },  { 2010, 10, 20 },  { 2010, 10, 21 },
  { 2010, 10, 22 },  { 2010, 10, 23 },  { 2010, 10, 24 },  { 2010, 10, 25 },
  { 2010, 10, 26 },  { 2010, 10, 27 },  { 2010, 10, 28 },  { 2010, 10, 29 },
  { 2010, 10, 30 },  { 2010, 10, 31 },  { 2010, 11,  1 },  { 2010, 11,  2 },
  { 2010, 11,  3 },  { 2010, 11,  4 },  { 2010, 11,  5 },  { 2010, 11,  6 },
  { 2010, 11,  7 },  { 2010, 11,  8 },  { 2010, 11,  9 },  { 2010, 11, 10 },
  { 2010, 11, 11 },  { 2010, 11, 12 },  { 2010, 11, 13 },  { 2010, 11, 14 },
  { 2010, 11, 15 },  { 2010, 11, 16 },  { 2010, 11, 17 },  { 2010, 11, 18 },
  { 2010, 11, 19 },  { 2010, 11, 20 },  { 2010, 11, 21 },  { 2010, 11, 22 },
  { 2010, 11, 23 },  { 2010, 11, 24 },  { 2010, 11, 25 },  { 2010, 11, 26 },
  { 2010, 11, 27 },  { 2010, 11, 28 },  { 2010, 11, 29 },  { 2010, 11, 30 },
  { 2010, 12,  1 },  { 2010, 12,  2 },  { 2010, 12,  3 },  { 2010, 12,  4 },
  { 2010, 12,  5 },  { 2010, 12,  6 },  { 2010, 12,  7 },  { 2010, 12,  8 },
  { 2010, 12,  9 },  { 2010, 12, 10 },  { 2010, 12, 11 },  { 2010, 12, 12 },
  { 2010, 12, 13 },  { 2010, 12, 14 },  { 2010, 12, 15 },  { 2010, 12, 16 },
  { 2010, 12, 17 },  { 2010, 12, 18 },  { 2010, 12, 19 },  { 2010, 12, 20 },
  { 2010, 12, 21 },  { 2010, 12, 22 },  { 2010, 12, 23 },  { 2010, 12, 24 },
  { 2010, 12, 25 },  { 2010, 12, 26 },  { 2010, 12, 27 },  { 2010, 12, 28 },
  { 2010, 12, 29 },  { 2010, 12, 30 },  { 2010, 12, 31 },  { 2011,  1,  1 },
  { 2011,  1,  2 },  { 2011,  1,  3 },  { 2011,  1,  4 },  { 2011,  1,  5 },
  { 2011,  1,  6 },  { 2011,  1,  7 },  { 2011,  1,  8 },  { 2011,  1,  9 },
  { 2011,  1, 10 },  { 2011,  1, 11 },  { 2011,  1, 12 },  { 2011,  1, 13 },
  { 2011,  1, 14 },  { 2011,  1, 15 },  { 2011,  1, 16 },  { 2011,  1, 17 },
  { 2011,  1, 18 },  { 2011,  1, 19 },  { 2011,  1, 20 },  { 2011,  1, 21 },
  { 2011,  1, 22 },  { 2011,  1, 23 },  { 2011,  1, 24 },  { 2011,  1, 25 },
  { 2011,  1, 26 },  { 2011,  1, 27 },  { 2011,  1, 28 },  { 2011,  1, 29 },
  { 2011,  1, 30 },  { 2011,  1, 31 },  { 2011,  2,  1 },  { 2011,  2,  2 },
  { 2011,  2,  3 },  { 2011,  2,  4 },  { 2011,  2,  5 },  { 2011,  2,  6 },
  { 2011,  2,  7 },  { 2011,  2,  8 },  { 2011,  2,  9 },  { 2011,  2, 10 },
  { 2011,  2, 11 },  { 2011,  2, 12 },  { 2011,  2, 13 },  { 2011,  2, 14 },
  { 2011,  2, 15 },  { 2011,  2, 16 },  { 2011,  2, 17 },  { 2011,  2, 18 },
  { 2011,  2, 19 },  { 2011,  2, 20 },  { 2011,  2, 21 },  { 2011,  2, 22 },
  { 2011,  2, 23 },  { 2011,  2, 24 },  { 2011,  2, 25 },  { 2011,  2, 26 },
  { 2011,  2, 27 },  { 2011,  2, 28 },  { 2011,  3,  1 },  { 2011,  3,  2 },
  { 2011,  3,  3 },  { 2011,  3,  4 },  { 2011,  3,  5 },  { 2011,  3,  6 },
  { 2011,  3,  7 },  { 2011,  3,  8 },  { 2011,  3,  9 },  { 2011,  3, 10 },
  { 2011,  3, 11 },  { 2011,  3, 12 },  { 2011,  3, 13 },  { 2011,  3, 14 },
  { 2011,  3, 15 },  { 2011,  3, 16 },  { 2011,  3, 17 },  { 2011,  3, 18 },
  { 2011,  3, 19 },  { 2011,  3, 20 },  { 2011,  3, 21 },  { 2011,  3, 22 },
  { 2011,  3, 23 },  { 2011,  3, 24 },  { 2011,  3, 25 },  { 2011,  3, 26 },
  { 2011,  3, 27 },  { 2011,  3, 28 },  { 2011,  3, 29 },  { 2011,  3, 30 },
  { 2011,  3, 31 },  { 2011,  4,  1 },  { 2011,  4,  2 },  { 2011,  4,  3 },
  { 2011,  4,  4 },  { 2011,  4,  5 },  { 2011,  4,  6 },  { 2011,  4,  7 },
  { 2011,  4,  8 },  { 2011,  4,  9 },  { 2011,  4, 10 },  { 2011,  4, 11 },
  { 2011,  4, 12 },  { 2011,  4, 13 },  { 2011,  4, 14 },  { 2011,  4, 15 },
  { 2011,  4, 16 },  { 2011,  4, 17 },  { 2011,  4, 18 },  { 2011,  4, 19 },
  { 2011,  4, 20 },  { 2011,  4, 21 },  { 2011,  4, 22 },  { 2011,  4, 23 },
  { 2011,  4, 24 },  { 2011,  4, 25 },  { 2011,  4, 26 },  { 2011,  4, 27 },
  { 2011,  4, 28 },  { 2011,  4, 29 },  { 2011,  4, 30 },  { 2011,  5,  1 },
  { 2011,  5,  2 },  { 2011,  5,  3 },  { 2011,  5,  4 },  { 2011,  5,  5 },
  { 2011,  5,  6 },  { 2011,  5,  7 },  { 2011,  5,  8 },  { 2011,  5,  9 },
  { 2011,  5, 10 },  { 2011,  5, 11 },  { 2011,  5, 12 },  { 2011,  5, 13 },
  { 2011,  5, 14 },  { 2011,  5, 15 },  { 2011,  5, 16 },  { 2011,  5, 17 },
  { 2011,  5, 18 },  { 2011,  5, 19 },  { 2011,  5, 20 },  { 2011,  5, 21 },
  { 2011,  5, 22 },  { 2011,  5, 23 },  { 2011,  5, 24 },  { 2011,  5, 25 },
  { 2011,  5, 26 },  { 2011,  5, 27 },  { 2011,  5, 28 },  { 2011,  5, 29 },
  { 2011,  5, 30 },  { 2011,  5, 31 },  { 2011,  6,  1 },  { 2011,  6,  2 },
  { 2011,  6,  3 },  { 2011,  6,  4 },  { 2011,  6,  5 },  { 2011,  6,  6 },
  { 2011,  6,  7 },  { 2011,  6,  8 },  { 2011,  6,  9 },  { 2011,  6, 10 },
  { 2011,  6, 11 },  { 2011,  6, 12 },  { 2011,  6, 13 },  { 2011,  6, 14 },
  { 2011,  6, 15 },  { 2011,  6, 16 },  { 2011,  6, 17 },  { 2011,  6, 18 },
  { 2011,  6, 19 },  { 2011,  6, 20 },  { 2011,  6, 21 },  { 2011,  6, 22 },
  { 2011,  6, 23 },  { 2011,  6, 24 },  { 2011,  6, 25 },  { 2011,  6, 26 },
  { 2011,  6, 27 },  { 2011,  6, 28 },  { 2011,  6, 29 },  { 2011,  6, 30 },
  { 2011,  7,  1 },  { 2011,  7,  2 },  { 2011,  7,  3 },  { 2011,  7,  4 },
  { 2011,  7,  5 },  { 2011,  7,  6 },  { 2011,  7,  7 },  { 2011,  7,  8 },
  { 2011,  7,  9 },  { 2011,  7, 10 },  { 2011,  7, 11 },  { 2011,  7, 12 },
  { 2011,  7, 13 },  { 2011,  7, 14 },  { 2011,  7, 15 },  { 2011,  7, 16 },
  { 2011,  7, 17 },  { 2011,  7, 18 },  { 2011,  7, 19 },  { 2011,  7, 20 },
  { 2011,  7, 21 },  { 2011,  7, 22 },  { 2011,  7, 23 },  { 2011,  7, 24 },
  { 2011,  7, 25 },  { 2011,  7, 26 },  { 2011,  7, 27 },  { 2011,  7, 28 },
  { 2011,  7, 29 },  { 2011,  7, 30 },  { 2011,  7, 31 },  { 2011,  8,  1 },
  { 2011,  8,  2 },  { 2011,  8,  3 },  { 2011,  8,  4 },  { 2011,  8,  5 },
  { 2011,  8,  6 },  { 2011,  8,  7 },  { 2011,  8,  8 },  { 2011,  8,  9 },
  { 2011,  8, 10 },  { 2011,  8, 11 },  { 2011,  8, 12 },  { 2011,  8, 13 },
  { 2011,  8, 14 },  { 2011,  8, 15 },  { 2011,  8, 16 },  { 2011,  8, 17 },
  { 2011,  8, 18 },  { 2011,  8, 19 },  { 2011,  8, 20 },  { 2011,  8, 21 },
  { 2011,  8, 22 },  { 2011,  8, 23 },  { 2011,  8, 24 },  { 2011,  8, 25 },
  { 2011,  8, 26 },  { 2011,  8, 27 },  { 2011,  8, 28 },  { 2011,  8, 29 },
  { 2011,  8, 30 },  { 2011,  8, 31 },  { 2011,  9,  1 },  { 2011,  9,  2 },
  { 2011,  9,  3 },  { 2011,  9,  4 },  { 2011,  9,  5 },  { 2011,  9,  6 },
  { 2011,  9,  7 },  { 2011,  9,  8 },  { 2011,  9,  9 },  { 2011,  9, 10 },
  { 2011,  9, 11 },  { 2011,  9, 12 },  { 2011,  9, 13 },  { 2011,  9, 14 },
  { 2011,  9, 15 },  { 2011,  9, 16 },  { 2011,  9, 17 },  { 2011,  9, 18 },
  { 2011,  9, 19 },  { 2011,  9, 20 },  { 2011,  9, 21 },  { 2011,  9, 22 },
  { 2011,  9, 23 },  { 2011,  9, 24 },  { 2011,  9, 25 },  { 2011,  9, 26 },
  { 2011,  9, 27 },  { 2011,  9, 28 },  { 2011,  9, 29 },  { 2011,  9, 30 },
  { 2011, 10,  1 },  { 2011, 10,  2 },  { 2011, 10,  3 },  { 2011, 10,  4 },
  { 2011, 10,  5 },  { 2011, 10,  6 },  { 2011, 10,  7 },  { 2011, 10,  8 },
  { 2011, 10,  9 },  { 2011, 10, 10 },  { 2011, 10, 11 },  { 2011, 10, 12 },
  { 2011, 10, 13 },  { 2011, 10, 14 },  { 2011, 10, 15 },  { 2011, 10, 16 },
  { 2011, 10, 17 },  { 2011, 10, 18 },  { 2011, 10, 19 },  { 2011, 10, 20 },
  { 2011, 10, 21 },  { 2011, 10, 22 },  { 2011, 10, 23 },  { 2011, 10, 24 },
  { 2011, 10, 25 },  { 2011, 10, 26 },  { 2011, 10, 27 },  { 2011, 10, 28 },
  { 2011, 10, 29 },  { 2011, 10, 30 },  { 2011, 10, 31 },  { 2011, 11,  1 },
  { 2011, 11,  2 },  { 2011, 11,  3 },  { 2011, 11,  4 },  { 2011, 11,  5 },
  { 2011, 11,  6 },  { 2011, 11,  7 },  { 2011, 11,  8 },  { 2011, 11,  9 },
  { 2011, 11, 10 },  { 2011, 11, 11 },  { 2011, 11, 12 },  { 2011, 11, 13 },
  { 2011, 11, 14 },  { 2011, 11, 15 },  { 2011, 11, 16 },  { 2011, 11, 17 },
  { 2011, 11, 18 },  { 2011, 11, 19 },  { 2011, 11, 20 },  { 2011, 11, 21 },
  { 2011, 11, 22 },  { 2011, 11, 23 },  { 2011, 11, 24 },  { 2011, 11, 25 },
  { 2011, 11, 26 },  { 2011, 11, 27 },  { 2011, 11, 28 },  { 2011, 11, 29 },
  { 2011, 11, 30 },  { 2011, 12,  1 },  { 2011, 12,  2 },  { 2011, 12,  3 },
  { 2011, 12,  4 },  { 2011, 12,  5 },  { 2011, 12,  6 },  { 2011, 12,  7 },
  { 2011, 12,  8 },  { 2011, 12,  9 },  { 2011, 12, 10 },  { 2011, 12, 11 },
  { 2011, 12, 12 },  { 2011, 12, 13 },  { 2011, 12, 14 },  { 2011, 12, 15 },
  { 2011, 12, 16 },  { 2011, 12, 17 },  { 2011, 12, 18 },  { 2011, 12, 19 },
  { 2011, 12, 20 },  { 2011, 12, 21 },  { 2011, 12, 22 },  { 2011, 12, 23 },
  { 2011, 12, 24 },  { 2011, 12, 25 },  { 2011, 12, 26 },  { 2011, 12, 27 },
  { 2011, 12, 28 },  { 2011, 12, 29 },  { 2011, 12, 30 },  { 2011, 12, 31 },
  { 2012,  1,  1 },  { 2012,  1,  2 },  { 2012,  1,  3 },  { 2012,  1,  4 },
  { 2012,  1,  5 },  { 2012,  1,  6 },  { 2012,  1,  7 },  { 2012,  1,  8 },
  { 2012,  1,  9 },  { 2012,  1, 10 },  { 2012,  1, 11 },  { 2012,  1, 12 },
  { 2012,  1, 13 },  { 2012,  1, 14 },  { 2012,  1, 15 },  { 2012,  1, 16 },
  { 2012,  1, 17 },  { 2012,  1, 18 },  { 2012,  1, 19 },  { 2012,  1, 20 },
  { 2012,  1, 21 },  { 2012,  1, 22 },  { 2012,  1, 23 },  { 2012,  1, 24 },
  { 2012,  1, 25 },  { 2012,  1, 26 },  { 2012,  1, 27 },  { 2012,  1, 28 },
  { 2012,  1, 29 },  { 2012,  1, 30 },  { 2012,  1, 31 },  { 2012,  2,  1 },
  { 2012,  2,  2 },  { 2012,  2,  3 },  { 2012,  2,  4 },  { 2012,  2,  5 },
  { 2012,  2,  6 },  { 2012,  2,  7 },  { 2012,  2,  8 },  { 2012,  2,  9 },
  { 2012,  2, 10 },  { 2012,  2, 11 },  { 2012,  2, 12 },  { 2012,  2, 13 },
  { 2012,  2, 14 },  { 2012,  2, 15 },  { 2012,  2, 16 },  { 2012,  2, 17 },
  { 2012,  2, 18 },  { 2012,  2, 19 },  { 2012,  2, 20 },  { 2012,  2, 21 },
  { 2012,  2, 22 },  { 2012,  2, 23 },  { 2012,  2, 24 },  { 2012,  2, 25 },
  { 2012,  2, 26 },  { 2012,  2, 27 },  { 2012,  2, 28 },  { 2012,  2, 29 },
  { 2012,  3,  1 },  { 2012,  3,  2 },  { 2012,  3,  3 },  { 2012,  3,  4 },
  { 2012,  3,  5 },  { 2012,  3,  6 },  { 2012,  3,  7 },  { 2012,  3,  8 },
  { 2012,  3,  9 },  { 2012,  3, 10 },  { 2012,  3, 11 },  { 2012,  3, 12 },
  { 2012,  3, 13 },  { 2012,  3, 14 },  { 2012,  3, 15 },  { 2012,  3, 16 },
  { 2012,  3, 17 },  { 2012,  3, 18 },  { 2012,  3, 19 },  { 2012,  3, 20 },
  { 2012,  3, 21 },  { 2012,  3, 22 },  { 2012,  3, 23 },  { 2012,  3, 24 },
  { 2012,  3, 25 },  { 2012,  3, 26 },  { 2012,  3, 27 },  { 2012,  3, 28 },
  { 2012,  3, 29 },  { 2012,  3, 30 },  { 2012,  3, 31 },  { 2012,  4,  1 },
  { 2012,  4,  2 },  { 2012,  4,  3 },  { 2012,  4,  4 },  { 2012,  4,  5 },
  { 2012,  4,  6 },  { 2012,  4,  7 },  { 2012,  4,  8 },  { 2012,  4,  9 },
  { 2012,  4, 10 },  { 2012,  4, 11 },  { 2012,  4, 12 },  { 2012,  4, 13 },
  { 2012,  4, 14 },  { 2012,  4, 15 },  { 2012,  4, 16 },  { 2012,  4, 17 },
  { 2012,  4, 18 },  { 2012,  4, 19 },  { 2012,  4, 20 },  { 2012,  4, 21 },
  { 2012,  4, 22 },  { 2012,  4, 23 },  { 2012,  4, 24 },  { 2012,  4, 25 },
  { 2012,  4, 26 },  { 2012,  4, 27 },  { 2012,  4, 28 },  { 2012,  4, 29 },
  { 2012,  4, 30 },  { 2012,  5,  1 },  { 2012,  5,  2 },  { 2012,  5,  3 },
  { 2012,  5,  4 },  { 2012,  5,  5 },  { 2012,  5,  6 },  { 2012,  5,  7 },
  { 2012,  5,  8 },  { 2012,  5,  9 },  { 2012,  5, 10 },  { 2012,  5, 11 },
  { 2012,  5, 12 },  { 2012,  5, 13 },  { 2012,  5, 14 },  { 2012,  5, 15 },
  { 2012,  5, 16 },  { 2012,  5, 17 },  { 2012,  5, 18 },  { 2012,  5, 19 },
  { 2012,  5, 20 },  { 2012,  5, 21 },  { 2012,  5, 22 },  { 2012,  5, 23 },
  { 2012,  5, 24 },  { 2012,  5, 25 },  { 2012,  5, 26 },  { 2012,  5, 27 },
  { 2012,  5, 28 },  { 2012,  5, 29 },  { 2012,  5, 30 },  { 2012,  5, 31 },
  { 2012,  6,  1 },  { 2012,  6,  2 },  { 2012,  6,  3 },  { 2012,  6,  4 },
  { 2012,  6,  5 },  { 2012,  6,  6 },  { 2012,  6,  7 },  { 2012,  6,  8 },
  { 2012,  6,  9 },  { 2012,  6, 10 },  { 2012,  6, 11 },  { 2012,  6, 12 },
  { 2012,  6, 13 },  { 2012,  6, 14 },  { 2012,  6, 15 },  { 2012,  6, 16 },
  { 2012,  6, 17 },  { 2012,  6, 18 },  { 2012,  6, 19 },  { 2012,  6, 20 },
  { 2012,  6, 21 },  { 2012,  6, 22 },  { 2012,  6, 23 },  { 2012,  6, 24 },
  { 2012,  6, 25 },  { 2012,  6, 26 },  { 2012,  6, 27 },  { 2012,  6, 28 },
  { 2012,  6, 29 },  { 2012,  6, 30 },  { 2012,  7,  1 },  { 2012,  7,  2 },
  { 2012,  7,  3 },  { 2012,  7,  4 },  { 2012,  7,  5 },  { 2012,  7,  6 },
  { 2012,  7,  7 },  { 2012,  7,  8 },  { 2012,  7,  9 },  { 2012,  7, 10 },
  { 2012,  7, 11 },  { 2012,  7, 12 },  { 2012,  7, 13 },  { 2012,  7, 14 },
  { 2012,  7, 15 },  { 2012,  7, 16 },  { 2012,  7, 17 },  { 2012,  7, 18 },
  { 2012,  7, 19 },  { 2012,  7, 20 },  { 2012,  7, 21 },  { 2012,  7, 22 },
  { 2012,  7, 23 },  { 2012,  7, 24 },  { 2012,  7, 25 },  { 2012,  7, 26 },
  { 2012,  7, 27 },  { 2012,  7, 28 },  { 2012,  7, 29 },  { 2012,  7, 30 },
  { 2012,  7, 31 },  { 2012,  8,  1 },  { 2012,  8,  2 },  { 2012,  8,  3 },
  { 2012,  8,  4 },  { 2012,  8,  5 },  { 2012,  8,  6 },  { 2012,  8,  7 },
  { 2012,  8,  8 },  { 2012,  8,  9 },  { 2012,  8, 10 },  { 2012,  8, 11 },
  { 2012,  8, 12 },  { 2012,  8, 13 },  { 2012,  8, 14 },  { 2012,  8, 15 },
  { 2012,  8, 16 },  { 2012,  8, 17 },  { 2012,  8, 18 },  { 2012,  8, 19 },
  { 2012,  8, 20 },  { 2012,  8, 21 },  { 2012,  8, 22 },  { 2012,  8, 23 },
  { 2012,  8, 24 },  { 2012,  8, 25 },  { 2012,  8, 26 },  { 2012,  8, 27 },
  { 2012,  8, 28 },  { 2012,  8, 29 },  { 2012,  8, 30 },  { 2012,  8, 31 },
  { 2012,  9,  1 },  { 2012,  9,  2 },  { 2012,  9,  3 },  { 2012,  9,  4 },
  { 2012,  9,  5 },  { 2012,  9,  6 },  { 2012,  9,  7 },  { 2012,  9,  8 },
  { 2012,  9,  9 },  { 2012,  9, 10 },  { 2012,  9, 11 },  { 2012,  9, 12 },
  { 2012,  9, 13 },  { 2012,  9, 14 },  { 2012,  9, 15 },  { 2012,  9, 16 },
  { 2012,  9, 17 },  { 2012,  9, 18 },  { 2012,  9, 19 },  { 2012,  9, 20 },
  { 2012,  9, 21 },  { 2012,  9, 22 },  { 2012,  9, 23 },  { 2012,  9, 24 },
  { 2012,  9, 25 },  { 2012,  9, 26 },  { 2012,  9, 27 },  { 2012,  9, 28 },
  { 2012,  9, 29 },  { 2012,  9, 30 },  { 2012, 10,  1 },  { 2012, 10,  2 },
  { 2012, 10,  3 },  { 2012, 10,  4 },  { 2012, 10,  5 },  { 2012, 10,  6 },
  { 2012, 10,  7 },  { 2012, 10,  8 },  { 2012, 10,  9 },  { 2012, 10, 10 },
  { 2012, 10, 11 },  { 2012, 10, 12 },  { 2012, 10, 13 },  { 2012, 10, 14 },
  { 2012, 10, 15 },  { 2012, 10, 16 },  { 2012, 10, 17 },  { 2012, 10, 18 },
  { 2012, 10, 19 },  { 2012, 10, 20 },  { 2012, 10, 21 },  { 2012, 10, 22 },
  { 2012, 10, 23 },  { 2012, 10, 24 },  { 2012, 10, 25 },  { 2012, 10, 26 },
  { 2012, 10, 27 },  { 2012, 10, 28 },  { 2012, 10, 29 },  { 2012, 10, 30 },
  { 2012, 10, 31 },  { 2012, 11,  1 },  { 2012, 11,  2 },  { 2012, 11,  3 },
  { 2012, 11,  4 },  { 2012, 11,  5 },  { 2012, 11,  6 },  { 2012, 11,  7 },
  { 2012, 11,  8 },  { 2012, 11,  9 },  { 2012, 11, 10 },  { 2012, 11, 11 },
  { 2012, 11, 12 },  { 2012, 11, 13 },  { 2012, 11, 14 },  { 2012, 11, 15 },
  { 2012, 11, 16 },  { 2012, 11, 17 },  { 2012, 11, 18 },  { 2012, 11, 19 },
  { 2012, 11, 20 },  { 2012, 11, 21 },  { 2012, 11, 22 },  { 2012, 11, 23 },
  { 2012, 11, 24 },  { 2012, 11, 25 },  { 2012, 11, 26 },  { 2012, 11, 27 },
  { 2012, 11, 28 },  { 2012, 11, 29 },  { 2012, 11, 30 },  { 2012, 12,  1 },
  { 2012, 12,  2 },  { 2012, 12,  3 },  { 2012, 12,  4 },  { 2012, 12,  5 },
  { 2012, 12,  6 },  { 2012, 12,  7 },  { 2012, 12,  8 },  { 2012, 12,  9 },
  { 2012, 12, 10 },  { 2012, 12, 11 },  { 2012, 12, 12 },  { 2012, 12, 13 },
  { 2012, 12, 14 },  { 2012, 12, 15 },  { 2012, 12, 16 },  { 2012, 12, 17 },
  { 2012, 12, 18 },  { 2012, 12, 19 },  { 2012, 12, 20 },  { 2012, 12, 21 },
  { 2012, 12, 22 },  { 2012, 12, 23 },  { 2012, 12, 24 },  { 2012, 12, 25 },
  { 2012, 12, 26 },  { 2012, 12, 27 },  { 2012, 12, 28 },  { 2012, 12, 29 },
  { 2012, 12, 30 },  { 2012, 12, 31 },  { 2013,  1,  1 },  { 2013,  1,  2 },
  { 2013,  1,  3 },  { 2013,  1,  4 },  { 2013,  1,  5 },  { 2013,  1,  6 },
  { 2013,  1,  7 },  { 2013,  1,  8 },  { 2013,  1,  9 },  { 2013,  1, 10 },
  { 2013,  1, 11 },  { 2013,  1, 12 },  { 2013,  1, 13 },  { 2013,  1, 14 },
  { 2013,  1, 15 },  { 2013,  1, 16 },  { 2013,  1, 17 },  { 2013,  1, 18 },
  { 2013,  1, 19 },  { 2013,  1, 20 },  { 2013,  1, 21 },  { 2013,  1, 22 },
  { 2013,  1, 23 },  { 2013,  1, 24 },  { 2013,  1, 25 },  { 2013,  1, 26 },
  { 2013,  1, 27 },  { 2013,  1, 28 },  { 2013,  1, 29 },  { 2013,  1, 30 },
  { 2013,  1, 31 },  { 2013,  2,  1 },  { 2013,  2,  2 },  { 2013,  2,  3 },
  { 2013,  2,  4 },  { 2013,  2,  5 },  { 2013,  2,  6 },  { 2013,  2,  7 },
  { 2013,  2,  8 },  { 2013,  2,  9 },  { 2013,  2, 10 },  { 2013,  2, 11 },
  { 2013,  2, 12 },  { 2013,  2, 13 },  { 2013,  2, 14 },  { 2013,  2, 15 },
  { 2013,  2, 16 },  { 2013,  2, 17 },  { 2013,  2, 18 },  { 2013,  2, 19 },
  { 2013,  2, 20 },  { 2013,  2, 21 },  { 2013,  2, 22 },  { 2013,  2, 23 },
  { 2013,  2, 24 },  { 2013,  2, 25 },  { 2013,  2, 26 },  { 2013,  2, 27 },
  { 2013,  2, 28 },  { 2013,  3,  1 },  { 2013,  3,  2 },  { 2013,  3,  3 },
  { 2013,  3,  4 },  { 2013,  3,  5 },  { 2013,  3,  6 },  { 2013,  3,  7 },
  { 2013,  3,  8 },  { 2013,  3,  9 },  { 2013,  3, 10 },  { 2013,  3, 11 },
  { 2013,  3, 12 },  { 2013,  3, 13 },  { 2013,  3, 14 },  { 2013,  3, 15 },
  { 2013,  3, 16 },  { 2013,  3, 17 },  { 2013,  3, 18 },  { 2013,  3, 19 },
  { 2013,  3, 20 },  { 2013,  3, 21 },  { 2013,  3, 22 },  { 2013,  3, 23 },
  { 2013,  3, 24 },  { 2013,  3, 25 },  { 2013,  3, 26 },  { 2013,  3, 27 },
  { 2013,  3, 28 },  { 2013,  3, 29 },  { 2013,  3, 30 },  { 2013,  3, 31 },
  { 2013,  4,  1 },  { 2013,  4,  2 },  { 2013,  4,  3 },  { 2013,  4,  4 },
  { 2013,  4,  5 },  { 2013,  4,  6 },  { 2013,  4,  7 },  { 2013,  4,  8 },
  { 2013,  4,  9 },  { 2013,  4, 10 },  { 2013,  4, 11 },  { 2013,  4, 12 },
  { 2013,  4, 13 },  { 2013,  4, 14 },  { 2013,  4, 15 },  { 2013,  4, 16 },
  { 2013,  4, 17 },  { 2013,  4, 18 },  { 2013,  4, 19 },  { 2013,  4, 20 },
  { 2013,  4, 21 },  { 2013,  4, 22 },  { 2013,  4, 23 },  { 2013,  4, 24 },
  { 2013,  4, 25 },  { 2013,  4, 26 },  { 2013,  4, 27 },  { 2013,  4, 28 },
  { 2013,  4, 29 },  { 2013,  4, 30 },  { 2013,  5,  1 },  { 2013,  5,  2 },
  { 2013,  5,  3 },  { 2013,  5,  4 },  { 2013,  5,  5 },  { 2013,  5,  6 },
  { 2013,  5,  7 },  { 2013,  5,  8 },  { 2013,  5,  9 },  { 2013,  5, 10 },
  { 2013,  5, 11 },  { 2013,  5, 12 },  { 2013,  5, 13 },  { 2013,  5, 14 },
  { 2013,  5, 15 },  { 2013,  5, 16 },  { 2013,  5, 17 },  { 2013,  5, 18 },
  { 2013,  5, 19 },  { 2013,  5, 20 },  { 2013,  5, 21 },  { 2013,  5, 22 },
  { 2013,  5, 23 },  { 2013,  5, 24 },  { 2013,  5, 25 },  { 2013,  5, 26 },
  { 2013,  5, 27 },  { 2013,  5, 28 },  { 2013,  5, 29 },  { 2013,  5, 30 },
  { 2013,  5, 31 },  { 2013,  6,  1 },  { 2013,  6,  2 },  { 2013,  6,  3 },
  { 2013,  6,  4 },  { 2013,  6,  5 },  { 2013,  6,  6 },  { 2013,  6,  7 },
  { 2013,  6,  8 },  { 2013,  6,  9 },  { 2013,  6, 10 },  { 2013,  6, 11 },
  { 2013,  6, 12 },  { 2013,  6, 13 },  { 2013,  6, 14 },  { 2013,  6, 15 },
  { 2013,  6, 16 },  { 2013,  6, 17 },  { 2013,  6, 18 },  { 2013,  6, 19 },
  { 2013,  6, 20 },  { 2013,  6, 21 },  { 2013,  6, 22 },  { 2013,  6, 23 },
  { 2013,  6, 24 },  { 2013,  6, 25 },  { 2013,  6, 26 },  { 2013,  6, 27 },
  { 2013,  6, 28 },  { 2013,  6, 29 },  { 2013,  6, 30 },  { 2013,  7,  1 },
  { 2013,  7,  2 },  { 2013,  7,  3 },  { 2013,  7,  4 },  { 2013,  7,  5 },
  { 2013,  7,  6 },  { 2013,  7,  7 },  { 2013,  7,  8 },  { 2013,  7,  9 },
  { 2013,  7, 10 },  { 2013,  7, 11 },  { 2013,  7, 12 },  { 2013,  7, 13 },
  { 2013,  7, 14 },  { 2013,  7, 15 },  { 2013,  7, 16 },  { 2013,  7, 17 },
  { 2013,  7, 18 },  { 2013,  7, 19 },  { 2013,  7, 20 },  { 2013,  7, 21 },
  { 2013,  7, 22 },  { 2013,  7, 23 },  { 2013,  7, 24 },  { 2013,  7, 25 },
  { 2013,  7, 26 },  { 2013,  7, 27 },  { 2013,  7, 28 },  { 2013,  7, 29 },
  { 2013,  7, 30 },  { 2013,  7, 31 },  { 2013,  8,  1 },  { 2013,  8,  2 },
  { 2013,  8,  3 },  { 2013,  8,  4 },  { 2013,  8,  5 },  { 2013,  8,  6 },
  { 2013,  8,  7 },  { 2013,  8,  8 },  { 2013,  8,  9 },  { 2013,  8, 10 },
  { 2013,  8, 11 },  { 2013,  8, 12 },  { 2013,  8, 13 },  { 2013,  8, 14 },
  { 2013,  8, 15 },  { 2013,  8, 16 },  { 2013,  8, 17 },  { 2013,  8, 18 },
  { 2013,  8, 19 },  { 2013,  8, 20 },  { 2013,  8, 21 },  { 2013,  8, 22 },
  { 2013,  8, 23 },  { 2013,  8, 24 },  { 2013,  8, 25 },  { 2013,  8, 26 },
  { 2013,  8, 27 },  { 2013,  8, 28 },  { 2013,  8, 29 },  { 2013,  8, 30 },
  { 2013,  8, 31 },  { 2013,  9,  1 },  { 2013,  9,  2 },  { 2013,  9,  3 },
  { 2013,  9,  4 },  { 2013,  9,  5 },  { 2013,  9,  6 },  { 2013,  9,  7 },
  { 2013,  9,  8 },  { 2013,  9,  9 },  { 2013,  9, 10 },  { 2013,  9, 11 },
  { 2013,  9, 12 },  { 2013,  9, 13 },  { 2013,  9, 14 },  { 2013,  9, 15 },
  { 2013,  9, 16 },  { 2013,  9, 17 },  { 2013,  9, 18 },  { 2013,  9, 19 },
  { 2013,  9, 20 },  { 2013,  9, 21 },  { 2013,  9, 22 },  { 2013,  9, 23 },
  { 2013,  9, 24 },  { 2013,  9, 25 },  { 2013,  9, 26 },  { 2013,  9, 27 },
  { 2013,  9, 28 },  { 2013,  9, 29 },  { 2013,  9, 30 },  { 2013, 10,  1 },
  { 2013, 10,  2 },  { 2013, 10,  3 },  { 2013, 10,  4 },  { 2013, 10,  5 },
  { 2013, 10,  6 },  { 2013, 10,  7 },  { 2013, 10,  8 },  { 2013, 10,  9 },
  { 2013, 10, 10 },  { 2013, 10, 11 },  { 2013, 10, 12 },  { 2013, 10, 13 },
  { 2013, 10, 14 },  { 2013, 10, 15 },  { 2013, 10, 16 },  { 2013, 10, 17 },
  { 2013, 10, 18 },  { 2013, 10, 19 },  { 2013, 10, 20 },  { 2013, 10, 21 },
  { 2013, 10, 22 },  { 2013, 10, 23 },  { 2013, 10, 24 },  { 2013, 10, 25 },
  { 2013, 10, 26 },  { 2013, 10, 27 },  { 2013, 10, 28 },  { 2013, 10, 29 },
  { 2013, 10, 30 },  { 2013, 10, 31 },  { 2013, 11,  1 },  { 2013, 11,  2 },
  { 2013, 11,  3 },  { 2013, 11,  4 },  { 2013, 11,  5 },  { 2013, 11,  6 },
  { 2013, 11,  7 },  { 2013, 11,  8 },  { 2013, 11,  9 },  { 2013, 11, 10 },
  { 2013, 11, 11 },  { 2013, 11, 12 },  { 2013, 11, 13 },  { 2013, 11, 14 },
  { 2013, 11, 15 },  { 2013, 11, 16 },  { 2013, 11, 17 },  { 2013, 11, 18 },
  { 2013, 11, 19 },  { 2013, 11, 20 },  { 2013, 11, 21 },  { 2013, 11, 22 },
  { 2013, 11, 23 },  { 2013, 11, 24 },  { 2013, 11, 25 },  { 2013, 11, 26 },
  { 2013, 11, 27 },  { 2013, 11, 28 },  { 2013, 11, 29 },  { 2013, 11, 30 },
  { 2013, 12,  1 },  { 2013, 12,  2 },  { 2013, 12,  3 },  { 2013, 12,  4 },
  { 2013, 12,  5 },  { 2013, 12,  6 },  { 2013, 12,  7 },  { 2013, 12,  8 },
  { 2013, 12,  9 },  { 2013, 12, 10 },  { 2013, 12, 11 },  { 2013, 12, 12 },
  { 2013, 12, 13 },  { 2013, 12, 14 },  { 2013, 12, 15 },  { 2013, 12, 16 },
  { 2013, 12, 17 },  { 2013, 12, 18 },  { 2013, 12, 19 },  { 2013, 12, 20 },
  { 2013, 12, 21 },  { 2013, 12, 22 },  { 2013, 12, 23 },  { 2013, 12, 24 },
  { 2013, 12, 25 },  { 2013, 12, 26 },  { 2013, 12, 27 },  { 2013, 12, 28 },
  { 2013, 12, 29 },  { 2013, 12, 30 },  { 2013, 12, 31 },  { 2014,  1,  1 },
  { 2014,  1,  2 },  { 2014,  1,  3 },  { 2014,  1,  4 },  { 2014,  1,  5 },
  { 2014,  1,  6 },  { 2014,  1,  7 },  { 2014,  1,  8 },  { 2014,  1,  9 },
  { 2014,  1, 10 },  { 2014,  1, 11 },  { 2014,  1, 12 },  { 2014,  1, 13 },
  { 2014,  1, 14 },  { 2014,  1, 15 },  { 2014,  1, 16 },  { 2014,  1, 17 },
  { 2014,  1, 18 },  { 2014,  1, 19 },  { 2014,  1, 20 },  { 2014,  1, 21 },
  { 2014,  1, 22 },  { 2014,  1, 23 },  { 2014,  1, 24 },  { 2014,  1, 25 },
  { 2014,  1, 26 },  { 2014,  1, 27 },  { 2014,  1, 28 },  { 2014,  1, 29 },
  { 2014,  1, 30 },  { 2014,  1, 31 },  { 2014,  2,  1 },  { 2014,  2,  2 },
  { 2014,  2,  3 },  { 2014,  2,  4 },  { 2014,  2,  5 },  { 2014,  2,  6 },
  { 2014,  2,  7 },  { 2014,  2,  8 },  { 2014,  2,  9 },  { 2014,  2, 10 },
  { 2014,  2, 11 },  { 2014,  2, 12 },  { 2014,  2, 13 },  { 2014,  2, 14 },
  { 2014,  2, 15 },  { 2014,  2, 16 },  { 2014,  2, 17 },  { 2014,  2, 18 },
  { 2014,  2, 19 },  { 2014,  2, 20 },  { 2014,  2, 21 },  { 2014,  2, 22 },
  { 2014,  2, 23 },  { 2014,  2, 24 },  { 2014,  2, 25 },  { 2014,  2, 26 },
  { 2014,  2, 27 },  { 2014,  2, 28 },  { 2014,  3,  1 },  { 2014,  3,  2 },
  { 2014,  3,  3 },  { 2014,  3,  4 },  { 2014,  3,  5 },  { 2014,  3,  6 },
  { 2014,  3,  7 },  { 2014,  3,  8 },  { 2014,  3,  9 },  { 2014,  3, 10 },
  { 2014,  3, 11 },  { 2014,  3, 12 },  { 2014,  3, 13 },  { 2014,  3, 14 },
  { 2014,  3, 15 },  { 2014,  3, 16 },  { 2014,  3, 17 },  { 2014,  3, 18 },
  { 2014,  3, 19 },  { 2014,  3, 20 },  { 2014,  3, 21 },  { 2014,  3, 22 },
  { 2014,  3, 23 },  { 2014,  3, 24 },  { 2014,  3, 25 },  { 2014,  3, 26 },
  { 2014,  3, 27 },  { 2014,  3, 28 },  { 2014,  3, 29 },  { 2014,  3, 30 },
  { 2014,  3, 31 },  { 2014,  4,  1 },  { 2014,  4,  2 },  { 2014,  4,  3 },
  { 2014,  4,  4 },  { 2014,  4,  5 },  { 2014,  4,  6 },  { 2014,  4,  7 },
  { 2014,  4,  8 },  { 2014,  4,  9 },  { 2014,  4, 10 },  { 2014,  4, 11 },
  { 2014,  4, 12 },  { 2014,  4, 13 },  { 2014,  4, 14 },  { 2014,  4, 15 },
  { 2014,  4, 16 },  { 2014,  4, 17 },  { 2014,  4, 18 },  { 2014,  4, 19 },
  { 2014,  4, 20 },  { 2014,  4, 21 },  { 2014,  4, 22 },  { 2014,  4, 23 },
  { 2014,  4, 24 },  { 2014,  4, 25 },  { 2014,  4, 26 },  { 2014,  4, 27 },
  { 2014,  4, 28 },  { 2014,  4, 29 },  { 2014,  4, 30 },  { 2014,  5,  1 },
  { 2014,  5,  2 },  { 2014,  5,  3 },  { 2014,  5,  4 },  { 2014,  5,  5 },
  { 2014,  5,  6 },  { 2014,  5,  7 },  { 2014,  5,  8 },  { 2014,  5,  9 },
  { 2014,  5, 10 },  { 2014,  5, 11 },  { 2014,  5, 12 },  { 2014,  5, 13 },
  { 2014,  5, 14 },  { 2014,  5, 15 },  { 2014,  5, 16 },  { 2014,  5, 17 },
  { 2014,  5, 18 },  { 2014,  5, 19 },  { 2014,  5, 20 },  { 2014,  5, 21 },
  { 2014,  5, 22 },  { 2014,  5, 23 },  { 2014,  5, 24 },  { 2014,  5, 25 },
  { 2014,  5, 26 },  { 2014,  5, 27 },  { 2014,  5, 28 },  { 2014,  5, 29 },
  { 2014,  5, 30 },  { 2014,  5, 31 },  { 2014,  6,  1 },  { 2014,  6,  2 },
  { 2014,  6,  3 },  { 2014,  6,  4 },  { 2014,  6,  5 },  { 2014,  6,  6 },
  { 2014,  6,  7 },  { 2014,  6,  8 },  { 2014,  6,  9 },  { 2014,  6, 10 },
  { 2014,  6, 11 },  { 2014,  6, 12 },  { 2014,  6, 13 },  { 2014,  6, 14 },
  { 2014,  6, 15 },  { 2014,  6, 16 },  { 2014,  6, 17 },  { 2014,  6, 18 },
  { 2014,  6, 19 },  { 2014,  6, 20 },  { 2014,  6, 21 },  { 2014,  6, 22 },
  { 2014,  6, 23 },  { 2014,  6, 24 },  { 2014,  6, 25 },  { 2014,  6, 26 },
  { 2014,  6, 27 },  { 2014,  6, 28 },  { 2014,  6, 29 },  { 2014,  6, 30 },
  { 2014,  7,  1 },  { 2014,  7,  2 },  { 2014,  7,  3 },  { 2014,  7,  4 },
  { 2014,  7,  5 },  { 2014,  7,  6 },  { 2014,  7,  7 },  { 2014,  7,  8 },
  { 2014,  7,  9 },  { 2014,  7, 10 },  { 2014,  7, 11 },  { 2014,  7, 12 },
  { 2014,  7, 13 },  { 2014,  7, 14 },  { 2014,  7, 15 },  { 2014,  7, 16 },
  { 2014,  7, 17 },  { 2014,  7, 18 },  { 2014,  7, 19 },  { 2014,  7, 20 },
  { 2014,  7, 21 },  { 2014,  7, 22 },  { 2014,  7, 23 },  { 2014,  7, 24 },
  { 2014,  7, 25 },  { 2014,  7, 26 },  { 2014,  7, 27 },  { 2014,  7, 28 },
  { 2014,  7, 29 },  { 2014,  7, 30 },  { 2014,  7, 31 },  { 2014,  8,  1 },
  { 2014,  8,  2 },  { 2014,  8,  3 },  { 2014,  8,  4 },  { 2014,  8,  5 },
  { 2014,  8,  6 },  { 2014,  8,  7 },  { 2014,  8,  8 },  { 2014,  8,  9 },
  { 2014,  8, 10 },  { 2014,  8, 11 },  { 2014,  8, 12 },  { 2014,  8, 13 },
  { 2014,  8, 14 },  { 2014,  8, 15 },  { 2014,  8, 16 },  { 2014,  8, 17 },
  { 2014,  8, 18 },  { 2014,  8, 19 },  { 2014,  8, 20 },  { 2014,  8, 21 },
  { 2014,  8, 22 },  { 2014,  8, 23 },  { 2014,  8, 24 },  { 2014,  8, 25 },
  { 2014,  8, 26 },  { 2014,  8, 27 },  { 2014,  8, 28 },  { 2014,  8, 29 },
  { 2014,  8, 30 },  { 2014,  8, 31 },  { 2014,  9,  1 },  { 2014,  9,  2 },
  { 2014,  9,  3 },  { 2014,  9,  4 },  { 2014,  9,  5 },  { 2014,  9,  6 },
  { 2014,  9,  7 },  { 2014,  9,  8 },  { 2014,  9,  9 },  { 2014,  9, 10 },
  { 2014,  9, 11 },  { 2014,  9, 12 },  { 2014,  9, 13 },  { 2014,  9, 14 },
  { 2014,  9, 15 },  { 2014,  9, 16 },  { 2014,  9, 17 },  { 2014,  9, 18 },
  { 2014,  9, 19 },  { 2014,  9, 20 },  { 2014,  9, 21 },  { 2014,  9, 22 },
  { 2014,  9, 23 },  { 2014,  9, 24 },  { 2014,  9, 25 },  { 2014,  9, 26 },
  { 2014,  9, 27 },  { 2014,  9, 28 },  { 2014,  9, 29 },  { 2014,  9, 30 },
  { 2014, 10,  1 },  { 2014, 10,  2 },  { 2014, 10,  3 },  { 2014, 10,  4 },
  { 2014, 10,  5 },  { 2014, 10,  6 },  { 2014, 10,  7 },  { 2014, 10,  8 },
  { 2014, 10,  9 },  { 2014, 10, 10 },  { 2014, 10, 11 },  { 2014, 10, 12 },
  { 2014, 10, 13 },  { 2014, 10, 14 },  { 2014, 10, 15 },  { 2014, 10, 16 },
  { 2014, 10, 17 },  { 2014, 10, 18 },  { 2014, 10, 19 },  { 2014, 10, 20 },
  { 2014, 10, 21 },  { 2014, 10, 22 },  { 2014, 10, 23 },  { 2014, 10, 24 },
  { 2014, 10, 25 },  { 2014, 10, 26 },  { 2014, 10, 27 },  { 2014, 10, 28 },
  { 2014, 10, 29 },  { 2014, 10, 30 },  { 2014, 10, 31 },  { 2014, 11,  1 },
  { 2014, 11,  2 },  { 2014, 11,  3 },  { 2014, 11,  4 },  { 2014, 11,  5 },
  { 2014, 11,  6 },  { 2014, 11,  7 },  { 2014, 11,  8 },  { 2014, 11,  9 },
  { 2014, 11, 10 },  { 2014, 11, 11 },  { 2014, 11, 12 },  { 2014, 11, 13 },
  { 2014, 11, 14 },  { 2014, 11, 15 },  { 2014, 11, 16 },  { 2014, 11, 17 },
  { 2014, 11, 18 },  { 2014, 11, 19 },  { 2014, 11, 20 },  { 2014, 11, 21 },
  { 2014, 11, 22 },  { 2014, 11, 23 },  { 2014, 11, 24 },  { 2014, 11, 25 },
  { 2014, 11, 26 },  { 2014, 11, 27 },  { 2014, 11, 28 },  { 2014, 11, 29 },
  { 2014, 11, 30 },  { 2014, 12,  1 },  { 2014, 12,  2 },  { 2014, 12,  3 },
  { 2014, 12,  4 },  { 2014, 12,  5 },  { 2014, 12,  6 },  { 2014, 12,  7 },
  { 2014, 12,  8 },  { 2014, 12,  9 },  { 2014, 12, 10 },  { 2014, 12, 11 },
  { 2014, 12, 12 },  { 2014, 12, 13 },  { 2014, 12, 14 },  { 2014, 12, 15 },
  { 2014, 12, 16 },  { 2014, 12, 17 },  { 2014, 12, 18 },  { 2014, 12, 19 },
  { 2014, 12, 20 },  { 2014, 12, 21 },  { 2014, 12, 22 },  { 2014, 12, 23 },
  { 2014, 12, 24 },  { 2014, 12, 25 },  { 2014, 12, 26 },  { 2014, 12, 27 },
  { 2014, 12, 28 },  { 2014, 12, 29 },  { 2014, 12, 30 },  { 2014, 12, 31 },
  { 2015,  1,  1 },  { 2015,  1,  2 },  { 2015,  1,  3 },  { 2015,  1,  4 },
  { 2015,  1,  5 },  { 2015,  1,  6 },  { 2015,  1,  7 },  { 2015,  1,  8 },
  { 2015,  1,  9 },  { 2015,  1, 10 },  { 2015,  1, 11 },  { 2015,  1, 12 },
  { 2015,  1, 13 },  { 2015,  1, 14 },  { 2015,  1, 15 },  { 2015,  1, 16 },
  { 2015,  1, 17 },  { 2015,  1, 18 },  { 2015,  1, 19 },  { 2015,  1, 20 },
  { 2015,  1, 21 },  { 2015,  1, 22 },  { 2015,  1, 23 },  { 2015,  1, 24 },
  { 2015,  1, 25 },  { 2015,  1, 26 },  { 2015,  1, 27 },  { 2015,  1, 28 },
  { 2015,  1, 29 },  { 2015,  1, 30 },  { 2015,  1, 31 },  { 2015,  2,  1 },
  { 2015,  2,  2 },  { 2015,  2,  3 },  { 2015,  2,  4 },  { 2015,  2,  5 },
  { 2015,  2,  6 },  { 2015,  2,  7 },  { 2015,  2,  8 },  { 2015,  2,  9 },
  { 2015,  2, 10 },  { 2015,  2, 11 },  { 2015,  2, 12 },  { 2015,  2, 13 },
  { 2015,  2, 14 },  { 2015,  2, 15 },  { 2015,  2, 16 },  { 2015,  2, 17 },
  { 2015,  2, 18 },  { 2015,  2, 19 },  { 2015,  2, 20 },  { 2015,  2, 21 },
  { 2015,  2, 22 },  { 2015,  2, 23 },  { 2015,  2, 24 },  { 2015,  2, 25 },
  { 2015,  2, 26 },  { 2015,  2, 27 },  { 2015,  2, 28 },  { 2015,  3,  1 },
  { 2015,  3,  2 },  { 2015,  3,  3 },  { 2015,  3,  4 },  { 2015,  3,  5 },
  { 2015,  3,  6 },  { 2015,  3,  7 },  { 2015,  3,  8 },  { 2015,  3,  9 },
  { 2015,  3, 10 },  { 2015,  3, 11 },  { 2015,  3, 12 },  { 2015,  3, 13 },
  { 2015,  3, 14 },  { 2015,  3, 15 },  { 2015,  3, 16 },  { 2015,  3, 17 },
  { 2015,  3, 18 },  { 2015,  3, 19 },  { 2015,  3, 20 },  { 2015,  3, 21 },
  { 2015,  3, 22 },  { 2015,  3, 23 },  { 2015,  3, 24 },  { 2015,  3, 25 },
  { 2015,  3, 26 },  { 2015,  3, 27 },  { 2015,  3, 28 },  { 2015,  3, 29 },
  { 2015,  3, 30 },  { 2015,  3, 31 },  { 2015,  4,  1 },  { 2015,  4,  2 },
  { 2015,  4,  3 },  { 2015,  4,  4 },  { 2015,  4,  5 },  { 2015,  4,  6 },
  { 2015,  4,  7 },  { 2015,  4,  8 },  { 2015,  4,  9 },  { 2015,  4, 10 },
  { 2015,  4, 11 },  { 2015,  4, 12 },  { 2015,  4, 13 },  { 2015,  4, 14 },
  { 2015,  4, 15 },  { 2015,  4, 16 },  { 2015,  4, 17 },  { 2015,  4, 18 },
  { 2015,  4, 19 },  { 2015,  4, 20 },  { 2015,  4, 21 },  { 2015,  4, 22 },
  { 2015,  4, 23 },  { 2015,  4, 24 },  { 2015,  4, 25 },  { 2015,  4, 26 },
  { 2015,  4, 27 },  { 2015,  4, 28 },  { 2015,  4, 29 },  { 2015,  4, 30 },
  { 2015,  5,  1 },  { 2015,  5,  2 },  { 2015,  5,  3 },  { 2015,  5,  4 },
  { 2015,  5,  5 },  { 2015,  5,  6 },  { 2015,  5,  7 },  { 2015,  5,  8 },
  { 2015,  5,  9 },  { 2015,  5, 10 },  { 2015,  5, 11 },  { 2015,  5, 12 },
  { 2015,  5, 13 },  { 2015,  5, 14 },  { 2015,  5, 15 },  { 2015,  5, 16 },
  { 2015,  5, 17 },  { 2015,  5, 18 },  { 2015,  5, 19 },  { 2015,  5, 20 },
  { 2015,  5, 21 },  { 2015,  5, 22 },  { 2015,  5, 23 },  { 2015,  5, 24 },
  { 2015,  5, 25 },  { 2015,  5, 26 },  { 2015,  5, 27 },  { 2015,  5, 28 },
  { 2015,  5, 29 },  { 2015,  5, 30 },  { 2015,  5, 31 },  { 2015,  6,  1 },
  { 2015,  6,  2 },  { 2015,  6,  3 },  { 2015,  6,  4 },  { 2015,  6,  5 },
  { 2015,  6,  6 },  { 2015,  6,  7 },  { 2015,  6,  8 },  { 2015,  6,  9 },
  { 2015,  6, 10 },  { 2015,  6, 11 },  { 2015,  6, 12 },  { 2015,  6, 13 },
  { 2015,  6, 14 },  { 2015,  6, 15 },  { 2015,  6, 16 },  { 2015,  6, 17 },
  { 2015,  6, 18 },  { 2015,  6, 19 },  { 2015,  6, 20 },  { 2015,  6, 21 },
  { 2015,  6, 22 },  { 2015,  6, 23 },  { 2015,  6, 24 },  { 2015,  6, 25 },
  { 2015,  6, 26 },  { 2015,  6, 27 },  { 2015,  6, 28 },  { 2015,  6, 29 },
  { 2015,  6, 30 },  { 2015,  7,  1 },  { 2015,  7,  2 },  { 2015,  7,  3 },
  { 2015,  7,  4 },  { 2015,  7,  5 },  { 2015,  7,  6 },  { 2015,  7,  7 },
  { 2015,  7,  8 },  { 2015,  7,  9 },  { 2015,  7, 10 },  { 2015,  7, 11 },
  { 2015,  7, 12 },  { 2015,  7, 13 },  { 2015,  7, 14 },  { 2015,  7, 15 },
  { 2015,  7, 16 },  { 2015,  7, 17 },  { 2015,  7, 18 },  { 2015,  7, 19 },
  { 2015,  7, 20 },  { 2015,  7, 21 },  { 2015,  7, 22 },  { 2015,  7, 23 },
  { 2015,  7, 24 },  { 2015,  7, 25 },  { 2015,  7, 26 },  { 2015,  7, 27 },
  { 2015,  7, 28 },  { 2015,  7, 29 },  { 2015,  7, 30 },  { 2015,  7, 31 },
  { 2015,  8,  1 },  { 2015,  8,  2 },  { 2015,  8,  3 },  { 2015,  8,  4 },
  { 2015,  8,  5 },  { 2015,  8,  6 },  { 2015,  8,  7 },  { 2015,  8,  8 },
  { 2015,  8,  9 },  { 2015,  8, 10 },  { 2015,  8, 11 },  { 2015,  8, 12 },
  { 2015,  8, 13 },  { 2015,  8, 14 },  { 2015,  8, 15 },  { 2015,  8, 16 },
  { 2015,  8, 17 },  { 2015,  8, 18 },  { 2015,  8, 19 },  { 2015,  8, 20 },
  { 2015,  8, 21 },  { 2015,  8, 22 },  { 2015,  8, 23 },  { 2015,  8, 24 },
  { 2015,  8, 25 },  { 2015,  8, 26 },  { 2015,  8, 27 },  { 2015,  8, 28 },
  { 2015,  8, 29 },  { 2015,  8, 30 },  { 2015,  8, 31 },  { 2015,  9,  1 },
  { 2015,  9,  2 },  { 2015,  9,  3 },  { 2015,  9,  4 },  { 2015,  9,  5 },
  { 2015,  9,  6 },  { 2015,  9,  7 },  { 2015,  9,  8 },  { 2015,  9,  9 },
  { 2015,  9, 10 },  { 2015,  9, 11 },  { 2015,  9, 12 },  { 2015,  9, 13 },
  { 2015,  9, 14 },  { 2015,  9, 15 },  { 2015,  9, 16 },  { 2015,  9, 17 },
  { 2015,  9, 18 },  { 2015,  9, 19 },  { 2015,  9, 20 },  { 2015,  9, 21 },
  { 2015,  9, 22 },  { 2015,  9, 23 },  { 2015,  9, 24 },  { 2015,  9, 25 },
  { 2015,  9, 26 },  { 2015,  9, 27 },  { 2015,  9, 28 },  { 2015,  9, 29 },
  { 2015,  9, 30 },  { 2015, 10,  1 },  { 2015, 10,  2 },  { 2015, 10,  3 },
  { 2015, 10,  4 },  { 2015, 10,  5 },  { 2015, 10,  6 },  { 2015, 10,  7 },
  { 2015, 10,  8 },  { 2015, 10,  9 },  { 2015, 10, 10 },  { 2015, 10, 11 },
  { 2015, 10, 12 },  { 2015, 10, 13 },  { 2015, 10, 14 },  { 2015, 10, 15 },
  { 2015, 10, 16 },  { 2015, 10, 17 },  { 2015, 10, 18 },  { 2015, 10, 19 },
  { 2015, 10, 20 },  { 2015, 10, 21 },  { 2015, 10, 22 },  { 2015, 10, 23 },
  { 2015, 10, 24 },  { 2015, 10, 25 },  { 2015, 10, 26 },  { 2015, 10, 27 },
  { 2015, 10, 28 },  { 2015, 10, 29 },  { 2015, 10, 30 },  { 2015, 10, 31 },
  { 2015, 11,  1 },  { 2015, 11,  2 },  { 2015, 11,  3 },  { 2015, 11,  4 },
  { 2015, 11,  5 },  { 2015, 11,  6 },  { 2015, 11,  7 },  { 2015, 11,  8 },
  { 2015, 11,  9 },  { 2015, 11, 10 },  { 2015, 11, 11 },  { 2015, 11, 12 },
  { 2015, 11, 13 },  { 2015, 11, 14 },  { 2015, 11, 15 },  { 2015, 11, 16 },
  { 2015, 11, 17 },  { 2015, 11, 18 },  { 2015, 11, 19 },  { 2015, 11, 20 },
  { 2015, 11, 21 },  { 2015, 11, 22 },  { 2015, 11, 23 },  { 2015, 11, 24 },
  { 2015, 11, 25 },  { 2015, 11, 26 },  { 2015, 11, 27 },  { 2015, 11, 28 },
  { 2015, 11, 29 },  { 2015, 11, 30 },  { 2015, 12,  1 },  { 2015, 12,  2 },
  { 2015, 12,  3 },  { 2015, 12,  4 },  { 2015, 12,  5 },  { 2015, 12,  6 },
  { 2015, 12,  7 },  { 2015, 12,  8 },  { 2015, 12,  9 },  { 2015, 12, 10 },
  { 2015, 12, 11 },  { 2015, 12, 12 },  { 2015, 12, 13 },  { 2015, 12, 14 },
  { 2015, 12, 15 },  { 2015, 12, 16 },  { 2015, 12, 17 },  { 2015, 12, 18 },
  { 2015, 12, 19 },  { 2015, 12, 20 },  { 2015, 12, 21 },  { 2015, 12, 22 },
  { 2015, 12, 23 },  { 2015, 12, 24 },  { 2015, 12, 25 },  { 2015, 12, 26 },
  { 2015, 12, 27 },  { 2015, 12, 28 },  { 2015, 12, 29 },  { 2015, 12, 30 },
  { 2015, 12, 31 },  { 2016,  1,  1 },  { 2016,  1,  2 },  { 2016,  1,  3 },
  { 2016,  1,  4 },  { 2016,  1,  5 },  { 2016,  1,  6 },  { 2016,  1,  7 },
  { 2016,  1,  8 },  { 2016,  1,  9 },  { 2016,  1, 10 },  { 2016,  1, 11 },
  { 2016,  1, 12 },  { 2016,  1, 13 },  { 2016,  1, 14 },  { 2016,  1, 15 },
  { 2016,  1, 16 },  { 2016,  1, 17 },  { 2016,  1, 18 },  { 2016,  1, 19 },
  { 2016,  1, 20 },  { 2016,  1, 21 },  { 2016,  1, 22 },  { 2016,  1, 23 },
  { 2016,  1, 24 },  { 2016,  1, 25 },  { 2016,  1, 26 },  { 2016,  1, 27 },
  { 2016,  1, 28 },  { 2016,  1, 29 },  { 2016,  1, 30 },  { 2016,  1, 31 },
  { 2016,  2,  1 },  { 2016,  2,  2 },  { 2016,  2,  3 },  { 2016,  2,  4 },
  { 2016,  2,  5 },  { 2016,  2,  6 },  { 2016,  2,  7 },  { 2016,  2,  8 },
  { 2016,  2,  9 },  { 2016,  2, 10 },  { 2016,  2, 11 },  { 2016,  2, 12 },
  { 2016,  2, 13 },  { 2016,  2, 14 },  { 2016,  2, 15 },  { 2016,  2, 16 },
  { 2016,  2, 17 },  { 2016,  2, 18 },  { 2016,  2, 19 },  { 2016,  2, 20 },
  { 2016,  2, 21 },  { 2016,  2, 22 },  { 2016,  2, 23 },  { 2016,  2, 24 },
  { 2016,  2, 25 },  { 2016,  2, 26 },  { 2016,  2, 27 },  { 2016,  2, 28 },
  { 2016,  2, 29 },  { 2016,  3,  1 },  { 2016,  3,  2 },  { 2016,  3,  3 },
  { 2016,  3,  4 },  { 2016,  3,  5 },  { 2016,  3,  6 },  { 2016,  3,  7 },
  { 2016,  3,  8 },  { 2016,  3,  9 },  { 2016,  3, 10 },  { 2016,  3, 11 },
  { 2016,  3, 12 },  { 2016,  3, 13 },  { 2016,  3, 14 },  { 2016,  3, 15 },
  { 2016,  3, 16 },  { 2016,  3, 17 },  { 2016,  3, 18 },  { 2016,  3, 19 },
  { 2016,  3, 20 },  { 2016,  3, 21 },  { 2016,  3, 22 },  { 2016,  3, 23 },
  { 2016,  3, 24 },  { 2016,  3, 25 },  { 2016,  3, 26 },  { 2016,  3, 27 },
  { 2016,  3, 28 },  { 2016,  3, 29 },  { 2016,  3, 30 },  { 2016,  3, 31 },
  { 2016,  4,  1 },  { 2016,  4,  2 },  { 2016,  4,  3 },  { 2016,  4,  4 },
  { 2016,  4,  5 },  { 2016,  4,  6 },  { 2016,  4,  7 },  { 2016,  4,  8 },
  { 2016,  4,  9 },  { 2016,  4, 10 },  { 2016,  4, 11 },  { 2016,  4, 12 },
  { 2016,  4, 13 },  { 2016,  4, 14 },  { 2016,  4, 15 },  { 2016,  4, 16 },
  { 2016,  4, 17 },  { 2016,  4, 18 },  { 2016,  4, 19 },  { 2016,  4, 20 },
  { 2016,  4, 21 },  { 2016,  4, 22 },  { 2016,  4, 23 },  { 2016,  4, 24 },
  { 2016,  4, 25 },  { 2016,  4, 26 },  { 2016,  4, 27 },  { 2016,  4, 28 },
  { 2016,  4, 29 },  { 2016,  4, 30 },  { 2016,  5,  1 },  { 2016,  5,  2 },
  { 2016,  5,  3 },  { 2016,  5,  4 },  { 2016,  5,  5 },  { 2016,  5,  6 },
  { 2016,  5,  7 },  { 2016,  5,  8 },  { 2016,  5,  9 },  { 2016,  5, 10 },
  { 2016,  5, 11 },  { 2016,  5, 12 },  { 2016,  5, 13 },  { 2016,  5, 14 },
  { 2016,  5, 15 },  { 2016,  5, 16 },  { 2016,  5, 17 },  { 2016,  5, 18 },
  { 2016,  5, 19 },  { 2016,  5, 20 },  { 2016,  5, 21 },  { 2016,  5, 22 },
  { 2016,  5, 23 },  { 2016,  5, 24 },  { 2016,  5, 25 },  { 2016,  5, 26 },
  { 2016,  5, 27 },  { 2016,  5, 28 },  { 2016,  5, 29 },  { 2016,  5, 30 },
  { 2016,  5, 31 },  { 2016,  6,  1 },  { 2016,  6,  2 },  { 2016,  6,  3 },
  { 2016,  6,  4 },  { 2016,  6,  5 },  { 2016,  6,  6 },  { 2016,  6,  7 },
  { 2016,  6,  8 },  { 2016,  6,  9 },  { 2016,  6, 10 },  { 2016,  6, 11 },
  { 2016,  6, 12 },  { 2016,  6, 13 },  { 2016,  6, 14 },  { 2016,  6, 15 },
  { 2016,  6, 16 },  { 2016,  6, 17 },  { 2016,  6, 18 },  { 2016,  6, 19 },
  { 2016,  6, 20 },  { 2016,  6, 21 },  { 2016,  6, 22 },  { 2016,  6, 23 },
  { 2016,  6, 24 },  { 2016,  6, 25 },  { 2016,  6, 26 },  { 2016,  6, 27 },
  { 2016,  6, 28 },  { 2016,  6, 29 },  { 2016,  6, 30 },  { 2016,  7,  1 },
  { 2016,  7,  2 },  { 2016,  7,  3 },  { 2016,  7,  4 },  { 2016,  7,  5 },
  { 2016,  7,  6 },  { 2016,  7,  7 },  { 2016,  7,  8 },  { 2016,  7,  9 },
  { 2016,  7, 10 },  { 2016,  7, 11 },  { 2016,  7, 12 },  { 2016,  7, 13 },
  { 2016,  7, 14 },  { 2016,  7, 15 },  { 2016,  7, 16 },  { 2016,  7, 17 },
  { 2016,  7, 18 },  { 2016,  7, 19 },  { 2016,  7, 20 },  { 2016,  7, 21 },
  { 2016,  7, 22 },  { 2016,  7, 23 },  { 2016,  7, 24 },  { 2016,  7, 25 },
  { 2016,  7, 26 },  { 2016,  7, 27 },  { 2016,  7, 28 },  { 2016,  7, 29 },
  { 2016,  7, 30 },  { 2016,  7, 31 },  { 2016,  8,  1 },  { 2016,  8,  2 },
  { 2016,  8,  3 },  { 2016,  8,  4 },  { 2016,  8,  5 },  { 2016,  8,  6 },
  { 2016,  8,  7 },  { 2016,  8,  8 },  { 2016,  8,  9 },  { 2016,  8, 10 },
  { 2016,  8, 11 },  { 2016,  8, 12 },  { 2016,  8, 13 },  { 2016,  8, 14 },
  { 2016,  8, 15 },  { 2016,  8, 16 },  { 2016,  8, 17 },  { 2016,  8, 18 },
  { 2016,  8, 19 },  { 2016,  8, 20 },  { 2016,  8, 21 },  { 2016,  8, 22 },
  { 2016,  8, 23 },  { 2016,  8, 24 },  { 2016,  8, 25 },  { 2016,  8, 26 },
  { 2016,  8, 27 },  { 2016,  8, 28 },  { 2016,  8, 29 },  { 2016,  8, 30 },
  { 2016,  8, 31 },  { 2016,  9,  1 },  { 2016,  9,  2 },  { 2016,  9,  3 },
  { 2016,  9,  4 },  { 2016,  9,  5 },  { 2016,  9,  6 },  { 2016,  9,  7 },
  { 2016,  9,  8 },  { 2016,  9,  9 },  { 2016,  9, 10 },  { 2016,  9, 11 },
  { 2016,  9, 12 },  { 2016,  9, 13 },  { 2016,  9, 14 },  { 2016,  9, 15 },
  { 2016,  9, 16 },  { 2016,  9, 17 },  { 2016,  9, 18 },  { 2016,  9, 19 },
  { 2016,  9, 20 },  { 2016,  9, 21 },  { 2016,  9, 22 },  { 2016,  9, 23 },
  { 2016,  9, 24 },  { 2016,  9, 25 },  { 2016,  9, 26 },  { 2016,  9, 27 },
  { 2016,  9, 28 },  { 2016,  9, 29 },  { 2016,  9, 30 },  { 2016, 10,  1 },
  { 2016, 10,  2 },  { 2016, 10,  3 },  { 2016, 10,  4 },  { 2016, 10,  5 },
  { 2016, 10,  6 },  { 2016, 10,  7 },  { 2016, 10,  8 },  { 2016, 10,  9 },
  { 2016, 10, 10 },  { 2016, 10, 11 },  { 2016, 10, 12 },  { 2016, 10, 13 },
  { 2016, 10, 14 },  { 2016, 10, 15 },  { 2016, 10, 16 },  { 2016, 10, 17 },
  { 2016, 10, 18 },  { 2016, 10, 19 },  { 2016, 10, 20 },  { 2016, 10, 21 },
  { 2016, 10, 22 },  { 2016, 10, 23 },  { 2016, 10, 24 },  { 2016, 10, 25 },
  { 2016, 10, 26 },  { 2016, 10, 27 },  { 2016, 10, 28 },  { 2016, 10, 29 },
  { 2016, 10, 30 },  { 2016, 10, 31 },  { 2016, 11,  1 },  { 2016, 11,  2 },
  { 2016, 11,  3 },  { 2016, 11,  4 },  { 2016, 11,  5 },  { 2016, 11,  6 },
  { 2016, 11,  7 },  { 2016, 11,  8 },  { 2016, 11,  9 },  { 2016, 11, 10 },
  { 2016, 11, 11 },  { 2016, 11, 12 },  { 2016, 11, 13 },  { 2016, 11, 14 },
  { 2016, 11, 15 },  { 2016, 11, 16 },  { 2016, 11, 17 },  { 2016, 11, 18 },
  { 2016, 11, 19 },  { 2016, 11, 20 },  { 2016, 11, 21 },  { 2016, 11, 22 },
  { 2016, 11, 23 },  { 2016, 11, 24 },  { 2016, 11, 25 },  { 2016, 11, 26 },
  { 2016, 11, 27 },  { 2016, 11, 28 },  { 2016, 11, 29 },  { 2016, 11, 30 },
  { 2016, 12,  1 },  { 2016, 12,  2 },  { 2016, 12,  3 },  { 2016, 12,  4 },
  { 2016, 12,  5 },  { 2016, 12,  6 },  { 2016, 12,  7 },  { 2016, 12,  8 },
  { 2016, 12,  9 },  { 2016, 12, 10 },  { 2016, 12, 11 },  { 2016, 12, 12 },
  { 2016, 12, 13 },  { 2016, 12, 14 },  { 2016, 12, 15 },  { 2016, 12, 16 },
  { 2016, 12, 17 },  { 2016, 12, 18 },  { 2016, 12, 19 },  { 2016, 12, 20 },
  { 2016, 12, 21 },  { 2016, 12, 22 },  { 2016, 12, 23 },  { 2016, 12, 24 },
  { 2016, 12, 25 },  { 2016, 12, 26 },  { 2016, 12, 27 },  { 2016, 12, 28 },
  { 2016, 12, 29 },  { 2016, 12, 30 },  { 2016, 12, 31 },  { 2017,  1,  1 },
  { 2017,  1,  2 },  { 2017,  1,  3 },  { 2017,  1,  4 },  { 2017,  1,  5 },
  { 2017,  1,  6 },  { 2017,  1,  7 },  { 2017,  1,  8 },  { 2017,  1,  9 },
  { 2017,  1, 10 },  { 2017,  1, 11 },  { 2017,  1, 12 },  { 2017,  1, 13 },
  { 2017,  1, 14 },  { 2017,  1, 15 },  { 2017,  1, 16 },  { 2017,  1, 17 },
  { 2017,  1, 18 },  { 2017,  1, 19 },  { 2017,  1, 20 },  { 2017,  1, 21 },
  { 2017,  1, 22 },  { 2017,  1, 23 },  { 2017,  1, 24 },  { 2017,  1, 25 },
  { 2017,  1, 26 },  { 2017,  1, 27 },  { 2017,  1, 28 },  { 2017,  1, 29 },
  { 2017,  1, 30 },  { 2017,  1, 31 },  { 2017,  2,  1 },  { 2017,  2,  2 },
  { 2017,  2,  3 },  { 2017,  2,  4 },  { 2017,  2,  5 },  { 2017,  2,  6 },
  { 2017,  2,  7 },  { 2017,  2,  8 },  { 2017,  2,  9 },  { 2017,  2, 10 },
  { 2017,  2, 11 },  { 2017,  2, 12 },  { 2017,  2, 13 },  { 2017,  2, 14 },
  { 2017,  2, 15 },  { 2017,  2, 16 },  { 2017,  2, 17 },  { 2017,  2, 18 },
  { 2017,  2, 19 },  { 2017,  2, 20 },  { 2017,  2, 21 },  { 2017,  2, 22 },
  { 2017,  2, 23 },  { 2017,  2, 24 },  { 2017,  2, 25 },  { 2017,  2, 26 },
  { 2017,  2, 27 },  { 2017,  2, 28 },  { 2017,  3,  1 },  { 2017,  3,  2 },
  { 2017,  3,  3 },  { 2017,  3,  4 },  { 2017,  3,  5 },  { 2017,  3,  6 },
  { 2017,  3,  7 },  { 2017,  3,  8 },  { 2017,  3,  9 },  { 2017,  3, 10 },
  { 2017,  3, 11 },  { 2017,  3, 12 },  { 2017,  3, 13 },  { 2017,  3, 14 },
  { 2017,  3, 15 },  { 2017,  3, 16 },  { 2017,  3, 17 },  { 2017,  3, 18 },
  { 2017,  3, 19 },  { 2017,  3, 20 },  { 2017,  3, 21 },  { 2017,  3, 22 },
  { 2017,  3, 23 },  { 2017,  3, 24 },  { 2017,  3, 25 },  { 2017,  3, 26 },
  { 2017,  3, 27 },  { 2017,  3, 28 },  { 2017,  3, 29 },  { 2017,  3, 30 },
  { 2017,  3, 31 },  { 2017,  4,  1 },  { 2017,  4,  2 },  { 2017,  4,  3 },
  { 2017,  4,  4 },  { 2017,  4,  5 },  { 2017,  4,  6 },  { 2017,  4,  7 },
  { 2017,  4,  8 },  { 2017,  4,  9 },  { 2017,  4, 10 },  { 2017,  4, 11 },
  { 2017,  4, 12 },  { 2017,  4, 13 },  { 2017,  4, 14 },  { 2017,  4, 15 },
  { 2017,  4, 16 },  { 2017,  4, 17 },  { 2017,  4, 18 },  { 2017,  4, 19 },
  { 2017,  4, 20 },  { 2017,  4, 21 },  { 2017,  4, 22 },  { 2017,  4, 23 },
  { 2017,  4, 24 },  { 2017,  4, 25 },  { 2017,  4, 26 },  { 2017,  4, 27 },
  { 2017,  4, 28 },  { 2017,  4, 29 },  { 2017,  4, 30 },  { 2017,  5,  1 },
  { 2017,  5,  2 },  { 2017,  5,  3 },  { 2017,  5,  4 },  { 2017,  5,  5 },
  { 2017,  5,  6 },  { 2017,  5,  7 },  { 2017,  5,  8 },  { 2017,  5,  9 },
  { 2017,  5, 10 },  { 2017,  5, 11 },  { 2017,  5, 12 },  { 2017,  5, 13 },
  { 2017,  5, 14 },  { 2017,  5, 15 },  { 2017,  5, 16 },  { 2017,  5, 17 },
  { 2017,  5, 18 },  { 2017,  5, 19 },  { 2017,  5, 20 },  { 2017,  5, 21 },
  { 2017,  5, 22 },  { 2017,  5, 23 },  { 2017,  5, 24 },  { 2017,  5, 25 },
  { 2017,  5, 26 },  { 2017,  5, 27 },  { 2017,  5, 28 },  { 2017,  5, 29 },
  { 2017,  5, 30 },  { 2017,  5, 31 },  { 2017,  6,  1 },  { 2017,  6,  2 },
  { 2017,  6,  3 },  { 2017,  6,  4 },  { 2017,  6,  5 },  { 2017,  6,  6 },
  { 2017,  6,  7 },  { 2017,  6,  8 },  { 2017,  6,  9 },  { 2017,  6, 10 },
  { 2017,  6, 11 },  { 2017,  6, 12 },  { 2017,  6, 13 },  { 2017,  6, 14 },
  { 2017,  6, 15 },  { 2017,  6, 16 },  { 2017,  6, 17 },  { 2017,  6, 18 },
  { 2017,  6, 19 },  { 2017,  6, 20 },  { 2017,  6, 21 },  { 2017,  6, 22 },
  { 2017,  6, 23 },  { 2017,  6, 24 },  { 2017,  6, 25 },  { 2017,  6, 26 },
  { 2017,  6, 27 },  { 2017,  6, 28 },  { 2017,  6, 29 },  { 2017,  6, 30 },
  { 2017,  7,  1 },  { 2017,  7,  2 },  { 2017,  7,  3 },  { 2017,  7,  4 },
  { 2017,  7,  5 },  { 2017,  7,  6 },  { 2017,  7,  7 },  { 2017,  7,  8 },
  { 2017,  7,  9 },  { 2017,  7, 10 },  { 2017,  7, 11 },  { 2017,  7, 12 },
  { 2017,  7, 13 },  { 2017,  7, 14 },  { 2017,  7, 15 },  { 2017,  7, 16 },
  { 2017,  7, 17 },  { 2017,  7, 18 },  { 2017,  7, 19 },  { 2017,  7, 20 },
  { 2017,  7, 21 },  { 2017,  7, 22 },  { 2017,  7, 23 },  { 2017,  7, 24 },
  { 2017,  7, 25 },  { 2017,  7, 26 },  { 2017,  7, 27 },  { 2017,  7, 28 },
  { 2017,  7, 29 },  { 2017,  7, 30 },  { 2017,  7, 31 },  { 2017,  8,  1 },
  { 2017,  8,  2 },  { 2017,  8,  3 },  { 2017,  8,  4 },  { 2017,  8,  5 },
  { 2017,  8,  6 },  { 2017,  8,  7 },  { 2017,  8,  8 },  { 2017,  8,  9 },
  { 2017,  8, 10 },  { 2017,  8, 11 },  { 2017,  8, 12 },  { 2017,  8, 13 },
  { 2017,  8, 14 },  { 2017,  8, 15 },  { 2017,  8, 16 },  { 2017,  8, 17 },
  { 2017,  8, 18 },  { 2017,  8, 19 },  { 2017,  8, 20 },  { 2017,  8, 21 },
  { 2017,  8, 22 },  { 2017,  8, 23 },  { 2017,  8, 24 },  { 2017,  8, 25 },
  { 2017,  8, 26 },  { 2017,  8, 27 },  { 2017,  8, 28 },  { 2017,  8, 29 },
  { 2017,  8, 30 },  { 2017,  8, 31 },  { 2017,  9,  1 },  { 2017,  9,  2 },
  { 2017,  9,  3 },  { 2017,  9,  4 },  { 2017,  9,  5 },  { 2017,  9,  6 },
  { 2017,  9,  7 },  { 2017,  9,  8 },  { 2017,  9,  9 },  { 2017,  9, 10 },
  { 2017,  9, 11 },  { 2017,  9, 12 },  { 2017,  9, 13 },  { 2017,  9, 14 },
  { 2017,  9, 15 },  { 2017,  9, 16 },  { 2017,  9, 17 },  { 2017,  9, 18 },
  { 2017,  9, 19 },  { 2017,  9, 20 },  { 2017,  9, 21 },  { 2017,  9, 22 },
  { 2017,  9, 23 },  { 2017,  9, 24 },  { 2017,  9, 25 },  { 2017,  9, 26 },
  { 2017,  9, 27 },  { 2017,  9, 28 },  { 2017,  9, 29 },  { 2017,  9, 30 },
  { 2017, 10,  1 },  { 2017, 10,  2 },  { 2017, 10,  3 },  { 2017, 10,  4 },
  { 2017, 10,  5 },  { 2017, 10,  6 },  { 2017, 10,  7 },  { 2017, 10,  8 },
  { 2017, 10,  9 },  { 2017, 10, 10 },  { 2017, 10, 11 },  { 2017, 10, 12 },
  { 2017, 10, 13 },  { 2017, 10, 14 },  { 2017, 10, 15 },  { 2017, 10, 16 },
  { 2017, 10, 17 },  { 2017, 10, 18 },  { 2017, 10, 19 },  { 2017, 10, 20 },
  { 2017, 10, 21 },  { 2017, 10, 22 },  { 2017, 10, 23 },  { 2017, 10, 24 },
  { 2017, 10, 25 },  { 2017, 10, 26 },  { 2017, 10, 27 },  { 2017, 10, 28 },
  { 2017, 10, 29 },  { 2017, 10, 30 },  { 2017, 10, 31 },  { 2017, 11,  1 },
  { 2017, 11,  2 },  { 2017, 11,  3 },  { 2017, 11,  4 },  { 2017, 11,  5 },
  { 2017, 11,  6 },  { 2017, 11,  7 },  { 2017, 11,  8 },  { 2017, 11,  9 },
  { 2017, 11, 10 },  { 2017, 11, 11 },  { 2017, 11, 12 },  { 2017, 11, 13 },
  { 2017, 11, 14 },  { 2017, 11, 15 },  { 2017, 11, 16 },  { 2017, 11, 17 },
  { 2017, 11, 18 },  { 2017, 11, 19 },  { 2017, 11, 20 },  { 2017, 11, 21 },
  { 2017, 11, 22 },  { 2017, 11, 23 },  { 2017, 11, 24 },  { 2017, 11, 25 },
  { 2017, 11, 26 },  { 2017, 11, 27 },  { 2017, 11, 28 },  { 2017, 11, 29 },
  { 2017, 11, 30 },  { 2017, 12,  1 },  { 2017, 12,  2 },  { 2017, 12,  3 },
  { 2017, 12,  4 },  { 2017, 12,  5 },  { 2017, 12,  6 },  { 2017, 12,  7 },
  { 2017, 12,  8 },  { 2017, 12,  9 },  { 2017, 12, 10 },  { 2017, 12, 11 },
  { 2017, 12, 12 },  { 2017, 12, 13 },  { 2017, 12, 14 },  { 2017, 12, 15 },
  { 2017, 12, 16 },  { 2017, 12, 17 },  { 2017, 12, 18 },  { 2017, 12, 19 },
  { 2017, 12, 20 },  { 2017, 12, 21 },  { 2017, 12, 22 },  { 2017, 12, 23 },
  { 2017, 12, 24 },  { 2017, 12, 25 },  { 2017, 12, 26 },  { 2017, 12, 27 },
  { 2017, 12, 28 },  { 2017, 12, 29 },  { 2017, 12, 30 },  { 2017, 12, 31 },
  { 2018,  1,  1 },  { 2018,  1,  2 },  { 2018,  1,  3 },  { 2018,  1,  4 },
  { 2018,  1,  5 },  { 2018,  1,  6 },  { 2018,  1,  7 },  { 2018,  1,  8 },
  { 2018,  1,  9 },  { 2018,  1, 10 },  { 2018,  1, 11 },  { 2018,  1, 12 },
  { 2018,  1, 13 },  { 2018,  1, 14 },  { 2018,  1, 15 },  { 2018,  1, 16 },
  { 2018,  1, 17 },  { 2018,  1, 18 },  { 2018,  1, 19 },  { 2018,  1, 20 },
  { 2018,  1, 21 },  { 2018,  1, 22 },  { 2018,  1, 23 },  { 2018,  1, 24 },
  { 2018,  1, 25 },  { 2018,  1, 26 },  { 2018,  1, 27 },  { 2018,  1, 28 },
  { 2018,  1, 29 },  { 2018,  1, 30 },  { 2018,  1, 31 },  { 2018,  2,  1 },
  { 2018,  2,  2 },  { 2018,  2,  3 },  { 2018,  2,  4 },  { 2018,  2,  5 },
  { 2018,  2,  6 },  { 2018,  2,  7 },  { 2018,  2,  8 },  { 2018,  2,  9 },
  { 2018,  2, 10 },  { 2018,  2, 11 },  { 2018,  2, 12 },  { 2018,  2, 13 },
  { 2018,  2, 14 },  { 2018,  2, 15 },  { 2018,  2, 16 },  { 2018,  2, 17 },
  { 2018,  2, 18 },  { 2018,  2, 19 },  { 2018,  2, 20 },  { 2018,  2, 21 },
  { 2018,  2, 22 },  { 2018,  2, 23 },  { 2018,  2, 24 },  { 2018,  2, 25 },
  { 2018,  2, 26 },  { 2018,  2, 27 },  { 2018,  2, 28 },  { 2018,  3,  1 },
  { 2018,  3,  2 },  { 2018,  3,  3 },  { 2018,  3,  4 },  { 2018,  3,  5 },
  { 2018,  3,  6 },  { 2018,  3,  7 },  { 2018,  3,  8 },  { 2018,  3,  9 },
  { 2018,  3, 10 },  { 2018,  3, 11 },  { 2018,  3, 12 },  { 2018,  3, 13 },
  { 2018,  3, 14 },  { 2018,  3, 15 },  { 2018,  3, 16 },  { 2018,  3, 17 },
  { 2018,  3, 18 },  { 2018,  3, 19 },  { 2018,  3, 20 },  { 2018,  3, 21 },
  { 2018,  3, 22 },  { 2018,  3, 23 },  { 2018,  3, 24 },  { 2018,  3, 25 },
  { 2018,  3, 26 },  { 2018,  3, 27 },  { 2018,  3, 28 },  { 2018,  3, 29 },
  { 2018,  3, 30 },  { 2018,  3, 31 },  { 2018,  4,  1 },  { 2018,  4,  2 },
  { 2018,  4,  3 },  { 2018,  4,  4 },  { 2018,  4,  5 },  { 2018,  4,  6 },
  { 2018,  4,  7 },  { 2018,  4,  8 },  { 2018,  4,  9 },  { 2018,  4, 10 },
  { 2018,  4, 11 },  { 2018,  4, 12 },  { 2018,  4, 13 },  { 2018,  4, 14 },
  { 2018,  4, 15 },  { 2018,  4, 16 },  { 2018,  4, 17 },  { 2018,  4, 18 },
  { 2018,  4, 19 },  { 2018,  4, 20 },  { 2018,  4, 21 },  { 2018,  4, 22 },
  { 2018,  4, 23 },  { 2018,  4, 24 },  { 2018,  4, 25 },  { 2018,  4, 26 },
  { 2018,  4, 27 },  { 2018,  4, 28 },  { 2018,  4, 29 },  { 2018,  4, 30 },
  { 2018,  5,  1 },  { 2018,  5,  2 },  { 2018,  5,  3 },  { 2018,  5,  4 },
  { 2018,  5,  5 },  { 2018,  5,  6 },  { 2018,  5,  7 },  { 2018,  5,  8 },
  { 2018,  5,  9 },  { 2018,  5, 10 },  { 2018,  5, 11 },  { 2018,  5, 12 },
  { 2018,  5, 13 },  { 2018,  5, 14 },  { 2018,  5, 15 },  { 2018,  5, 16 },
  { 2018,  5, 17 },  { 2018,  5, 18 },  { 2018,  5, 19 },  { 2018,  5, 20 },
  { 2018,  5, 21 },  { 2018,  5, 22 },  { 2018,  5, 23 },  { 2018,  5, 24 },
  { 2018,  5, 25 },  { 2018,  5, 26 },  { 2018,  5, 27 },  { 2018,  5, 28 },
  { 2018,  5, 29 },  { 2018,  5, 30 },  { 2018,  5, 31 },  { 2018,  6,  1 },
  { 2018,  6,  2 },  { 2018,  6,  3 },  { 2018,  6,  4 },  { 2018,  6,  5 },
  { 2018,  6,  6 },  { 2018,  6,  7 },  { 2018,  6,  8 },  { 2018,  6,  9 },
  { 2018,  6, 10 },  { 2018,  6, 11 },  { 2018,  6, 12 },  { 2018,  6, 13 },
  { 2018,  6, 14 },  { 2018,  6, 15 },  { 2018,  6, 16 },  { 2018,  6, 17 },
  { 2018,  6, 18 },  { 2018,  6, 19 },  { 2018,  6, 20 },  { 2018,  6, 21 },
  { 2018,  6, 22 },  { 2018,  6, 23 },  { 2018,  6, 24 },  { 2018,  6, 25 },
  { 2018,  6, 26 },  { 2018,  6, 27 },  { 2018,  6, 28 },  { 2018,  6, 29 },
  { 2018,  6, 30 },  { 2018,  7,  1 },  { 2018,  7,  2 },  { 2018,  7,  3 },
  { 2018,  7,  4 },  { 2018,  7,  5 },  { 2018,  7,  6 },  { 2018,  7,  7 },
  { 2018,  7,  8 },  { 2018,  7,  9 },  { 2018,  7, 10 },  { 2018,  7, 11 },
  { 2018,  7, 12 },  { 2018,  7, 13 },  { 2018,  7, 14 },  { 2018,  7, 15 },
  { 2018,  7, 16 },  { 2018,  7, 17 },  { 2018,  7, 18 },  { 2018,  7, 19 },
  { 2018,  7, 20 },  { 2018,  7, 21 },  { 2018,  7, 22 },  { 2018,  7, 23 },
  { 2018,  7, 24 },  { 2018,  7, 25 },  { 2018,  7, 26 },  { 2018,  7, 27 },
  { 2018,  7, 28 },  { 2018,  7, 29 },  { 2018,  7, 30 },  { 2018,  7, 31 },
  { 2018,  8,  1 },  { 2018,  8,  2 },  { 2018,  8,  3 },  { 2018,  8,  4 },
  { 2018,  8,  5 },  { 2018,  8,  6 },  { 2018,  8,  7 },  { 2018,  8,  8 },
  { 2018,  8,  9 },  { 2018,  8, 10 },  { 2018,  8, 11 },  { 2018,  8, 12 },
  { 2018,  8, 13 },  { 2018,  8, 14 },  { 2018,  8, 15 },  { 2018,  8, 16 },
  { 2018,  8, 17 },  { 2018,  8, 18 },  { 2018,  8, 19 },  { 2018,  8, 20 },
  { 2018,  8, 21 },  { 2018,  8, 22 },  { 2018,  8, 23 },  { 2018,  8, 24 },
  { 2018,  8, 25 },  { 2018,  8, 26 },  { 2018,  8, 27 },  { 2018,  8, 28 },
  { 2018,  8, 29 },  { 2018,  8, 30 },  { 2018,  8, 31 },  { 2018,  9,  1 },
  { 2018,  9,  2 },  { 2018,  9,  3 },  { 2018,  9,  4 },  { 2018,  9,  5 },
  { 2018,  9,  6 },  { 2018,  9,  7 },  { 2018,  9,  8 },  { 2018,  9,  9 },
  { 2018,  9, 10 },  { 2018,  9, 11 },  { 2018,  9, 12 },  { 2018,  9, 13 },
  { 2018,  9, 14 },  { 2018,  9, 15 },  { 2018,  9, 16 },  { 2018,  9, 17 },
  { 2018,  9, 18 },  { 2018,  9, 19 },  { 2018,  9, 20 },  { 2018,  9, 21 },
  { 2018,  9, 22 },  { 2018,  9, 23 },  { 2018,  9, 24 },  { 2018,  9, 25 },
  { 2018,  9, 26 },  { 2018,  9, 27 },  { 2018,  9, 28 },  { 2018,  9, 29 },
  { 2018,  9, 30 },  { 2018, 10,  1 },  { 2018, 10,  2 },  { 2018, 10,  3 },
  { 2018, 10,  4 },  { 2018, 10,  5 },  { 2018, 10,  6 },  { 2018, 10,  7 },
  { 2018, 10,  8 },  { 2018, 10,  9 },  { 2018, 10, 10 },  { 2018, 10, 11 },
  { 2018, 10, 12 },  { 2018, 10, 13 },  { 2018, 10, 14 },  { 2018, 10, 15 },
  { 2018, 10, 16 },  { 2018, 10, 17 },  { 2018, 10, 18 },  { 2018, 10, 19 },
  { 2018, 10, 20 },  { 2018, 10, 21 },  { 2018, 10, 22 },  { 2018, 10, 23 },
  { 2018, 10, 24 },  { 2018, 10, 25 },  { 2018, 10, 26 },  { 2018, 10, 27 },
  { 2018, 10, 28 },  { 2018, 10, 29 },  { 2018, 10, 30 },  { 2018, 10, 31 },
  { 2018, 11,  1 },  { 2018, 11,  2 },  { 2018, 11,  3 },  { 2018, 11,  4 },
  { 2018, 11,  5 },  { 2018, 11,  6 },  { 2018, 11,  7 },  { 2018, 11,  8 },
  { 2018, 11,  9 },  { 2018, 11, 10 },  { 2018, 11, 11 },  { 2018, 11, 12 },
  { 2018, 11, 13 },  { 2018, 11, 14 },  { 2018, 11, 15 },  { 2018, 11, 16 },
  { 2018, 11, 17 },  { 2018, 11, 18 },  { 2018, 11, 19 },  { 2018, 11, 20 },
  { 2018, 11, 21 },  { 2018, 11, 22 },  { 2018, 11, 23 },  { 2018, 11, 24 },
  { 2018, 11, 25 },  { 2018, 11, 26 },  { 2018, 11, 27 },  { 2018, 11, 28 },
  { 2018, 11, 29 },  { 2018, 11, 30 },  { 2018, 12,  1 },  { 2018, 12,  2 },
  { 2018, 12,  3 },  { 2018, 12,  4 },  { 2018, 12,  5 },  { 2018, 12,  6 },
  { 2018, 12,  7 },  { 2018, 12,  8 },  { 2018, 12,  9 },  { 2018, 12, 10 },
  { 2018, 12, 11 },  { 2018, 12, 12 },  { 2018, 12, 13 },  { 2018, 12, 14 },
  { 2018, 12, 15 },  { 2018, 12, 16 },  { 2018, 12, 17 },  { 2018, 12, 18 },
  { 2018, 12, 19 },  { 2018, 12, 20 },  { 2018, 12, 21 },  { 2018, 12, 22 },
  { 2018, 12, 23 },  { 2018, 12, 24 },  { 2018, 12, 25 },  { 2018, 12, 26 },
  { 2018, 12, 27 },  { 2018, 12, 28 },  { 2018, 12, 29 },  { 2018, 12, 30 },
  { 2018, 12, 31 },  { 2019,  1,  1 },  { 2019,  1,  2 },  { 2019,  1,  3 },
  { 2019,  1,  4 },  { 2019,  1,  5 },  { 2019,  1,  6 },  { 2019,  1,  7 },
  { 2019,  1,  8 },  { 2019,  1,  9 },  { 2019,  1, 10 },  { 2019,  1, 11 },
  { 2019,  1, 12 },  { 2019,  1, 13 },  { 2019,  1, 14 },  { 2019,  1, 15 },
  { 2019,  1, 16 },  { 2019,  1, 17 },  { 2019,  1, 18 },  { 2019,  1, 19 },
  { 2019,  1, 20 },  { 2019,  1, 21 },  { 2019,  1, 22 },  { 2019,  1, 23 },
  { 2019,  1, 24 },  { 2019,  1, 25 },  { 2019,  1, 26 },  { 2019,  1, 27 },
  { 2019,  1, 28 },  { 2019,  1, 29 },  { 2019,  1, 30 },  { 2019,  1, 31 },
  { 2019,  2,  1 },  { 2019,  2,  2 },  { 2019,  2,  3 },  { 2019,  2,  4 },
  { 2019,  2,  5 },  { 2019,  2,  6 },  { 2019,  2,  7 },  { 2019,  2,  8 },
  { 2019,  2,  9 },  { 2019,  2, 10 },  { 2019,  2, 11 },  { 2019,  2, 12 },
  { 2019,  2, 13 },  { 2019,  2, 14 },  { 2019,  2, 15 },  { 2019,  2, 16 },
  { 2019,  2, 17 },  { 2019,  2, 18 },  { 2019,  2, 19 },  { 2019,  2, 20 },
  { 2019,  2, 21 },  { 2019,  2, 22 },  { 2019,  2, 23 },  { 2019,  2, 24 },
  { 2019,  2, 25 },  { 2019,  2, 26 },  { 2019,  2, 27 },  { 2019,  2, 28 },
  { 2019,  3,  1 },  { 2019,  3,  2 },  { 2019,  3,  3 },  { 2019,  3,  4 },
  { 2019,  3,  5 },  { 2019,  3,  6 },  { 2019,  3,  7 },  { 2019,  3,  8 },
  { 2019,  3,  9 },  { 2019,  3, 10 },  { 2019,  3, 11 },  { 2019,  3, 12 },
  { 2019,  3, 13 },  { 2019,  3, 14 },  { 2019,  3, 15 },  { 2019,  3, 16 },
  { 2019,  3, 17 },  { 2019,  3, 18 },  { 2019,  3, 19 },  { 2019,  3, 20 },
  { 2019,  3, 21 },  { 2019,  3, 22 },  { 2019,  3, 23 },  { 2019,  3, 24 },
  { 2019,  3, 25 },  { 2019,  3, 26 },  { 2019,  3, 27 },  { 2019,  3, 28 },
  { 2019,  3, 29 },  { 2019,  3, 30 },  { 2019,  3, 31 },  { 2019,  4,  1 },
  { 2019,  4,  2 },  { 2019,  4,  3 },  { 2019,  4,  4 },  { 2019,  4,  5 },
  { 2019,  4,  6 },  { 2019,  4,  7 },  { 2019,  4,  8 },  { 2019,  4,  9 },
  { 2019,  4, 10 },  { 2019,  4, 11 },  { 2019,  4, 12 },  { 2019,  4, 13 },
  { 2019,  4, 14 },  { 2019,  4, 15 },  { 2019,  4, 16 },  { 2019,  4, 17 },
  { 2019,  4, 18 },  { 2019,  4, 19 },  { 2019,  4, 20 },  { 2019,  4, 21 },
  { 2019,  4, 22 },  { 2019,  4, 23 },  { 2019,  4, 24 },  { 2019,  4, 25 },
  { 2019,  4, 26 },  { 2019,  4, 27 },  { 2019,  4, 28 },  { 2019,  4, 29 },
  { 2019,  4, 30 },  { 2019,  5,  1 },  { 2019,  5,  2 },  { 2019,  5,  3 },
  { 2019,  5,  4 },  { 2019,  5,  5 },  { 2019,  5,  6 },  { 2019,  5,  7 },
  { 2019,  5,  8 },  { 2019,  5,  9 },  { 2019,  5, 10 },  { 2019,  5, 11 },
  { 2019,  5, 12 },  { 2019,  5, 13 },  { 2019,  5, 14 },  { 2019,  5, 15 },
  { 2019,  5, 16 },  { 2019,  5, 17 },  { 2019,  5, 18 },  { 2019,  5, 19 },
  { 2019,  5, 20 },  { 2019,  5, 21 },  { 2019,  5, 22 },  { 2019,  5, 23 },
  { 2019,  5, 24 },  { 2019,  5, 25 },  { 2019,  5, 26 },  { 2019,  5, 27 },
  { 2019,  5, 28 },  { 2019,  5, 29 },  { 2019,  5, 30 },  { 2019,  5, 31 },
  { 2019,  6,  1 },  { 2019,  6,  2 },  { 2019,  6,  3 },  { 2019,  6,  4 },
  { 2019,  6,  5 },  { 2019,  6,  6 },  { 2019,  6,  7 },  { 2019,  6,  8 },
  { 2019,  6,  9 },  { 2019,  6, 10 },  { 2019,  6, 11 },  { 2019,  6, 12 },
  { 2019,  6, 13 },  { 2019,  6, 14 },  { 2019,  6, 15 },  { 2019,  6, 16 },
  { 2019,  6, 17 },  { 2019,  6, 18 },  { 2019,  6, 19 },  { 2019,  6, 20 },
  { 2019,  6, 21 },  { 2019,  6, 22 },  { 2019,  6, 23 },  { 2019,  6, 24 },
  { 2019,  6, 25 },  { 2019,  6, 26 },  { 2019,  6, 27 },  { 2019,  6, 28 },
  { 2019,  6, 29 },  { 2019,  6, 30 },  { 2019,  7,  1 },  { 2019,  7,  2 },
  { 2019,  7,  3 },  { 2019,  7,  4 },  { 2019,  7,  5 },  { 2019,  7,  6 },
  { 2019,  7,  7 },  { 2019,  7,  8 },  { 2019,  7,  9 },  { 2019,  7, 10 },
  { 2019,  7, 11 },  { 2019,  7, 12 },  { 2019,  7, 13 },  { 2019,  7, 14 },
  { 2019,  7, 15 },  { 2019,  7, 16 },  { 2019,  7, 17 },  { 2019,  7, 18 },
  { 2019,  7, 19 },  { 2019,  7, 20 },  { 2019,  7, 21 },  { 2019,  7, 22 },
  { 2019,  7, 23 },  { 2019,  7, 24 },  { 2019,  7, 25 },  { 2019,  7, 26 },
  { 2019,  7, 27 },  { 2019,  7, 28 },  { 2019,  7, 29 },  { 2019,  7, 30 },
  { 2019,  7, 31 },  { 2019,  8,  1 },  { 2019,  8,  2 },  { 2019,  8,  3 },
  { 2019,  8,  4 },  { 2019,  8,  5 },  { 2019,  8,  6 },  { 2019,  8,  7 },
  { 2019,  8,  8 },  { 2019,  8,  9 },  { 2019,  8, 10 },  { 2019,  8, 11 },
  { 2019,  8, 12 },  { 2019,  8, 13 },  { 2019,  8, 14 },  { 2019,  8, 15 },
  { 2019,  8, 16 },  { 2019,  8, 17 },  { 2019,  8, 18 },  { 2019,  8, 19 },
  { 2019,  8, 20 },  { 2019,  8, 21 },  { 2019,  8, 22 },  { 2019,  8, 23 },
  { 2019,  8, 24 },  { 2019,  8, 25 },  { 2019,  8, 26 },  { 2019,  8, 27 },
  { 2019,  8, 28 },  { 2019,  8, 29 },  { 2019,  8, 30 },  { 2019,  8, 31 },
  { 2019,  9,  1 },  { 2019,  9,  2 },  { 2019,  9,  3 },  { 2019,  9,  4 },
  { 2019,  9,  5 },  { 2019,  9,  6 },  { 2019,  9,  7 },  { 2019,  9,  8 },
  { 2019,  9,  9 },  { 2019,  9, 10 },  { 2019,  9, 11 },  { 2019,  9, 12 },
  { 2019,  9, 13 },  { 2019,  9, 14 },  { 2019,  9, 15 },  { 2019,  9, 16 },
  { 2019,  9, 17 },  { 2019,  9, 18 },  { 2019,  9, 19 },  { 2019,  9, 20 },
  { 2019,  9, 21 },  { 2019,  9, 22 },  { 2019,  9, 23 },  { 2019,  9, 24 },
  { 2019,  9, 25 },  { 2019,  9, 26 },  { 2019,  9, 27 },  { 2019,  9, 28 },
  { 2019,  9, 29 },  { 2019,  9, 30 },  { 2019, 10,  1 },  { 2019, 10,  2 },
  { 2019, 10,  3 },  { 2019, 10,  4 },  { 2019, 10,  5 },  { 2019, 10,  6 },
  { 2019, 10,  7 },  { 2019, 10,  8 },  { 2019, 10,  9 },  { 2019, 10, 10 },
  { 2019, 10, 11 },  { 2019, 10, 12 },  { 2019, 10, 13 },  { 2019, 10, 14 },
  { 2019, 10, 15 },  { 2019, 10, 16 },  { 2019, 10, 17 },  { 2019, 10, 18 },
  { 2019, 10, 19 },  { 2019, 10, 20 },  { 2019, 10, 21 },  { 2019, 10, 22 },
  { 2019, 10, 23 },  { 2019, 10, 24 },  { 2019, 10, 25 },  { 2019, 10, 26 },
  { 2019, 10, 27 },  { 2019, 10, 28 },  { 2019, 10, 29 },  { 2019, 10, 30 },
  { 2019, 10, 31 },  { 2019, 11,  1 },  { 2019, 11,  2 },  { 2019, 11,  3 },
  { 2019, 11,  4 },  { 2019, 11,  5 },  { 2019, 11,  6 },  { 2019, 11,  7 },
  { 2019, 11,  8 },  { 2019, 11,  9 },  { 2019, 11, 10 },  { 2019, 11, 11 },
  { 2019, 11, 12 },  { 2019, 11, 13 },  { 2019, 11, 14 },  { 2019, 11, 15 },
  { 2019, 11, 16 },  { 2019, 11, 17 },  { 2019, 11, 18 },  { 2019, 11, 19 },
  { 2019, 11, 20 },  { 2019, 11, 21 },  { 2019, 11, 22 },  { 2019, 11, 23 },
  { 2019, 11, 24 },  { 2019, 11, 25 },  { 2019, 11, 26 },  { 2019, 11, 27 },
  { 2019, 11, 28 },  { 2019, 11, 29 },  { 2019, 11, 30 },  { 2019, 12,  1 },
  { 2019, 12,  2 },  { 2019, 12,  3 },  { 2019, 12,  4 },  { 2019, 12,  5 },
  { 2019, 12,  6 },  { 2019, 12,  7 },  { 2019, 12,  8 },  { 2019, 12,  9 },
  { 2019, 12, 10 },  { 2019, 12, 11 },  { 2019, 12, 12 },  { 2019, 12, 13 },
  { 2019, 12, 14 },  { 2019, 12, 15 },  { 2019, 12, 16 },  { 2019, 12, 17 },
  { 2019, 12, 18 },  { 2019, 12, 19 },  { 2019, 12, 20 },  { 2019, 12, 21 },
  { 2019, 12, 22 },  { 2019, 12, 23 },  { 2019, 12, 24 },  { 2019, 12, 25 },
  { 2019, 12, 26 },  { 2019, 12, 27 },  { 2019, 12, 28 },  { 2019, 12, 29 },
  { 2019, 12, 30 },  { 2019, 12, 31 },  { 2020,  1,  1 },  { 2020,  1,  2 },
  { 2020,  1,  3 },  { 2020,  1,  4 },  { 2020,  1,  5 },  { 2020,  1,  6 },
  { 2020,  1,  7 },  { 2020,  1,  8 },  { 2020,  1,  9 },  { 2020,  1, 10 },
  { 2020,  1, 11 },  { 2020,  1, 12 },  { 2020,  1, 13 },  { 2020,  1, 14 },
  { 2020,  1, 15 },  { 2020,  1, 16 },  { 2020,  1, 17 },  { 2020,  1, 18 },
  { 2020,  1, 19 },  { 2020,  1, 20 },  { 2020,  1, 21 },  { 2020,  1, 22 },
  { 2020,  1, 23 },  { 2020,  1, 24 },  { 2020,  1, 25 },  { 2020,  1, 26 },
  { 2020,  1, 27 },  { 2020,  1, 28 },  { 2020,  1, 29 },  { 2020,  1, 30 },
  { 2020,  1, 31 },  { 2020,  2,  1 },  { 2020,  2,  2 },  { 2020,  2,  3 },
  { 2020,  2,  4 },  { 2020,  2,  5 },  { 2020,  2,  6 },  { 2020,  2,  7 },
  { 2020,  2,  8 },  { 2020,  2,  9 },  { 2020,  2, 10 },  { 2020,  2, 11 },
  { 2020,  2, 12 },  { 2020,  2, 13 },  { 2020,  2, 14 },  { 2020,  2, 15 },
  { 2020,  2, 16 },  { 2020,  2, 17 },  { 2020,  2, 18 },  { 2020,  2, 19 },
  { 2020,  2, 20 },  { 2020,  2, 21 },  { 2020,  2, 22 },  { 2020,  2, 23 },
  { 2020,  2, 24 },  { 2020,  2, 25 },  { 2020,  2, 26 },  { 2020,  2, 27 },
  { 2020,  2, 28 },  { 2020,  2, 29 },  { 2020,  3,  1 },  { 2020,  3,  2 },
  { 2020,  3,  3 },  { 2020,  3,  4 },  { 2020,  3,  5 },  { 2020,  3,  6 },
  { 2020,  3,  7 },  { 2020,  3,  8 },  { 2020,  3,  9 },  { 2020,  3, 10 },
  { 2020,  3, 11 },  { 2020,  3, 12 },  { 2020,  3, 13 },  { 2020,  3, 14 },
  { 2020,  3, 15 },  { 2020,  3, 16 },  { 2020,  3, 17 },  { 2020,  3, 18 },
  { 2020,  3, 19 },  { 2020,  3, 20 },  { 2020,  3, 21 },  { 2020,  3, 22 },
  { 2020,  3, 23 },  { 2020,  3, 24 },  { 2020,  3, 25 },  { 2020,  3, 26 },
  { 2020,  3, 27 },  { 2020,  3, 28 },  { 2020,  3, 29 },  { 2020,  3, 30 },
  { 2020,  3, 31 },  { 2020,  4,  1 },  { 2020,  4,  2 },  { 2020,  4,  3 },
  { 2020,  4,  4 },  { 2020,  4,  5 },  { 2020,  4,  6 },  { 2020,  4,  7 },
  { 2020,  4,  8 },  { 2020,  4,  9 },  { 2020,  4, 10 },  { 2020,  4, 11 },
  { 2020,  4, 12 },  { 2020,  4, 13 },  { 2020,  4, 14 },  { 2020,  4, 15 },
  { 2020,  4, 16 },  { 2020,  4, 17 },  { 2020,  4, 18 },  { 2020,  4, 19 },
  { 2020,  4, 20 },  { 2020,  4, 21 },  { 2020,  4, 22 },  { 2020,  4, 23 },
  { 2020,  4, 24 },  { 2020,  4, 25 },  { 2020,  4, 26 },  { 2020,  4, 27 },
  { 2020,  4, 28 },  { 2020,  4, 29 },  { 2020,  4, 30 },  { 2020,  5,  1 },
  { 2020,  5,  2 },  { 2020,  5,  3 },  { 2020,  5,  4 },  { 2020,  5,  5 },
  { 2020,  5,  6 },  { 2020,  5,  7 },  { 2020,  5,  8 },  { 2020,  5,  9 },
  { 2020,  5, 10 },  { 2020,  5, 11 },  { 2020,  5, 12 },  { 2020,  5, 13 },
  { 2020,  5, 14 },  { 2020,  5, 15 },  { 2020,  5, 16 },  { 2020,  5, 17 },
  { 2020,  5, 18 },  { 2020,  5, 19 },  { 2020,  5, 20 },  { 2020,  5, 21 },
  { 2020,  5, 22 },  { 2020,  5, 23 },  { 2020,  5, 24 },  { 2020,  5, 25 },
  { 2020,  5, 26 },  { 2020,  5, 27 },  { 2020,  5, 28 },  { 2020,  5, 29 },
  { 2020,  5, 30 },  { 2020,  5, 31 },  { 2020,  6,  1 },  { 2020,  6,  2 },
  { 2020,  6,  3 },  { 2020,  6,  4 },  { 2020,  6,  5 },  { 2020,  6,  6 },
  { 2020,  6,  7 },  { 2020,  6,  8 },  { 2020,  6,  9 },  { 2020,  6, 10 },
  { 2020,  6, 11 },  { 2020,  6, 12 },  { 2020,  6, 13 },  { 2020,  6, 14 },
  { 2020,  6, 15 },  { 2020,  6, 16 },  { 2020,  6, 17 },  { 2020,  6, 18 },
  { 2020,  6, 19 },  { 2020,  6, 20 },  { 2020,  6, 21 },  { 2020,  6, 22 },
  { 2020,  6, 23 },  { 2020,  6, 24 },  { 2020,  6, 25 },  { 2020,  6, 26 },
  { 2020,  6, 27 },  { 2020,  6, 28 },  { 2020,  6, 29 },  { 2020,  6, 30 },
  { 2020,  7,  1 },  { 2020,  7,  2 },  { 2020,  7,  3 },  { 2020,  7,  4 },
  { 2020,  7,  5 },  { 2020,  7,  6 },  { 2020,  7,  7 },  { 2020,  7,  8 },
  { 2020,  7,  9 },  { 2020,  7, 10 },  { 2020,  7, 11 },  { 2020,  7, 12 },
  { 2020,  7, 13 },  { 2020,  7, 14 },  { 2020,  7, 15 },  { 2020,  7, 16 },
  { 2020,  7, 17 },  { 2020,  7, 18 },  { 2020,  7, 19 },  { 2020,  7, 20 },
  { 2020,  7, 21 },  { 2020,  7, 22 },  { 2020,  7, 23 },  { 2020,  7, 24 },
  { 2020,  7, 25 },  { 2020,  7, 26 },  { 2020,  7, 27 },  { 2020,  7, 28 },
  { 2020,  7, 29 },  { 2020,  7, 30 },  { 2020,  7, 31 },  { 2020,  8,  1 },
  { 2020,  8,  2 },  { 2020,  8,  3 },  { 2020,  8,  4 },  { 2020,  8,  5 },
  { 2020,  8,  6 },  { 2020,  8,  7 },  { 2020,  8,  8 },  { 2020,  8,  9 },
  { 2020,  8, 10 },  { 2020,  8, 11 },  { 2020,  8, 12 },  { 2020,  8, 13 },
  { 2020,  8, 14 },  { 2020,  8, 15 },  { 2020,  8, 16 },  { 2020,  8, 17 },
  { 2020,  8, 18 },  { 2020,  8, 19 },  { 2020,  8, 20 },  { 2020,  8, 21 },
  { 2020,  8, 22 },  { 2020,  8, 23 },  { 2020,  8, 24 },  { 2020,  8, 25 },
  { 2020,  8, 26 },  { 2020,  8, 27 },  { 2020,  8, 28 },  { 2020,  8, 29 },
  { 2020,  8, 30 },  { 2020,  8, 31 },  { 2020,  9,  1 },  { 2020,  9,  2 },
  { 2020,  9,  3 },  { 2020,  9,  4 },  { 2020,  9,  5 },  { 2020,  9,  6 },
  { 2020,  9,  7 },  { 2020,  9,  8 },  { 2020,  9,  9 },  { 2020,  9, 10 },
  { 2020,  9, 11 },  { 2020,  9, 12 },  { 2020,  9, 13 },  { 2020,  9, 14 },
  { 2020,  9, 15 },  { 2020,  9, 16 },  { 2020,  9, 17 },  { 2020,  9, 18 },
  { 2020,  9, 19 },  { 2020,  9, 20 },  { 2020,  9, 21 },  { 2020,  9, 22 },
  { 2020,  9, 23 },  { 2020,  9, 24 },  { 2020,  9, 25 },  { 2020,  9, 26 },
  { 2020,  9, 27 },  { 2020,  9, 28 },  { 2020,  9, 29 },  { 2020,  9, 30 },
  { 2020, 10,  1 },  { 2020, 10,  2 },  { 2020, 10,  3 },  { 2020, 10,  4 },
  { 2020, 10,  5 },  { 2020, 10,  6 },  { 2020, 10,  7 },  { 2020, 10,  8 },
  { 2020, 10,  9 },  { 2020, 10, 10 },  { 2020, 10, 11 },  { 2020, 10, 12 },
  { 2020, 10, 13 },  { 2020, 10, 14 },  { 2020, 10, 15 },  { 2020, 10, 16 },
  { 2020, 10, 17 },  { 2020, 10, 18 },  { 2020, 10, 19 },  { 2020, 10, 20 },
  { 2020, 10, 21 },  { 2020, 10, 22 },  { 2020, 10, 23 },  { 2020, 10, 24 },
  { 2020, 10, 25 },  { 2020, 10, 26 },  { 2020, 10, 27 },  { 2020, 10, 28 },
  { 2020, 10, 29 },  { 2020, 10, 30 },  { 2020, 10, 31 },  { 2020, 11,  1 },
  { 2020, 11,  2 },  { 2020, 11,  3 },  { 2020, 11,  4 },  { 2020, 11,  5 },
  { 2020, 11,  6 },  { 2020, 11,  7 },  { 2020, 11,  8 },  { 2020, 11,  9 },
  { 2020, 11, 10 },  { 2020, 11, 11 },  { 2020, 11, 12 },  { 2020, 11, 13 },
  { 2020, 11, 14 },  { 2020, 11, 15 },  { 2020, 11, 16 },  { 2020, 11, 17 },
  { 2020, 11, 18 },  { 2020, 11, 19 },  { 2020, 11, 20 },  { 2020, 11, 21 },
  { 2020, 11, 22 },  { 2020, 11, 23 },  { 2020, 11, 24 },  { 2020, 11, 25 },
  { 2020, 11, 26 },  { 2020, 11, 27 },  { 2020, 11, 28 },  { 2020, 11, 29 },
  { 2020, 11, 30 },  { 2020, 12,  1 },  { 2020, 12,  2 },  { 2020, 12,  3 },
  { 2020, 12,  4 },  { 2020, 12,  5 },  { 2020, 12,  6 },  { 2020, 12,  7 },
  { 2020, 12,  8 },  { 2020, 12,  9 },  { 2020, 12, 10 },  { 2020, 12, 11 },
  { 2020, 12, 12 },  { 2020, 12, 13 },  { 2020, 12, 14 },  { 2020, 12, 15 },
  { 2020, 12, 16 },  { 2020, 12, 17 },  { 2020, 12, 18 },  { 2020, 12, 19 },
  { 2020, 12, 20 },  { 2020, 12, 21 },  { 2020, 12, 22 },  { 2020, 12, 23 },
  { 2020, 12, 24 },  { 2020, 12, 25 },  { 2020, 12, 26 },  { 2020, 12, 27 },
  { 2020, 12, 28 },  { 2020, 12, 29 },  { 2020, 12, 30 },  { 2020, 12, 31 },
  { 2021,  1,  1 },  { 2021,  1,  2 },  { 2021,  1,  3 },  { 2021,  1,  4 },
  { 2021,  1,  5 },  { 2021,  1,  6 },  { 2021,  1,  7 },  { 2021,  1,  8 },
  { 2021,  1,  9 },  { 2021,  1, 10 },  { 2021,  1, 11 },  { 2021,  1, 12 },
  { 2021,  1, 13 },  { 2021,  1, 14 },  { 2021,  1, 15 },  { 2021,  1, 16 },
  { 2021,  1, 17 },  { 2021,  1, 18 },  { 2021,  1, 19 },  { 2021,  1, 20 },
  { 2021,  1, 21 },  { 2021,  1, 22 },  { 2021,  1, 23 },  { 2021,  1, 24 },
  { 2021,  1, 25 },  { 2021,  1, 26 },  { 2021,  1, 27 },  { 2021,  1, 28 },
  { 2021,  1, 29 },  { 2021,  1, 30 },  { 2021,  1, 31 },  { 2021,  2,  1 },
  { 2021,  2,  2 },  { 2021,  2,  3 },  { 2021,  2,  4 },  { 2021,  2,  5 },
  { 2021,  2,  6 },  { 2021,  2,  7 },  { 2021,  2,  8 },  { 2021,  2,  9 },
  { 2021,  2, 10 },  { 2021,  2, 11 },  { 2021,  2, 12 },  { 2021,  2, 13 },
  { 2021,  2, 14 },  { 2021,  2, 15 },  { 2021,  2, 16 },  { 2021,  2, 17 },
  { 2021,  2, 18 },  { 2021,  2, 19 },  { 2021,  2, 20 },  { 2021,  2, 21 },
  { 2021,  2, 22 },  { 2021,  2, 23 },  { 2021,  2, 24 },  { 2021,  2, 25 },
  { 2021,  2, 26 },  { 2021,  2, 27 },  { 2021,  2, 28 },  { 2021,  3,  1 },
  { 2021,  3,  2 },  { 2021,  3,  3 },  { 2021,  3,  4 },  { 2021,  3,  5 },
  { 2021,  3,  6 },  { 2021,  3,  7 },  { 2021,  3,  8 },  { 2021,  3,  9 },
  { 2021,  3, 10 },  { 2021,  3, 11 },  { 2021,  3, 12 },  { 2021,  3, 13 },
  { 2021,  3, 14 },  { 2021,  3, 15 },  { 2021,  3, 16 },  { 2021,  3, 17 },
  { 2021,  3, 18 },  { 2021,  3, 19 },  { 2021,  3, 20 },  { 2021,  3, 21 },
  { 2021,  3, 22 },  { 2021,  3, 23 },  { 2021,  3, 24 },  { 2021,  3, 25 },
  { 2021,  3, 26 },  { 2021,  3, 27 },  { 2021,  3, 28 },  { 2021,  3, 29 },
  { 2021,  3, 30 },  { 2021,  3, 31 },  { 2021,  4,  1 },  { 2021,  4,  2 },
  { 2021,  4,  3 },  { 2021,  4,  4 },  { 2021,  4,  5 },  { 2021,  4,  6 },
  { 2021,  4,  7 },  { 2021,  4,  8 },  { 2021,  4,  9 },  { 2021,  4, 10 },
  { 2021,  4, 11 },  { 2021,  4, 12 },  { 2021,  4, 13 },  { 2021,  4, 14 },
  { 2021,  4, 15 },  { 2021,  4, 16 },  { 2021,  4, 17 },  { 2021,  4, 18 },
  { 2021,  4, 19 },  { 2021,  4, 20 },  { 2021,  4, 21 },  { 2021,  4, 22 },
  { 2021,  4, 23 },  { 2021,  4, 24 },  { 2021,  4, 25 },  { 2021,  4, 26 },
  { 2021,  4, 27 },  { 2021,  4, 28 },  { 2021,  4, 29 },  { 2021,  4, 30 },
  { 2021,  5,  1 },  { 2021,  5,  2 },  { 2021,  5,  3 },  { 2021,  5,  4 },
  { 2021,  5,  5 },  { 2021,  5,  6 },  { 2021,  5,  7 },  { 2021,  5,  8 },
  { 2021,  5,  9 },  { 2021,  5, 10 },  { 2021,  5, 11 },  { 2021,  5, 12 },
  { 2021,  5, 13 },  { 2021,  5, 14 },  { 2021,  5, 15 },  { 2021,  5, 16 },
  { 2021,  5, 17 },  { 2021,  5, 18 },  { 2021,  5, 19 },  { 2021,  5, 20 },
  { 2021,  5, 21 },  { 2021,  5, 22 },  { 2021,  5, 23 },  { 2021,  5, 24 },
  { 2021,  5, 25 },  { 2021,  5, 26 },  { 2021,  5, 27 },  { 2021,  5, 28 },
  { 2021,  5, 29 },  { 2021,  5, 30 },  { 2021,  5, 31 },  { 2021,  6,  1 },
  { 2021,  6,  2 },  { 2021,  6,  3 },  { 2021,  6,  4 },  { 2021,  6,  5 },
  { 2021,  6,  6 },  { 2021,  6,  7 },  { 2021,  6,  8 },  { 2021,  6,  9 },
  { 2021,  6, 10 },  { 2021,  6, 11 },  { 2021,  6, 12 },  { 2021,  6, 13 },
  { 2021,  6, 14 },  { 2021,  6, 15 },  { 2021,  6, 16 },  { 2021,  6, 17 },
  { 2021,  6, 18 },  { 2021,  6, 19 },  { 2021,  6, 20 },  { 2021,  6, 21 },
  { 2021,  6, 22 },  { 2021,  6, 23 },  { 2021,  6, 24 },  { 2021,  6, 25 },
  { 2021,  6, 26 },  { 2021,  6, 27 },  { 2021,  6, 28 },  { 2021,  6, 29 },
  { 2021,  6, 30 },  { 2021,  7,  1 },  { 2021,  7,  2 },  { 2021,  7,  3 },
  { 2021,  7,  4 },  { 2021,  7,  5 },  { 2021,  7,  6 },  { 2021,  7,  7 },
  { 2021,  7,  8 },  { 2021,  7,  9 },  { 2021,  7, 10 },  { 2021,  7, 11 },
  { 2021,  7, 12 },  { 2021,  7, 13 },  { 2021,  7, 14 },  { 2021,  7, 15 },
  { 2021,  7, 16 },  { 2021,  7, 17 },  { 2021,  7, 18 },  { 2021,  7, 19 },
  { 2021,  7, 20 },  { 2021,  7, 21 },  { 2021,  7, 22 },  { 2021,  7, 23 },
  { 2021,  7, 24 },  { 2021,  7, 25 },  { 2021,  7, 26 },  { 2021,  7, 27 },
  { 2021,  7, 28 },  { 2021,  7, 29 },  { 2021,  7, 30 },  { 2021,  7, 31 },
  { 2021,  8,  1 },  { 2021,  8,  2 },  { 2021,  8,  3 },  { 2021,  8,  4 },
  { 2021,  8,  5 },  { 2021,  8,  6 },  { 2021,  8,  7 },  { 2021,  8,  8 },
  { 2021,  8,  9 },  { 2021,  8, 10 },  { 2021,  8, 11 },  { 2021,  8, 12 },
  { 2021,  8, 13 },  { 2021,  8, 14 },  { 2021,  8, 15 },  { 2021,  8, 16 },
  { 2021,  8, 17 },  { 2021,  8, 18 },  { 2021,  8, 19 },  { 2021,  8, 20 },
  { 2021,  8, 21 },  { 2021,  8, 22 },  { 2021,  8, 23 },  { 2021,  8, 24 },
  { 2021,  8, 25 },  { 2021,  8, 26 },  { 2021,  8, 27 },  { 2021,  8, 28 },
  { 2021,  8, 29 },  { 2021,  8, 30 },  { 2021,  8, 31 },  { 2021,  9,  1 },
  { 2021,  9,  2 },  { 2021,  9,  3 },  { 2021,  9,  4 },  { 2021,  9,  5 },
  { 2021,  9,  6 },  { 2021,  9,  7 },  { 2021,  9,  8 },  { 2021,  9,  9 },
  { 2021,  9, 10 },  { 2021,  9, 11 },  { 2021,  9, 12 },  { 2021,  9, 13 },
  { 2021,  9, 14 },  { 2021,  9, 15 },  { 2021,  9, 16 },  { 2021,  9, 17 },
  { 2021,  9, 18 },  { 2021,  9, 19 },  { 2021,  9, 20 },  { 2021,  9, 21 },
  { 2021,  9, 22 },  { 2021,  9, 23 },  { 2021,  9, 24 },  { 2021,  9, 25 },
  { 2021,  9, 26 },  { 2021,  9, 27 },  { 2021,  9, 28 },  { 2021,  9, 29 },
  { 2021,  9, 30 },  { 2021, 10,  1 },  { 2021, 10,  2 },  { 2021, 10,  3 },
  { 2021, 10,  4 },  { 2021, 10,  5 },  { 2021, 10,  6 },  { 2021, 10,  7 },
  { 2021, 10,  8 },  { 2021, 10,  9 },  { 2021, 10, 10 },  { 2021, 10, 11 },
  { 2021, 10, 12 },  { 2021, 10, 13 },  { 2021, 10, 14 },  { 2021, 10, 15 },
  { 2021, 10, 16 },  { 2021, 10, 17 },  { 2021, 10, 18 },  { 2021, 10, 19 },
  { 2021, 10, 20 },  { 2021, 10, 21 },  { 2021, 10, 22 },  { 2021, 10, 23 },
  { 2021, 10, 24 },  { 2021, 10, 25 },  { 2021, 10, 26 },  { 2021, 10, 27 },
  { 2021, 10, 28 },  { 2021, 10, 29 },  { 2021, 10, 30 },  { 2021, 10, 31 },
  { 2021, 11,  1 },  { 2021, 11,  2 },  { 2021, 11,  3 },  { 2021, 11,  4 },
  { 2021, 11,  5 },  { 2021, 11,  6 },  { 2021, 11,  7 },  { 2021, 11,  8 },
  { 2021, 11,  9 },  { 2021, 11, 10 },  { 2021, 11, 11 },  { 2021, 11, 12 },
  { 2021, 11, 13 },  { 2021, 11, 14 },  { 2021, 11, 15 },  { 2021, 11, 16 },
  { 2021, 11, 17 },  { 2021, 11, 18 },  { 2021, 11, 19 },  { 2021, 11, 20 },
  { 2021, 11, 21 },  { 2021, 11, 22 },  { 2021, 11, 23 },  { 2021, 11, 24 },
  { 2021, 11, 25 },  { 2021, 11, 26 },  { 2021, 11, 27 },  { 2021, 11, 28 },
  { 2021, 11, 29 },  { 2021, 11, 30 },  { 2021, 12,  1 },  { 2021, 12,  2 },
  { 2021, 12,  3 },  { 2021, 12,  4 },  { 2021, 12,  5 },  { 2021, 12,  6 },
  { 2021, 12,  7 },  { 2021, 12,  8 },  { 2021, 12,  9 },  { 2021, 12, 10 },
  { 2021, 12, 11 },  { 2021, 12, 12 },  { 2021, 12, 13 },  { 2021, 12, 14 },
  { 2021, 12, 15 },  { 2021, 12, 16 },  { 2021, 12, 17 },  { 2021, 12, 18 },
  { 2021, 12, 19 },  { 2021, 12, 20 },  { 2021, 12, 21 },  { 2021, 12, 22 },
  { 2021, 12, 23 },  { 2021, 12, 24 },  { 2021, 12, 25 },  { 2021, 12, 26 },
  { 2021, 12, 27 },  { 2021, 12, 28 },  { 2021, 12, 29 },  { 2021, 12, 30 },
  { 2021, 12, 31 },  { 2022,  1,  1 },  { 2022,  1,  2 },  { 2022,  1,  3 },
  { 2022,  1,  4 },  { 2022,  1,  5 },  { 2022,  1,  6 },  { 2022,  1,  7 },
  { 2022,  1,  8 },  { 2022,  1,  9 },  { 2022,  1, 10 },  { 2022,  1, 11 },
  { 2022,  1, 12 },  { 2022,  1, 13 },  { 2022,  1, 14 },  { 2022,  1, 15 },
  { 2022,  1, 16 },  { 2022,  1, 17 },  { 2022,  1, 18 },  { 2022,  1, 19 },
  { 2022,  1, 20 },  { 2022,  1, 21 },  { 2022,  1, 22 },  { 2022,  1, 23 },
  { 2022,  1, 24 },  { 2022,  1, 25 },  { 2022,  1, 26 },  { 2022,  1, 27 },
  { 2022,  1, 28 },  { 2022,  1, 29 },  { 2022,  1, 30 },  { 2022,  1, 31 },
  { 2022,  2,  1 },  { 2022,  2,  2 },  { 2022,  2,  3 },  { 2022,  2,  4 },
  { 2022,  2,  5 },  { 2022,  2,  6 },  { 2022,  2,  7 },  { 2022,  2,  8 },
  { 2022,  2,  9 },  { 2022,  2, 10 },  { 2022,  2, 11 },  { 2022,  2, 12 },
  { 2022,  2, 13 },  { 2022,  2, 14 },  { 2022,  2, 15 },  { 2022,  2, 16 },
  { 2022,  2, 17 },  { 2022,  2, 18 },  { 2022,  2, 19 },  { 2022,  2, 20 },
  { 2022,  2, 21 },  { 2022,  2, 22 },  { 2022,  2, 23 },  { 2022,  2, 24 },
  { 2022,  2, 25 },  { 2022,  2, 26 },  { 2022,  2, 27 },  { 2022,  2, 28 },
  { 2022,  3,  1 },  { 2022,  3,  2 },  { 2022,  3,  3 },  { 2022,  3,  4 },
  { 2022,  3,  5 },  { 2022,  3,  6 },  { 2022,  3,  7 },  { 2022,  3,  8 },
  { 2022,  3,  9 },  { 2022,  3, 10 },  { 2022,  3, 11 },  { 2022,  3, 12 },
  { 2022,  3, 13 },  { 2022,  3, 14 },  { 2022,  3, 15 },  { 2022,  3, 16 },
  { 2022,  3, 17 },  { 2022,  3, 18 },  { 2022,  3, 19 },  { 2022,  3, 20 },
  { 2022,  3, 21 },  { 2022,  3, 22 },  { 2022,  3, 23 },  { 2022,  3, 24 },
  { 2022,  3, 25 },  { 2022,  3, 26 },  { 2022,  3, 27 },  { 2022,  3, 28 },
  { 2022,  3, 29 },  { 2022,  3, 30 },  { 2022,  3, 31 },  { 2022,  4,  1 },
  { 2022,  4,  2 },  { 2022,  4,  3 },  { 2022,  4,  4 },  { 2022,  4,  5 },
  { 2022,  4,  6 },  { 2022,  4,  7 },  { 2022,  4,  8 },  { 2022,  4,  9 },
  { 2022,  4, 10 },  { 2022,  4, 11 },  { 2022,  4, 12 },  { 2022,  4, 13 },
  { 2022,  4, 14 },  { 2022,  4, 15 },  { 2022,  4, 16 },  { 2022,  4, 17 },
  { 2022,  4, 18 },  { 2022,  4, 19 },  { 2022,  4, 20 },  { 2022,  4, 21 },
  { 2022,  4, 22 },  { 2022,  4, 23 },  { 2022,  4, 24 },  { 2022,  4, 25 },
  { 2022,  4, 26 },  { 2022,  4, 27 },  { 2022,  4, 28 },  { 2022,  4, 29 },
  { 2022,  4, 30 },  { 2022,  5,  1 },  { 2022,  5,  2 },  { 2022,  5,  3 },
  { 2022,  5,  4 },  { 2022,  5,  5 },  { 2022,  5,  6 },  { 2022,  5,  7 },
  { 2022,  5,  8 },  { 2022,  5,  9 },  { 2022,  5, 10 },  { 2022,  5, 11 },
  { 2022,  5, 12 },  { 2022,  5, 13 },  { 2022,  5, 14 },  { 2022,  5, 15 },
  { 2022,  5, 16 },  { 2022,  5, 17 },  { 2022,  5, 18 },  { 2022,  5, 19 },
  { 2022,  5, 20 },  { 2022,  5, 21 },  { 2022,  5, 22 },  { 2022,  5, 23 },
  { 2022,  5, 24 },  { 2022,  5, 25 },  { 2022,  5, 26 },  { 2022,  5, 27 },
  { 2022,  5, 28 },  { 2022,  5, 29 },  { 2022,  5, 30 },  { 2022,  5, 31 },
  { 2022,  6,  1 },  { 2022,  6,  2 },  { 2022,  6,  3 },  { 2022,  6,  4 },
  { 2022,  6,  5 },  { 2022,  6,  6 },  { 2022,  6,  7 },  { 2022,  6,  8 },
  { 2022,  6,  9 },  { 2022,  6, 10 },  { 2022,  6, 11 },  { 2022,  6, 12 },
  { 2022,  6, 13 },  { 2022,  6, 14 },  { 2022,  6, 15 },  { 2022,  6, 16 },
  { 2022,  6, 17 },  { 2022,  6, 18 },  { 2022,  6, 19 },  { 2022,  6, 20 },
  { 2022,  6, 21 },  { 2022,  6, 22 },  { 2022,  6, 23 },  { 2022,  6, 24 },
  { 2022,  6, 25 },  { 2022,  6, 26 },  { 2022,  6, 27 },  { 2022,  6, 28 },
  { 2022,  6, 29 },  { 2022,  6, 30 },  { 2022,  7,  1 },  { 2022,  7,  2 },
  { 2022,  7,  3 },  { 2022,  7,  4 },  { 2022,  7,  5 },  { 2022,  7,  6 },
  { 2022,  7,  7 },  { 2022,  7,  8 },  { 2022,  7,  9 },  { 2022,  7, 10 },
  { 2022,  7, 11 },  { 2022,  7, 12 },  { 2022,  7, 13 },  { 2022,  7, 14 },
  { 2022,  7, 15 },  { 2022,  7, 16 },  { 2022,  7, 17 },  { 2022,  7, 18 },
  { 2022,  7, 19 },  { 2022,  7, 20 },  { 2022,  7, 21 },  { 2022,  7, 22 },
  { 2022,  7, 23 },  { 2022,  7, 24 },  { 2022,  7, 25 },  { 2022,  7, 26 },
  { 2022,  7, 27 },  { 2022,  7, 28 },  { 2022,  7, 29 },  { 2022,  7, 30 },
  { 2022,  7, 31 },  { 2022,  8,  1 },  { 2022,  8,  2 },  { 2022,  8,  3 },
  { 2022,  8,  4 },  { 2022,  8,  5 },  { 2022,  8,  6 },  { 2022,  8,  7 },
  { 2022,  8,  8 },  { 2022,  8,  9 },  { 2022,  8, 10 },  { 2022,  8, 11 },
  { 2022,  8, 12 },  { 2022,  8, 13 },  { 2022,  8, 14 },  { 2022,  8, 15 },
  { 2022,  8, 16 },  { 2022,  8, 17 },  { 2022,  8, 18 },  { 2022,  8, 19 },
  { 2022,  8, 20 },  { 2022,  8, 21 },  { 2022,  8, 22 },  { 2022,  8, 23 },
  { 2022,  8, 24 },  { 2022,  8, 25 },  { 2022,  8, 26 },  { 2022,  8, 27 },
  { 2022,  8, 28 },  { 2022,  8, 29 },  { 2022,  8, 30 },  { 2022,  8, 31 },
  { 2022,  9,  1 },  { 2022,  9,  2 },  { 2022,  9,  3 },  { 2022,  9,  4 },
  { 2022,  9,  5 },  { 2022,  9,  6 },  { 2022,  9,  7 },  { 2022,  9,  8 },
  { 2022,  9,  9 },  { 2022,  9, 10 },  { 2022,  9, 11 },  { 2022,  9, 12 },
  { 2022,  9, 13 },  { 2022,  9, 14 },  { 2022,  9, 15 },  { 2022,  9, 16 },
  { 2022,  9, 17 },  { 2022,  9, 18 },  { 2022,  9, 19 },  { 2022,  9, 20 },
  { 2022,  9, 21 },  { 2022,  9, 22 },  { 2022,  9, 23 },  { 2022,  9, 24 },
  { 2022,  9, 25 },  { 2022,  9, 26 },  { 2022,  9, 27 },  { 2022,  9, 28 },
  { 2022,  9, 29 },  { 2022,  9, 30 },  { 2022, 10,  1 },  { 2022, 10,  2 },
  { 2022, 10,  3 },  { 2022, 10,  4 },  { 2022, 10,  5 },  { 2022, 10,  6 },
  { 2022, 10,  7 },  { 2022, 10,  8 },  { 2022, 10,  9 },  { 2022, 10, 10 },
  { 2022, 10, 11 },  { 2022, 10, 12 },  { 2022, 10, 13 },  { 2022, 10, 14 },
  { 2022, 10, 15 },  { 2022, 10, 16 },  { 2022, 10, 17 },  { 2022, 10, 18 },
  { 2022, 10, 19 },  { 2022, 10, 20 },  { 2022, 10, 21 },  { 2022, 10, 22 },
  { 2022, 10, 23 },  { 2022, 10, 24 },  { 2022, 10, 25 },  { 2022, 10, 26 },
  { 2022, 10, 27 },  { 2022, 10, 28 },  { 2022, 10, 29 },  { 2022, 10, 30 },
  { 2022, 10, 31 },  { 2022, 11,  1 },  { 2022, 11,  2 },  { 2022, 11,  3 },
  { 2022, 11,  4 },  { 2022, 11,  5 },  { 2022, 11,  6 },  { 2022, 11,  7 },
  { 2022, 11,  8 },  { 2022, 11,  9 },  { 2022, 11, 10 },  { 2022, 11, 11 },
  { 2022, 11, 12 },  { 2022, 11, 13 },  { 2022, 11, 14 },  { 2022, 11, 15 },
  { 2022, 11, 16 },  { 2022, 11, 17 },  { 2022, 11, 18 },  { 2022, 11, 19 },
  { 2022, 11, 20 },  { 2022, 11, 21 },  { 2022, 11, 22 },  { 2022, 11, 23 },
  { 2022, 11, 24 },  { 2022, 11, 25 },  { 2022, 11, 26 },  { 2022, 11, 27 },
  { 2022, 11, 28 },  { 2022, 11, 29 },  { 2022, 11, 30 },  { 2022, 12,  1 },
  { 2022, 12,  2 },  { 2022, 12,  3 },  { 2022, 12,  4 },  { 2022, 12,  5 },
  { 2022, 12,  6 },  { 2022, 12,  7 },  { 2022, 12,  8 },  { 2022, 12,  9 },
  { 2022, 12, 10 },  { 2022, 12, 11 },  { 2022, 12, 12 },  { 2022, 12, 13 },
  { 2022, 12, 14 },  { 2022, 12, 15 },  { 2022, 12, 16 },  { 2022, 12, 17 },
  { 2022, 12, 18 },  { 2022, 12, 19 },  { 2022, 12, 20 },  { 2022, 12, 21 },
  { 2022, 12, 22 },  { 2022, 12, 23 },  { 2022, 12, 24 },  { 2022, 12, 25 },
  { 2022, 12, 26 },  { 2022, 12, 27 },  { 2022, 12, 28 },  { 2022, 12, 29 },
  { 2022, 12, 30 },  { 2022, 12, 31 },  { 2023,  1,  1 },  { 2023,  1,  2 },
  { 2023,  1,  3 },  { 2023,  1,  4 },  { 2023,  1,  5 },  { 2023,  1,  6 },
  { 2023,  1,  7 },  { 2023,  1,  8 },  { 2023,  1,  9 },  { 2023,  1, 10 },
  { 2023,  1, 11 },  { 2023,  1, 12 },  { 2023,  1, 13 },  { 2023,  1, 14 },
  { 2023,  1, 15 },  { 2023,  1, 16 },  { 2023,  1, 17 },  { 2023,  1, 18 },
  { 2023,  1, 19 },  { 2023,  1, 20 },  { 2023,  1, 21 },  { 2023,  1, 22 },
  { 2023,  1, 23 },  { 2023,  1, 24 },  { 2023,  1, 25 },  { 2023,  1, 26 },
  { 2023,  1, 27 },  { 2023,  1, 28 },  { 2023,  1, 29 },  { 2023,  1, 30 },
  { 2023,  1, 31 },  { 2023,  2,  1 },  { 2023,  2,  2 },  { 2023,  2,  3 },
  { 2023,  2,  4 },  { 2023,  2,  5 },  { 2023,  2,  6 },  { 2023,  2,  7 },
  { 2023,  2,  8 },  { 2023,  2,  9 },  { 2023,  2, 10 },  { 2023,  2, 11 },
  { 2023,  2, 12 },  { 2023,  2, 13 },  { 2023,  2, 14 },  { 2023,  2, 15 },
  { 2023,  2, 16 },  { 2023,  2, 17 },  { 2023,  2, 18 },  { 2023,  2, 19 },
  { 2023,  2, 20 },  { 2023,  2, 21 },  { 2023,  2, 22 },  { 2023,  2, 23 },
  { 2023,  2, 24 },  { 2023,  2, 25 },  { 2023,  2, 26 },  { 2023,  2, 27 },
  { 2023,  2, 28 },  { 2023,  3,  1 },  { 2023,  3,  2 },  { 2023,  3,  3 },
  { 2023,  3,  4 },  { 2023,  3,  5 },  { 2023,  3,  6 },  { 2023,  3,  7 },
  { 2023,  3,  8 },  { 2023,  3,  9 },  { 2023,  3, 10 },  { 2023,  3, 11 },
  { 2023,  3, 12 },  { 2023,  3, 13 },  { 2023,  3, 14 },  { 2023,  3, 15 },
  { 2023,  3, 16 },  { 2023,  3, 17 },  { 2023,  3, 18 },  { 2023,  3, 19 },
  { 2023,  3, 20 },  { 2023,  3, 21 },  { 2023,  3, 22 },  { 2023,  3, 23 },
  { 2023,  3, 24 },  { 2023,  3, 25 },  { 2023,  3, 26 },  { 2023,  3, 27 },
  { 2023,  3, 28 },  { 2023,  3, 29 },  { 2023,  3, 30 },  { 2023,  3, 31 },
  { 2023,  4,  1 },  { 2023,  4,  2 },  { 2023,  4,  3 },  { 2023,  4,  4 },
  { 2023,  4,  5 },  { 2023,  4,  6 },  { 2023,  4,  7 },  { 2023,  4,  8 },
  { 2023,  4,  9 },  { 2023,  4, 10 },  { 2023,  4, 11 },  { 2023,  4, 12 },
  { 2023,  4, 13 },  { 2023,  4, 14 },  { 2023,  4, 15 },  { 2023,  4, 16 },
  { 2023,  4, 17 },  { 2023,  4, 18 },  { 2023,  4, 19 },  { 2023,  4, 20 },
  { 2023,  4, 21 },  { 2023,  4, 22 },  { 2023,  4, 23 },  { 2023,  4, 24 },
  { 2023,  4, 25 },  { 2023,  4, 26 },  { 2023,  4, 27 },  { 2023,  4, 28 },
  { 2023,  4, 29 },  { 2023,  4, 30 },  { 2023,  5,  1 },  { 2023,  5,  2 },
  { 2023,  5,  3 },  { 2023,  5,  4 },  { 2023,  5,  5 },  { 2023,  5,  6 },
  { 2023,  5,  7 },  { 2023,  5,  8 },  { 2023,  5,  9 },  { 2023,  5, 10 },
  { 2023,  5, 11 },  { 2023,  5, 12 },  { 2023,  5, 13 },  { 2023,  5, 14 },
  { 2023,  5, 15 },  { 2023,  5, 16 },  { 2023,  5, 17 },  { 2023,  5, 18 },
  { 2023,  5, 19 },  { 2023,  5, 20 },  { 2023,  5, 21 },  { 2023,  5, 22 },
  { 2023,  5, 23 },  { 2023,  5, 24 },  { 2023,  5, 25 },  { 2023,  5, 26 },
  { 2023,  5, 27 },  { 2023,  5, 28 },  { 2023,  5, 29 },  { 2023,  5, 30 },
  { 2023,  5, 31 },  { 2023,  6,  1 },  { 2023,  6,  2 },  { 2023,  6,  3 },
  { 2023,  6,  4 },  { 2023,  6,  5 },  { 2023,  6,  6 },  { 2023,  6,  7 },
  { 2023,  6,  8 },  { 2023,  6,  9 },  { 2023,  6, 10 },  { 2023,  6, 11 },
  { 2023,  6, 12 },  { 2023,  6, 13 },  { 2023,  6, 14 },  { 2023,  6, 15 },
  { 2023,  6, 16 },  { 2023,  6, 17 },  { 2023,  6, 18 },  { 2023,  6, 19 },
  { 2023,  6, 20 },  { 2023,  6, 21 },  { 2023,  6, 22 },  { 2023,  6, 23 },
  { 2023,  6, 24 },  { 2023,  6, 25 },  { 2023,  6, 26 },  { 2023,  6, 27 },
  { 2023,  6, 28 },  { 2023,  6, 29 },  { 2023,  6, 30 },  { 2023,  7,  1 },
  { 2023,  7,  2 },  { 2023,  7,  3 },  { 2023,  7,  4 },  { 2023,  7,  5 },
  { 2023,  7,  6 },  { 2023,  7,  7 },  { 2023,  7,  8 },  { 2023,  7,  9 },
  { 2023,  7, 10 },  { 2023,  7, 11 },  { 2023,  7, 12 },  { 2023,  7, 13 },
  { 2023,  7, 14 },  { 2023,  7, 15 },  { 2023,  7, 16 },  { 2023,  7, 17 },
  { 2023,  7, 18 },  { 2023,  7, 19 },  { 2023,  7, 20 },  { 2023,  7, 21 },
  { 2023,  7, 22 },  { 2023,  7, 23 },  { 2023,  7, 24 },  { 2023,  7, 25 },
  { 2023,  7, 26 },  { 2023,  7, 27 },  { 2023,  7, 28 },  { 2023,  7, 29 },
  { 2023,  7, 30 },  { 2023,  7, 31 },  { 2023,  8,  1 },  { 2023,  8,  2 },
  { 2023,  8,  3 },  { 2023,  8,  4 },  { 2023,  8,  5 },  { 2023,  8,  6 },
  { 2023,  8,  7 },  { 2023,  8,  8 },  { 2023,  8,  9 },  { 2023,  8, 10 },
  { 2023,  8, 11 },  { 2023,  8, 12 },  { 2023,  8, 13 },  { 2023,  8, 14 },
  { 2023,  8, 15 },  { 2023,  8, 16 },  { 2023,  8, 17 },  { 2023,  8, 18 },
  { 2023,  8, 19 },  { 2023,  8, 20 },  { 2023,  8, 21 },  { 2023,  8, 22 },
  { 2023,  8, 23 },  { 2023,  8, 24 },  { 2023,  8, 25 },  { 2023,  8, 26 },
  { 2023,  8, 27 },  { 2023,  8, 28 },  { 2023,  8, 29 },  { 2023,  8, 30 },
  { 2023,  8, 31 },  { 2023,  9,  1 },  { 2023,  9,  2 },  { 2023,  9,  3 },
  { 2023,  9,  4 },  { 2023,  9,  5 },  { 2023,  9,  6 },  { 2023,  9,  7 },
  { 2023,  9,  8 },  { 2023,  9,  9 },  { 2023,  9, 10 },  { 2023,  9, 11 },
  { 2023,  9, 12 },  { 2023,  9, 13 },  { 2023,  9, 14 },  { 2023,  9, 15 },
  { 2023,  9, 16 },  { 2023,  9, 17 },  { 2023,  9, 18 },  { 2023,  9, 19 },
  { 2023,  9, 20 },  { 2023,  9, 21 },  { 2023,  9, 22 },  { 2023,  9, 23 },
  { 2023,  9, 24 },  { 2023,  9, 25 },  { 2023,  9, 26 },  { 2023,  9, 27 },
  { 2023,  9, 28 },  { 2023,  9, 29 },  { 2023,  9, 30 },  { 2023, 10,  1 },
  { 2023, 10,  2 },  { 2023, 10,  3 },  { 2023, 10,  4 },  { 2023, 10,  5 },
  { 2023, 10,  6 },  { 2023, 10,  7 },  { 2023, 10,  8 },  { 2023, 10,  9 },
  { 2023, 10, 10 },  { 2023, 10, 11 },  { 2023, 10, 12 },  { 2023, 10, 13 },
  { 2023, 10, 14 },  { 2023, 10, 15 },  { 2023, 10, 16 },  { 2023, 10, 17 },
  { 2023, 10, 18 },  { 2023, 10, 19 },  { 2023, 10, 20 },  { 2023, 10, 21 },
  { 2023, 10, 22 },  { 2023, 10, 23 },  { 2023, 10, 24 },  { 2023, 10, 25 },
  { 2023, 10, 26 },  { 2023, 10, 27 },  { 2023, 10, 28 },  { 2023, 10, 29 },
  { 2023, 10, 30 },  { 2023, 10, 31 },  { 2023, 11,  1 },  { 2023, 11,  2 },
  { 2023, 11,  3 },  { 2023, 11,  4 },  { 2023, 11,  5 },  { 2023, 11,  6 },
  { 2023, 11,  7 },  { 2023, 11,  8 },  { 2023, 11,  9 },  { 2023, 11, 10 },
  { 2023, 11, 11 },  { 2023, 11, 12 },  { 2023, 11, 13 },  { 2023, 11, 14 },
  { 2023, 11, 15 },  { 2023, 11, 16 },  { 2023, 11, 17 },  { 2023, 11, 18 },
  { 2023, 11, 19 },  { 2023, 11, 20 },  { 2023, 11, 21 },  { 2023, 11, 22 },
  { 2023, 11, 23 },  { 2023, 11, 24 },  { 2023, 11, 25 },  { 2023, 11, 26 },
  { 2023, 11, 27 },  { 2023, 11, 28 },  { 2023, 11, 29 },  { 2023, 11, 30 },
  { 2023, 12,  1 },  { 2023, 12,  2 },  { 2023, 12,  3 },  { 2023, 12,  4 },
  { 2023, 12,  5 },  { 2023, 12,  6 },  { 2023, 12,  7 },  { 2023, 12,  8 },
  { 2023, 12,  9 },  { 2023, 12, 10 },  { 2023, 12, 11 },  { 2023, 12, 12 },
  { 2023, 12, 13 },  { 2023, 12, 14 },  { 2023, 12, 15 },  { 2023, 12, 16 },
  { 2023, 12, 17 },  { 2023, 12, 18 },  { 2023, 12, 19 },  { 2023, 12, 20 },
  { 2023, 12, 21 },  { 2023, 12, 22 },  { 2023, 12, 23 },  { 2023, 12, 24 },
  { 2023, 12, 25 },  { 2023, 12, 26 },  { 2023, 12, 27 },  { 2023, 12, 28 },
  { 2023, 12, 29 },  { 2023, 12, 30 },  { 2023, 12, 31 },  { 2024,  1,  1 },
  { 2024,  1,  2 },  { 2024,  1,  3 },  { 2024,  1,  4 },  { 2024,  1,  5 },
  { 2024,  1,  6 },  { 2024,  1,  7 },  { 2024,  1,  8 },  { 2024,  1,  9 },
  { 2024,  1, 10 },  { 2024,  1, 11 },  { 2024,  1, 12 },  { 2024,  1, 13 },
  { 2024,  1, 14 },  { 2024,  1, 15 },  { 2024,  1, 16 },  { 2024,  1, 17 },
  { 2024,  1, 18 },  { 2024,  1, 19 },  { 2024,  1, 20 },  { 2024,  1, 21 },
  { 2024,  1, 22 },  { 2024,  1, 23 },  { 2024,  1, 24 },  { 2024,  1, 25 },
  { 2024,  1, 26 },  { 2024,  1, 27 },  { 2024,  1, 28 },  { 2024,  1, 29 },
  { 2024,  1, 30 },  { 2024,  1, 31 },  { 2024,  2,  1 },  { 2024,  2,  2 },
  { 2024,  2,  3 },  { 2024,  2,  4 },  { 2024,  2,  5 },  { 2024,  2,  6 },
  { 2024,  2,  7 },  { 2024,  2,  8 },  { 2024,  2,  9 },  { 2024,  2, 10 },
  { 2024,  2, 11 },  { 2024,  2, 12 },  { 2024,  2, 13 },  { 2024,  2, 14 },
  { 2024,  2, 15 },  { 2024,  2, 16 },  { 2024,  2, 17 },  { 2024,  2, 18 },
  { 2024,  2, 19 },  { 2024,  2, 20 },  { 2024,  2, 21 },  { 2024,  2, 22 },
  { 2024,  2, 23 },  { 2024,  2, 24 },  { 2024,  2, 25 },  { 2024,  2, 26 },
  { 2024,  2, 27 },  { 2024,  2, 28 },  { 2024,  2, 29 },  { 2024,  3,  1 },
  { 2024,  3,  2 },  { 2024,  3,  3 },  { 2024,  3,  4 },  { 2024,  3,  5 },
  { 2024,  3,  6 },  { 2024,  3,  7 },  { 2024,  3,  8 },  { 2024,  3,  9 },
  { 2024,  3, 10 },  { 2024,  3, 11 },  { 2024,  3, 12 },  { 2024,  3, 13 },
  { 2024,  3, 14 },  { 2024,  3, 15 },  { 2024,  3, 16 },  { 2024,  3, 17 },
  { 2024,  3, 18 },  { 2024,  3, 19 },  { 2024,  3, 20 },  { 2024,  3, 21 },
  { 2024,  3, 22 },  { 2024,  3, 23 },  { 2024,  3, 24 },  { 2024,  3, 25 },
  { 2024,  3, 26 },  { 2024,  3, 27 },  { 2024,  3, 28 },  { 2024,  3, 29 },
  { 2024,  3, 30 },  { 2024,  3, 31 },  { 2024,  4,  1 },  { 2024,  4,  2 },
  { 2024,  4,  3 },  { 2024,  4,  4 },  { 2024,  4,  5 },  { 2024,  4,  6 },
  { 2024,  4,  7 },  { 2024,  4,  8 },  { 2024,  4,  9 },  { 2024,  4, 10 },
  { 2024,  4, 11 },  { 2024,  4, 12 },  { 2024,  4, 13 },  { 2024,  4, 14 },
  { 2024,  4, 15 },  { 2024,  4, 16 },  { 2024,  4, 17 },  { 2024,  4, 18 },
  { 2024,  4, 19 },  { 2024,  4, 20 },  { 2024,  4, 21 },  { 2024,  4, 22 },
  { 2024,  4, 23 },  { 2024,  4, 24 },  { 2024,  4, 25 },  { 2024,  4, 26 },
  { 2024,  4, 27 },  { 2024,  4, 28 },  { 2024,  4, 29 },  { 2024,  4, 30 },
  { 2024,  5,  1 },  { 2024,  5,  2 },  { 2024,  5,  3 },  { 2024,  5,  4 },
  { 2024,  5,  5 },  { 2024,  5,  6 },  { 2024,  5,  7 },  { 2024,  5,  8 },
  { 2024,  5,  9 },  { 2024,  5, 10 },  { 2024,  5, 11 },  { 2024,  5, 12 },
  { 2024,  5, 13 },  { 2024,  5, 14 },  { 2024,  5, 15 },  { 2024,  5, 16 },
  { 2024,  5, 17 },  { 2024,  5, 18 },  { 2024,  5, 19 },  { 2024,  5, 20 },
  { 2024,  5, 21 },  { 2024,  5, 22 },  { 2024,  5, 23 },  { 2024,  5, 24 },
  { 2024,  5, 25 },  { 2024,  5, 26 },  { 2024,  5, 27 },  { 2024,  5, 28 },
  { 2024,  5, 29 },  { 2024,  5, 30 },  { 2024,  5, 31 },  { 2024,  6,  1 },
  { 2024,  6,  2 },  { 2024,  6,  3 },  { 2024,  6,  4 },  { 2024,  6,  5 },
  { 2024,  6,  6 },  { 2024,  6,  7 },  { 2024,  6,  8 },  { 2024,  6,  9 },
  { 2024,  6, 10 },  { 2024,  6, 11 },  { 2024,  6, 12 },  { 2024,  6, 13 },
  { 2024,  6, 14 },  { 2024,  6, 15 },  { 2024,  6, 16 },  { 2024,  6, 17 },
  { 2024,  6, 18 },  { 2024,  6, 19 },  { 2024,  6, 20 },  { 2024,  6, 21 },
  { 2024,  6, 22 },  { 2024,  6, 23 },  { 2024,  6, 24 },  { 2024,  6, 25 },
  { 2024,  6, 26 },  { 2024,  6, 27 },  { 2024,  6, 28 },  { 2024,  6, 29 },
  { 2024,  6, 30 },  { 2024,  7,  1 },  { 2024,  7,  2 },  { 2024,  7,  3 },
  { 2024,  7,  4 },  { 2024,  7,  5 },  { 2024,  7,  6 },  { 2024,  7,  7 },
  { 2024,  7,  8 },  { 2024,  7,  9 },  { 2024,  7, 10 },  { 2024,  7, 11 },
  { 2024,  7, 12 },  { 2024,  7, 13 },  { 2024,  7, 14 },  { 2024,  7, 15 },
  { 2024,  7, 16 },  { 2024,  7, 17 },  { 2024,  7, 18 },  { 2024,  7, 19 },
  { 2024,  7, 20 },  { 2024,  7, 21 },  { 2024,  7, 22 },  { 2024,  7, 23 },
  { 2024,  7, 24 },  { 2024,  7, 25 },  { 2024,  7, 26 },  { 2024,  7, 27 },
  { 2024,  7, 28 },  { 2024,  7, 29 },  { 2024,  7, 30 },  { 2024,  7, 31 },
  { 2024,  8,  1 },  { 2024,  8,  2 },  { 2024,  8,  3 },  { 2024,  8,  4 },
  { 2024,  8,  5 },  { 2024,  8,  6 },  { 2024,  8,  7 },  { 2024,  8,  8 },
  { 2024,  8,  9 },  { 2024,  8, 10 },  { 2024,  8, 11 },  { 2024,  8, 12 },
  { 2024,  8, 13 },  { 2024,  8, 14 },  { 2024,  8, 15 },  { 2024,  8, 16 },
  { 2024,  8, 17 },  { 2024,  8, 18 },  { 2024,  8, 19 },  { 2024,  8, 20 },
  { 2024,  8, 21 },  { 2024,  8, 22 },  { 2024,  8, 23 },  { 2024,  8, 24 },
  { 2024,  8, 25 },  { 2024,  8, 26 },  { 2024,  8, 27 },  { 2024,  8, 28 },
  { 2024,  8, 29 },  { 2024,  8, 30 },  { 2024,  8, 31 },  { 2024,  9,  1 },
  { 2024,  9,  2 },  { 2024,  9,  3 },  { 2024,  9,  4 },  { 2024,  9,  5 },
  { 2024,  9,  6 },  { 2024,  9,  7 },  { 2024,  9,  8 },  { 2024,  9,  9 },
  { 2024,  9, 10 },  { 2024,  9, 11 },  { 2024,  9, 12 },  { 2024,  9, 13 },
  { 2024,  9, 14 },  { 2024,  9, 15 },  { 2024,  9, 16 },  { 2024,  9, 17 },
  { 2024,  9, 18 },  { 2024,  9, 19 },  { 2024,  9, 20 },  { 2024,  9, 21 },
  { 2024,  9, 22 },  { 2024,  9, 23 },  { 2024,  9, 24 },  { 2024,  9, 25 },
  { 2024,  9, 26 },  { 2024,  9, 27 },  { 2024,  9, 28 },  { 2024,  9, 29 },
  { 2024,  9, 30 },  { 2024, 10,  1 },  { 2024, 10,  2 },  { 2024, 10,  3 },
  { 2024, 10,  4 },  { 2024, 10,  5 },  { 2024, 10,  6 },  { 2024, 10,  7 },
  { 2024, 10,  8 },  { 2024, 10,  9 },  { 2024, 10, 10 },  { 2024, 10, 11 },
  { 2024, 10, 12 },  { 2024, 10, 13 },  { 2024, 10, 14 },  { 2024, 10, 15 },
  { 2024, 10, 16 },  { 2024, 10, 17 },  { 2024, 10, 18 },  { 2024, 10, 19 },
  { 2024, 10, 20 },  { 2024, 10, 21 },  { 2024, 10, 22 },  { 2024, 10, 23 },
  { 2024, 10, 24 },  { 2024, 10, 25 },  { 2024, 10, 26 },  { 2024, 10, 27 },
  { 2024, 10, 28 },  { 2024, 10, 29 },  { 2024, 10, 30 },  { 2024, 10, 31 },
  { 2024, 11,  1 },  { 2024, 11,  2 },  { 2024, 11,  3 },  { 2024, 11,  4 },
  { 2024, 11,  5 },  { 2024, 11,  6 },  { 2024, 11,  7 },  { 2024, 11,  8 },
  { 2024, 11,  9 },  { 2024, 11, 10 },  { 2024, 11, 11 },  { 2024, 11, 12 },
  { 2024, 11, 13 },  { 2024, 11, 14 },  { 2024, 11, 15 },  { 2024, 11, 16 },
  { 2024, 11, 17 },  { 2024, 11, 18 },  { 2024, 11, 19 },  { 2024, 11, 20 },
  { 2024, 11, 21 },  { 2024, 11, 22 },  { 2024, 11, 23 },  { 2024, 11, 24 },
  { 2024, 11, 25 },  { 2024, 11, 26 },  { 2024, 11, 27 },  { 2024, 11, 28 },
  { 2024, 11, 29 },  { 2024, 11, 30 },  { 2024, 12,  1 },  { 2024, 12,  2 },
  { 2024, 12,  3 },  { 2024, 12,  4 },  { 2024, 12,  5 },  { 2024, 12,  6 },
  { 2024, 12,  7 },  { 2024, 12,  8 },  { 2024, 12,  9 },  { 2024, 12, 10 },
  { 2024, 12, 11 },  { 2024, 12, 12 },  { 2024, 12, 13 },  { 2024, 12, 14 },
  { 2024, 12, 15 },  { 2024, 12, 16 },  { 2024, 12, 17 },  { 2024, 12, 18 },
  { 2024, 12, 19 },  { 2024, 12, 20 },  { 2024, 12, 21 },  { 2024, 12, 22 },
  { 2024, 12, 23 },  { 2024, 12, 24 },  { 2024, 12, 25 },  { 2024, 12, 26 },
  { 2024, 12, 27 },  { 2024, 12, 28 },  { 2024, 12, 29 },  { 2024, 12, 30 },
  { 2024, 12, 31 },  { 2025,  1,  1 },  { 2025,  1,  2 },  { 2025,  1,  3 },
  { 2025,  1,  4 },  { 2025,  1,  5 },  { 2025,  1,  6 },  { 2025,  1,  7 },
  { 2025,  1,  8 },  { 2025,  1,  9 },  { 2025,  1, 10 },  { 2025,  1, 11 },
  { 2025,  1, 12 },  { 2025,  1, 13 },  { 2025,  1, 14 },  { 2025,  1, 15 },
  { 2025,  1, 16 },  { 2025,  1, 17 },  { 2025,  1, 18 },  { 2025,  1, 19 },
  { 2025,  1, 20 },  { 2025,  1, 21 },  { 2025,  1, 22 },  { 2025,  1, 23 },
  { 2025,  1, 24 },  { 2025,  1, 25 },  { 2025,  1, 26 },  { 2025,  1, 27 },
  { 2025,  1, 28 },  { 2025,  1, 29 },  { 2025,  1, 30 },  { 2025,  1, 31 },
  { 2025,  2,  1 },  { 2025,  2,  2 },  { 2025,  2,  3 },  { 2025,  2,  4 },
  { 2025,  2,  5 },  { 2025,  2,  6 },  { 2025,  2,  7 },  { 2025,  2,  8 },
  { 2025,  2,  9 },  { 2025,  2, 10 },  { 2025,  2, 11 },  { 2025,  2, 12 },
  { 2025,  2, 13 },  { 2025,  2, 14 },  { 2025,  2, 15 },  { 2025,  2, 16 },
  { 2025,  2, 17 },  { 2025,  2, 18 },  { 2025,  2, 19 },  { 2025,  2, 20 },
  { 2025,  2, 21 },  { 2025,  2, 22 },  { 2025,  2, 23 },  { 2025,  2, 24 },
  { 2025,  2, 25 },  { 2025,  2, 26 },  { 2025,  2, 27 },  { 2025,  2, 28 },
  { 2025,  3,  1 },  { 2025,  3,  2 },  { 2025,  3,  3 },  { 2025,  3,  4 },
  { 2025,  3,  5 },  { 2025,  3,  6 },  { 2025,  3,  7 },  { 2025,  3,  8 },
  { 2025,  3,  9 },  { 2025,  3, 10 },  { 2025,  3, 11 },  { 2025,  3, 12 },
  { 2025,  3, 13 },  { 2025,  3, 14 },  { 2025,  3, 15 },  { 2025,  3, 16 },
  { 2025,  3, 17 },  { 2025,  3, 18 },  { 2025,  3, 19 },  { 2025,  3, 20 },
  { 2025,  3, 21 },  { 2025,  3, 22 },  { 2025,  3, 23 },  { 2025,  3, 24 },
  { 2025,  3, 25 },  { 2025,  3, 26 },  { 2025,  3, 27 },  { 2025,  3, 28 },
  { 2025,  3, 29 },  { 2025,  3, 30 },  { 2025,  3, 31 },  { 2025,  4,  1 },
  { 2025,  4,  2 },  { 2025,  4,  3 },  { 2025,  4,  4 },  { 2025,  4,  5 },
  { 2025,  4,  6 },  { 2025,  4,  7 },  { 2025,  4,  8 },  { 2025,  4,  9 },
  { 2025,  4, 10 },  { 2025,  4, 11 },  { 2025,  4, 12 },  { 2025,  4, 13 },
  { 2025,  4, 14 },  { 2025,  4, 15 },  { 2025,  4, 16 },  { 2025,  4, 17 },
  { 2025,  4, 18 },  { 2025,  4, 19 },  { 2025,  4, 20 },  { 2025,  4, 21 },
  { 2025,  4, 22 },  { 2025,  4, 23 },  { 2025,  4, 24 },  { 2025,  4, 25 },
  { 2025,  4, 26 },  { 2025,  4, 27 },  { 2025,  4, 28 },  { 2025,  4, 29 },
  { 2025,  4, 30 },  { 2025,  5,  1 },  { 2025,  5,  2 },  { 2025,  5,  3 },
  { 2025,  5,  4 },  { 2025,  5,  5 },  { 2025,  5,  6 },  { 2025,  5,  7 },
  { 2025,  5,  8 },  { 2025,  5,  9 },  { 2025,  5, 10 },  { 2025,  5, 11 },
  { 2025,  5, 12 },  { 2025,  5, 13 },  { 2025,  5, 14 },  { 2025,  5, 15 },
  { 2025,  5, 16 },  { 2025,  5, 17 },  { 2025,  5, 18 },  { 2025,  5, 19 },
  { 2025,  5, 20 },  { 2025,  5, 21 },  { 2025,  5, 22 },  { 2025,  5, 23 },
  { 2025,  5, 24 },  { 2025,  5, 25 },  { 2025,  5, 26 },  { 2025,  5, 27 },
  { 2025,  5, 28 },  { 2025,  5, 29 },  { 2025,  5, 30 },  { 2025,  5, 31 },
  { 2025,  6,  1 },  { 2025,  6,  2 },  { 2025,  6,  3 },  { 2025,  6,  4 },
  { 2025,  6,  5 },  { 2025,  6,  6 },  { 2025,  6,  7 },  { 2025,  6,  8 },
  { 2025,  6,  9 },  { 2025,  6, 10 },  { 2025,  6, 11 },  { 2025,  6, 12 },
  { 2025,  6, 13 },  { 2025,  6, 14 },  { 2025,  6, 15 },  { 2025,  6, 16 },
  { 2025,  6, 17 },  { 2025,  6, 18 },  { 2025,  6, 19 },  { 2025,  6, 20 },
  { 2025,  6, 21 },  { 2025,  6, 22 },  { 2025,  6, 23 },  { 2025,  6, 24 },
  { 2025,  6, 25 },  { 2025,  6, 26 },  { 2025,  6, 27 },  { 2025,  6, 28 },
  { 2025,  6, 29 },  { 2025,  6, 30 },  { 2025,  7,  1 },  { 2025,  7,  2 },
  { 2025,  7,  3 },  { 2025,  7,  4 },  { 2025,  7,  5 },  { 2025,  7,  6 },
  { 2025,  7,  7 },  { 2025,  7,  8 },  { 2025,  7,  9 },  { 2025,  7, 10 },
  { 2025,  7, 11 },  { 2025,  7, 12 },  { 2025,  7, 13 },  { 2025,  7, 14 },
  { 2025,  7, 15 },  { 2025,  7, 16 },  { 2025,  7, 17 },  { 2025,  7, 18 },
  { 2025,  7, 19 },  { 2025,  7, 20 },  { 2025,  7, 21 },  { 2025,  7, 22 },
  { 2025,  7, 23 },  { 2025,  7, 24 },  { 2025,  7, 25 },  { 2025,  7, 26 },
  { 2025,  7, 27 },  { 2025,  7, 28 },  { 2025,  7, 29 },  { 2025,  7, 30 },
  { 2025,  7, 31 },  { 2025,  8,  1 },  { 2025,  8,  2 },  { 2025,  8,  3 },
  { 2025,  8,  4 },  { 2025,  8,  5 },  { 2025,  8,  6 },  { 2025,  8,  7 },
  { 2025,  8,  8 },  { 2025,  8,  9 },  { 2025,  8, 10 },  { 2025,  8, 11 },
  { 2025,  8, 12 },  { 2025,  8, 13 },  { 2025,  8, 14 },  { 2025,  8, 15 },
  { 2025,  8, 16 },  { 2025,  8, 17 },  { 2025,  8, 18 },  { 2025,  8, 19 },
  { 2025,  8, 20 },  { 2025,  8, 21 },  { 2025,  8, 22 },  { 2025,  8, 23 },
  { 2025,  8, 24 },  { 2025,  8, 25 },  { 2025,  8, 26 },  { 2025,  8, 27 },
  { 2025,  8, 28 },  { 2025,  8, 29 },  { 2025,  8, 30 },  { 2025,  8, 31 },
  { 2025,  9,  1 },  { 2025,  9,  2 },  { 2025,  9,  3 },  { 2025,  9,  4 },
  { 2025,  9,  5 },  { 2025,  9,  6 },  { 2025,  9,  7 },  { 2025,  9,  8 },
  { 2025,  9,  9 },  { 2025,  9, 10 },  { 2025,  9, 11 },  { 2025,  9, 12 },
  { 2025,  9, 13 },  { 2025,  9, 14 },  { 2025,  9, 15 },  { 2025,  9, 16 },
  { 2025,  9, 17 },  { 2025,  9, 18 },  { 2025,  9, 19 },  { 2025,  9, 20 },
  { 2025,  9, 21 },  { 2025,  9, 22 },  { 2025,  9, 23 },  { 2025,  9, 24 },
  { 2025,  9, 25 },  { 2025,  9, 26 },  { 2025,  9, 27 },  { 2025,  9, 28 },
  { 2025,  9, 29 },  { 2025,  9, 30 },  { 2025, 10,  1 },  { 2025, 10,  2 },
  { 2025, 10,  3 },  { 2025, 10,  4 },  { 2025, 10,  5 },  { 2025, 10,  6 },
  { 2025, 10,  7 },  { 2025, 10,  8 },  { 2025, 10,  9 },  { 2025, 10, 10 },
  { 2025, 10, 11 },  { 2025, 10, 12 },  { 2025, 10, 13 },  { 2025, 10, 14 },
  { 2025, 10, 15 },  { 2025, 10, 16 },  { 2025, 10, 17 },  { 2025, 10, 18 },
  { 2025, 10, 19 },  { 2025, 10, 20 },  { 2025, 10, 21 },  { 2025, 10, 22 },
  { 2025, 10, 23 },  { 2025, 10, 24 },  { 2025, 10, 25 },  { 2025, 10, 26 },
  { 2025, 10, 27 },  { 2025, 10, 28 },  { 2025, 10, 29 },  { 2025, 10, 30 },
  { 2025, 10, 31 },  { 2025, 11,  1 },  { 2025, 11,  2 },  { 2025, 11,  3 },
  { 2025, 11,  4 },  { 2025, 11,  5 },  { 2025, 11,  6 },  { 2025, 11,  7 },
  { 2025, 11,  8 },  { 2025, 11,  9 },  { 2025, 11, 10 },  { 2025, 11, 11 },
  { 2025, 11, 12 },  { 2025, 11, 13 },  { 2025, 11, 14 },  { 2025, 11, 15 },
  { 2025, 11, 16 },  { 2025, 11, 17 },  { 2025, 11, 18 },  { 2025, 11, 19 },
  { 2025, 11, 20 },  { 2025, 11, 21 },  { 2025, 11, 22 },  { 2025, 11, 23 },
  { 2025, 11, 24 },  { 2025, 11, 25 },  { 2025, 11, 26 },  { 2025, 11, 27 },
  { 2025, 11, 28 },  { 2025, 11, 29 },  { 2025, 11, 30 },  { 2025, 12,  1 },
  { 2025, 12,  2 },  { 2025, 12,  3 },  { 2025, 12,  4 },  { 2025, 12,  5 },
  { 2025, 12,  6 },  { 2025, 12,  7 },  { 2025, 12,  8 },  { 2025, 12,  9 },
  { 2025, 12, 10 },  { 2025, 12, 11 },  { 2025, 12, 12 },  { 2025, 12, 13 },
  { 2025, 12, 14 },  { 2025, 12, 15 },  { 2025, 12, 16 },  { 2025, 12, 17 },
  { 2025, 12, 18 },  { 2025, 12, 19 },  { 2025, 12, 20 },  { 2025, 12, 21 },
  { 2025, 12, 22 },  { 2025, 12, 23 },  { 2025, 12, 24 },  { 2025, 12, 25 },
  { 2025, 12, 26 },  { 2025, 12, 27 },  { 2025, 12, 28 },  { 2025, 12, 29 },
  { 2025, 12, 30 },  { 2025, 12, 31 },  { 2026,  1,  1 },  { 2026,  1,  2 },
  { 2026,  1,  3 },  { 2026,  1,  4 },  { 2026,  1,  5 },  { 2026,  1,  6 },
  { 2026,  1,  7 },  { 2026,  1,  8 },  { 2026,  1,  9 },  { 2026,  1, 10 },
  { 2026,  1, 11 },  { 2026,  1, 12 },  { 2026,  1, 13 },  { 2026,  1, 14 },
  { 2026,  1, 15 },  { 2026,  1, 16 },  { 2026,  1, 17 },  { 2026,  1, 18 },
  { 2026,  1, 19 },  { 2026,  1, 20 },  { 2026,  1, 21 },  { 2026,  1, 22 },
  { 2026,  1, 23 },  { 2026,  1, 24 },  { 2026,  1, 25 },  { 2026,  1, 26 },
  { 2026,  1, 27 },  { 2026,  1, 28 },  { 2026,  1, 29 },  { 2026,  1, 30 },
  { 2026,  1, 31 },  { 2026,  2,  1 },  { 2026,  2,  2 },  { 2026,  2,  3 },
  { 2026,  2,  4 },  { 2026,  2,  5 },  { 2026,  2,  6 },  { 2026,  2,  7 },
  { 2026,  2,  8 },  { 2026,  2,  9 },  { 2026,  2, 10 },  { 2026,  2, 11 },
  { 2026,  2, 12 },  { 2026,  2, 13 },  { 2026,  2, 14 },  { 2026,  2, 15 },
  { 2026,  2, 16 },  { 2026,  2, 17 },  { 2026,  2, 18 },  { 2026,  2, 19 },
  { 2026,  2, 20 },  { 2026,  2, 21 },  { 2026,  2, 22 },  { 2026,  2, 23 },
  { 2026,  2, 24 },  { 2026,  2, 25 },  { 2026,  2, 26 },  { 2026,  2, 27 },
  { 2026,  2, 28 },  { 2026,  3,  1 },  { 2026,  3,  2 },  { 2026,  3,  3 },
  { 2026,  3,  4 },  { 2026,  3,  5 },  { 2026,  3,  6 },  { 2026,  3,  7 },
  { 2026,  3,  8 },  { 2026,  3,  9 },  { 2026,  3, 10 },  { 2026,  3, 11 },
  { 2026,  3, 12 },  { 2026,  3, 13 },  { 2026,  3, 14 },  { 2026,  3, 15 },
  { 2026,  3, 16 },  { 2026,  3, 17 },  { 2026,  3, 18 },  { 2026,  3, 19 },
  { 2026,  3, 20 },  { 2026,  3, 21 },  { 2026,  3, 22 },  { 2026,  3, 23 },
  { 2026,  3, 24 },  { 2026,  3, 25 },  { 2026,  3, 26 },  { 2026,  3, 27 },
  { 2026,  3, 28 },  { 2026,  3, 29 },  { 2026,  3, 30 },  { 2026,  3, 31 },
  { 2026,  4,  1 },  { 2026,  4,  2 },  { 2026,  4,  3 },  { 2026,  4,  4 },
  { 2026,  4,  5 },  { 2026,  4,  6 },  { 2026,  4,  7 },  { 2026,  4,  8 },
  { 2026,  4,  9 },  { 2026,  4, 10 },  { 2026,  4, 11 },  { 2026,  4, 12 },
  { 2026,  4, 13 },  { 2026,  4, 14 },  { 2026,  4, 15 },  { 2026,  4, 16 },
  { 2026,  4, 17 },  { 2026,  4, 18 },  { 2026,  4, 19 },  { 2026,  4, 20 },
  { 2026,  4, 21 },  { 2026,  4, 22 },  { 2026,  4, 23 },  { 2026,  4, 24 },
  { 2026,  4, 25 },  { 2026,  4, 26 },  { 2026,  4, 27 },  { 2026,  4, 28 },
  { 2026,  4, 29 },  { 2026,  4, 30 },  { 2026,  5,  1 },  { 2026,  5,  2 },
  { 2026,  5,  3 },  { 2026,  5,  4 },  { 2026,  5,  5 },  { 2026,  5,  6 },
  { 2026,  5,  7 },  { 2026,  5,  8 },  { 2026,  5,  9 },  { 2026,  5, 10 },
  { 2026,  5, 11 },  { 2026,  5, 12 },  { 2026,  5, 13 },  { 2026,  5, 14 },
  { 2026,  5, 15 },  { 2026,  5, 16 },  { 2026,  5, 17 },  { 2026,  5, 18 },
  { 2026,  5, 19 },  { 2026,  5, 20 },  { 2026,  5, 21 },  { 2026,  5, 22 },
  { 2026,  5, 23 },  { 2026,  5, 24 },  { 2026,  5, 25 },  { 2026,  5, 26 },
  { 2026,  5, 27 },  { 2026,  5, 28 },  { 2026,  5, 29 },  { 2026,  5, 30 },
  { 2026,  5, 31 },  { 2026,  6,  1 },  { 2026,  6,  2 },  { 2026,  6,  3 },
  { 2026,  6,  4 },  { 2026,  6,  5 },  { 2026,  6,  6 },  { 2026,  6,  7 },
  { 2026,  6,  8 },  { 2026,  6,  9 },  { 2026,  6, 10 },  { 2026,  6, 11 },
  { 2026,  6, 12 },  { 2026,  6, 13 },  { 2026,  6, 14 },  { 2026,  6, 15 },
  { 2026,  6, 16 },  { 2026,  6, 17 },  { 2026,  6, 18 },  { 2026,  6, 19 },
  { 2026,  6, 20 },  { 2026,  6, 21 },  { 2026,  6, 22 },  { 2026,  6, 23 },
  { 2026,  6, 24 },  { 2026,  6, 25 },  { 2026,  6, 26 },  { 2026,  6, 27 },
  { 2026,  6, 28 },  { 2026,  6, 29 },  { 2026,  6, 30 },  { 2026,  7,  1 },
  { 2026,  7,  2 },  { 2026,  7,  3 },  { 2026,  7,  4 },  { 2026,  7,  5 },
  { 2026,  7,  6 },  { 2026,  7,  7 },  { 2026,  7,  8 },  { 2026,  7,  9 },
  { 2026,  7, 10 },  { 2026,  7, 11 },  { 2026,  7, 12 },  { 2026,  7, 13 },
  { 2026,  7, 14 },  { 2026,  7, 15 },  { 2026,  7, 16 },  { 2026,  7, 17 },
  { 2026,  7, 18 },  { 2026,  7, 19 },  { 2026,  7, 20 },  { 2026,  7, 21 },
  { 2026,  7, 22 },  { 2026,  7, 23 },  { 2026,  7, 24 },  { 2026,  7, 25 },
  { 2026,  7, 26 },  { 2026,  7, 27 },  { 2026,  7, 28 },  { 2026,  7, 29 },
  { 2026,  7, 30 },  { 2026,  7, 31 },  { 2026,  8,  1 },  { 2026,  8,  2 },
  { 2026,  8,  3 },  { 2026,  8,  4 },  { 2026,  8,  5 },  { 2026,  8,  6 },
  { 2026,  8,  7 },  { 2026,  8,  8 },  { 2026,  8,  9 },  { 2026,  8, 10 },
  { 2026,  8, 11 },  { 2026,  8, 12 },  { 2026,  8, 13 },  { 2026,  8, 14 },
  { 2026,  8, 15 },  { 2026,  8, 16 },  { 2026,  8, 17 },  { 2026,  8, 18 },
  { 2026,  8, 19 },  { 2026,  8, 20 },  { 2026,  8, 21 },  { 2026,  8, 22 },
  { 2026,  8, 23 },  { 2026,  8, 24 },  { 2026,  8, 25 },  { 2026,  8, 26 },
  { 2026,  8, 27 },  { 2026,  8, 28 },  { 2026,  8, 29 },  { 2026,  8, 30 },
  { 2026,  8, 31 },  { 2026,  9,  1 },  { 2026,  9,  2 },  { 2026,  9,  3 },
  { 2026,  9,  4 },  { 2026,  9,  5 },  { 2026,  9,  6 },  { 2026,  9,  7 },
  { 2026,  9,  8 },  { 2026,  9,  9 },  { 2026,  9, 10 },  { 2026,  9, 11 },
  { 2026,  9, 12 },  { 2026,  9, 13 },  { 2026,  9, 14 },  { 2026,  9, 15 },
  { 2026,  9, 16 },  { 2026,  9, 17 },  { 2026,  9, 18 },  { 2026,  9, 19 },
  { 2026,  9, 20 },  { 2026,  9, 21 },  { 2026,  9, 22 },  { 2026,  9, 23 },
  { 2026,  9, 24 },  { 2026,  9, 25 },  { 2026,  9, 26 },  { 2026,  9, 27 },
  { 2026,  9, 28 },  { 2026,  9, 29 },  { 2026,  9, 30 },  { 2026, 10,  1 },
  { 2026, 10,  2 },  { 2026, 10,  3 },  { 2026, 10,  4 },  { 2026, 10,  5 },
  { 2026, 10,  6 },  { 2026, 10,  7 },  { 2026, 10,  8 },  { 2026, 10,  9 },
  { 2026, 10, 10 },  { 2026, 10, 11 },  { 2026, 10, 12 },  { 2026, 10, 13 },
  { 2026, 10, 14 },  { 2026, 10, 15 },  { 2026, 10, 16 },  { 2026, 10, 17 },
  { 2026, 10, 18 },  { 2026, 10, 19 },  { 2026, 10, 20 },  { 2026, 10, 21 },
  { 2026, 10, 22 },  { 2026, 10, 23 },  { 2026, 10, 24 },  { 2026, 10, 25 },
  { 2026, 10, 26 },  { 2026, 10, 27 },  { 2026, 10, 28 },  { 2026, 10, 29 },
  { 2026, 10, 30 },  { 2026, 10, 31 },  { 2026, 11,  1 },  { 2026, 11,  2 },
  { 2026, 11,  3 },  { 2026, 11,  4 },  { 2026, 11,  5 },  { 2026, 11,  6 },
  { 2026, 11,  7 },  { 2026, 11,  8 },  { 2026, 11,  9 },  { 2026, 11, 10 },
  { 2026, 11, 11 },  { 2026, 11, 12 },  { 2026, 11, 13 },  { 2026, 11, 14 },
  { 2026, 11, 15 },  { 2026, 11, 16 },  { 2026, 11, 17 },  { 2026, 11, 18 },
  { 2026, 11, 19 },  { 2026, 11, 20 },  { 2026, 11, 21 },  { 2026, 11, 22 },
  { 2026, 11, 23 },  { 2026, 11, 24 },  { 2026, 11, 25 },  { 2026, 11, 26 },
  { 2026, 11, 27 },  { 2026, 11, 28 },  { 2026, 11, 29 },  { 2026, 11, 30 },
  { 2026, 12,  1 },  { 2026, 12,  2 },  { 2026, 12,  3 },  { 2026, 12,  4 },
  { 2026, 12,  5 },  { 2026, 12,  6 },  { 2026, 12,  7 },  { 2026, 12,  8 },
  { 2026, 12,  9 },  { 2026, 12, 10 },  { 2026, 12, 11 },  { 2026, 12, 12 },
  { 2026, 12, 13 },  { 2026, 12, 14 },  { 2026, 12, 15 },  { 2026, 12, 16 },
  { 2026, 12, 17 },  { 2026, 12, 18 },  { 2026, 12, 19 },  { 2026, 12, 20 },
  { 2026, 12, 21 },  { 2026, 12, 22 },  { 2026, 12, 23 },  { 2026, 12, 24 },
  { 2026, 12, 25 },  { 2026, 12, 26 },  { 2026, 12, 27 },  { 2026, 12, 28 },
  { 2026, 12, 29 },  { 2026, 12, 30 },  { 2026, 12, 31 },  { 2027,  1,  1 },
  { 2027,  1,  2 },  { 2027,  1,  3 },  { 2027,  1,  4 },  { 2027,  1,  5 },
  { 2027,  1,  6 },  { 2027,  1,  7 },  { 2027,  1,  8 },  { 2027,  1,  9 },
  { 2027,  1, 10 },  { 2027,  1, 11 },  { 2027,  1, 12 },  { 2027,  1, 13 },
  { 2027,  1, 14 },  { 2027,  1, 15 },  { 2027,  1, 16 },  { 2027,  1, 17 },
  { 2027,  1, 18 },  { 2027,  1, 19 },  { 2027,  1, 20 },  { 2027,  1, 21 },
  { 2027,  1, 22 },  { 2027,  1, 23 },  { 2027,  1, 24 },  { 2027,  1, 25 },
  { 2027,  1, 26 },  { 2027,  1, 27 },  { 2027,  1, 28 },  { 2027,  1, 29 },
  { 2027,  1, 30 },  { 2027,  1, 31 },  { 2027,  2,  1 },  { 2027,  2,  2 },
  { 2027,  2,  3 },  { 2027,  2,  4 },  { 2027,  2,  5 },  { 2027,  2,  6 },
  { 2027,  2,  7 },  { 2027,  2,  8 },  { 2027,  2,  9 },  { 2027,  2, 10 },
  { 2027,  2, 11 },  { 2027,  2, 12 },  { 2027,  2, 13 },  { 2027,  2, 14 },
  { 2027,  2, 15 },  { 2027,  2, 16 },  { 2027,  2, 17 },  { 2027,  2, 18 },
  { 2027,  2, 19 },  { 2027,  2, 20 },  { 2027,  2, 21 },  { 2027,  2, 22 },
  { 2027,  2, 23 },  { 2027,  2, 24 },  { 2027,  2, 25 },  { 2027,  2, 26 },
  { 2027,  2, 27 },  { 2027,  2, 28 },  { 2027,  3,  1 },  { 2027,  3,  2 },
  { 2027,  3,  3 },  { 2027,  3,  4 },  { 2027,  3,  5 },  { 2027,  3,  6 },
  { 2027,  3,  7 },  { 2027,  3,  8 },  { 2027,  3,  9 },  { 2027,  3, 10 },
  { 2027,  3, 11 },  { 2027,  3, 12 },  { 2027,  3, 13 },  { 2027,  3, 14 },
  { 2027,  3, 15 },  { 2027,  3, 16 },  { 2027,  3, 17 },  { 2027,  3, 18 },
  { 2027,  3, 19 },  { 2027,  3, 20 },  { 2027,  3, 21 },  { 2027,  3, 22 },
  { 2027,  3, 23 },  { 2027,  3, 24 },  { 2027,  3, 25 },  { 2027,  3, 26 },
  { 2027,  3, 27 },  { 2027,  3, 28 },  { 2027,  3, 29 },  { 2027,  3, 30 },
  { 2027,  3, 31 },  { 2027,  4,  1 },  { 2027,  4,  2 },  { 2027,  4,  3 },
  { 2027,  4,  4 },  { 2027,  4,  5 },  { 2027,  4,  6 },  { 2027,  4,  7 },
  { 2027,  4,  8 },  { 2027,  4,  9 },  { 2027,  4, 10 },  { 2027,  4, 11 },
  { 2027,  4, 12 },  { 2027,  4, 13 },  { 2027,  4, 14 },  { 2027,  4, 15 },
  { 2027,  4, 16 },  { 2027,  4, 17 },  { 2027,  4, 18 },  { 2027,  4, 19 },
  { 2027,  4, 20 },  { 2027,  4, 21 },  { 2027,  4, 22 },  { 2027,  4, 23 },
  { 2027,  4, 24 },  { 2027,  4, 25 },  { 2027,  4, 26 },  { 2027,  4, 27 },
  { 2027,  4, 28 },  { 2027,  4, 29 },  { 2027,  4, 30 },  { 2027,  5,  1 },
  { 2027,  5,  2 },  { 2027,  5,  3 },  { 2027,  5,  4 },  { 2027,  5,  5 },
  { 2027,  5,  6 },  { 2027,  5,  7 },  { 2027,  5,  8 },  { 2027,  5,  9 },
  { 2027,  5, 10 },  { 2027,  5, 11 },  { 2027,  5, 12 },  { 2027,  5, 13 },
  { 2027,  5, 14 },  { 2027,  5, 15 },  { 2027,  5, 16 },  { 2027,  5, 17 },
  { 2027,  5, 18 },  { 2027,  5, 19 },  { 2027,  5, 20 },  { 2027,  5, 21 },
  { 2027,  5, 22 },  { 2027,  5, 23 },  { 2027,  5, 24 },  { 2027,  5, 25 },
  { 2027,  5, 26 },  { 2027,  5, 27 },  { 2027,  5, 28 },  { 2027,  5, 29 },
  { 2027,  5, 30 },  { 2027,  5, 31 },  { 2027,  6,  1 },  { 2027,  6,  2 },
  { 2027,  6,  3 },  { 2027,  6,  4 },  { 2027,  6,  5 },  { 2027,  6,  6 },
  { 2027,  6,  7 },  { 2027,  6,  8 },  { 2027,  6,  9 },  { 2027,  6, 10 },
  { 2027,  6, 11 },  { 2027,  6, 12 },  { 2027,  6, 13 },  { 2027,  6, 14 },
  { 2027,  6, 15 },  { 2027,  6, 16 },  { 2027,  6, 17 },  { 2027,  6, 18 },
  { 2027,  6, 19 },  { 2027,  6, 20 },  { 2027,  6, 21 },  { 2027,  6, 22 },
  { 2027,  6, 23 },  { 2027,  6, 24 },  { 2027,  6, 25 },  { 2027,  6, 26 },
  { 2027,  6, 27 },  { 2027,  6, 28 },  { 2027,  6, 29 },  { 2027,  6, 30 },
  { 2027,  7,  1 },  { 2027,  7,  2 },  { 2027,  7,  3 },  { 2027,  7,  4 },
  { 2027,  7,  5 },  { 2027,  7,  6 },  { 2027,  7,  7 },  { 2027,  7,  8 },
  { 2027,  7,  9 },  { 2027,  7, 10 },  { 2027,  7, 11 },  { 2027,  7, 12 },
  { 2027,  7, 13 },  { 2027,  7, 14 },  { 2027,  7, 15 },  { 2027,  7, 16 },
  { 2027,  7, 17 },  { 2027,  7, 18 },  { 2027,  7, 19 },  { 2027,  7, 20 },
  { 2027,  7, 21 },  { 2027,  7, 22 },  { 2027,  7, 23 },  { 2027,  7, 24 },
  { 2027,  7, 25 },  { 2027,  7, 26 },  { 2027,  7, 27 },  { 2027,  7, 28 },
  { 2027,  7, 29 },  { 2027,  7, 30 },  { 2027,  7, 31 },  { 2027,  8,  1 },
  { 2027,  8,  2 },  { 2027,  8,  3 },  { 2027,  8,  4 },  { 2027,  8,  5 },
  { 2027,  8,  6 },  { 2027,  8,  7 },  { 2027,  8,  8 },  { 2027,  8,  9 },
  { 2027,  8, 10 },  { 2027,  8, 11 },  { 2027,  8, 12 },  { 2027,  8, 13 },
  { 2027,  8, 14 },  { 2027,  8, 15 },  { 2027,  8, 16 },  { 2027,  8, 17 },
  { 2027,  8, 18 },  { 2027,  8, 19 },  { 2027,  8, 20 },  { 2027,  8, 21 },
  { 2027,  8, 22 },  { 2027,  8, 23 },  { 2027,  8, 24 },  { 2027,  8, 25 },
  { 2027,  8, 26 },  { 2027,  8, 27 },  { 2027,  8, 28 },  { 2027,  8, 29 },
  { 2027,  8, 30 },  { 2027,  8, 31 },  { 2027,  9,  1 },  { 2027,  9,  2 },
  { 2027,  9,  3 },  { 2027,  9,  4 },  { 2027,  9,  5 },  { 2027,  9,  6 },
  { 2027,  9,  7 },  { 2027,  9,  8 },  { 2027,  9,  9 },  { 2027,  9, 10 },
  { 2027,  9, 11 },  { 2027,  9, 12 },  { 2027,  9, 13 },  { 2027,  9, 14 },
  { 2027,  9, 15 },  { 2027,  9, 16 },  { 2027,  9, 17 },  { 2027,  9, 18 },
  { 2027,  9, 19 },  { 2027,  9, 20 },  { 2027,  9, 21 },  { 2027,  9, 22 },
  { 2027,  9, 23 },  { 2027,  9, 24 },  { 2027,  9, 25 },  { 2027,  9, 26 },
  { 2027,  9, 27 },  { 2027,  9, 28 },  { 2027,  9, 29 },  { 2027,  9, 30 },
  { 2027, 10,  1 },  { 2027, 10,  2 },  { 2027, 10,  3 },  { 2027, 10,  4 },
  { 2027, 10,  5 },  { 2027, 10,  6 },  { 2027, 10,  7 },  { 2027, 10,  8 },
  { 2027, 10,  9 },  { 2027, 10, 10 },  { 2027, 10, 11 },  { 2027, 10, 12 },
  { 2027, 10, 13 },  { 2027, 10, 14 },  { 2027, 10, 15 },  { 2027, 10, 16 },
  { 2027, 10, 17 },  { 2027, 10, 18 },  { 2027, 10, 19 },  { 2027, 10, 20 },
  { 2027, 10, 21 },  { 2027, 10, 22 },  { 2027, 10, 23 },  { 2027, 10, 24 },
  { 2027, 10, 25 },  { 2027, 10, 26 },  { 2027, 10, 27 },  { 2027, 10, 28 },
  { 2027, 10, 29 },  { 2027, 10, 30 },  { 2027, 10, 31 },  { 2027, 11,  1 },
  { 2027, 11,  2 },  { 2027, 11,  3 },  { 2027, 11,  4 },  { 2027, 11,  5 },
  { 2027, 11,  6 },  { 2027, 11,  7 },  { 2027, 11,  8 },  { 2027, 11,  9 },
  { 2027, 11, 10 },  { 2027, 11, 11 },  { 2027, 11, 12 },  { 2027, 11, 13 },
  { 2027, 11, 14 },  { 2027, 11, 15 },  { 2027, 11, 16 },  { 2027, 11, 17 },
  { 2027, 11, 18 },  { 2027, 11, 19 },  { 2027, 11, 20 },  { 2027, 11, 21 },
  { 2027, 11, 22 },  { 2027, 11, 23 },  { 2027, 11, 24 },  { 2027, 11, 25 },
  { 2027, 11, 26 },  { 2027, 11, 27 },  { 2027, 11, 28 },  { 2027, 11, 29 },
  { 2027, 11, 30 },  { 2027, 12,  1 },  { 2027, 12,  2 },  { 2027, 12,  3 },
  { 2027, 12,  4 },  { 2027, 12,  5 },  { 2027, 12,  6 },  { 2027, 12,  7 },
  { 2027, 12,  8 },  { 2027, 12,  9 },  { 2027, 12, 10 },  { 2027, 12, 11 },
  { 2027, 12, 12 },  { 2027, 12, 13 },  { 2027, 12, 14 },  { 2027, 12, 15 },
  { 2027, 12, 16 },  { 2027, 12, 17 },  { 2027, 12, 18 },  { 2027, 12, 19 },
  { 2027, 12, 20 },  { 2027, 12, 21 },  { 2027, 12, 22 },  { 2027, 12, 23 },
  { 2027, 12, 24 },  { 2027, 12, 25 },  { 2027, 12, 26 },  { 2027, 12, 27 },
  { 2027, 12, 28 },  { 2027, 12, 29 },  { 2027, 12, 30 },  { 2027, 12, 31 },
  { 2028,  1,  1 },  { 2028,  1,  2 },  { 2028,  1,  3 },  { 2028,  1,  4 },
  { 2028,  1,  5 },  { 2028,  1,  6 },  { 2028,  1,  7 },  { 2028,  1,  8 },
  { 2028,  1,  9 },  { 2028,  1, 10 },  { 2028,  1, 11 },  { 2028,  1, 12 },
  { 2028,  1, 13 },  { 2028,  1, 14 },  { 2028,  1, 15 },  { 2028,  1, 16 },
  { 2028,  1, 17 },  { 2028,  1, 18 },  { 2028,  1, 19 },  { 2028,  1, 20 },
  { 2028,  1, 21 },  { 2028,  1, 22 },  { 2028,  1, 23 },  { 2028,  1, 24 },
  { 2028,  1, 25 },  { 2028,  1, 26 },  { 2028,  1, 27 },  { 2028,  1, 28 },
  { 2028,  1, 29 },  { 2028,  1, 30 },  { 2028,  1, 31 },  { 2028,  2,  1 },
  { 2028,  2,  2 },  { 2028,  2,  3 },  { 2028,  2,  4 },  { 2028,  2,  5 },
  { 2028,  2,  6 },  { 2028,  2,  7 },  { 2028,  2,  8 },  { 2028,  2,  9 },
  { 2028,  2, 10 },  { 2028,  2, 11 },  { 2028,  2, 12 },  { 2028,  2, 13 },
  { 2028,  2, 14 },  { 2028,  2, 15 },  { 2028,  2, 16 },  { 2028,  2, 17 },
  { 2028,  2, 18 },  { 2028,  2, 19 },  { 2028,  2, 20 },  { 2028,  2, 21 },
  { 2028,  2, 22 },  { 2028,  2, 23 },  { 2028,  2, 24 },  { 2028,  2, 25 },
  { 2028,  2, 26 },  { 2028,  2, 27 },  { 2028,  2, 28 },  { 2028,  2, 29 },
  { 2028,  3,  1 },  { 2028,  3,  2 },  { 2028,  3,  3 },  { 2028,  3,  4 },
  { 2028,  3,  5 },  { 2028,  3,  6 },  { 2028,  3,  7 },  { 2028,  3,  8 },
  { 2028,  3,  9 },  { 2028,  3, 10 },  { 2028,  3, 11 },  { 2028,  3, 12 },
  { 2028,  3, 13 },  { 2028,  3, 14 },  { 2028,  3, 15 },  { 2028,  3, 16 },
  { 2028,  3, 17 },  { 2028,  3, 18 },  { 2028,  3, 19 },  { 2028,  3, 20 },
  { 2028,  3, 21 },  { 2028,  3, 22 },  { 2028,  3, 23 },  { 2028,  3, 24 },
  { 2028,  3, 25 },  { 2028,  3, 26 },  { 2028,  3, 27 },  { 2028,  3, 28 },
  { 2028,  3, 29 },  { 2028,  3, 30 },  { 2028,  3, 31 },  { 2028,  4,  1 },
  { 2028,  4,  2 },  { 2028,  4,  3 },  { 2028,  4,  4 },  { 2028,  4,  5 },
  { 2028,  4,  6 },  { 2028,  4,  7 },  { 2028,  4,  8 },  { 2028,  4,  9 },
  { 2028,  4, 10 },  { 2028,  4, 11 },  { 2028,  4, 12 },  { 2028,  4, 13 },
  { 2028,  4, 14 },  { 2028,  4, 15 },  { 2028,  4, 16 },  { 2028,  4, 17 },
  { 2028,  4, 18 },  { 2028,  4, 19 },  { 2028,  4, 20 },  { 2028,  4, 21 },
  { 2028,  4, 22 },  { 2028,  4, 23 },  { 2028,  4, 24 },  { 2028,  4, 25 },
  { 2028,  4, 26 },  { 2028,  4, 27 },  { 2028,  4, 28 },  { 2028,  4, 29 },
  { 2028,  4, 30 },  { 2028,  5,  1 },  { 2028,  5,  2 },  { 2028,  5,  3 },
  { 2028,  5,  4 },  { 2028,  5,  5 },  { 2028,  5,  6 },  { 2028,  5,  7 },
  { 2028,  5,  8 },  { 2028,  5,  9 },  { 2028,  5, 10 },  { 2028,  5, 11 },
  { 2028,  5, 12 },  { 2028,  5, 13 },  { 2028,  5, 14 },  { 2028,  5, 15 },
  { 2028,  5, 16 },  { 2028,  5, 17 },  { 2028,  5, 18 },  { 2028,  5, 19 },
  { 2028,  5, 20 },  { 2028,  5, 21 },  { 2028,  5, 22 },  { 2028,  5, 23 },
  { 2028,  5, 24 },  { 2028,  5, 25 },  { 2028,  5, 26 },  { 2028,  5, 27 },
  { 2028,  5, 28 },  { 2028,  5, 29 },  { 2028,  5, 30 },  { 2028,  5, 31 },
  { 2028,  6,  1 },  { 2028,  6,  2 },  { 2028,  6,  3 },  { 2028,  6,  4 },
  { 2028,  6,  5 },  { 2028,  6,  6 },  { 2028,  6,  7 },  { 2028,  6,  8 },
  { 2028,  6,  9 },  { 2028,  6, 10 },  { 2028,  6, 11 },  { 2028,  6, 12 },
  { 2028,  6, 13 },  { 2028,  6, 14 },  { 2028,  6, 15 },  { 2028,  6, 16 },
  { 2028,  6, 17 },  { 2028,  6, 18 },  { 2028,  6, 19 },  { 2028,  6, 20 },
  { 2028,  6, 21 },  { 2028,  6, 22 },  { 2028,  6, 23 },  { 2028,  6, 24 },
  { 2028,  6, 25 },  { 2028,  6, 26 },  { 2028,  6, 27 },  { 2028,  6, 28 },
  { 2028,  6, 29 },  { 2028,  6, 30 },  { 2028,  7,  1 },  { 2028,  7,  2 },
  { 2028,  7,  3 },  { 2028,  7,  4 },  { 2028,  7,  5 },  { 2028,  7,  6 },
  { 2028,  7,  7 },  { 2028,  7,  8 },  { 2028,  7,  9 },  { 2028,  7, 10 },
  { 2028,  7, 11 },  { 2028,  7, 12 },  { 2028,  7, 13 },  { 2028,  7, 14 },
  { 2028,  7, 15 },  { 2028,  7, 16 },  { 2028,  7, 17 },  { 2028,  7, 18 },
  { 2028,  7, 19 },  { 2028,  7, 20 },  { 2028,  7, 21 },  { 2028,  7, 22 },
  { 2028,  7, 23 },  { 2028,  7, 24 },  { 2028,  7, 25 },  { 2028,  7, 26 },
  { 2028,  7, 27 },  { 2028,  7, 28 },  { 2028,  7, 29 },  { 2028,  7, 30 },
  { 2028,  7, 31 },  { 2028,  8,  1 },  { 2028,  8,  2 },  { 2028,  8,  3 },
  { 2028,  8,  4 },  { 2028,  8,  5 },  { 2028,  8,  6 },  { 2028,  8,  7 },
  { 2028,  8,  8 },  { 2028,  8,  9 },  { 2028,  8, 10 },  { 2028,  8, 11 },
  { 2028,  8, 12 },  { 2028,  8, 13 },  { 2028,  8, 14 },  { 2028,  8, 15 },
  { 2028,  8, 16 },  { 2028,  8, 17 },  { 2028,  8, 18 },  { 2028,  8, 19 },
  { 2028,  8, 20 },  { 2028,  8, 21 },  { 2028,  8, 22 },  { 2028,  8, 23 },
  { 2028,  8, 24 },  { 2028,  8, 25 },  { 2028,  8, 26 },  { 2028,  8, 27 },
  { 2028,  8, 28 },  { 2028,  8, 29 },  { 2028,  8, 30 },  { 2028,  8, 31 },
  { 2028,  9,  1 },  { 2028,  9,  2 },  { 2028,  9,  3 },  { 2028,  9,  4 },
  { 2028,  9,  5 },  { 2028,  9,  6 },  { 2028,  9,  7 },  { 2028,  9,  8 },
  { 2028,  9,  9 },  { 2028,  9, 10 },  { 2028,  9, 11 },  { 2028,  9, 12 },
  { 2028,  9, 13 },  { 2028,  9, 14 },  { 2028,  9, 15 },  { 2028,  9, 16 },
  { 2028,  9, 17 },  { 2028,  9, 18 },  { 2028,  9, 19 },  { 2028,  9, 20 },
  { 2028,  9, 21 },  { 2028,  9, 22 },  { 2028,  9, 23 },  { 2028,  9, 24 },
  { 2028,  9, 25 },  { 2028,  9, 26 },  { 2028,  9, 27 },  { 2028,  9, 28 },
  { 2028,  9, 29 },  { 2028,  9, 30 },  { 2028, 10,  1 },  { 2028, 10,  2 },
  { 2028, 10,  3 },  { 2028, 10,  4 },  { 2028, 10,  5 },  { 2028, 10,  6 },
  { 2028, 10,  7 },  { 2028, 10,  8 },  { 2028, 10,  9 },  { 2028, 10, 10 },
  { 2028, 10, 11 },  { 2028, 10, 12 },  { 2028, 10, 13 },  { 2028, 10, 14 },
  { 2028, 10, 15 },  { 2028, 10, 16 },  { 2028, 10, 17 },  { 2028, 10, 18 },
  { 2028, 10, 19 },  { 2028, 10, 20 },  { 2028, 10, 21 },  { 2028, 10, 22 },
  { 2028, 10, 23 },  { 2028, 10, 24 },  { 2028, 10, 25 },  { 2028, 10, 26 },
  { 2028, 10, 27 },  { 2028, 10, 28 },  { 2028, 10, 29 },  { 2028, 10, 30 },
  { 2028, 10, 31 },  { 2028, 11,  1 },  { 2028, 11,  2 },  { 2028, 11,  3 },
  { 2028, 11,  4 },  { 2028, 11,  5 },  { 2028, 11,  6 },  { 2028, 11,  7 },
  { 2028, 11,  8 },  { 2028, 11,  9 },  { 2028, 11, 10 },  { 2028, 11, 11 },
  { 2028, 11, 12 },  { 2028, 11, 13 },  { 2028, 11, 14 },  { 2028, 11, 15 },
  { 2028, 11, 16 },  { 2028, 11, 17 },  { 2028, 11, 18 },  { 2028, 11, 19 },
  { 2028, 11, 20 },  { 2028, 11, 21 },  { 2028, 11, 22 },  { 2028, 11, 23 },
  { 2028, 11, 24 },  { 2028, 11, 25 },  { 2028, 11, 26 },  { 2028, 11, 27 },
  { 2028, 11, 28 },  { 2028, 11, 29 },  { 2028, 11, 30 },  { 2028, 12,  1 },
  { 2028, 12,  2 },  { 2028, 12,  3 },  { 2028, 12,  4 },  { 2028, 12,  5 },
  { 2028, 12,  6 },  { 2028, 12,  7 },  { 2028, 12,  8 },  { 2028, 12,  9 },
  { 2028, 12, 10 },  { 2028, 12, 11 },  { 2028, 12, 12 },  { 2028, 12, 13 },
  { 2028, 12, 14 },  { 2028, 12, 15 },  { 2028, 12, 16 },  { 2028, 12, 17 },
  { 2028, 12, 18 },  { 2028, 12, 19 },  { 2028, 12, 20 },  { 2028, 12, 21 },
  { 2028, 12, 22 },  { 2028, 12, 23 },  { 2028, 12, 24 },  { 2028, 12, 25 },
  { 2028, 12, 26 },  { 2028, 12, 27 },  { 2028, 12, 28 },  { 2028, 12, 29 },
  { 2028, 12, 30 },  { 2028, 12, 31 },  { 2029,  1,  1 },  { 2029,  1,  2 },
  { 2029,  1,  3 },  { 2029,  1,  4 },  { 2029,  1,  5 },  { 2029,  1,  6 },
  { 2029,  1,  7 },  { 2029,  1,  8 },  { 2029,  1,  9 },  { 2029,  1, 10 },
  { 2029,  1, 11 },  { 2029,  1, 12 },  { 2029,  1, 13 },  { 2029,  1, 14 },
  { 2029,  1, 15 },  { 2029,  1, 16 },  { 2029,  1, 17 },  { 2029,  1, 18 },
  { 2029,  1, 19 },  { 2029,  1, 20 },  { 2029,  1, 21 },  { 2029,  1, 22 },
  { 2029,  1, 23 },  { 2029,  1, 24 },  { 2029,  1, 25 },  { 2029,  1, 26 },
  { 2029,  1, 27 },  { 2029,  1, 28 },  { 2029,  1, 29 },  { 2029,  1, 30 },
  { 2029,  1, 31 },  { 2029,  2,  1 },  { 2029,  2,  2 },  { 2029,  2,  3 },
  { 2029,  2,  4 },  { 2029,  2,  5 },  { 2029,  2,  6 },  { 2029,  2,  7 },
  { 2029,  2,  8 },  { 2029,  2,  9 },  { 2029,  2, 10 },  { 2029,  2, 11 },
  { 2029,  2, 12 },  { 2029,  2, 13 },  { 2029,  2, 14 },  { 2029,  2, 15 },
  { 2029,  2, 16 },  { 2029,  2, 17 },  { 2029,  2, 18 },  { 2029,  2, 19 },
  { 2029,  2, 20 },  { 2029,  2, 21 },  { 2029,  2, 22 },  { 2029,  2, 23 },
  { 2029,  2, 24 },  { 2029,  2, 25 },  { 2029,  2, 26 },  { 2029,  2, 27 },
  { 2029,  2, 28 },  { 2029,  3,  1 },  { 2029,  3,  2 },  { 2029,  3,  3 },
  { 2029,  3,  4 },  { 2029,  3,  5 },  { 2029,  3,  6 },  { 2029,  3,  7 },
  { 2029,  3,  8 },  { 2029,  3,  9 },  { 2029,  3, 10 },  { 2029,  3, 11 },
  { 2029,  3, 12 },  { 2029,  3, 13 },  { 2029,  3, 14 },  { 2029,  3, 15 },
  { 2029,  3, 16 },  { 2029,  3, 17 },  { 2029,  3, 18 },  { 2029,  3, 19 },
  { 2029,  3, 20 },  { 2029,  3, 21 },  { 2029,  3, 22 },  { 2029,  3, 23 },
  { 2029,  3, 24 },  { 2029,  3, 25 },  { 2029,  3, 26 },  { 2029,  3, 27 },
  { 2029,  3, 28 },  { 2029,  3, 29 },  { 2029,  3, 30 },  { 2029,  3, 31 },
  { 2029,  4,  1 },  { 2029,  4,  2 },  { 2029,  4,  3 },  { 2029,  4,  4 },
  { 2029,  4,  5 },  { 2029,  4,  6 },  { 2029,  4,  7 },  { 2029,  4,  8 },
  { 2029,  4,  9 },  { 2029,  4, 10 },  { 2029,  4, 11 },  { 2029,  4, 12 },
  { 2029,  4, 13 },  { 2029,  4, 14 },  { 2029,  4, 15 },  { 2029,  4, 16 },
  { 2029,  4, 17 },  { 2029,  4, 18 },  { 2029,  4, 19 },  { 2029,  4, 20 },
  { 2029,  4, 21 },  { 2029,  4, 22 },  { 2029,  4, 23 },  { 2029,  4, 24 },
  { 2029,  4, 25 },  { 2029,  4, 26 },  { 2029,  4, 27 },  { 2029,  4, 28 },
  { 2029,  4, 29 },  { 2029,  4, 30 },  { 2029,  5,  1 },  { 2029,  5,  2 },
  { 2029,  5,  3 },  { 2029,  5,  4 },  { 2029,  5,  5 },  { 2029,  5,  6 },
  { 2029,  5,  7 },  { 2029,  5,  8 },  { 2029,  5,  9 },  { 2029,  5, 10 },
  { 2029,  5, 11 },  { 2029,  5, 12 },  { 2029,  5, 13 },  { 2029,  5, 14 },
  { 2029,  5, 15 },  { 2029,  5, 16 },  { 2029,  5, 17 },  { 2029,  5, 18 },
  { 2029,  5, 19 },  { 2029,  5, 20 },  { 2029,  5, 21 },  { 2029,  5, 22 },
  { 2029,  5, 23 },  { 2029,  5, 24 },  { 2029,  5, 25 },  { 2029,  5, 26 },
  { 2029,  5, 27 },  { 2029,  5, 28 },  { 2029,  5, 29 },  { 2029,  5, 30 },
  { 2029,  5, 31 },  { 2029,  6,  1 },  { 2029,  6,  2 },  { 2029,  6,  3 },
  { 2029,  6,  4 },  { 2029,  6,  5 },  { 2029,  6,  6 },  { 2029,  6,  7 },
  { 2029,  6,  8 },  { 2029,  6,  9 },  { 2029,  6, 10 },  { 2029,  6, 11 },
  { 2029,  6, 12 },  { 2029,  6, 13 },  { 2029,  6, 14 },  { 2029,  6, 15 },
  { 2029,  6, 16 },  { 2029,  6, 17 },  { 2029,  6, 18 },  { 2029,  6, 19 },
  { 2029,  6, 20 },  { 2029,  6, 21 },  { 2029,  6, 22 },  { 2029,  6, 23 },
  { 2029,  6, 24 },  { 2029,  6, 25 },  { 2029,  6, 26 },  { 2029,  6, 27 },
  { 2029,  6, 28 },  { 2029,  6, 29 },  { 2029,  6, 30 },  { 2029,  7,  1 },
  { 2029,  7,  2 },  { 2029,  7,  3 },  { 2029,  7,  4 },  { 2029,  7,  5 },
  { 2029,  7,  6 },  { 2029,  7,  7 },  { 2029,  7,  8 },  { 2029,  7,  9 },
  { 2029,  7, 10 },  { 2029,  7, 11 },  { 2029,  7, 12 },  { 2029,  7, 13 },
  { 2029,  7, 14 },  { 2029,  7, 15 },  { 2029,  7, 16 },  { 2029,  7, 17 },
  { 2029,  7, 18 },  { 2029,  7, 19 },  { 2029,  7, 20 },  { 2029,  7, 21 },
  { 2029,  7, 22 },  { 2029,  7, 23 },  { 2029,  7, 24 },  { 2029,  7, 25 },
  { 2029,  7, 26 },  { 2029,  7, 27 },  { 2029,  7, 28 },  { 2029,  7, 29 },
  { 2029,  7, 30 },  { 2029,  7, 31 },  { 2029,  8,  1 },  { 2029,  8,  2 },
  { 2029,  8,  3 },  { 2029,  8,  4 },  { 2029,  8,  5 },  { 2029,  8,  6 },
  { 2029,  8,  7 },  { 2029,  8,  8 },  { 2029,  8,  9 },  { 2029,  8, 10 },
  { 2029,  8, 11 },  { 2029,  8, 12 },  { 2029,  8, 13 },  { 2029,  8, 14 },
  { 2029,  8, 15 },  { 2029,  8, 16 },  { 2029,  8, 17 },  { 2029,  8, 18 },
  { 2029,  8, 19 },  { 2029,  8, 20 },  { 2029,  8, 21 },  { 2029,  8, 22 },
  { 2029,  8, 23 },  { 2029,  8, 24 },  { 2029,  8, 25 },  { 2029,  8, 26 },
  { 2029,  8, 27 },  { 2029,  8, 28 },  { 2029,  8, 29 },  { 2029,  8, 30 },
  { 2029,  8, 31 },  { 2029,  9,  1 },  { 2029,  9,  2 },  { 2029,  9,  3 },
  { 2029,  9,  4 },  { 2029,  9,  5 },  { 2029,  9,  6 },  { 2029,  9,  7 },
  { 2029,  9,  8 },  { 2029,  9,  9 },  { 2029,  9, 10 },  { 2029,  9, 11 },
  { 2029,  9, 12 },  { 2029,  9, 13 },  { 2029,  9, 14 },  { 2029,  9, 15 },
  { 2029,  9, 16 },  { 2029,  9, 17 },  { 2029,  9, 18 },  { 2029,  9, 19 },
  { 2029,  9, 20 },  { 2029,  9, 21 },  { 2029,  9, 22 },  { 2029,  9, 23 },
  { 2029,  9, 24 },  { 2029,  9, 25 },  { 2029,  9, 26 },  { 2029,  9, 27 },
  { 2029,  9, 28 },  { 2029,  9, 29 },  { 2029,  9, 30 },  { 2029, 10,  1 },
  { 2029, 10,  2 },  { 2029, 10,  3 },  { 2029, 10,  4 },  { 2029, 10,  5 },
  { 2029, 10,  6 },  { 2029, 10,  7 },  { 2029, 10,  8 },  { 2029, 10,  9 },
  { 2029, 10, 10 },  { 2029, 10, 11 },  { 2029, 10, 12 },  { 2029, 10, 13 },
  { 2029, 10, 14 },  { 2029, 10, 15 },  { 2029, 10, 16 },  { 2029, 10, 17 },
  { 2029, 10, 18 },  { 2029, 10, 19 },  { 2029, 10, 20 },  { 2029, 10, 21 },
  { 2029, 10, 22 },  { 2029, 10, 23 },  { 2029, 10, 24 },  { 2029, 10, 25 },
  { 2029, 10, 26 },  { 2029, 10, 27 },  { 2029, 10, 28 },  { 2029, 10, 29 },
  { 2029, 10, 30 },  { 2029, 10, 31 },  { 2029, 11,  1 },  { 2029, 11,  2 },
  { 2029, 11,  3 },  { 2029, 11,  4 },  { 2029, 11,  5 },  { 2029, 11,  6 },
  { 2029, 11,  7 },  { 2029, 11,  8 },  { 2029, 11,  9 },  { 2029, 11, 10 },
  { 2029, 11, 11 },  { 2029, 11, 12 },  { 2029, 11, 13 },  { 2029, 11, 14 },
  { 2029, 11, 15 },  { 2029, 11, 16 },  { 2029, 11, 17 },  { 2029, 11, 18 },
  { 2029, 11, 19 },  { 2029, 11, 20 },  { 2029, 11, 21 },  { 2029, 11, 22 },
  { 2029, 11, 23 },  { 2029, 11, 24 },  { 2029, 11, 25 },  { 2029, 11, 26 },
  { 2029, 11, 27 },  { 2029, 11, 28 },  { 2029, 11, 29 },  { 2029, 11, 30 },
  { 2029, 12,  1 },  { 2029, 12,  2 },  { 2029, 12,  3 },  { 2029, 12,  4 },
  { 2029, 12,  5 },  { 2029, 12,  6 },  { 2029, 12,  7 },  { 2029, 12,  8 },
  { 2029, 12,  9 },  { 2029, 12, 10 },  { 2029, 12, 11 },  { 2029, 12, 12 },
  { 2029, 12, 13 },  { 2029, 12, 14 },  { 2029, 12, 15 },  { 2029, 12, 16 },
  { 2029, 12, 17 },  { 2029, 12, 18 },  { 2029, 12, 19 },  { 2029, 12, 20 },
  { 2029, 12, 21 },  { 2029, 12, 22 },  { 2029, 12, 23 },  { 2029, 12, 24 },
  { 2029, 12, 25 },  { 2029, 12, 26 },  { 2029, 12, 27 },  { 2029, 12, 28 },
  { 2029, 12, 29 },  { 2029, 12, 30 },  { 2029, 12, 31 },  { 2030,  1,  1 },
  { 2030,  1,  2 },  { 2030,  1,  3 },  { 2030,  1,  4 },  { 2030,  1,  5 },
  { 2030,  1,  6 },  { 2030,  1,  7 },  { 2030,  1,  8 },  { 2030,  1,  9 },
  { 2030,  1, 10 },  { 2030,  1, 11 },  { 2030,  1, 12 },  { 2030,  1, 13 },
  { 2030,  1, 14 },  { 2030,  1, 15 },  { 2030,  1, 16 },  { 2030,  1, 17 },
  { 2030,  1, 18 },  { 2030,  1, 19 },  { 2030,  1, 20 },  { 2030,  1, 21 },
  { 2030,  1, 22 },  { 2030,  1, 23 },  { 2030,  1, 24 },  { 2030,  1, 25 },
  { 2030,  1, 26 },  { 2030,  1, 27 },  { 2030,  1, 28 },  { 2030,  1, 29 },
  { 2030,  1, 30 },  { 2030,  1, 31 },  { 2030,  2,  1 },  { 2030,  2,  2 },
  { 2030,  2,  3 },  { 2030,  2,  4 },  { 2030,  2,  5 },  { 2030,  2,  6 },
  { 2030,  2,  7 },  { 2030,  2,  8 },  { 2030,  2,  9 },  { 2030,  2, 10 },
  { 2030,  2, 11 },  { 2030,  2, 12 },  { 2030,  2, 13 },  { 2030,  2, 14 },
  { 2030,  2, 15 },  { 2030,  2, 16 },  { 2030,  2, 17 },  { 2030,  2, 18 },
  { 2030,  2, 19 },  { 2030,  2, 20 },  { 2030,  2, 21 },  { 2030,  2, 22 },
  { 2030,  2, 23 },  { 2030,  2, 24 },  { 2030,  2, 25 },  { 2030,  2, 26 },
  { 2030,  2, 27 },  { 2030,  2, 28 },  { 2030,  3,  1 },  { 2030,  3,  2 },
  { 2030,  3,  3 },  { 2030,  3,  4 },  { 2030,  3,  5 },  { 2030,  3,  6 },
  { 2030,  3,  7 },  { 2030,  3,  8 },  { 2030,  3,  9 },  { 2030,  3, 10 },
  { 2030,  3, 11 },  { 2030,  3, 12 },  { 2030,  3, 13 },  { 2030,  3, 14 },
  { 2030,  3, 15 },  { 2030,  3, 16 },  { 2030,  3, 17 },  { 2030,  3, 18 },
  { 2030,  3, 19 },  { 2030,  3, 20 },  { 2030,  3, 21 },  { 2030,  3, 22 },
  { 2030,  3, 23 },  { 2030,  3, 24 },  { 2030,  3, 25 },  { 2030,  3, 26 },
  { 2030,  3, 27 },  { 2030,  3, 28 },  { 2030,  3, 29 },  { 2030,  3, 30 },
  { 2030,  3, 31 },  { 2030,  4,  1 },  { 2030,  4,  2 },  { 2030,  4,  3 },
  { 2030,  4,  4 },  { 2030,  4,  5 },  { 2030,  4,  6 },  { 2030,  4,  7 },
  { 2030,  4,  8 },  { 2030,  4,  9 },  { 2030,  4, 10 },  { 2030,  4, 11 },
  { 2030,  4, 12 },  { 2030,  4, 13 },  { 2030,  4, 14 },  { 2030,  4, 15 },
  { 2030,  4, 16 },  { 2030,  4, 17 },  { 2030,  4, 18 },  { 2030,  4, 19 },
  { 2030,  4, 20 },  { 2030,  4, 21 },  { 2030,  4, 22 },  { 2030,  4, 23 },
  { 2030,  4, 24 },  { 2030,  4, 25 },  { 2030,  4, 26 },  { 2030,  4, 27 },
  { 2030,  4, 28 },  { 2030,  4, 29 },  { 2030,  4, 30 },  { 2030,  5,  1 },
  { 2030,  5,  2 },  { 2030,  5,  3 },  { 2030,  5,  4 },  { 2030,  5,  5 },
  { 2030,  5,  6 },  { 2030,  5,  7 },  { 2030,  5,  8 },  { 2030,  5,  9 },
  { 2030,  5, 10 },  { 2030,  5, 11 },  { 2030,  5, 12 },  { 2030,  5, 13 },
  { 2030,  5, 14 },  { 2030,  5, 15 },  { 2030,  5, 16 },  { 2030,  5, 17 },
  { 2030,  5, 18 },  { 2030,  5, 19 },  { 2030,  5, 20 },  { 2030,  5, 21 },
  { 2030,  5, 22 },  { 2030,  5, 23 },  { 2030,  5, 24 },  { 2030,  5, 25 },
  { 2030,  5, 26 },  { 2030,  5, 27 },  { 2030,  5, 28 },  { 2030,  5, 29 },
  { 2030,  5, 30 },  { 2030,  5, 31 },  { 2030,  6,  1 },  { 2030,  6,  2 },
  { 2030,  6,  3 },  { 2030,  6,  4 },  { 2030,  6,  5 },  { 2030,  6,  6 },
  { 2030,  6,  7 },  { 2030,  6,  8 },  { 2030,  6,  9 },  { 2030,  6, 10 },
  { 2030,  6, 11 },  { 2030,  6, 12 },  { 2030,  6, 13 },  { 2030,  6, 14 },
  { 2030,  6, 15 },  { 2030,  6, 16 },  { 2030,  6, 17 },  { 2030,  6, 18 },
  { 2030,  6, 19 },  { 2030,  6, 20 },  { 2030,  6, 21 },  { 2030,  6, 22 },
  { 2030,  6, 23 },  { 2030,  6, 24 },  { 2030,  6, 25 },  { 2030,  6, 26 },
  { 2030,  6, 27 },  { 2030,  6, 28 },  { 2030,  6, 29 },  { 2030,  6, 30 },
  { 2030,  7,  1 },  { 2030,  7,  2 },  { 2030,  7,  3 },  { 2030,  7,  4 },
  { 2030,  7,  5 },  { 2030,  7,  6 },  { 2030,  7,  7 },  { 2030,  7,  8 },
  { 2030,  7,  9 },  { 2030,  7, 10 },  { 2030,  7, 11 },  { 2030,  7, 12 },
  { 2030,  7, 13 },  { 2030,  7, 14 },  { 2030,  7, 15 },  { 2030,  7, 16 },
  { 2030,  7, 17 },  { 2030,  7, 18 },  { 2030,  7, 19 },  { 2030,  7, 20 },
  { 2030,  7, 21 },  { 2030,  7, 22 },  { 2030,  7, 23 },  { 2030,  7, 24 },
  { 2030,  7, 25 },  { 2030,  7, 26 },  { 2030,  7, 27 },  { 2030,  7, 28 },
  { 2030,  7, 29 },  { 2030,  7, 30 },  { 2030,  7, 31 },  { 2030,  8,  1 },
  { 2030,  8,  2 },  { 2030,  8,  3 },  { 2030,  8,  4 },  { 2030,  8,  5 },
  { 2030,  8,  6 },  { 2030,  8,  7 },  { 2030,  8,  8 },  { 2030,  8,  9 },
  { 2030,  8, 10 },  { 2030,  8, 11 },  { 2030,  8, 12 },  { 2030,  8, 13 },
  { 2030,  8, 14 },  { 2030,  8, 15 },  { 2030,  8, 16 },  { 2030,  8, 17 },
  { 2030,  8, 18 },  { 2030,  8, 19 },  { 2030,  8, 20 },  { 2030,  8, 21 },
  { 2030,  8, 22 },  { 2030,  8, 23 },  { 2030,  8, 24 },  { 2030,  8, 25 },
  { 2030,  8, 26 },  { 2030,  8, 27 },  { 2030,  8, 28 },  { 2030,  8, 29 },
  { 2030,  8, 30 },  { 2030,  8, 31 },  { 2030,  9,  1 },  { 2030,  9,  2 },
  { 2030,  9,  3 },  { 2030,  9,  4 },  { 2030,  9,  5 },  { 2030,  9,  6 },
  { 2030,  9,  7 },  { 2030,  9,  8 },  { 2030,  9,  9 },  { 2030,  9, 10 },
  { 2030,  9, 11 },  { 2030,  9, 12 },  { 2030,  9, 13 },  { 2030,  9, 14 },
  { 2030,  9, 15 },  { 2030,  9, 16 },  { 2030,  9, 17 },  { 2030,  9, 18 },
  { 2030,  9, 19 },  { 2030,  9, 20 },  { 2030,  9, 21 },  { 2030,  9, 22 },
  { 2030,  9, 23 },  { 2030,  9, 24 },  { 2030,  9, 25 },  { 2030,  9, 26 },
  { 2030,  9, 27 },  { 2030,  9, 28 },  { 2030,  9, 29 },  { 2030,  9, 30 },
  { 2030, 10,  1 },  { 2030, 10,  2 },  { 2030, 10,  3 },  { 2030, 10,  4 },
  { 2030, 10,  5 },  { 2030, 10,  6 },  { 2030, 10,  7 },  { 2030, 10,  8 },
  { 2030, 10,  9 },  { 2030, 10, 10 },  { 2030, 10, 11 },  { 2030, 10, 12 },
  { 2030, 10, 13 },  { 2030, 10, 14 },  { 2030, 10, 15 },  { 2030, 10, 16 },
  { 2030, 10, 17 },  { 2030, 10, 18 },  { 2030, 10, 19 },  { 2030, 10, 20 },
  { 2030, 10, 21 },  { 2030, 10, 22 },  { 2030, 10, 23 },  { 2030, 10, 24 },
  { 2030, 10, 25 },  { 2030, 10, 26 },  { 2030, 10, 27 },  { 2030, 10, 28 },
  { 2030, 10, 29 },  { 2030, 10, 30 },  { 2030, 10, 31 },  { 2030, 11,  1 },
  { 2030, 11,  2 },  { 2030, 11,  3 },  { 2030, 11,  4 },  { 2030, 11,  5 },
  { 2030, 11,  6 },  { 2030, 11,  7 },  { 2030, 11,  8 },  { 2030, 11,  9 },
  { 2030, 11, 10 },  { 2030, 11, 11 },  { 2030, 11, 12 },  { 2030, 11, 13 },
  { 2030, 11, 14 },  { 2030, 11, 15 },  { 2030, 11, 16 },  { 2030, 11, 17 },
  { 2030, 11, 18 },  { 2030, 11, 19 },  { 2030, 11, 20 },  { 2030, 11, 21 },
  { 2030, 11, 22 },  { 2030, 11, 23 },  { 2030, 11, 24 },  { 2030, 11, 25 },
  { 2030, 11, 26 },  { 2030, 11, 27 },  { 2030, 11, 28 },  { 2030, 11, 29 },
  { 2030, 11, 30 },  { 2030, 12,  1 },  { 2030, 12,  2 },  { 2030, 12,  3 },
  { 2030, 12,  4 },  { 2030, 12,  5 },  { 2030, 12,  6 },  { 2030, 12,  7 },
  { 2030, 12,  8 },  { 2030, 12,  9 },  { 2030, 12, 10 },  { 2030, 12, 11 },
  { 2030, 12, 12 },  { 2030, 12, 13 },  { 2030, 12, 14 },  { 2030, 12, 15 },
  { 2030, 12, 16 },  { 2030, 12, 17 },  { 2030, 12, 18 },  { 2030, 12, 19 },
  { 2030, 12, 20 },  { 2030, 12, 21 },  { 2030, 12, 22 },  { 2030, 12, 23 },
  { 2030, 12, 24 },  { 2030, 12, 25 },  { 2030, 12, 26 },  { 2030, 12, 27 },
  { 2030, 12, 28 },  { 2030, 12, 29 },  { 2030, 12, 30 },  { 2030, 12, 31 },
  { 2031,  1,  1 },  { 2031,  1,  2 },  { 2031,  1,  3 },  { 2031,  1,  4 },
  { 2031,  1,  5 },  { 2031,  1,  6 },  { 2031,  1,  7 },  { 2031,  1,  8 },
  { 2031,  1,  9 },  { 2031,  1, 10 },  { 2031,  1, 11 },  { 2031,  1, 12 },
  { 2031,  1, 13 },  { 2031,  1, 14 },  { 2031,  1, 15 },  { 2031,  1, 16 },
  { 2031,  1, 17 },  { 2031,  1, 18 },  { 2031,  1, 19 },  { 2031,  1, 20 },
  { 2031,  1, 21 },  { 2031,  1, 22 },  { 2031,  1, 23 },  { 2031,  1, 24 },
  { 2031,  1, 25 },  { 2031,  1, 26 },  { 2031,  1, 27 },  { 2031,  1, 28 },
  { 2031,  1, 29 },  { 2031,  1, 30 },  { 2031,  1, 31 },  { 2031,  2,  1 },
  { 2031,  2,  2 },  { 2031,  2,  3 },  { 2031,  2,  4 },  { 2031,  2,  5 },
  { 2031,  2,  6 },  { 2031,  2,  7 },  { 2031,  2,  8 },  { 2031,  2,  9 },
  { 2031,  2, 10 },  { 2031,  2, 11 },  { 2031,  2, 12 },  { 2031,  2, 13 },
  { 2031,  2, 14 },  { 2031,  2, 15 },  { 2031,  2, 16 },  { 2031,  2, 17 },
  { 2031,  2, 18 },  { 2031,  2, 19 },  { 2031,  2, 20 },  { 2031,  2, 21 },
  { 2031,  2, 22 },  { 2031,  2, 23 },  { 2031,  2, 24 },  { 2031,  2, 25 },
  { 2031,  2, 26 },  { 2031,  2, 27 },  { 2031,  2, 28 },  { 2031,  3,  1 },
  { 2031,  3,  2 },  { 2031,  3,  3 },  { 2031,  3,  4 },  { 2031,  3,  5 },
  { 2031,  3,  6 },  { 2031,  3,  7 },  { 2031,  3,  8 },  { 2031,  3,  9 },
  { 2031,  3, 10 },  { 2031,  3, 11 },  { 2031,  3, 12 },  { 2031,  3, 13 },
  { 2031,  3, 14 },  { 2031,  3, 15 },  { 2031,  3, 16 },  { 2031,  3, 17 },
  { 2031,  3, 18 },  { 2031,  3, 19 },  { 2031,  3, 20 },  { 2031,  3, 21 },
  { 2031,  3, 22 },  { 2031,  3, 23 },  { 2031,  3, 24 },  { 2031,  3, 25 },
  { 2031,  3, 26 },  { 2031,  3, 27 },  { 2031,  3, 28 },  { 2031,  3, 29 },
  { 2031,  3, 30 },  { 2031,  3, 31 },  { 2031,  4,  1 },  { 2031,  4,  2 },
  { 2031,  4,  3 },  { 2031,  4,  4 },  { 2031,  4,  5 },  { 2031,  4,  6 },
  { 2031,  4,  7 },  { 2031,  4,  8 },  { 2031,  4,  9 },  { 2031,  4, 10 },
  { 2031,  4, 11 },  { 2031,  4, 12 },  { 2031,  4, 13 },  { 2031,  4, 14 },
  { 2031,  4, 15 },  { 2031,  4, 16 },  { 2031,  4, 17 },  { 2031,  4, 18 },
  { 2031,  4, 19 },  { 2031,  4, 20 },  { 2031,  4, 21 },  { 2031,  4, 22 },
  { 2031,  4, 23 },  { 2031,  4, 24 },  { 2031,  4, 25 },  { 2031,  4, 26 },
  { 2031,  4, 27 },  { 2031,  4, 28 },  { 2031,  4, 29 },  { 2031,  4, 30 },
  { 2031,  5,  1 },  { 2031,  5,  2 },  { 2031,  5,  3 },  { 2031,  5,  4 },
  { 2031,  5,  5 },  { 2031,  5,  6 },  { 2031,  5,  7 },  { 2031,  5,  8 },
  { 2031,  5,  9 },  { 2031,  5, 10 },  { 2031,  5, 11 },  { 2031,  5, 12 },
  { 2031,  5, 13 },  { 2031,  5, 14 },  { 2031,  5, 15 },  { 2031,  5, 16 },
  { 2031,  5, 17 },  { 2031,  5, 18 },  { 2031,  5, 19 },  { 2031,  5, 20 },
  { 2031,  5, 21 },  { 2031,  5, 22 },  { 2031,  5, 23 },  { 2031,  5, 24 },
  { 2031,  5, 25 },  { 2031,  5, 26 },  { 2031,  5, 27 },  { 2031,  5, 28 },
  { 2031,  5, 29 },  { 2031,  5, 30 },  { 2031,  5, 31 },  { 2031,  6,  1 },
  { 2031,  6,  2 },  { 2031,  6,  3 },  { 2031,  6,  4 },  { 2031,  6,  5 },
  { 2031,  6,  6 },  { 2031,  6,  7 },  { 2031,  6,  8 },  { 2031,  6,  9 },
  { 2031,  6, 10 },  { 2031,  6, 11 },  { 2031,  6, 12 },  { 2031,  6, 13 },
  { 2031,  6, 14 },  { 2031,  6, 15 },  { 2031,  6, 16 },  { 2031,  6, 17 },
  { 2031,  6, 18 },  { 2031,  6, 19 },  { 2031,  6, 20 },  { 2031,  6, 21 },
  { 2031,  6, 22 },  { 2031,  6, 23 },  { 2031,  6, 24 },  { 2031,  6, 25 },
  { 2031,  6, 26 },  { 2031,  6, 27 },  { 2031,  6, 28 },  { 2031,  6, 29 },
  { 2031,  6, 30 },  { 2031,  7,  1 },  { 2031,  7,  2 },  { 2031,  7,  3 },
  { 2031,  7,  4 },  { 2031,  7,  5 },  { 2031,  7,  6 },  { 2031,  7,  7 },
  { 2031,  7,  8 },  { 2031,  7,  9 },  { 2031,  7, 10 },  { 2031,  7, 11 },
  { 2031,  7, 12 },  { 2031,  7, 13 },  { 2031,  7, 14 },  { 2031,  7, 15 },
  { 2031,  7, 16 },  { 2031,  7, 17 },  { 2031,  7, 18 },  { 2031,  7, 19 },
  { 2031,  7, 20 },  { 2031,  7, 21 },  { 2031,  7, 22 },  { 2031,  7, 23 },
  { 2031,  7, 24 },  { 2031,  7, 25 },  { 2031,  7, 26 },  { 2031,  7, 27 },
  { 2031,  7, 28 },  { 2031,  7, 29 },  { 2031,  7, 30 },  { 2031,  7, 31 },
  { 2031,  8,  1 },  { 2031,  8,  2 },  { 2031,  8,  3 },  { 2031,  8,  4 },
  { 2031,  8,  5 },  { 2031,  8,  6 },  { 2031,  8,  7 },  { 2031,  8,  8 },
  { 2031,  8,  9 },  { 2031,  8, 10 },  { 2031,  8, 11 },  { 2031,  8, 12 },
  { 2031,  8, 13 },  { 2031,  8, 14 },  { 2031,  8, 15 },  { 2031,  8, 16 },
  { 2031,  8, 17 },  { 2031,  8, 18 },  { 2031,  8, 19 },  { 2031,  8, 20 },
  { 2031,  8, 21 },  { 2031,  8, 22 },  { 2031,  8, 23 },  { 2031,  8, 24 },
  { 2031,  8, 25 },  { 2031,  8, 26 },  { 2031,  8, 27 },  { 2031,  8, 28 },
  { 2031,  8, 29 },  { 2031,  8, 30 },  { 2031,  8, 31 },  { 2031,  9,  1 },
  { 2031,  9,  2 },  { 2031,  9,  3 },  { 2031,  9,  4 },  { 2031,  9,  5 },
  { 2031,  9,  6 },  { 2031,  9,  7 },  { 2031,  9,  8 },  { 2031,  9,  9 },
  { 2031,  9, 10 },  { 2031,  9, 11 },  { 2031,  9, 12 },  { 2031,  9, 13 },
  { 2031,  9, 14 },  { 2031,  9, 15 },  { 2031,  9, 16 },  { 2031,  9, 17 },
  { 2031,  9, 18 },  { 2031,  9, 19 },  { 2031,  9, 20 },  { 2031,  9, 21 },
  { 2031,  9, 22 },  { 2031,  9, 23 },  { 2031,  9, 24 },  { 2031,  9, 25 },
  { 2031,  9, 26 },  { 2031,  9, 27 },  { 2031,  9, 28 },  { 2031,  9, 29 },
  { 2031,  9, 30 },  { 2031, 10,  1 },  { 2031, 10,  2 },  { 2031, 10,  3 },
  { 2031, 10,  4 },  { 2031, 10,  5 },  { 2031, 10,  6 },  { 2031, 10,  7 },
  { 2031, 10,  8 },  { 2031, 10,  9 },  { 2031, 10, 10 },  { 2031, 10, 11 },
  { 2031, 10, 12 },  { 2031, 10, 13 },  { 2031, 10, 14 },  { 2031, 10, 15 },
  { 2031, 10, 16 },  { 2031, 10, 17 },  { 2031, 10, 18 },  { 2031, 10, 19 },
  { 2031, 10, 20 },  { 2031, 10, 21 },  { 2031, 10, 22 },  { 2031, 10, 23 },
  { 2031, 10, 24 },  { 2031, 10, 25 },  { 2031, 10, 26 },  { 2031, 10, 27 },
  { 2031, 10, 28 },  { 2031, 10, 29 },  { 2031, 10, 30 },  { 2031, 10, 31 },
  { 2031, 11,  1 },  { 2031, 11,  2 },  { 2031, 11,  3 },  { 2031, 11,  4 },
  { 2031, 11,  5 },  { 2031, 11,  6 },  { 2031, 11,  7 },  { 2031, 11,  8 },
  { 2031, 11,  9 },  { 2031, 11, 10 },  { 2031, 11, 11 },  { 2031, 11, 12 },
  { 2031, 11, 13 },  { 2031, 11, 14 },  { 2031, 11, 15 },  { 2031, 11, 16 },
  { 2031, 11, 17 },  { 2031, 11, 18 },  { 2031, 11, 19 },  { 2031, 11, 20 },
  { 2031, 11, 21 },  { 2031, 11, 22 },  { 2031, 11, 23 },  { 2031, 11, 24 },
  { 2031, 11, 25 },  { 2031, 11, 26 },  { 2031, 11, 27 },  { 2031, 11, 28 },
  { 2031, 11, 29 },  { 2031, 11, 30 },  { 2031, 12,  1 },  { 2031, 12,  2 },
  { 2031, 12,  3 },  { 2031, 12,  4 },  { 2031, 12,  5 },  { 2031, 12,  6 },
  { 2031, 12,  7 },  { 2031, 12,  8 },  { 2031, 12,  9 },  { 2031, 12, 10 },
  { 2031, 12, 11 },  { 2031, 12, 12 },  { 2031, 12, 13 },  { 2031, 12, 14 },
  { 2031, 12, 15 },  { 2031, 12, 16 },  { 2031, 12, 17 },  { 2031, 12, 18 },
  { 2031, 12, 19 },  { 2031, 12, 20 },  { 2031, 12, 21 },  { 2031, 12, 22 },
  { 2031, 12, 23 },  { 2031, 12, 24 },  { 2031, 12, 25 },  { 2031, 12, 26 },
  { 2031, 12, 27 },  { 2031, 12, 28 },  { 2031, 12, 29 },  { 2031, 12, 30 },
  { 2031, 12, 31 },  { 2032,  1,  1 },  { 2032,  1,  2 },  { 2032,  1,  3 },
  { 2032,  1,  4 },  { 2032,  1,  5 },  { 2032,  1,  6 },  { 2032,  1,  7 },
  { 2032,  1,  8 },  { 2032,  1,  9 },  { 2032,  1, 10 },  { 2032,  1, 11 },
  { 2032,  1, 12 },  { 2032,  1, 13 },  { 2032,  1, 14 },  { 2032,  1, 15 },
  { 2032,  1, 16 },  { 2032,  1, 17 },  { 2032,  1, 18 },  { 2032,  1, 19 },
  { 2032,  1, 20 },  { 2032,  1, 21 },  { 2032,  1, 22 },  { 2032,  1, 23 },
  { 2032,  1, 24 },  { 2032,  1, 25 },  { 2032,  1, 26 },  { 2032,  1, 27 },
  { 2032,  1, 28 },  { 2032,  1, 29 },  { 2032,  1, 30 },  { 2032,  1, 31 },
  { 2032,  2,  1 },  { 2032,  2,  2 },  { 2032,  2,  3 },  { 2032,  2,  4 },
  { 2032,  2,  5 },  { 2032,  2,  6 },  { 2032,  2,  7 },  { 2032,  2,  8 },
  { 2032,  2,  9 },  { 2032,  2, 10 },  { 2032,  2, 11 },  { 2032,  2, 12 },
  { 2032,  2, 13 },  { 2032,  2, 14 },  { 2032,  2, 15 },  { 2032,  2, 16 },
  { 2032,  2, 17 },  { 2032,  2, 18 },  { 2032,  2, 19 },  { 2032,  2, 20 },
  { 2032,  2, 21 },  { 2032,  2, 22 },  { 2032,  2, 23 },  { 2032,  2, 24 },
  { 2032,  2, 25 },  { 2032,  2, 26 },  { 2032,  2, 27 },  { 2032,  2, 28 },
  { 2032,  2, 29 },  { 2032,  3,  1 },  { 2032,  3,  2 },  { 2032,  3,  3 },
  { 2032,  3,  4 },  { 2032,  3,  5 },  { 2032,  3,  6 },  { 2032,  3,  7 },
  { 2032,  3,  8 },  { 2032,  3,  9 },  { 2032,  3, 10 },  { 2032,  3, 11 },
  { 2032,  3, 12 },  { 2032,  3, 13 },  { 2032,  3, 14 },  { 2032,  3, 15 },
  { 2032,  3, 16 },  { 2032,  3, 17 },  { 2032,  3, 18 },  { 2032,  3, 19 },
  { 2032,  3, 20 },  { 2032,  3, 21 },  { 2032,  3, 22 },  { 2032,  3, 23 },
  { 2032,  3, 24 },  { 2032,  3, 25 },  { 2032,  3, 26 },  { 2032,  3, 27 },
  { 2032,  3, 28 },  { 2032,  3, 29 },  { 2032,  3, 30 },  { 2032,  3, 31 },
  { 2032,  4,  1 },  { 2032,  4,  2 },  { 2032,  4,  3 },  { 2032,  4,  4 },
  { 2032,  4,  5 },  { 2032,  4,  6 },  { 2032,  4,  7 },  { 2032,  4,  8 },
  { 2032,  4,  9 },  { 2032,  4, 10 },  { 2032,  4, 11 },  { 2032,  4, 12 },
  { 2032,  4, 13 },  { 2032,  4, 14 },  { 2032,  4, 15 },  { 2032,  4, 16 },
  { 2032,  4, 17 },  { 2032,  4, 18 },  { 2032,  4, 19 },  { 2032,  4, 20 },
  { 2032,  4, 21 },  { 2032,  4, 22 },  { 2032,  4, 23 },  { 2032,  4, 24 },
  { 2032,  4, 25 },  { 2032,  4, 26 },  { 2032,  4, 27 },  { 2032,  4, 28 },
  { 2032,  4, 29 },  { 2032,  4, 30 },  { 2032,  5,  1 },  { 2032,  5,  2 },
  { 2032,  5,  3 },  { 2032,  5,  4 },  { 2032,  5,  5 },  { 2032,  5,  6 },
  { 2032,  5,  7 },  { 2032,  5,  8 },  { 2032,  5,  9 },  { 2032,  5, 10 },
  { 2032,  5, 11 },  { 2032,  5, 12 },  { 2032,  5, 13 },  { 2032,  5, 14 },
  { 2032,  5, 15 },  { 2032,  5, 16 },  { 2032,  5, 17 },  { 2032,  5, 18 },
  { 2032,  5, 19 },  { 2032,  5, 20 },  { 2032,  5, 21 },  { 2032,  5, 22 },
  { 2032,  5, 23 },  { 2032,  5, 24 },  { 2032,  5, 25 },  { 2032,  5, 26 },
  { 2032,  5, 27 },  { 2032,  5, 28 },  { 2032,  5, 29 },  { 2032,  5, 30 },
  { 2032,  5, 31 },  { 2032,  6,  1 },  { 2032,  6,  2 },  { 2032,  6,  3 },
  { 2032,  6,  4 },  { 2032,  6,  5 },  { 2032,  6,  6 },  { 2032,  6,  7 },
  { 2032,  6,  8 },  { 2032,  6,  9 },  { 2032,  6, 10 },  { 2032,  6, 11 },
  { 2032,  6, 12 },  { 2032,  6, 13 },  { 2032,  6, 14 },  { 2032,  6, 15 },
  { 2032,  6, 16 },  { 2032,  6, 17 },  { 2032,  6, 18 },  { 2032,  6, 19 },
  { 2032,  6, 20 },  { 2032,  6, 21 },  { 2032,  6, 22 },  { 2032,  6, 23 },
  { 2032,  6, 24 },  { 2032,  6, 25 },  { 2032,  6, 26 },  { 2032,  6, 27 },
  { 2032,  6, 28 },  { 2032,  6, 29 },  { 2032,  6, 30 },  { 2032,  7,  1 },
  { 2032,  7,  2 },  { 2032,  7,  3 },  { 2032,  7,  4 },  { 2032,  7,  5 },
  { 2032,  7,  6 },  { 2032,  7,  7 },  { 2032,  7,  8 },  { 2032,  7,  9 },
  { 2032,  7, 10 },  { 2032,  7, 11 },  { 2032,  7, 12 },  { 2032,  7, 13 },
  { 2032,  7, 14 },  { 2032,  7, 15 },  { 2032,  7, 16 },  { 2032,  7, 17 },
  { 2032,  7, 18 },  { 2032,  7, 19 },  { 2032,  7, 20 },  { 2032,  7, 21 },
  { 2032,  7, 22 },  { 2032,  7, 23 },  { 2032,  7, 24 },  { 2032,  7, 25 },
  { 2032,  7, 26 },  { 2032,  7, 27 },  { 2032,  7, 28 },  { 2032,  7, 29 },
  { 2032,  7, 30 },  { 2032,  7, 31 },  { 2032,  8,  1 },  { 2032,  8,  2 },
  { 2032,  8,  3 },  { 2032,  8,  4 },  { 2032,  8,  5 },  { 2032,  8,  6 },
  { 2032,  8,  7 },  { 2032,  8,  8 },  { 2032,  8,  9 },  { 2032,  8, 10 },
  { 2032,  8, 11 },  { 2032,  8, 12 },  { 2032,  8, 13 },  { 2032,  8, 14 },
  { 2032,  8, 15 },  { 2032,  8, 16 },  { 2032,  8, 17 },  { 2032,  8, 18 },
  { 2032,  8, 19 },  { 2032,  8, 20 },  { 2032,  8, 21 },  { 2032,  8, 22 },
  { 2032,  8, 23 },  { 2032,  8, 24 },  { 2032,  8, 25 },  { 2032,  8, 26 },
  { 2032,  8, 27 },  { 2032,  8, 28 },  { 2032,  8, 29 },  { 2032,  8, 30 },
  { 2032,  8, 31 },  { 2032,  9,  1 },  { 2032,  9,  2 },  { 2032,  9,  3 },
  { 2032,  9,  4 },  { 2032,  9,  5 },  { 2032,  9,  6 },  { 2032,  9,  7 },
  { 2032,  9,  8 },  { 2032,  9,  9 },  { 2032,  9, 10 },  { 2032,  9, 11 },
  { 2032,  9, 12 },  { 2032,  9, 13 },  { 2032,  9, 14 },  { 2032,  9, 15 },
  { 2032,  9, 16 },  { 2032,  9, 17 },  { 2032,  9, 18 },  { 2032,  9, 19 },
  { 2032,  9, 20 },  { 2032,  9, 21 },  { 2032,  9, 22 },  { 2032,  9, 23 },
  { 2032,  9, 24 },  { 2032,  9, 25 },  { 2032,  9, 26 },  { 2032,  9, 27 },
  { 2032,  9, 28 },  { 2032,  9, 29 },  { 2032,  9, 30 },  { 2032, 10,  1 },
  { 2032, 10,  2 },  { 2032, 10,  3 },  { 2032, 10,  4 },  { 2032, 10,  5 },
  { 2032, 10,  6 },  { 2032, 10,  7 },  { 2032, 10,  8 },  { 2032, 10,  9 },
  { 2032, 10, 10 },  { 2032, 10, 11 },  { 2032, 10, 12 },  { 2032, 10, 13 },
  { 2032, 10, 14 },  { 2032, 10, 15 },  { 2032, 10, 16 },  { 2032, 10, 17 },
  { 2032, 10, 18 },  { 2032, 10, 19 },  { 2032, 10, 20 },  { 2032, 10, 21 },
  { 2032, 10, 22 },  { 2032, 10, 23 },  { 2032, 10, 24 },  { 2032, 10, 25 },
  { 2032, 10, 26 },  { 2032, 10, 27 },  { 2032, 10, 28 },  { 2032, 10, 29 },
  { 2032, 10, 30 },  { 2032, 10, 31 },  { 2032, 11,  1 },  { 2032, 11,  2 },
  { 2032, 11,  3 },  { 2032, 11,  4 },  { 2032, 11,  5 },  { 2032, 11,  6 },
  { 2032, 11,  7 },  { 2032, 11,  8 },  { 2032, 11,  9 },  { 2032, 11, 10 },
  { 2032, 11, 11 },  { 2032, 11, 12 },  { 2032, 11, 13 },  { 2032, 11, 14 },
  { 2032, 11, 15 },  { 2032, 11, 16 },  { 2032, 11, 17 },  { 2032, 11, 18 },
  { 2032, 11, 19 },  { 2032, 11, 20 },  { 2032, 11, 21 },  { 2032, 11, 22 },
  { 2032, 11, 23 },  { 2032, 11, 24 },  { 2032, 11, 25 },  { 2032, 11, 26 },
  { 2032, 11, 27 },  { 2032, 11, 28 },  { 2032, 11, 29 },  { 2032, 11, 30 },
  { 2032, 12,  1 },  { 2032, 12,  2 },  { 2032, 12,  3 },  { 2032, 12,  4 },
  { 2032, 12,  5 },  { 2032, 12,  6 },  { 2032, 12,  7 },  { 2032, 12,  8 },
  { 2032, 12,  9 },  { 2032, 12, 10 },  { 2032, 12, 11 },  { 2032, 12, 12 },
  { 2032, 12, 13 },  { 2032, 12, 14 },  { 2032, 12, 15 },  { 2032, 12, 16 },
  { 2032, 12, 17 },  { 2032, 12, 18 },  { 2032, 12, 19 },  { 2032, 12, 20 },
  { 2032, 12, 21 },  { 2032, 12, 22 },  { 2032, 12, 23 },  { 2032, 12, 24 },
  { 2032, 12, 25 },  { 2032, 12, 26 },  { 2032, 12, 27 },  { 2032, 12, 28 },
  { 2032, 12, 29 },  { 2032, 12, 30 },  { 2032, 12, 31 },  { 2033,  1,  1 },
  { 2033,  1,  2 },  { 2033,  1,  3 },  { 2033,  1,  4 },  { 2033,  1,  5 },
  { 2033,  1,  6 },  { 2033,  1,  7 },  { 2033,  1,  8 },  { 2033,  1,  9 },
  { 2033,  1, 10 },  { 2033,  1, 11 },  { 2033,  1, 12 },  { 2033,  1, 13 },
  { 2033,  1, 14 },  { 2033,  1, 15 },  { 2033,  1, 16 },  { 2033,  1, 17 },
  { 2033,  1, 18 },  { 2033,  1, 19 },  { 2033,  1, 20 },  { 2033,  1, 21 },
  { 2033,  1, 22 },  { 2033,  1, 23 },  { 2033,  1, 24 },  { 2033,  1, 25 },
  { 2033,  1, 26 },  { 2033,  1, 27 },  { 2033,  1, 28 },  { 2033,  1, 29 },
  { 2033,  1, 30 },  { 2033,  1, 31 },  { 2033,  2,  1 },  { 2033,  2,  2 },
  { 2033,  2,  3 },  { 2033,  2,  4 },  { 2033,  2,  5 },  { 2033,  2,  6 },
  { 2033,  2,  7 },  { 2033,  2,  8 },  { 2033,  2,  9 },  { 2033,  2, 10 },
  { 2033,  2, 11 },  { 2033,  2, 12 },  { 2033,  2, 13 },  { 2033,  2, 14 },
  { 2033,  2, 15 },  { 2033,  2, 16 },  { 2033,  2, 17 },  { 2033,  2, 18 },
  { 2033,  2, 19 },  { 2033,  2, 20 },  { 2033,  2, 21 },  { 2033,  2, 22 },
  { 2033,  2, 23 },  { 2033,  2, 24 },  { 2033,  2, 25 },  { 2033,  2, 26 },
  { 2033,  2, 27 },  { 2033,  2, 28 },  { 2033,  3,  1 },  { 2033,  3,  2 },
  { 2033,  3,  3 },  { 2033,  3,  4 },  { 2033,  3,  5 },  { 2033,  3,  6 },
  { 2033,  3,  7 },  { 2033,  3,  8 },  { 2033,  3,  9 },  { 2033,  3, 10 },
  { 2033,  3, 11 },  { 2033,  3, 12 },  { 2033,  3, 13 },  { 2033,  3, 14 },
  { 2033,  3, 15 },  { 2033,  3, 16 },  { 2033,  3, 17 },  { 2033,  3, 18 },
  { 2033,  3, 19 },  { 2033,  3, 20 },  { 2033,  3, 21 },  { 2033,  3, 22 },
  { 2033,  3, 23 },  { 2033,  3, 24 },  { 2033,  3, 25 },  { 2033,  3, 26 },
  { 2033,  3, 27 },  { 2033,  3, 28 },  { 2033,  3, 29 },  { 2033,  3, 30 },
  { 2033,  3, 31 },  { 2033,  4,  1 },  { 2033,  4,  2 },  { 2033,  4,  3 },
  { 2033,  4,  4 },  { 2033,  4,  5 },  { 2033,  4,  6 },  { 2033,  4,  7 },
  { 2033,  4,  8 },  { 2033,  4,  9 },  { 2033,  4, 10 },  { 2033,  4, 11 },
  { 2033,  4, 12 },  { 2033,  4, 13 },  { 2033,  4, 14 },  { 2033,  4, 15 },
  { 2033,  4, 16 },  { 2033,  4, 17 },  { 2033,  4, 18 },  { 2033,  4, 19 },
  { 2033,  4, 20 },  { 2033,  4, 21 },  { 2033,  4, 22 },  { 2033,  4, 23 },
  { 2033,  4, 24 },  { 2033,  4, 25 },  { 2033,  4, 26 },  { 2033,  4, 27 },
  { 2033,  4, 28 },  { 2033,  4, 29 },  { 2033,  4, 30 },  { 2033,  5,  1 },
  { 2033,  5,  2 },  { 2033,  5,  3 },  { 2033,  5,  4 },  { 2033,  5,  5 },
  { 2033,  5,  6 },  { 2033,  5,  7 },  { 2033,  5,  8 },  { 2033,  5,  9 },
  { 2033,  5, 10 },  { 2033,  5, 11 },  { 2033,  5, 12 },  { 2033,  5, 13 },
  { 2033,  5, 14 },  { 2033,  5, 15 },  { 2033,  5, 16 },  { 2033,  5, 17 },
  { 2033,  5, 18 },  { 2033,  5, 19 },  { 2033,  5, 20 },  { 2033,  5, 21 },
  { 2033,  5, 22 },  { 2033,  5, 23 },  { 2033,  5, 24 },  { 2033,  5, 25 },
  { 2033,  5, 26 },  { 2033,  5, 27 },  { 2033,  5, 28 },  { 2033,  5, 29 },
  { 2033,  5, 30 },  { 2033,  5, 31 },  { 2033,  6,  1 },  { 2033,  6,  2 },
  { 2033,  6,  3 },  { 2033,  6,  4 },  { 2033,  6,  5 },  { 2033,  6,  6 },
  { 2033,  6,  7 },  { 2033,  6,  8 },  { 2033,  6,  9 },  { 2033,  6, 10 },
  { 2033,  6, 11 },  { 2033,  6, 12 },  { 2033,  6, 13 },  { 2033,  6, 14 },
  { 2033,  6, 15 },  { 2033,  6, 16 },  { 2033,  6, 17 },  { 2033,  6, 18 },
  { 2033,  6, 19 },  { 2033,  6, 20 },  { 2033,  6, 21 },  { 2033,  6, 22 },
  { 2033,  6, 23 },  { 2033,  6, 24 },  { 2033,  6, 25 },  { 2033,  6, 26 },
  { 2033,  6, 27 },  { 2033,  6, 28 },  { 2033,  6, 29 },  { 2033,  6, 30 },
  { 2033,  7,  1 },  { 2033,  7,  2 },  { 2033,  7,  3 },  { 2033,  7,  4 },
  { 2033,  7,  5 },  { 2033,  7,  6 },  { 2033,  7,  7 },  { 2033,  7,  8 },
  { 2033,  7,  9 },  { 2033,  7, 10 },  { 2033,  7, 11 },  { 2033,  7, 12 },
  { 2033,  7, 13 },  { 2033,  7, 14 },  { 2033,  7, 15 },  { 2033,  7, 16 },
  { 2033,  7, 17 },  { 2033,  7, 18 },  { 2033,  7, 19 },  { 2033,  7, 20 },
  { 2033,  7, 21 },  { 2033,  7, 22 },  { 2033,  7, 23 },  { 2033,  7, 24 },
  { 2033,  7, 25 },  { 2033,  7, 26 },  { 2033,  7, 27 },  { 2033,  7, 28 },
  { 2033,  7, 29 },  { 2033,  7, 30 },  { 2033,  7, 31 },  { 2033,  8,  1 },
  { 2033,  8,  2 },  { 2033,  8,  3 },  { 2033,  8,  4 },  { 2033,  8,  5 },
  { 2033,  8,  6 },  { 2033,  8,  7 },  { 2033,  8,  8 },  { 2033,  8,  9 },
  { 2033,  8, 10 },  { 2033,  8, 11 },  { 2033,  8, 12 },  { 2033,  8, 13 },
  { 2033,  8, 14 },  { 2033,  8, 15 },  { 2033,  8, 16 },  { 2033,  8, 17 },
  { 2033,  8, 18 },  { 2033,  8, 19 },  { 2033,  8, 20 },  { 2033,  8, 21 },
  { 2033,  8, 22 },  { 2033,  8, 23 },  { 2033,  8, 24 },  { 2033,  8, 25 },
  { 2033,  8, 26 },  { 2033,  8, 27 },  { 2033,  8, 28 },  { 2033,  8, 29 },
  { 2033,  8, 30 },  { 2033,  8, 31 },  { 2033,  9,  1 },  { 2033,  9,  2 },
  { 2033,  9,  3 },  { 2033,  9,  4 },  { 2033,  9,  5 },  { 2033,  9,  6 },
  { 2033,  9,  7 },  { 2033,  9,  8 },  { 2033,  9,  9 },  { 2033,  9, 10 },
  { 2033,  9, 11 },  { 2033,  9, 12 },  { 2033,  9, 13 },  { 2033,  9, 14 },
  { 2033,  9, 15 },  { 2033,  9, 16 },  { 2033,  9, 17 },  { 2033,  9, 18 },
  { 2033,  9, 19 },  { 2033,  9, 20 },  { 2033,  9, 21 },  { 2033,  9, 22 },
  { 2033,  9, 23 },  { 2033,  9, 24 },  { 2033,  9, 25 },  { 2033,  9, 26 },
  { 2033,  9, 27 },  { 2033,  9, 28 },  { 2033,  9, 29 },  { 2033,  9, 30 },
  { 2033, 10,  1 },  { 2033, 10,  2 },  { 2033, 10,  3 },  { 2033, 10,  4 },
  { 2033, 10,  5 },  { 2033, 10,  6 },  { 2033, 10,  7 },  { 2033, 10,  8 },
  { 2033, 10,  9 },  { 2033, 10, 10 },  { 2033, 10, 11 },  { 2033, 10, 12 },
  { 2033, 10, 13 },  { 2033, 10, 14 },  { 2033, 10, 15 },  { 2033, 10, 16 },
  { 2033, 10, 17 },  { 2033, 10, 18 },  { 2033, 10, 19 },  { 2033, 10, 20 },
  { 2033, 10, 21 },  { 2033, 10, 22 },  { 2033, 10, 23 },  { 2033, 10, 24 },
  { 2033, 10, 25 },  { 2033, 10, 26 },  { 2033, 10, 27 },  { 2033, 10, 28 },
  { 2033, 10, 29 },  { 2033, 10, 30 },  { 2033, 10, 31 },  { 2033, 11,  1 },
  { 2033, 11,  2 },  { 2033, 11,  3 },  { 2033, 11,  4 },  { 2033, 11,  5 },
  { 2033, 11,  6 },  { 2033, 11,  7 },  { 2033, 11,  8 },  { 2033, 11,  9 },
  { 2033, 11, 10 },  { 2033, 11, 11 },  { 2033, 11, 12 },  { 2033, 11, 13 },
  { 2033, 11, 14 },  { 2033, 11, 15 },  { 2033, 11, 16 },  { 2033, 11, 17 },
  { 2033, 11, 18 },  { 2033, 11, 19 },  { 2033, 11, 20 },  { 2033, 11, 21 },
  { 2033, 11, 22 },  { 2033, 11, 23 },  { 2033, 11, 24 },  { 2033, 11, 25 },
  { 2033, 11, 26 },  { 2033, 11, 27 },  { 2033, 11, 28 },  { 2033, 11, 29 },
  { 2033, 11, 30 },  { 2033, 12,  1 },  { 2033, 12,  2 },  { 2033, 12,  3 },
  { 2033, 12,  4 },  { 2033, 12,  5 },  { 2033, 12,  6 },  { 2033, 12,  7 },
  { 2033, 12,  8 },  { 2033, 12,  9 },  { 2033, 12, 10 },  { 2033, 12, 11 },
  { 2033, 12, 12 },  { 2033, 12, 13 },  { 2033, 12, 14 },  { 2033, 12, 15 },
  { 2033, 12, 16 },  { 2033, 12, 17 },  { 2033, 12, 18 },  { 2033, 12, 19 },
  { 2033, 12, 20 },  { 2033, 12, 21 },  { 2033, 12, 22 },  { 2033, 12, 23 },
  { 2033, 12, 24 },  { 2033, 12, 25 },  { 2033, 12, 26 },  { 2033, 12, 27 },
  { 2033, 12, 28 },  { 2033, 12, 29 },  { 2033, 12, 30 },  { 2033, 12, 31 },
  { 2034,  1,  1 },  { 2034,  1,  2 },  { 2034,  1,  3 },  { 2034,  1,  4 },
  { 2034,  1,  5 },  { 2034,  1,  6 },  { 2034,  1,  7 },  { 2034,  1,  8 },
  { 2034,  1,  9 },  { 2034,  1, 10 },  { 2034,  1, 11 },  { 2034,  1, 12 },
  { 2034,  1, 13 },  { 2034,  1, 14 },  { 2034,  1, 15 },  { 2034,  1, 16 },
  { 2034,  1, 17 },  { 2034,  1, 18 },  { 2034,  1, 19 },  { 2034,  1, 20 },
  { 2034,  1, 21 },  { 2034,  1, 22 },  { 2034,  1, 23 },  { 2034,  1, 24 },
  { 2034,  1, 25 },  { 2034,  1, 26 },  { 2034,  1, 27 },  { 2034,  1, 28 },
  { 2034,  1, 29 },  { 2034,  1, 30 },  { 2034,  1, 31 },  { 2034,  2,  1 },
  { 2034,  2,  2 },  { 2034,  2,  3 },  { 2034,  2,  4 },  { 2034,  2,  5 },
  { 2034,  2,  6 },  { 2034,  2,  7 },  { 2034,  2,  8 },  { 2034,  2,  9 },
  { 2034,  2, 10 },  { 2034,  2, 11 },  { 2034,  2, 12 },  { 2034,  2, 13 },
  { 2034,  2, 14 },  { 2034,  2, 15 },  { 2034,  2, 16 },  { 2034,  2, 17 },
  { 2034,  2, 18 },  { 2034,  2, 19 },  { 2034,  2, 20 },  { 2034,  2, 21 },
  { 2034,  2, 22 },  { 2034,  2, 23 },  { 2034,  2, 24 },  { 2034,  2, 25 },
  { 2034,  2, 26 },  { 2034,  2, 27 },  { 2034,  2, 28 },  { 2034,  3,  1 },
  { 2034,  3,  2 },  { 2034,  3,  3 },  { 2034,  3,  4 },  { 2034,  3,  5 },
  { 2034,  3,  6 },  { 2034,  3,  7 },  { 2034,  3,  8 },  { 2034,  3,  9 },
  { 2034,  3, 10 },  { 2034,  3, 11 },  { 2034,  3, 12 },  { 2034,  3, 13 },
  { 2034,  3, 14 },  { 2034,  3, 15 },  { 2034,  3, 16 },  { 2034,  3, 17 },
  { 2034,  3, 18 },  { 2034,  3, 19 },  { 2034,  3, 20 },  { 2034,  3, 21 },
  { 2034,  3, 22 },  { 2034,  3, 23 },  { 2034,  3, 24 },  { 2034,  3, 25 },
  { 2034,  3, 26 },  { 2034,  3, 27 },  { 2034,  3, 28 },  { 2034,  3, 29 },
  { 2034,  3, 30 },  { 2034,  3, 31 },  { 2034,  4,  1 },  { 2034,  4,  2 },
  { 2034,  4,  3 },  { 2034,  4,  4 },  { 2034,  4,  5 },  { 2034,  4,  6 },
  { 2034,  4,  7 },  { 2034,  4,  8 },  { 2034,  4,  9 },  { 2034,  4, 10 },
  { 2034,  4, 11 },  { 2034,  4, 12 },  { 2034,  4, 13 },  { 2034,  4, 14 },
  { 2034,  4, 15 },  { 2034,  4, 16 },  { 2034,  4, 17 },  { 2034,  4, 18 },
  { 2034,  4, 19 },  { 2034,  4, 20 },  { 2034,  4, 21 },  { 2034,  4, 22 },
  { 2034,  4, 23 },  { 2034,  4, 24 },  { 2034,  4, 25 },  { 2034,  4, 26 },
  { 2034,  4, 27 },  { 2034,  4, 28 },  { 2034,  4, 29 },  { 2034,  4, 30 },
  { 2034,  5,  1 },  { 2034,  5,  2 },  { 2034,  5,  3 },  { 2034,  5,  4 },
  { 2034,  5,  5 },  { 2034,  5,  6 },  { 2034,  5,  7 },  { 2034,  5,  8 },
  { 2034,  5,  9 },  { 2034,  5, 10 },  { 2034,  5, 11 },  { 2034,  5, 12 },
  { 2034,  5, 13 },  { 2034,  5, 14 },  { 2034,  5, 15 },  { 2034,  5, 16 },
  { 2034,  5, 17 },  { 2034,  5, 18 },  { 2034,  5, 19 },  { 2034,  5, 20 },
  { 2034,  5, 21 },  { 2034,  5, 22 },  { 2034,  5, 23 },  { 2034,  5, 24 },
  { 2034,  5, 25 },  { 2034,  5, 26 },  { 2034,  5, 27 },  { 2034,  5, 28 },
  { 2034,  5, 29 },  { 2034,  5, 30 },  { 2034,  5, 31 },  { 2034,  6,  1 },
  { 2034,  6,  2 },  { 2034,  6,  3 },  { 2034,  6,  4 },  { 2034,  6,  5 },
  { 2034,  6,  6 },  { 2034,  6,  7 },  { 2034,  6,  8 },  { 2034,  6,  9 },
  { 2034,  6, 10 },  { 2034,  6, 11 },  { 2034,  6, 12 },  { 2034,  6, 13 },
  { 2034,  6, 14 },  { 2034,  6, 15 },  { 2034,  6, 16 },  { 2034,  6, 17 },
  { 2034,  6, 18 },  { 2034,  6, 19 },  { 2034,  6, 20 },  { 2034,  6, 21 },
  { 2034,  6, 22 },  { 2034,  6, 23 },  { 2034,  6, 24 },  { 2034,  6, 25 },
  { 2034,  6, 26 },  { 2034,  6, 27 },  { 2034,  6, 28 },  { 2034,  6, 29 },
  { 2034,  6, 30 },  { 2034,  7,  1 },  { 2034,  7,  2 },  { 2034,  7,  3 },
  { 2034,  7,  4 },  { 2034,  7,  5 },  { 2034,  7,  6 },  { 2034,  7,  7 },
  { 2034,  7,  8 },  { 2034,  7,  9 },  { 2034,  7, 10 },  { 2034,  7, 11 },
  { 2034,  7, 12 },  { 2034,  7, 13 },  { 2034,  7, 14 },  { 2034,  7, 15 },
  { 2034,  7, 16 },  { 2034,  7, 17 },  { 2034,  7, 18 },  { 2034,  7, 19 },
  { 2034,  7, 20 },  { 2034,  7, 21 },  { 2034,  7, 22 },  { 2034,  7, 23 },
  { 2034,  7, 24 },  { 2034,  7, 25 },  { 2034,  7, 26 },  { 2034,  7, 27 },
  { 2034,  7, 28 },  { 2034,  7, 29 },  { 2034,  7, 30 },  { 2034,  7, 31 },
  { 2034,  8,  1 },  { 2034,  8,  2 },  { 2034,  8,  3 },  { 2034,  8,  4 },
  { 2034,  8,  5 },  { 2034,  8,  6 },  { 2034,  8,  7 },  { 2034,  8,  8 },
  { 2034,  8,  9 },  { 2034,  8, 10 },  { 2034,  8, 11 },  { 2034,  8, 12 },
  { 2034,  8, 13 },  { 2034,  8, 14 },  { 2034,  8, 15 },  { 2034,  8, 16 },
  { 2034,  8, 17 },  { 2034,  8, 18 },  { 2034,  8, 19 },  { 2034,  8, 20 },
  { 2034,  8, 21 },  { 2034,  8, 22 },  { 2034,  8, 23 },  { 2034,  8, 24 },
  { 2034,  8, 25 },  { 2034,  8, 26 },  { 2034,  8, 27 },  { 2034,  8, 28 },
  { 2034,  8, 29 },  { 2034,  8, 30 },  { 2034,  8, 31 },  { 2034,  9,  1 },
  { 2034,  9,  2 },  { 2034,  9,  3 },  { 2034,  9,  4 },  { 2034,  9,  5 },
  { 2034,  9,  6 },  { 2034,  9,  7 },  { 2034,  9,  8 },  { 2034,  9,  9 },
  { 2034,  9, 10 },  { 2034,  9, 11 },  { 2034,  9, 12 },  { 2034,  9, 13 },
  { 2034,  9, 14 },  { 2034,  9, 15 },  { 2034,  9, 16 },  { 2034,  9, 17 },
  { 2034,  9, 18 },  { 2034,  9, 19 },  { 2034,  9, 20 },  { 2034,  9, 21 },
  { 2034,  9, 22 },  { 2034,  9, 23 },  { 2034,  9, 24 },  { 2034,  9, 25 },
  { 2034,  9, 26 },  { 2034,  9, 27 },  { 2034,  9, 28 },  { 2034,  9, 29 },
  { 2034,  9, 30 },  { 2034, 10,  1 },  { 2034, 10,  2 },  { 2034, 10,  3 },
  { 2034, 10,  4 },  { 2034, 10,  5 },  { 2034, 10,  6 },  { 2034, 10,  7 },
  { 2034, 10,  8 },  { 2034, 10,  9 },  { 2034, 10, 10 },  { 2034, 10, 11 },
  { 2034, 10, 12 },  { 2034, 10, 13 },  { 2034, 10, 14 },  { 2034, 10, 15 },
  { 2034, 10, 16 },  { 2034, 10, 17 },  { 2034, 10, 18 },  { 2034, 10, 19 },
  { 2034, 10, 20 },  { 2034, 10, 21 },  { 2034, 10, 22 },  { 2034, 10, 23 },
  { 2034, 10, 24 },  { 2034, 10, 25 },  { 2034, 10, 26 },  { 2034, 10, 27 },
  { 2034, 10, 28 },  { 2034, 10, 29 },  { 2034, 10, 30 },  { 2034, 10, 31 },
  { 2034, 11,  1 },  { 2034, 11,  2 },  { 2034, 11,  3 },  { 2034, 11,  4 },
  { 2034, 11,  5 },  { 2034, 11,  6 },  { 2034, 11,  7 },  { 2034, 11,  8 },
  { 2034, 11,  9 },  { 2034, 11, 10 },  { 2034, 11, 11 },  { 2034, 11, 12 },
  { 2034, 11, 13 },  { 2034, 11, 14 },  { 2034, 11, 15 },  { 2034, 11, 16 },
  { 2034, 11, 17 },  { 2034, 11, 18 },  { 2034, 11, 19 },  { 2034, 11, 20 },
  { 2034, 11, 21 },  { 2034, 11, 22 },  { 2034, 11, 23 },  { 2034, 11, 24 },
  { 2034, 11, 25 },  { 2034, 11, 26 },  { 2034, 11, 27 },  { 2034, 11, 28 },
  { 2034, 11, 29 },  { 2034, 11, 30 },  { 2034, 12,  1 },  { 2034, 12,  2 },
  { 2034, 12,  3 },  { 2034, 12,  4 },  { 2034, 12,  5 },  { 2034, 12,  6 },
  { 2034, 12,  7 },  { 2034, 12,  8 },  { 2034, 12,  9 },  { 2034, 12, 10 },
  { 2034, 12, 11 },  { 2034, 12, 12 },  { 2034, 12, 13 },  { 2034, 12, 14 },
  { 2034, 12, 15 },  { 2034, 12, 16 },  { 2034, 12, 17 },  { 2034, 12, 18 },
  { 2034, 12, 19 },  { 2034, 12, 20 },  { 2034, 12, 21 },  { 2034, 12, 22 },
  { 2034, 12, 23 },  { 2034, 12, 24 },  { 2034, 12, 25 },  { 2034, 12, 26 },
  { 2034, 12, 27 },  { 2034, 12, 28 },  { 2034, 12, 29 },  { 2034, 12, 30 },
  { 2034, 12, 31 },  { 2035,  1,  1 },  { 2035,  1,  2 },  { 2035,  1,  3 },
  { 2035,  1,  4 },  { 2035,  1,  5 },  { 2035,  1,  6 },  { 2035,  1,  7 },
  { 2035,  1,  8 },  { 2035,  1,  9 },  { 2035,  1, 10 },  { 2035,  1, 11 },
  { 2035,  1, 12 },  { 2035,  1, 13 },  { 2035,  1, 14 },  { 2035,  1, 15 },
  { 2035,  1, 16 },  { 2035,  1, 17 },  { 2035,  1, 18 },  { 2035,  1, 19 },
  { 2035,  1, 20 },  { 2035,  1, 21 },  { 2035,  1, 22 },  { 2035,  1, 23 },
  { 2035,  1, 24 },  { 2035,  1, 25 },  { 2035,  1, 26 },  { 2035,  1, 27 },
  { 2035,  1, 28 },  { 2035,  1, 29 },  { 2035,  1, 30 },  { 2035,  1, 31 },
  { 2035,  2,  1 },  { 2035,  2,  2 },  { 2035,  2,  3 },  { 2035,  2,  4 },
  { 2035,  2,  5 },  { 2035,  2,  6 },  { 2035,  2,  7 },  { 2035,  2,  8 },
  { 2035,  2,  9 },  { 2035,  2, 10 },  { 2035,  2, 11 },  { 2035,  2, 12 },
  { 2035,  2, 13 },  { 2035,  2, 14 },  { 2035,  2, 15 },  { 2035,  2, 16 },
  { 2035,  2, 17 },  { 2035,  2, 18 },  { 2035,  2, 19 },  { 2035,  2, 20 },
  { 2035,  2, 21 },  { 2035,  2, 22 },  { 2035,  2, 23 },  { 2035,  2, 24 },
  { 2035,  2, 25 },  { 2035,  2, 26 },  { 2035,  2, 27 },  { 2035,  2, 28 },
  { 2035,  3,  1 },  { 2035,  3,  2 },  { 2035,  3,  3 },  { 2035,  3,  4 },
  { 2035,  3,  5 },  { 2035,  3,  6 },  { 2035,  3,  7 },  { 2035,  3,  8 },
  { 2035,  3,  9 },  { 2035,  3, 10 },  { 2035,  3, 11 },  { 2035,  3, 12 },
  { 2035,  3, 13 },  { 2035,  3, 14 },  { 2035,  3, 15 },  { 2035,  3, 16 },
  { 2035,  3, 17 },  { 2035,  3, 18 },  { 2035,  3, 19 },  { 2035,  3, 20 },
  { 2035,  3, 21 },  { 2035,  3, 22 },  { 2035,  3, 23 },  { 2035,  3, 24 },
  { 2035,  3, 25 },  { 2035,  3, 26 },  { 2035,  3, 27 },  { 2035,  3, 28 },
  { 2035,  3, 29 },  { 2035,  3, 30 },  { 2035,  3, 31 },  { 2035,  4,  1 },
  { 2035,  4,  2 },  { 2035,  4,  3 },  { 2035,  4,  4 },  { 2035,  4,  5 },
  { 2035,  4,  6 },  { 2035,  4,  7 },  { 2035,  4,  8 },  { 2035,  4,  9 },
  { 2035,  4, 10 },  { 2035,  4, 11 },  { 2035,  4, 12 },  { 2035,  4, 13 },
  { 2035,  4, 14 },  { 2035,  4, 15 },  { 2035,  4, 16 },  { 2035,  4, 17 },
  { 2035,  4, 18 },  { 2035,  4, 19 },  { 2035,  4, 20 },  { 2035,  4, 21 },
  { 2035,  4, 22 },  { 2035,  4, 23 },  { 2035,  4, 24 },  { 2035,  4, 25 },
  { 2035,  4, 26 },  { 2035,  4, 27 },  { 2035,  4, 28 },  { 2035,  4, 29 },
  { 2035,  4, 30 },  { 2035,  5,  1 },  { 2035,  5,  2 },  { 2035,  5,  3 },
  { 2035,  5,  4 },  { 2035,  5,  5 },  { 2035,  5,  6 },  { 2035,  5,  7 },
  { 2035,  5,  8 },  { 2035,  5,  9 },  { 2035,  5, 10 },  { 2035,  5, 11 },
  { 2035,  5, 12 },  { 2035,  5, 13 },  { 2035,  5, 14 },  { 2035,  5, 15 },
  { 2035,  5, 16 },  { 2035,  5, 17 },  { 2035,  5, 18 },  { 2035,  5, 19 },
  { 2035,  5, 20 },  { 2035,  5, 21 },  { 2035,  5, 22 },  { 2035,  5, 23 },
  { 2035,  5, 24 },  { 2035,  5, 25 },  { 2035,  5, 26 },  { 2035,  5, 27 },
  { 2035,  5, 28 },  { 2035,  5, 29 },  { 2035,  5, 30 },  { 2035,  5, 31 },
  { 2035,  6,  1 },  { 2035,  6,  2 },  { 2035,  6,  3 },  { 2035,  6,  4 },
  { 2035,  6,  5 },  { 2035,  6,  6 },  { 2035,  6,  7 },  { 2035,  6,  8 },
  { 2035,  6,  9 },  { 2035,  6, 10 },  { 2035,  6, 11 },  { 2035,  6, 12 },
  { 2035,  6, 13 },  { 2035,  6, 14 },  { 2035,  6, 15 },  { 2035,  6, 16 },
  { 2035,  6, 17 },  { 2035,  6, 18 },  { 2035,  6, 19 },  { 2035,  6, 20 },
  { 2035,  6, 21 },  { 2035,  6, 22 },  { 2035,  6, 23 },  { 2035,  6, 24 },
  { 2035,  6, 25 },  { 2035,  6, 26 },  { 2035,  6, 27 },  { 2035,  6, 28 },
  { 2035,  6, 29 },  { 2035,  6, 30 },  { 2035,  7,  1 },  { 2035,  7,  2 },
  { 2035,  7,  3 },  { 2035,  7,  4 },  { 2035,  7,  5 },  { 2035,  7,  6 },
  { 2035,  7,  7 },  { 2035,  7,  8 },  { 2035,  7,  9 },  { 2035,  7, 10 },
  { 2035,  7, 11 },  { 2035,  7, 12 },  { 2035,  7, 13 },  { 2035,  7, 14 },
  { 2035,  7, 15 },  { 2035,  7, 16 },  { 2035,  7, 17 },  { 2035,  7, 18 },
  { 2035,  7, 19 },  { 2035,  7, 20 },  { 2035,  7, 21 },  { 2035,  7, 22 },
  { 2035,  7, 23 },  { 2035,  7, 24 },  { 2035,  7, 25 },  { 2035,  7, 26 },
  { 2035,  7, 27 },  { 2035,  7, 28 },  { 2035,  7, 29 },  { 2035,  7, 30 },
  { 2035,  7, 31 },  { 2035,  8,  1 },  { 2035,  8,  2 },  { 2035,  8,  3 },
  { 2035,  8,  4 },  { 2035,  8,  5 },  { 2035,  8,  6 },  { 2035,  8,  7 },
  { 2035,  8,  8 },  { 2035,  8,  9 },  { 2035,  8, 10 },  { 2035,  8, 11 },
  { 2035,  8, 12 },  { 2035,  8, 13 },  { 2035,  8, 14 },  { 2035,  8, 15 },
  { 2035,  8, 16 },  { 2035,  8, 17 },  { 2035,  8, 18 },  { 2035,  8, 19 },
  { 2035,  8, 20 },  { 2035,  8, 21 },  { 2035,  8, 22 },  { 2035,  8, 23 },
  { 2035,  8, 24 },  { 2035,  8, 25 },  { 2035,  8, 26 },  { 2035,  8, 27 },
  { 2035,  8, 28 },  { 2035,  8, 29 },  { 2035,  8, 30 },  { 2035,  8, 31 },
  { 2035,  9,  1 },  { 2035,  9,  2 },  { 2035,  9,  3 },  { 2035,  9,  4 },
  { 2035,  9,  5 },  { 2035,  9,  6 },  { 2035,  9,  7 },  { 2035,  9,  8 },
  { 2035,  9,  9 },  { 2035,  9, 10 },  { 2035,  9, 11 },  { 2035,  9, 12 },
  { 2035,  9, 13 },  { 2035,  9, 14 },  { 2035,  9, 15 },  { 2035,  9, 16 },
  { 2035,  9, 17 },  { 2035,  9, 18 },  { 2035,  9, 19 },  { 2035,  9, 20 },
  { 2035,  9, 21 },  { 2035,  9, 22 },  { 2035,  9, 23 },  { 2035,  9, 24 },
  { 2035,  9, 25 },  { 2035,  9, 26 },  { 2035,  9, 27 },  { 2035,  9, 28 },
  { 2035,  9, 29 },  { 2035,  9, 30 },  { 2035, 10,  1 },  { 2035, 10,  2 },
  { 2035, 10,  3 },  { 2035, 10,  4 },  { 2035, 10,  5 },  { 2035, 10,  6 },
  { 2035, 10,  7 },  { 2035, 10,  8 },  { 2035, 10,  9 },  { 2035, 10, 10 },
  { 2035, 10, 11 },  { 2035, 10, 12 },  { 2035, 10, 13 },  { 2035, 10, 14 },
  { 2035, 10, 15 },  { 2035, 10, 16 },  { 2035, 10, 17 },  { 2035, 10, 18 },
  { 2035, 10, 19 },  { 2035, 10, 20 },  { 2035, 10, 21 },  { 2035, 10, 22 },
  { 2035, 10, 23 },  { 2035, 10, 24 },  { 2035, 10, 25 },  { 2035, 10, 26 },
  { 2035, 10, 27 },  { 2035, 10, 28 },  { 2035, 10, 29 },  { 2035, 10, 30 },
  { 2035, 10, 31 },  { 2035, 11,  1 },  { 2035, 11,  2 },  { 2035, 11,  3 },
  { 2035, 11,  4 },  { 2035, 11,  5 },  { 2035, 11,  6 },  { 2035, 11,  7 },
  { 2035, 11,  8 },  { 2035, 11,  9 },  { 2035, 11, 10 },  { 2035, 11, 11 },
  { 2035, 11, 12 },  { 2035, 11, 13 },  { 2035, 11, 14 },  { 2035, 11, 15 },
  { 2035, 11, 16 },  { 2035, 11, 17 },  { 2035, 11, 18 },  { 2035, 11, 19 },
  { 2035, 11, 20 },  { 2035, 11, 21 },  { 2035, 11, 22 },  { 2035, 11, 23 },
  { 2035, 11, 24 },  { 2035, 11, 25 },  { 2035, 11, 26 },  { 2035, 11, 27 },
  { 2035, 11, 28 },  { 2035, 11, 29 },  { 2035, 11, 30 },  { 2035, 12,  1 },
  { 2035, 12,  2 },  { 2035, 12,  3 },  { 2035, 12,  4 },  { 2035, 12,  5 },
  { 2035, 12,  6 },  { 2035, 12,  7 },  { 2035, 12,  8 },  { 2035, 12,  9 },
  { 2035, 12, 10 },  { 2035, 12, 11 },  { 2035, 12, 12 },  { 2035, 12, 13 },
  { 2035, 12, 14 },  { 2035, 12, 15 },  { 2035, 12, 16 },  { 2035, 12, 17 },
  { 2035, 12, 18 },  { 2035, 12, 19 },  { 2035, 12, 20 },  { 2035, 12, 21 },
  { 2035, 12, 22 },  { 2035, 12, 23 },  { 2035, 12, 24 },  { 2035, 12, 25 },
  { 2035, 12, 26 },  { 2035, 12, 27 },  { 2035, 12, 28 },  { 2035, 12, 29 },
  { 2035, 12, 30 },  { 2035, 12, 31 },  { 2036,  1,  1 },  { 2036,  1,  2 },
  { 2036,  1,  3 },  { 2036,  1,  4 },  { 2036,  1,  5 },  { 2036,  1,  6 },
  { 2036,  1,  7 },  { 2036,  1,  8 },  { 2036,  1,  9 },  { 2036,  1, 10 },
  { 2036,  1, 11 },  { 2036,  1, 12 },  { 2036,  1, 13 },  { 2036,  1, 14 },
  { 2036,  1, 15 },  { 2036,  1, 16 },  { 2036,  1, 17 },  { 2036,  1, 18 },
  { 2036,  1, 19 },  { 2036,  1, 20 },  { 2036,  1, 21 },  { 2036,  1, 22 },
  { 2036,  1, 23 },  { 2036,  1, 24 },  { 2036,  1, 25 },  { 2036,  1, 26 },
  { 2036,  1, 27 },  { 2036,  1, 28 },  { 2036,  1, 29 },  { 2036,  1, 30 },
  { 2036,  1, 31 },  { 2036,  2,  1 },  { 2036,  2,  2 },  { 2036,  2,  3 },
  { 2036,  2,  4 },  { 2036,  2,  5 },  { 2036,  2,  6 },  { 2036,  2,  7 },
  { 2036,  2,  8 },  { 2036,  2,  9 },  { 2036,  2, 10 },  { 2036,  2, 11 },
  { 2036,  2, 12 },  { 2036,  2, 13 },  { 2036,  2, 14 },  { 2036,  2, 15 },
  { 2036,  2, 16 },  { 2036,  2, 17 },  { 2036,  2, 18 },  { 2036,  2, 19 },
  { 2036,  2, 20 },  { 2036,  2, 21 },  { 2036,  2, 22 },  { 2036,  2, 23 },
  { 2036,  2, 24 },  { 2036,  2, 25 },  { 2036,  2, 26 },  { 2036,  2, 27 },
  { 2036,  2, 28 },  { 2036,  2, 29 },  { 2036,  3,  1 },  { 2036,  3,  2 },
  { 2036,  3,  3 },  { 2036,  3,  4 },  { 2036,  3,  5 },  { 2036,  3,  6 },
  { 2036,  3,  7 },  { 2036,  3,  8 },  { 2036,  3,  9 },  { 2036,  3, 10 },
  { 2036,  3, 11 },  { 2036,  3, 12 },  { 2036,  3, 13 },  { 2036,  3, 14 },
  { 2036,  3, 15 },  { 2036,  3, 16 },  { 2036,  3, 17 },  { 2036,  3, 18 },
  { 2036,  3, 19 },  { 2036,  3, 20 },  { 2036,  3, 21 },  { 2036,  3, 22 },
  { 2036,  3, 23 },  { 2036,  3, 24 },  { 2036,  3, 25 },  { 2036,  3, 26 },
  { 2036,  3, 27 },  { 2036,  3, 28 },  { 2036,  3, 29 },  { 2036,  3, 30 },
  { 2036,  3, 31 },  { 2036,  4,  1 },  { 2036,  4,  2 },  { 2036,  4,  3 },
  { 2036,  4,  4 },  { 2036,  4,  5 },  { 2036,  4,  6 },  { 2036,  4,  7 },
  { 2036,  4,  8 },  { 2036,  4,  9 },  { 2036,  4, 10 },  { 2036,  4, 11 },
  { 2036,  4, 12 },  { 2036,  4, 13 },  { 2036,  4, 14 },  { 2036,  4, 15 },
  { 2036,  4, 16 },  { 2036,  4, 17 },  { 2036,  4, 18 },  { 2036,  4, 19 },
  { 2036,  4, 20 },  { 2036,  4, 21 },  { 2036,  4, 22 },  { 2036,  4, 23 },
  { 2036,  4, 24 },  { 2036,  4, 25 },  { 2036,  4, 26 },  { 2036,  4, 27 },
  { 2036,  4, 28 },  { 2036,  4, 29 },  { 2036,  4, 30 },  { 2036,  5,  1 },
  { 2036,  5,  2 },  { 2036,  5,  3 },  { 2036,  5,  4 },  { 2036,  5,  5 },
  { 2036,  5,  6 },  { 2036,  5,  7 },  { 2036,  5,  8 },  { 2036,  5,  9 },
  { 2036,  5, 10 },  { 2036,  5, 11 },  { 2036,  5, 12 },  { 2036,  5, 13 },
  { 2036,  5, 14 },  { 2036,  5, 15 },  { 2036,  5, 16 },  { 2036,  5, 17 },
  { 2036,  5, 18 },  { 2036,  5, 19 },  { 2036,  5, 20 },  { 2036,  5, 21 },
  { 2036,  5, 22 },  { 2036,  5, 23 },  { 2036,  5, 24 },  { 2036,  5, 25 },
  { 2036,  5, 26 },  { 2036,  5, 27 },  { 2036,  5, 28 },  { 2036,  5, 29 },
  { 2036,  5, 30 },  { 2036,  5, 31 },  { 2036,  6,  1 },  { 2036,  6,  2 },
  { 2036,  6,  3 },  { 2036,  6,  4 },  { 2036,  6,  5 },  { 2036,  6,  6 },
  { 2036,  6,  7 },  { 2036,  6,  8 },  { 2036,  6,  9 },  { 2036,  6, 10 },
  { 2036,  6, 11 },  { 2036,  6, 12 },  { 2036,  6, 13 },  { 2036,  6, 14 },
  { 2036,  6, 15 },  { 2036,  6, 16 },  { 2036,  6, 17 },  { 2036,  6, 18 },
  { 2036,  6, 19 },  { 2036,  6, 20 },  { 2036,  6, 21 },  { 2036,  6, 22 },
  { 2036,  6, 23 },  { 2036,  6, 24 },  { 2036,  6, 25 },  { 2036,  6, 26 },
  { 2036,  6, 27 },  { 2036,  6, 28 },  { 2036,  6, 29 },  { 2036,  6, 30 },
  { 2036,  7,  1 },  { 2036,  7,  2 },  { 2036,  7,  3 },  { 2036,  7,  4 },
  { 2036,  7,  5 },  { 2036,  7,  6 },  { 2036,  7,  7 },  { 2036,  7,  8 },
  { 2036,  7,  9 },  { 2036,  7, 10 },  { 2036,  7, 11 },  { 2036,  7, 12 },
  { 2036,  7, 13 },  { 2036,  7, 14 },  { 2036,  7, 15 },  { 2036,  7, 16 },
  { 2036,  7, 17 },  { 2036,  7, 18 },  { 2036,  7, 19 },  { 2036,  7, 20 },
  { 2036,  7, 21 },  { 2036,  7, 22 },  { 2036,  7, 23 },  { 2036,  7, 24 },
  { 2036,  7, 25 },  { 2036,  7, 26 },  { 2036,  7, 27 },  { 2036,  7, 28 },
  { 2036,  7, 29 },  { 2036,  7, 30 },  { 2036,  7, 31 },  { 2036,  8,  1 },
  { 2036,  8,  2 },  { 2036,  8,  3 },  { 2036,  8,  4 },  { 2036,  8,  5 },
  { 2036,  8,  6 },  { 2036,  8,  7 },  { 2036,  8,  8 },  { 2036,  8,  9 },
  { 2036,  8, 10 },  { 2036,  8, 11 },  { 2036,  8, 12 },  { 2036,  8, 13 },
  { 2036,  8, 14 },  { 2036,  8, 15 },  { 2036,  8, 16 },  { 2036,  8, 17 },
  { 2036,  8, 18 },  { 2036,  8, 19 },  { 2036,  8, 20 },  { 2036,  8, 21 },
  { 2036,  8, 22 },  { 2036,  8, 23 },  { 2036,  8, 24 },  { 2036,  8, 25 },
  { 2036,  8, 26 },  { 2036,  8, 27 },  { 2036,  8, 28 },  { 2036,  8, 29 },
  { 2036,  8, 30 },  { 2036,  8, 31 },  { 2036,  9,  1 },  { 2036,  9,  2 },
  { 2036,  9,  3 },  { 2036,  9,  4 },  { 2036,  9,  5 },  { 2036,  9,  6 },
  { 2036,  9,  7 },  { 2036,  9,  8 },  { 2036,  9,  9 },  { 2036,  9, 10 },
  { 2036,  9, 11 },  { 2036,  9, 12 },  { 2036,  9, 13 },  { 2036,  9, 14 },
  { 2036,  9, 15 },  { 2036,  9, 16 },  { 2036,  9, 17 },  { 2036,  9, 18 },
  { 2036,  9, 19 },  { 2036,  9, 20 },  { 2036,  9, 21 },  { 2036,  9, 22 },
  { 2036,  9, 23 },  { 2036,  9, 24 },  { 2036,  9, 25 },  { 2036,  9, 26 },
  { 2036,  9, 27 },  { 2036,  9, 28 },  { 2036,  9, 29 },  { 2036,  9, 30 },
  { 2036, 10,  1 },  { 2036, 10,  2 },  { 2036, 10,  3 },  { 2036, 10,  4 },
  { 2036, 10,  5 },  { 2036, 10,  6 },  { 2036, 10,  7 },  { 2036, 10,  8 },
  { 2036, 10,  9 },  { 2036, 10, 10 },  { 2036, 10, 11 },  { 2036, 10, 12 },
  { 2036, 10, 13 },  { 2036, 10, 14 },  { 2036, 10, 15 },  { 2036, 10, 16 },
  { 2036, 10, 17 },  { 2036, 10, 18 },  { 2036, 10, 19 },  { 2036, 10, 20 },
  { 2036, 10, 21 },  { 2036, 10, 22 },  { 2036, 10, 23 },  { 2036, 10, 24 },
  { 2036, 10, 25 },  { 2036, 10, 26 },  { 2036, 10, 27 },  { 2036, 10, 28 },
  { 2036, 10, 29 },  { 2036, 10, 30 },  { 2036, 10, 31 },  { 2036, 11,  1 },
  { 2036, 11,  2 },  { 2036, 11,  3 },  { 2036, 11,  4 },  { 2036, 11,  5 },
  { 2036, 11,  6 },  { 2036, 11,  7 },  { 2036, 11,  8 },  { 2036, 11,  9 },
  { 2036, 11, 10 },  { 2036, 11, 11 },  { 2036, 11, 12 },  { 2036, 11, 13 },
  { 2036, 11, 14 },  { 2036, 11, 15 },  { 2036, 11, 16 },  { 2036, 11, 17 },
  { 2036, 11, 18 },  { 2036, 11, 19 },  { 2036, 11, 20 },  { 2036, 11, 21 },
  { 2036, 11, 22 },  { 2036, 11, 23 },  { 2036, 11, 24 },  { 2036, 11, 25 },
  { 2036, 11, 26 },  { 2036, 11, 27 },  { 2036, 11, 28 },  { 2036, 11, 29 },
  { 2036, 11, 30 },  { 2036, 12,  1 },  { 2036, 12,  2 },  { 2036, 12,  3 },
  { 2036, 12,  4 },  { 2036, 12,  5 },  { 2036, 12,  6 },  { 2036, 12,  7 },
  { 2036, 12,  8 },  { 2036, 12,  9 },  { 2036, 12, 10 },  { 2036, 12, 11 },
  { 2036, 12, 12 },  { 2036, 12, 13 },  { 2036, 12, 14 },  { 2036, 12, 15 },
  { 2036, 12, 16 },  { 2036, 12, 17 },  { 2036, 12, 18 },  { 2036, 12, 19 },
  { 2036, 12, 20 },  { 2036, 12, 21 },  { 2036, 12, 22 },  { 2036, 12, 23 },
  { 2036, 12, 24 },  { 2036, 12, 25 },  { 2036, 12, 26 },  { 2036, 12, 27 },
  { 2036, 12, 28 },  { 2036, 12, 29 },  { 2036, 12, 30 },  { 2036, 12, 31 },
  { 2037,  1,  1 },  { 2037,  1,  2 },  { 2037,  1,  3 },  { 2037,  1,  4 },
  { 2037,  1,  5 },  { 2037,  1,  6 },  { 2037,  1,  7 },  { 2037,  1,  8 },
  { 2037,  1,  9 },  { 2037,  1, 10 },  { 2037,  1, 11 },  { 2037,  1, 12 },
  { 2037,  1, 13 },  { 2037,  1, 14 },  { 2037,  1, 15 },  { 2037,  1, 16 },
  { 2037,  1, 17 },  { 2037,  1, 18 },  { 2037,  1, 19 },  { 2037,  1, 20 },
  { 2037,  1, 21 },  { 2037,  1, 22 },  { 2037,  1, 23 },  { 2037,  1, 24 },
  { 2037,  1, 25 },  { 2037,  1, 26 },  { 2037,  1, 27 },  { 2037,  1, 28 },
  { 2037,  1, 29 },  { 2037,  1, 30 },  { 2037,  1, 31 },  { 2037,  2,  1 },
  { 2037,  2,  2 },  { 2037,  2,  3 },  { 2037,  2,  4 },  { 2037,  2,  5 },
  { 2037,  2,  6 },  { 2037,  2,  7 },  { 2037,  2,  8 },  { 2037,  2,  9 },
  { 2037,  2, 10 },  { 2037,  2, 11 },  { 2037,  2, 12 },  { 2037,  2, 13 },
  { 2037,  2, 14 },  { 2037,  2, 15 },  { 2037,  2, 16 },  { 2037,  2, 17 },
  { 2037,  2, 18 },  { 2037,  2, 19 },  { 2037,  2, 20 },  { 2037,  2, 21 },
  { 2037,  2, 22 },  { 2037,  2, 23 },  { 2037,  2, 24 },  { 2037,  2, 25 },
  { 2037,  2, 26 },  { 2037,  2, 27 },  { 2037,  2, 28 },  { 2037,  3,  1 },
  { 2037,  3,  2 },  { 2037,  3,  3 },  { 2037,  3,  4 },  { 2037,  3,  5 },
  { 2037,  3,  6 },  { 2037,  3,  7 },  { 2037,  3,  8 },  { 2037,  3,  9 },
  { 2037,  3, 10 },  { 2037,  3, 11 },  { 2037,  3, 12 },  { 2037,  3, 13 },
  { 2037,  3, 14 },  { 2037,  3, 15 },  { 2037,  3, 16 },  { 2037,  3, 17 },
  { 2037,  3, 18 },  { 2037,  3, 19 },  { 2037,  3, 20 },  { 2037,  3, 21 },
  { 2037,  3, 22 },  { 2037,  3, 23 },  { 2037,  3, 24 },  { 2037,  3, 25 },
  { 2037,  3, 26 },  { 2037,  3, 27 },  { 2037,  3, 28 },  { 2037,  3, 29 },
  { 2037,  3, 30 },  { 2037,  3, 31 },  { 2037,  4,  1 },  { 2037,  4,  2 },
  { 2037,  4,  3 },  { 2037,  4,  4 },  { 2037,  4,  5 },  { 2037,  4,  6 },
  { 2037,  4,  7 },  { 2037,  4,  8 },  { 2037,  4,  9 },  { 2037,  4, 10 },
  { 2037,  4, 11 },  { 2037,  4, 12 },  { 2037,  4, 13 },  { 2037,  4, 14 },
  { 2037,  4, 15 },  { 2037,  4, 16 },  { 2037,  4, 17 },  { 2037,  4, 18 },
  { 2037,  4, 19 },  { 2037,  4, 20 },  { 2037,  4, 21 },  { 2037,  4, 22 },
  { 2037,  4, 23 },  { 2037,  4, 24 },  { 2037,  4, 25 },  { 2037,  4, 26 },
  { 2037,  4, 27 },  { 2037,  4, 28 },  { 2037,  4, 29 },  { 2037,  4, 30 },
  { 2037,  5,  1 },  { 2037,  5,  2 },  { 2037,  5,  3 },  { 2037,  5,  4 },
  { 2037,  5,  5 },  { 2037,  5,  6 },  { 2037,  5,  7 },  { 2037,  5,  8 },
  { 2037,  5,  9 },  { 2037,  5, 10 },  { 2037,  5, 11 },  { 2037,  5, 12 },
  { 2037,  5, 13 },  { 2037,  5, 14 },  { 2037,  5, 15 },  { 2037,  5, 16 },
  { 2037,  5, 17 },  { 2037,  5, 18 },  { 2037,  5, 19 },  { 2037,  5, 20 },
  { 2037,  5, 21 },  { 2037,  5, 22 },  { 2037,  5, 23 },  { 2037,  5, 24 },
  { 2037,  5, 25 },  { 2037,  5, 26 },  { 2037,  5, 27 },  { 2037,  5, 28 },
  { 2037,  5, 29 },  { 2037,  5, 30 },  { 2037,  5, 31 },  { 2037,  6,  1 },
  { 2037,  6,  2 },  { 2037,  6,  3 },  { 2037,  6,  4 },  { 2037,  6,  5 },
  { 2037,  6,  6 },  { 2037,  6,  7 },  { 2037,  6,  8 },  { 2037,  6,  9 },
  { 2037,  6, 10 },  { 2037,  6, 11 },  { 2037,  6, 12 },  { 2037,  6, 13 },
  { 2037,  6, 14 },  { 2037,  6, 15 },  { 2037,  6, 16 },  { 2037,  6, 17 },
  { 2037,  6, 18 },  { 2037,  6, 19 },  { 2037,  6, 20 },  { 2037,  6, 21 },
  { 2037,  6, 22 },  { 2037,  6, 23 },  { 2037,  6, 24 },  { 2037,  6, 25 },
  { 2037,  6, 26 },  { 2037,  6, 27 },  { 2037,  6, 28 },  { 2037,  6, 29 },
  { 2037,  6, 30 },  { 2037,  7,  1 },  { 2037,  7,  2 },  { 2037,  7,  3 },
  { 2037,  7,  4 },  { 2037,  7,  5 },  { 2037,  7,  6 },  { 2037,  7,  7 },
  { 2037,  7,  8 },  { 2037,  7,  9 },  { 2037,  7, 10 },  { 2037,  7, 11 },
  { 2037,  7, 12 },  { 2037,  7, 13 },  { 2037,  7, 14 },  { 2037,  7, 15 },
  { 2037,  7, 16 },  { 2037,  7, 17 },  { 2037,  7, 18 },  { 2037,  7, 19 },
  { 2037,  7, 20 },  { 2037,  7, 21 },  { 2037,  7, 22 },  { 2037,  7, 23 },
  { 2037,  7, 24 },  { 2037,  7, 25 },  { 2037,  7, 26 },  { 2037,  7, 27 },
  { 2037,  7, 28 },  { 2037,  7, 29 },  { 2037,  7, 30 },  { 2037,  7, 31 },
  { 2037,  8,  1 },  { 2037,  8,  2 },  { 2037,  8,  3 },  { 2037,  8,  4 },
  { 2037,  8,  5 },  { 2037,  8,  6 },  { 2037,  8,  7 },  { 2037,  8,  8 },
  { 2037,  8,  9 },  { 2037,  8, 10 },  { 2037,  8, 11 },  { 2037,  8, 12 },
  { 2037,  8, 13 },  { 2037,  8, 14 },  { 2037,  8, 15 },  { 2037,  8, 16 },
  { 2037,  8, 17 },  { 2037,  8, 18 },  { 2037,  8, 19 },  { 2037,  8, 20 },
  { 2037,  8, 21 },  { 2037,  8, 22 },  { 2037,  8, 23 },  { 2037,  8, 24 },
  { 2037,  8, 25 },  { 2037,  8, 26 },  { 2037,  8, 27 },  { 2037,  8, 28 },
  { 2037,  8, 29 },  { 2037,  8, 30 },  { 2037,  8, 31 },  { 2037,  9,  1 },
  { 2037,  9,  2 },  { 2037,  9,  3 },  { 2037,  9,  4 },  { 2037,  9,  5 },
  { 2037,  9,  6 },  { 2037,  9,  7 },  { 2037,  9,  8 },  { 2037,  9,  9 },
  { 2037,  9, 10 },  { 2037,  9, 11 },  { 2037,  9, 12 },  { 2037,  9, 13 },
  { 2037,  9, 14 },  { 2037,  9, 15 },  { 2037,  9, 16 },  { 2037,  9, 17 },
  { 2037,  9, 18 },  { 2037,  9, 19 },  { 2037,  9, 20 },  { 2037,  9, 21 },
  { 2037,  9, 22 },  { 2037,  9, 23 },  { 2037,  9, 24 },  { 2037,  9, 25 },
  { 2037,  9, 26 },  { 2037,  9, 27 },  { 2037,  9, 28 },  { 2037,  9, 29 },
  { 2037,  9, 30 },  { 2037, 10,  1 },  { 2037, 10,  2 },  { 2037, 10,  3 },
  { 2037, 10,  4 },  { 2037, 10,  5 },  { 2037, 10,  6 },  { 2037, 10,  7 },
  { 2037, 10,  8 },  { 2037, 10,  9 },  { 2037, 10, 10 },  { 2037, 10, 11 },
  { 2037, 10, 12 },  { 2037, 10, 13 },  { 2037, 10, 14 },  { 2037, 10, 15 },
  { 2037, 10, 16 },  { 2037, 10, 17 },  { 2037, 10, 18 },  { 2037, 10, 19 },
  { 2037, 10, 20 },  { 2037, 10, 21 },  { 2037, 10, 22 },  { 2037, 10, 23 },
  { 2037, 10, 24 },  { 2037, 10, 25 },  { 2037, 10, 26 },  { 2037, 10, 27 },
  { 2037, 10, 28 },  { 2037, 10, 29 },  { 2037, 10, 30 },  { 2037, 10, 31 },
  { 2037, 11,  1 },  { 2037, 11,  2 },  { 2037, 11,  3 },  { 2037, 11,  4 },
  { 2037, 11,  5 },  { 2037, 11,  6 },  { 2037, 11,  7 },  { 2037, 11,  8 },
  { 2037, 11,  9 },  { 2037, 11, 10 },  { 2037, 11, 11 },  { 2037, 11, 12 },
  { 2037, 11, 13 },  { 2037, 11, 14 },  { 2037, 11, 15 },  { 2037, 11, 16 },
  { 2037, 11, 17 },  { 2037, 11, 18 },  { 2037, 11, 19 },  { 2037, 11, 20 },
  { 2037, 11, 21 },  { 2037, 11, 22 },  { 2037, 11, 23 },  { 2037, 11, 24 },
  { 2037, 11, 25 },  { 2037, 11, 26 },  { 2037, 11, 27 },  { 2037, 11, 28 },
  { 2037, 11, 29 },  { 2037, 11, 30 },  { 2037, 12,  1 },  { 2037, 12,  2 },
  { 2037, 12,  3 },  { 2037, 12,  4 },  { 2037, 12,  5 },  { 2037, 12,  6 },
  { 2037, 12,  7 },  { 2037, 12,  8 },  { 2037, 12,  9 },  { 2037, 12, 10 },
  { 2037, 12, 11 },  { 2037, 12, 12 },  { 2037, 12, 13 },  { 2037, 12, 14 },
  { 2037, 12, 15 },  { 2037, 12, 16 },  { 2037, 12, 17 },  { 2037, 12, 18 },
  { 2037, 12, 19 },  { 2037, 12, 20 },  { 2037, 12, 21 },  { 2037, 12, 22 },
  { 2037, 12, 23 },  { 2037, 12, 24 },  { 2037, 12, 25 },  { 2037, 12, 26 },
  { 2037, 12, 27 },  { 2037, 12, 28 },  { 2037, 12, 29 },  { 2037, 12, 30 },
  { 2037, 12, 31 },  { 2038,  1,  1 },  { 2038,  1,  2 },  { 2038,  1,  3 },
  { 2038,  1,  4 },  { 2038,  1,  5 },  { 2038,  1,  6 },  { 2038,  1,  7 },
  { 2038,  1,  8 },  { 2038,  1,  9 },  { 2038,  1, 10 },  { 2038,  1, 11 },
  { 2038,  1, 12 },  { 2038,  1, 13 },  { 2038,  1, 14 },  { 2038,  1, 15 },
  { 2038,  1, 16 },  { 2038,  1, 17 },  { 2038,  1, 18 },  { 2038,  1, 19 },
  { 2038,  1, 20 },  { 2038,  1, 21 },  { 2038,  1, 22 },  { 2038,  1, 23 },
  { 2038,  1, 24 },  { 2038,  1, 25 },  { 2038,  1, 26 },  { 2038,  1, 27 },
  { 2038,  1, 28 },  { 2038,  1, 29 },  { 2038,  1, 30 },  { 2038,  1, 31 },
  { 2038,  2,  1 },  { 2038,  2,  2 },  { 2038,  2,  3 },  { 2038,  2,  4 },
  { 2038,  2,  5 },  { 2038,  2,  6 },  { 2038,  2,  7 },  { 2038,  2,  8 },
  { 2038,  2,  9 },  { 2038,  2, 10 },  { 2038,  2, 11 },  { 2038,  2, 12 },
  { 2038,  2, 13 },  { 2038,  2, 14 },  { 2038,  2, 15 },  { 2038,  2, 16 },
  { 2038,  2, 17 },  { 2038,  2, 18 },  { 2038,  2, 19 },  { 2038,  2, 20 },
  { 2038,  2, 21 },  { 2038,  2, 22 },  { 2038,  2, 23 },  { 2038,  2, 24 },
  { 2038,  2, 25 },  { 2038,  2, 26 },  { 2038,  2, 27 },  { 2038,  2, 28 },
  { 2038,  3,  1 },  { 2038,  3,  2 },  { 2038,  3,  3 },  { 2038,  3,  4 },
  { 2038,  3,  5 },  { 2038,  3,  6 },  { 2038,  3,  7 },  { 2038,  3,  8 },
  { 2038,  3,  9 },  { 2038,  3, 10 },  { 2038,  3, 11 },  { 2038,  3, 12 },
  { 2038,  3, 13 },  { 2038,  3, 14 },  { 2038,  3, 15 },  { 2038,  3, 16 },
  { 2038,  3, 17 },  { 2038,  3, 18 },  { 2038,  3, 19 },  { 2038,  3, 20 },
  { 2038,  3, 21 },  { 2038,  3, 22 },  { 2038,  3, 23 },  { 2038,  3, 24 },
  { 2038,  3, 25 },  { 2038,  3, 26 },  { 2038,  3, 27 },  { 2038,  3, 28 },
  { 2038,  3, 29 },  { 2038,  3, 30 },  { 2038,  3, 31 },  { 2038,  4,  1 },
  { 2038,  4,  2 },  { 2038,  4,  3 },  { 2038,  4,  4 },  { 2038,  4,  5 },
  { 2038,  4,  6 },  { 2038,  4,  7 },  { 2038,  4,  8 },  { 2038,  4,  9 },
  { 2038,  4, 10 },  { 2038,  4, 11 },  { 2038,  4, 12 },  { 2038,  4, 13 },
  { 2038,  4, 14 },  { 2038,  4, 15 },  { 2038,  4, 16 },  { 2038,  4, 17 },
  { 2038,  4, 18 },  { 2038,  4, 19 },  { 2038,  4, 20 },  { 2038,  4, 21 },
  { 2038,  4, 22 },  { 2038,  4, 23 },  { 2038,  4, 24 },  { 2038,  4, 25 },
  { 2038,  4, 26 },  { 2038,  4, 27 },  { 2038,  4, 28 },  { 2038,  4, 29 },
  { 2038,  4, 30 },  { 2038,  5,  1 },  { 2038,  5,  2 },  { 2038,  5,  3 },
  { 2038,  5,  4 },  { 2038,  5,  5 },  { 2038,  5,  6 },  { 2038,  5,  7 },
  { 2038,  5,  8 },  { 2038,  5,  9 },  { 2038,  5, 10 },  { 2038,  5, 11 },
  { 2038,  5, 12 },  { 2038,  5, 13 },  { 2038,  5, 14 },  { 2038,  5, 15 },
  { 2038,  5, 16 },  { 2038,  5, 17 },  { 2038,  5, 18 },  { 2038,  5, 19 },
  { 2038,  5, 20 },  { 2038,  5, 21 },  { 2038,  5, 22 },  { 2038,  5, 23 },
  { 2038,  5, 24 },  { 2038,  5, 25 },  { 2038,  5, 26 },  { 2038,  5, 27 },
  { 2038,  5, 28 },  { 2038,  5, 29 },  { 2038,  5, 30 },  { 2038,  5, 31 },
  { 2038,  6,  1 },  { 2038,  6,  2 },  { 2038,  6,  3 },  { 2038,  6,  4 },
  { 2038,  6,  5 },  { 2038,  6,  6 },  { 2038,  6,  7 },  { 2038,  6,  8 },
  { 2038,  6,  9 },  { 2038,  6, 10 },  { 2038,  6, 11 },  { 2038,  6, 12 },
  { 2038,  6, 13 },  { 2038,  6, 14 },  { 2038,  6, 15 },  { 2038,  6, 16 },
  { 2038,  6, 17 },  { 2038,  6, 18 },  { 2038,  6, 19 },  { 2038,  6, 20 },
  { 2038,  6, 21 },  { 2038,  6, 22 },  { 2038,  6, 23 },  { 2038,  6, 24 },
  { 2038,  6, 25 },  { 2038,  6, 26 },  { 2038,  6, 27 },  { 2038,  6, 28 },
  { 2038,  6, 29 },  { 2038,  6, 30 },  { 2038,  7,  1 },  { 2038,  7,  2 },
  { 2038,  7,  3 },  { 2038,  7,  4 },  { 2038,  7,  5 },  { 2038,  7,  6 },
  { 2038,  7,  7 },  { 2038,  7,  8 },  { 2038,  7,  9 },  { 2038,  7, 10 },
  { 2038,  7, 11 },  { 2038,  7, 12 },  { 2038,  7, 13 },  { 2038,  7, 14 },
  { 2038,  7, 15 },  { 2038,  7, 16 },  { 2038,  7, 17 },  { 2038,  7, 18 },
  { 2038,  7, 19 },  { 2038,  7, 20 },  { 2038,  7, 21 },  { 2038,  7, 22 },
  { 2038,  7, 23 },  { 2038,  7, 24 },  { 2038,  7, 25 },  { 2038,  7, 26 },
  { 2038,  7, 27 },  { 2038,  7, 28 },  { 2038,  7, 29 },  { 2038,  7, 30 },
  { 2038,  7, 31 },  { 2038,  8,  1 },  { 2038,  8,  2 },  { 2038,  8,  3 },
  { 2038,  8,  4 },  { 2038,  8,  5 },  { 2038,  8,  6 },  { 2038,  8,  7 },
  { 2038,  8,  8 },  { 2038,  8,  9 },  { 2038,  8, 10 },  { 2038,  8, 11 },
  { 2038,  8, 12 },  { 2038,  8, 13 },  { 2038,  8, 14 },  { 2038,  8, 15 },
  { 2038,  8, 16 },  { 2038,  8, 17 },  { 2038,  8, 18 },  { 2038,  8, 19 },
  { 2038,  8, 20 },  { 2038,  8, 21 },  { 2038,  8, 22 },  { 2038,  8, 23 },
  { 2038,  8, 24 },  { 2038,  8, 25 },  { 2038,  8, 26 },  { 2038,  8, 27 },
  { 2038,  8, 28 },  { 2038,  8, 29 },  { 2038,  8, 30 },  { 2038,  8, 31 },
  { 2038,  9,  1 },  { 2038,  9,  2 },  { 2038,  9,  3 },  { 2038,  9,  4 },
  { 2038,  9,  5 },  { 2038,  9,  6 },  { 2038,  9,  7 },  { 2038,  9,  8 },
  { 2038,  9,  9 },  { 2038,  9, 10 },  { 2038,  9, 11 },  { 2038,  9, 12 },
  { 2038,  9, 13 },  { 2038,  9, 14 },  { 2038,  9, 15 },  { 2038,  9, 16 },
  { 2038,  9, 17 },  { 2038,  9, 18 },  { 2038,  9, 19 },  { 2038,  9, 20 },
  { 2038,  9, 21 },  { 2038,  9, 22 },  { 2038,  9, 23 },  { 2038,  9, 24 },
  { 2038,  9, 25 },  { 2038,  9, 26 },  { 2038,  9, 27 },  { 2038,  9, 28 },
  { 2038,  9, 29 },  { 2038,  9, 30 },  { 2038, 10,  1 },  { 2038, 10,  2 },
  { 2038, 10,  3 },  { 2038, 10,  4 },  { 2038, 10,  5 },  { 2038, 10,  6 },
  { 2038, 10,  7 },  { 2038, 10,  8 },  { 2038, 10,  9 },  { 2038, 10, 10 },
  { 2038, 10, 11 },  { 2038, 10, 12 },  { 2038, 10, 13 },  { 2038, 10, 14 },
  { 2038, 10, 15 },  { 2038, 10, 16 },  { 2038, 10, 17 },  { 2038, 10, 18 },
  { 2038, 10, 19 },  { 2038, 10, 20 },  { 2038, 10, 21 },  { 2038, 10, 22 },
  { 2038, 10, 23 },  { 2038, 10, 24 },  { 2038, 10, 25 },  { 2038, 10, 26 },
  { 2038, 10, 27 },  { 2038, 10, 28 },  { 2038, 10, 29 },  { 2038, 10, 30 },
  { 2038, 10, 31 },  { 2038, 11,  1 },  { 2038, 11,  2 },  { 2038, 11,  3 },
  { 2038, 11,  4 },  { 2038, 11,  5 },  { 2038, 11,  6 },  { 2038, 11,  7 },
  { 2038, 11,  8 },  { 2038, 11,  9 },  { 2038, 11, 10 },  { 2038, 11, 11 },
  { 2038, 11, 12 },  { 2038, 11, 13 },  { 2038, 11, 14 },  { 2038, 11, 15 },
  { 2038, 11, 16 },  { 2038, 11, 17 },  { 2038, 11, 18 },  { 2038, 11, 19 },
  { 2038, 11, 20 },  { 2038, 11, 21 },  { 2038, 11, 22 },  { 2038, 11, 23 },
  { 2038, 11, 24 },  { 2038, 11, 25 },  { 2038, 11, 26 },  { 2038, 11, 27 },
  { 2038, 11, 28 },  { 2038, 11, 29 },  { 2038, 11, 30 },  { 2038, 12,  1 },
  { 2038, 12,  2 },  { 2038, 12,  3 },  { 2038, 12,  4 },  { 2038, 12,  5 },
  { 2038, 12,  6 },  { 2038, 12,  7 },  { 2038, 12,  8 },  { 2038, 12,  9 },
  { 2038, 12, 10 },  { 2038, 12, 11 },  { 2038, 12, 12 },  { 2038, 12, 13 },
  { 2038, 12, 14 },  { 2038, 12, 15 },  { 2038, 12, 16 },  { 2038, 12, 17 },
  { 2038, 12, 18 },  { 2038, 12, 19 },  { 2038, 12, 20 },  { 2038, 12, 21 },
  { 2038, 12, 22 },  { 2038, 12, 23 },  { 2038, 12, 24 },  { 2038, 12, 25 },
  { 2038, 12, 26 },  { 2038, 12, 27 },  { 2038, 12, 28 },  { 2038, 12, 29 },
  { 2038, 12, 30 },  { 2038, 12, 31 },  { 2039,  1,  1 },  { 2039,  1,  2 },
  { 2039,  1,  3 },  { 2039,  1,  4 },  { 2039,  1,  5 },  { 2039,  1,  6 },
  { 2039,  1,  7 },  { 2039,  1,  8 },  { 2039,  1,  9 },  { 2039,  1, 10 },
  { 2039,  1, 11 },  { 2039,  1, 12 },  { 2039,  1, 13 },  { 2039,  1, 14 },
  { 2039,  1, 15 },  { 2039,  1, 16 },  { 2039,  1, 17 },  { 2039,  1, 18 },
  { 2039,  1, 19 },  { 2039,  1, 20 },  { 2039,  1, 21 },  { 2039,  1, 22 },
  { 2039,  1, 23 },  { 2039,  1, 24 },  { 2039,  1, 25 },  { 2039,  1, 26 },
  { 2039,  1, 27 },  { 2039,  1, 28 },  { 2039,  1, 29 },  { 2039,  1, 30 },
  { 2039,  1, 31 },  { 2039,  2,  1 },  { 2039,  2,  2 },  { 2039,  2,  3 },
  { 2039,  2,  4 },  { 2039,  2,  5 },  { 2039,  2,  6 },  { 2039,  2,  7 },
  { 2039,  2,  8 },  { 2039,  2,  9 },  { 2039,  2, 10 },  { 2039,  2, 11 },
  { 2039,  2, 12 },  { 2039,  2, 13 },  { 2039,  2, 14 },  { 2039,  2, 15 },
  { 2039,  2, 16 },  { 2039,  2, 17 },  { 2039,  2, 18 },  { 2039,  2, 19 },
  { 2039,  2, 20 },  { 2039,  2, 21 },  { 2039,  2, 22 },  { 2039,  2, 23 },
  { 2039,  2, 24 },  { 2039,  2, 25 },  { 2039,  2, 26 },  { 2039,  2, 27 },
  { 2039,  2, 28 },  { 2039,  3,  1 },  { 2039,  3,  2 },  { 2039,  3,  3 },
  { 2039,  3,  4 },  { 2039,  3,  5 },  { 2039,  3,  6 },  { 2039,  3,  7 },
  { 2039,  3,  8 },  { 2039,  3,  9 },  { 2039,  3, 10 },  { 2039,  3, 11 },
  { 2039,  3, 12 },  { 2039,  3, 13 },  { 2039,  3, 14 },  { 2039,  3, 15 },
  { 2039,  3, 16 },  { 2039,  3, 17 },  { 2039,  3, 18 },  { 2039,  3, 19 },
  { 2039,  3, 20 },  { 2039,  3, 21 },  { 2039,  3, 22 },  { 2039,  3, 23 },
  { 2039,  3, 24 },  { 2039,  3, 25 },  { 2039,  3, 26 },  { 2039,  3, 27 },
  { 2039,  3, 28 },  { 2039,  3, 29 },  { 2039,  3, 30 },  { 2039,  3, 31 },
  { 2039,  4,  1 },  { 2039,  4,  2 },  { 2039,  4,  3 },  { 2039,  4,  4 },
  { 2039,  4,  5 },  { 2039,  4,  6 },  { 2039,  4,  7 },  { 2039,  4,  8 },
  { 2039,  4,  9 },  { 2039,  4, 10 },  { 2039,  4, 11 },  { 2039,  4, 12 },
  { 2039,  4, 13 },  { 2039,  4, 14 },  { 2039,  4, 15 },  { 2039,  4, 16 },
  { 2039,  4, 17 },  { 2039,  4, 18 },  { 2039,  4, 19 },  { 2039,  4, 20 },
  { 2039,  4, 21 },  { 2039,  4, 22 },  { 2039,  4, 23 },  { 2039,  4, 24 },
  { 2039,  4, 25 },  { 2039,  4, 26 },  { 2039,  4, 27 },  { 2039,  4, 28 },
  { 2039,  4, 29 },  { 2039,  4, 30 },  { 2039,  5,  1 },  { 2039,  5,  2 },
  { 2039,  5,  3 },  { 2039,  5,  4 },  { 2039,  5,  5 },  { 2039,  5,  6 },
  { 2039,  5,  7 },  { 2039,  5,  8 },  { 2039,  5,  9 },  { 2039,  5, 10 },
  { 2039,  5, 11 },  { 2039,  5, 12 },  { 2039,  5, 13 },  { 2039,  5, 14 },
  { 2039,  5, 15 },  { 2039,  5, 16 },  { 2039,  5, 17 },  { 2039,  5, 18 },
  { 2039,  5, 19 },  { 2039,  5, 20 },  { 2039,  5, 21 },  { 2039,  5, 22 },
  { 2039,  5, 23 },  { 2039,  5, 24 },  { 2039,  5, 25 },  { 2039,  5, 26 },
  { 2039,  5, 27 },  { 2039,  5, 28 },  { 2039,  5, 29 },  { 2039,  5, 30 },
  { 2039,  5, 31 },  { 2039,  6,  1 },  { 2039,  6,  2 },  { 2039,  6,  3 },
  { 2039,  6,  4 },  { 2039,  6,  5 },  { 2039,  6,  6 },  { 2039,  6,  7 },
  { 2039,  6,  8 },  { 2039,  6,  9 },  { 2039,  6, 10 },  { 2039,  6, 11 },
  { 2039,  6, 12 },  { 2039,  6, 13 },  { 2039,  6, 14 },  { 2039,  6, 15 },
  { 2039,  6, 16 },  { 2039,  6, 17 },  { 2039,  6, 18 },  { 2039,  6, 19 },
  { 2039,  6, 20 },  { 2039,  6, 21 },  { 2039,  6, 22 },  { 2039,  6, 23 },
  { 2039,  6, 24 },  { 2039,  6, 25 },  { 2039,  6, 26 },  { 2039,  6, 27 },
  { 2039,  6, 28 },  { 2039,  6, 29 },  { 2039,  6, 30 },  { 2039,  7,  1 },
  { 2039,  7,  2 },  { 2039,  7,  3 },  { 2039,  7,  4 },  { 2039,  7,  5 },
  { 2039,  7,  6 },  { 2039,  7,  7 },  { 2039,  7,  8 },  { 2039,  7,  9 },
  { 2039,  7, 10 },  { 2039,  7, 11 },  { 2039,  7, 12 },  { 2039,  7, 13 },
  { 2039,  7, 14 },  { 2039,  7, 15 },  { 2039,  7, 16 },  { 2039,  7, 17 },
  { 2039,  7, 18 },  { 2039,  7, 19 },  { 2039,  7, 20 },  { 2039,  7, 21 },
  { 2039,  7, 22 },  { 2039,  7, 23 },  { 2039,  7, 24 },  { 2039,  7, 25 },
  { 2039,  7, 26 },  { 2039,  7, 27 },  { 2039,  7, 28 },  { 2039,  7, 29 },
  { 2039,  7, 30 },  { 2039,  7, 31 },  { 2039,  8,  1 },  { 2039,  8,  2 },
  { 2039,  8,  3 },  { 2039,  8,  4 },  { 2039,  8,  5 },  { 2039,  8,  6 },
  { 2039,  8,  7 },  { 2039,  8,  8 },  { 2039,  8,  9 },  { 2039,  8, 10 },
  { 2039,  8, 11 },  { 2039,  8, 12 },  { 2039,  8, 13 },  { 2039,  8, 14 },
  { 2039,  8, 15 },  { 2039,  8, 16 },  { 2039,  8, 17 },  { 2039,  8, 18 },
  { 2039,  8, 19 },  { 2039,  8, 20 },  { 2039,  8, 21 },  { 2039,  8, 22 },
  { 2039,  8, 23 },  { 2039,  8, 24 },  { 2039,  8, 25 },  { 2039,  8, 26 },
  { 2039,  8, 27 },  { 2039,  8, 28 },  { 2039,  8, 29 },  { 2039,  8, 30 },
  { 2039,  8, 31 },  { 2039,  9,  1 },  { 2039,  9,  2 },  { 2039,  9,  3 },
  { 2039,  9,  4 },  { 2039,  9,  5 },  { 2039,  9,  6 },  { 2039,  9,  7 },
  { 2039,  9,  8 },  { 2039,  9,  9 },  { 2039,  9, 10 },  { 2039,  9, 11 },
  { 2039,  9, 12 },  { 2039,  9, 13 },  { 2039,  9, 14 },  { 2039,  9, 15 },
  { 2039,  9, 16 },  { 2039,  9, 17 },  { 2039,  9, 18 },  { 2039,  9, 19 },
  { 2039,  9, 20 },  { 2039,  9, 21 },  { 2039,  9, 22 },  { 2039,  9, 23 },
  { 2039,  9, 24 },  { 2039,  9, 25 },  { 2039,  9, 26 },  { 2039,  9, 27 },
  { 2039,  9, 28 },  { 2039,  9, 29 },  { 2039,  9, 30 },  { 2039, 10,  1 },
  { 2039, 10,  2 },  { 2039, 10,  3 },  { 2039, 10,  4 },  { 2039, 10,  5 },
  { 2039, 10,  6 },  { 2039, 10,  7 },  { 2039, 10,  8 },  { 2039, 10,  9 },
  { 2039, 10, 10 },  { 2039, 10, 11 },  { 2039, 10, 12 },  { 2039, 10, 13 },
  { 2039, 10, 14 },  { 2039, 10, 15 },  { 2039, 10, 16 },  { 2039, 10, 17 },
  { 2039, 10, 18 },  { 2039, 10, 19 },  { 2039, 10, 20 },  { 2039, 10, 21 },
  { 2039, 10, 22 },  { 2039, 10, 23 },  { 2039, 10, 24 },  { 2039, 10, 25 },
  { 2039, 10, 26 },  { 2039, 10, 27 },  { 2039, 10, 28 },  { 2039, 10, 29 },
  { 2039, 10, 30 },  { 2039, 10, 31 },  { 2039, 11,  1 },  { 2039, 11,  2 },
  { 2039, 11,  3 },  { 2039, 11,  4 },  { 2039, 11,  5 },  { 2039, 11,  6 },
  { 2039, 11,  7 },  { 2039, 11,  8 },  { 2039, 11,  9 },  { 2039, 11, 10 },
  { 2039, 11, 11 },  { 2039, 11, 12 },  { 2039, 11, 13 },  { 2039, 11, 14 },
  { 2039, 11, 15 },  { 2039, 11, 16 },  { 2039, 11, 17 },  { 2039, 11, 18 },
  { 2039, 11, 19 },  { 2039, 11, 20 },  { 2039, 11, 21 },  { 2039, 11, 22 },
  { 2039, 11, 23 },  { 2039, 11, 24 },  { 2039, 11, 25 },  { 2039, 11, 26 },
  { 2039, 11, 27 },  { 2039, 11, 28 },  { 2039, 11, 29 },  { 2039, 11, 30 },
  { 2039, 12,  1 },  { 2039, 12,  2 },  { 2039, 12,  3 },  { 2039, 12,  4 },
  { 2039, 12,  5 },  { 2039, 12,  6 },  { 2039, 12,  7 },  { 2039, 12,  8 },
  { 2039, 12,  9 },  { 2039, 12, 10 },  { 2039, 12, 11 },  { 2039, 12, 12 },
  { 2039, 12, 13 },  { 2039, 12, 14 },  { 2039, 12, 15 },  { 2039, 12, 16 },
  { 2039, 12, 17 },  { 2039, 12, 18 },  { 2039, 12, 19 },  { 2039, 12, 20 },
  { 2039, 12, 21 },  { 2039, 12, 22 },  { 2039, 12, 23 },  { 2039, 12, 24 },
  { 2039, 12, 25 },  { 2039, 12, 26 },  { 2039, 12, 27 },  { 2039, 12, 28 },
  { 2039, 12, 29 },  { 2039, 12, 30 },  { 2039, 12, 31 },  { 2040,  1,  1 },
  { 2040,  1,  2 },  { 2040,  1,  3 },  { 2040,  1,  4 },  { 2040,  1,  5 },
  { 2040,  1,  6 },  { 2040,  1,  7 },  { 2040,  1,  8 },  { 2040,  1,  9 },
  { 2040,  1, 10 },  { 2040,  1, 11 },  { 2040,  1, 12 },  { 2040,  1, 13 },
  { 2040,  1, 14 },  { 2040,  1, 15 },  { 2040,  1, 16 },  { 2040,  1, 17 },
  { 2040,  1, 18 },  { 2040,  1, 19 },  { 2040,  1, 20 },  { 2040,  1, 21 },
  { 2040,  1, 22 },  { 2040,  1, 23 },  { 2040,  1, 24 },  { 2040,  1, 25 },
  { 2040,  1, 26 },  { 2040,  1, 27 },  { 2040,  1, 28 },  { 2040,  1, 29 },
  { 2040,  1, 30 },  { 2040,  1, 31 },  { 2040,  2,  1 },  { 2040,  2,  2 },
  { 2040,  2,  3 },  { 2040,  2,  4 },  { 2040,  2,  5 },  { 2040,  2,  6 },
  { 2040,  2,  7 },  { 2040,  2,  8 },  { 2040,  2,  9 },  { 2040,  2, 10 },
  { 2040,  2, 11 },  { 2040,  2, 12 },  { 2040,  2, 13 },  { 2040,  2, 14 },
  { 2040,  2, 15 },  { 2040,  2, 16 },  { 2040,  2, 17 },  { 2040,  2, 18 },
  { 2040,  2, 19 },  { 2040,  2, 20 },  { 2040,  2, 21 },  { 2040,  2, 22 },
  { 2040,  2, 23 },  { 2040,  2, 24 },  { 2040,  2, 25 },  { 2040,  2, 26 },
  { 2040,  2, 27 },  { 2040,  2, 28 },  { 2040,  2, 29 },  { 2040,  3,  1 },
  { 2040,  3,  2 },  { 2040,  3,  3 },  { 2040,  3,  4 },  { 2040,  3,  5 },
  { 2040,  3,  6 },  { 2040,  3,  7 },  { 2040,  3,  8 },  { 2040,  3,  9 },
  { 2040,  3, 10 },  { 2040,  3, 11 },  { 2040,  3, 12 },  { 2040,  3, 13 },
  { 2040,  3, 14 },  { 2040,  3, 15 },  { 2040,  3, 16 },  { 2040,  3, 17 },
  { 2040,  3, 18 },  { 2040,  3, 19 },  { 2040,  3, 20 },  { 2040,  3, 21 },
  { 2040,  3, 22 },  { 2040,  3, 23 },  { 2040,  3, 24 },  { 2040,  3, 25 },
  { 2040,  3, 26 },  { 2040,  3, 27 },  { 2040,  3, 28 },  { 2040,  3, 29 },
  { 2040,  3, 30 },  { 2040,  3, 31 },  { 2040,  4,  1 },  { 2040,  4,  2 },
  { 2040,  4,  3 },  { 2040,  4,  4 },  { 2040,  4,  5 },  { 2040,  4,  6 },
  { 2040,  4,  7 },  { 2040,  4,  8 },  { 2040,  4,  9 },  { 2040,  4, 10 },
  { 2040,  4, 11 },  { 2040,  4, 12 },  { 2040,  4, 13 },  { 2040,  4, 14 },
  { 2040,  4, 15 },  { 2040,  4, 16 },  { 2040,  4, 17 },  { 2040,  4, 18 },
  { 2040,  4, 19 },  { 2040,  4, 20 },  { 2040,  4, 21 },  { 2040,  4, 22 },
  { 2040,  4, 23 },  { 2040,  4, 24 },  { 2040,  4, 25 },  { 2040,  4, 26 },
  { 2040,  4, 27 },  { 2040,  4, 28 },  { 2040,  4, 29 },  { 2040,  4, 30 },
  { 2040,  5,  1 },  { 2040,  5,  2 },  { 2040,  5,  3 },  { 2040,  5,  4 },
  { 2040,  5,  5 },  { 2040,  5,  6 },  { 2040,  5,  7 },  { 2040,  5,  8 },
  { 2040,  5,  9 },  { 2040,  5, 10 },  { 2040,  5, 11 },  { 2040,  5, 12 },
  { 2040,  5, 13 },  { 2040,  5, 14 },  { 2040,  5, 15 },  { 2040,  5, 16 },
  { 2040,  5, 17 },  { 2040,  5, 18 },  { 2040,  5, 19 },  { 2040,  5, 20 },
  { 2040,  5, 21 },  { 2040,  5, 22 },  { 2040,  5, 23 },  { 2040,  5, 24 },
  { 2040,  5, 25 },  { 2040,  5, 26 },  { 2040,  5, 27 },  { 2040,  5, 28 },
  { 2040,  5, 29 },  { 2040,  5, 30 },  { 2040,  5, 31 },  { 2040,  6,  1 },
  { 2040,  6,  2 },  { 2040,  6,  3 },  { 2040,  6,  4 },  { 2040,  6,  5 },
  { 2040,  6,  6 },  { 2040,  6,  7 },  { 2040,  6,  8 },  { 2040,  6,  9 },
  { 2040,  6, 10 },  { 2040,  6, 11 },  { 2040,  6, 12 },  { 2040,  6, 13 },
  { 2040,  6, 14 },  { 2040,  6, 15 },  { 2040,  6, 16 },  { 2040,  6, 17 },
  { 2040,  6, 18 },  { 2040,  6, 19 },  { 2040,  6, 20 },  { 2040,  6, 21 },
  { 2040,  6, 22 },  { 2040,  6, 23 },  { 2040,  6, 24 },  { 2040,  6, 25 },
  { 2040,  6, 26 },  { 2040,  6, 27 },  { 2040,  6, 28 },  { 2040,  6, 29 },
  { 2040,  6, 30 },  { 2040,  7,  1 },  { 2040,  7,  2 },  { 2040,  7,  3 },
  { 2040,  7,  4 },  { 2040,  7,  5 },  { 2040,  7,  6 },  { 2040,  7,  7 },
  { 2040,  7,  8 },  { 2040,  7,  9 },  { 2040,  7, 10 },  { 2040,  7, 11 },
  { 2040,  7, 12 },  { 2040,  7, 13 },  { 2040,  7, 14 },  { 2040,  7, 15 },
  { 2040,  7, 16 },  { 2040,  7, 17 },  { 2040,  7, 18 },  { 2040,  7, 19 },
  { 2040,  7, 20 },  { 2040,  7, 21 },  { 2040,  7, 22 },  { 2040,  7, 23 },
  { 2040,  7, 24 },  { 2040,  7, 25 },  { 2040,  7, 26 },  { 2040,  7, 27 },
  { 2040,  7, 28 },  { 2040,  7, 29 },  { 2040,  7, 30 },  { 2040,  7, 31 },
  { 2040,  8,  1 },  { 2040,  8,  2 },  { 2040,  8,  3 },  { 2040,  8,  4 },
  { 2040,  8,  5 },  { 2040,  8,  6 },  { 2040,  8,  7 },  { 2040,  8,  8 },
  { 2040,  8,  9 },  { 2040,  8, 10 },  { 2040,  8, 11 },  { 2040,  8, 12 },
  { 2040,  8, 13 },  { 2040,  8, 14 },  { 2040,  8, 15 },  { 2040,  8, 16 },
  { 2040,  8, 17 },  { 2040,  8, 18 },  { 2040,  8, 19 },  { 2040,  8, 20 },
  { 2040,  8, 21 },  { 2040,  8, 22 },  { 2040,  8, 23 },  { 2040,  8, 24 },
  { 2040,  8, 25 },  { 2040,  8, 26 },  { 2040,  8, 27 },  { 2040,  8, 28 },
  { 2040,  8, 29 },  { 2040,  8, 30 },  { 2040,  8, 31 },  { 2040,  9,  1 },
  { 2040,  9,  2 },  { 2040,  9,  3 },  { 2040,  9,  4 },  { 2040,  9,  5 },
  { 2040,  9,  6 },  { 2040,  9,  7 },  { 2040,  9,  8 },  { 2040,  9,  9 },
  { 2040,  9, 10 },  { 2040,  9, 11 },  { 2040,  9, 12 },  { 2040,  9, 13 },
  { 2040,  9, 14 },  { 2040,  9, 15 },  { 2040,  9, 16 },  { 2040,  9, 17 },
  { 2040,  9, 18 },  { 2040,  9, 19 },  { 2040,  9, 20 },  { 2040,  9, 21 },
  { 2040,  9, 22 },  { 2040,  9, 23 },  { 2040,  9, 24 },  { 2040,  9, 25 },
  { 2040,  9, 26 },  { 2040,  9, 27 },  { 2040,  9, 28 },  { 2040,  9, 29 },
  { 2040,  9, 30 },  { 2040, 10,  1 },  { 2040, 10,  2 },  { 2040, 10,  3 },
  { 2040, 10,  4 },  { 2040, 10,  5 },  { 2040, 10,  6 },  { 2040, 10,  7 },
  { 2040, 10,  8 },  { 2040, 10,  9 },  { 2040, 10, 10 },  { 2040, 10, 11 },
  { 2040, 10, 12 },  { 2040, 10, 13 },  { 2040, 10, 14 },  { 2040, 10, 15 },
  { 2040, 10, 16 },  { 2040, 10, 17 },  { 2040, 10, 18 },  { 2040, 10, 19 },
  { 2040, 10, 20 },  { 2040, 10, 21 },  { 2040, 10, 22 },  { 2040, 10, 23 },
  { 2040, 10, 24 },  { 2040, 10, 25 },  { 2040, 10, 26 },  { 2040, 10, 27 },
  { 2040, 10, 28 },  { 2040, 10, 29 },  { 2040, 10, 30 },  { 2040, 10, 31 },
  { 2040, 11,  1 },  { 2040, 11,  2 },  { 2040, 11,  3 },  { 2040, 11,  4 },
  { 2040, 11,  5 },  { 2040, 11,  6 },  { 2040, 11,  7 },  { 2040, 11,  8 },
  { 2040, 11,  9 },  { 2040, 11, 10 },  { 2040, 11, 11 },  { 2040, 11, 12 },
  { 2040, 11, 13 },  { 2040, 11, 14 },  { 2040, 11, 15 },  { 2040, 11, 16 },
  { 2040, 11, 17 },  { 2040, 11, 18 },  { 2040, 11, 19 },  { 2040, 11, 20 },
  { 2040, 11, 21 },  { 2040, 11, 22 },  { 2040, 11, 23 },  { 2040, 11, 24 },
  { 2040, 11, 25 },  { 2040, 11, 26 },  { 2040, 11, 27 },  { 2040, 11, 28 },
  { 2040, 11, 29 },  { 2040, 11, 30 },  { 2040, 12,  1 },  { 2040, 12,  2 },
  { 2040, 12,  3 },  { 2040, 12,  4 },  { 2040, 12,  5 },  { 2040, 12,  6 },
  { 2040, 12,  7 },  { 2040, 12,  8 },  { 2040, 12,  9 },  { 2040, 12, 10 },
  { 2040, 12, 11 },  { 2040, 12, 12 },  { 2040, 12, 13 },  { 2040, 12, 14 },
  { 2040, 12, 15 },  { 2040, 12, 16 },  { 2040, 12, 17 },  { 2040, 12, 18 },
  { 2040, 12, 19 },  { 2040, 12, 20 },  { 2040, 12, 21 },  { 2040, 12, 22 },
  { 2040, 12, 23 },  { 2040, 12, 24 },  { 2040, 12, 25 },  { 2040, 12, 26 },
  { 2040, 12, 27 },  { 2040, 12, 28 },  { 2040, 12, 29 },  { 2040, 12, 30 },
  { 2040, 12, 31 },
};

const int ProlepticDateImpUtil::s_cachedSerialDate[61][13] = {
  { 0, 722814, 722845, 722874, 722905, 722935, 722966,
       722996, 723027, 723058, 723088, 723119, 723149,   },
  { 0, 723180, 723211, 723239, 723270, 723300, 723331,
       723361, 723392, 723423, 723453, 723484, 723514,   },
  { 0, 723545, 723576, 723604, 723635, 723665, 723696,
       723726, 723757, 723788, 723818, 723849, 723879,   },
  { 0, 723910, 723941, 723969, 724000, 724030, 724061,
       724091, 724122, 724153, 724183, 724214, 724244,   },
  { 0, 724275, 724306, 724335, 724366, 724396, 724427,
       724457, 724488, 724519, 724549, 724580, 724610,   },
  { 0, 724641, 724672, 724700, 724731, 724761, 724792,
       724822, 724853, 724884, 724914, 724945, 724975,   },
  { 0, 725006, 725037, 725065, 725096, 725126, 725157,
       725187, 725218, 725249, 725279, 725310, 725340,   },
  { 0, 725371, 725402, 725430, 725461, 725491, 725522,
       725552, 725583, 725614, 725644, 725675, 725705,   },
  { 0, 725736, 725767, 725796, 725827, 725857, 725888,
       725918, 725949, 725980, 726010, 726041, 726071,   },
  { 0, 726102, 726133, 726161, 726192, 726222, 726253,
       726283, 726314, 726345, 726375, 726406, 726436,   },
  { 0, 726467, 726498, 726526, 726557, 726587, 726618,
       726648, 726679, 726710, 726740, 726771, 726801,   },
  { 0, 726832, 726863, 726891, 726922, 726952, 726983,
       727013, 727044, 727075, 727105, 727136, 727166,   },
  { 0, 727197, 727228, 727257, 727288, 727318, 727349,
       727379, 727410, 727441, 727471, 727502, 727532,   },
  { 0, 727563, 727594, 727622, 727653, 727683, 727714,
       727744, 727775, 727806, 727836, 727867, 727897,   },
  { 0, 727928, 727959, 727987, 728018, 728048, 728079,
       728109, 728140, 728171, 728201, 728232, 728262,   },
  { 0, 728293, 728324, 728352, 728383, 728413, 728444,
       728474, 728505, 728536, 728566, 728597, 728627,   },
  { 0, 728658, 728689, 728718, 728749, 728779, 728810,
       728840, 728871, 728902, 728932, 728963, 728993,   },
  { 0, 729024, 729055, 729083, 729114, 729144, 729175,
       729205, 729236, 729267, 729297, 729328, 729358,   },
  { 0, 729389, 729420, 729448, 729479, 729509, 729540,
       729570, 729601, 729632, 729662, 729693, 729723,   },
  { 0, 729754, 729785, 729813, 729844, 729874, 729905,
       729935, 729966, 729997, 730027, 730058, 730088,   },
  { 0, 730119, 730150, 730179, 730210, 730240, 730271,
       730301, 730332, 730363, 730393, 730424, 730454,   },
  { 0, 730485, 730516, 730544, 730575, 730605, 730636,
       730666, 730697, 730728, 730758, 730789, 730819,   },
  { 0, 730850, 730881, 730909, 730940, 730970, 731001,
       731031, 731062, 731093, 731123, 731154, 731184,   },
  { 0, 731215, 731246, 731274, 731305, 731335, 731366,
       731396, 731427, 731458, 731488, 731519, 731549,   },
  { 0, 731580, 731611, 731640, 731671, 731701, 731732,
       731762, 731793, 731824, 731854, 731885, 731915,   },
  { 0, 731946, 731977, 732005, 732036, 732066, 732097,
       732127, 732158, 732189, 732219, 732250, 732280,   },
  { 0, 732311, 732342, 732370, 732401, 732431, 732462,
       732492, 732523, 732554, 732584, 732615, 732645,   },
  { 0, 732676, 732707, 732735, 732766, 732796, 732827,
       732857, 732888, 732919, 732949, 732980, 733010,   },
  { 0, 733041, 733072, 733101, 733132, 733162, 733193,
       733223, 733254, 733285, 733315, 733346, 733376,   },
  { 0, 733407, 733438, 733466, 733497, 733527, 733558,
       733588, 733619, 733650, 733680, 733711, 733741,   },
  { 0, 733772, 733803, 733831, 733862, 733892, 733923,
       733953, 733984, 734015, 734045, 734076, 734106,   },
  { 0, 734137, 734168, 734196, 734227, 734257, 734288,
       734318, 734349, 734380, 734410, 734441, 734471,   },
  { 0, 734502, 734533, 734562, 734593, 734623, 734654,
       734684, 734715, 734746, 734776, 734807, 734837,   },
  { 0, 734868, 734899, 734927, 734958, 734988, 735019,
       735049, 735080, 735111, 735141, 735172, 735202,   },
  { 0, 735233, 735264, 735292, 735323, 735353, 735384,
       735414, 735445, 735476, 735506, 735537, 735567,   },
  { 0, 735598, 735629, 735657, 735688, 735718, 735749,
       735779, 735810, 735841, 735871, 735902, 735932,   },
  { 0, 735963, 735994, 736023, 736054, 736084, 736115,
       736145, 736176, 736207, 736237, 736268, 736298,   },
  { 0, 736329, 736360, 736388, 736419, 736449, 736480,
       736510, 736541, 736572, 736602, 736633, 736663,   },
  { 0, 736694, 736725, 736753, 736784, 736814, 736845,
       736875, 736906, 736937, 736967, 736998, 737028,   },
  { 0, 737059, 737090, 737118, 737149, 737179, 737210,
       737240, 737271, 737302, 737332, 737363, 737393,   },
  { 0, 737424, 737455, 737484, 737515, 737545, 737576,
       737606, 737637, 737668, 737698, 737729, 737759,   },
  { 0, 737790, 737821, 737849, 737880, 737910, 737941,
       737971, 738002, 738033, 738063, 738094, 738124,   },
  { 0, 738155, 738186, 738214, 738245, 738275, 738306,
       738336, 738367, 738398, 738428, 738459, 738489,   },
  { 0, 738520, 738551, 738579, 738610, 738640, 738671,
       738701, 738732, 738763, 738793, 738824, 738854,   },
  { 0, 738885, 738916, 738945, 738976, 739006, 739037,
       739067, 739098, 739129, 739159, 739190, 739220,   },
  { 0, 739251, 739282, 739310, 739341, 739371, 739402,
       739432, 739463, 739494, 739524, 739555, 739585,   },
  { 0, 739616, 739647, 739675, 739706, 739736, 739767,
       739797, 739828, 739859, 739889, 739920, 739950,   },
  { 0, 739981, 740012, 740040, 740071, 740101, 740132,
       740162, 740193, 740224, 740254, 740285, 740315,   },
  { 0, 740346, 740377, 740406, 740437, 740467, 740498,
       740528, 740559, 740590, 740620, 740651, 740681,   },
  { 0, 740712, 740743, 740771, 740802, 740832, 740863,
       740893, 740924, 740955, 740985, 741016, 741046,   },
  { 0, 741077, 741108, 741136, 741167, 741197, 741228,
       741258, 741289, 741320, 741350, 741381, 741411,   },
  { 0, 741442, 741473, 741501, 741532, 741562, 741593,
       741623, 741654, 741685, 741715, 741746, 741776,   },
  { 0, 741807, 741838, 741867, 741898, 741928, 741959,
       741989, 742020, 742051, 742081, 742112, 742142,   },
  { 0, 742173, 742204, 742232, 742263, 742293, 742324,
       742354, 742385, 742416, 742446, 742477, 742507,   },
  { 0, 742538, 742569, 742597, 742628, 742658, 742689,
       742719, 742750, 742781, 742811, 742842, 742872,   },
  { 0, 742903, 742934, 742962, 742993, 743023, 743054,
       743084, 743115, 743146, 743176, 743207, 743237,   },
  { 0, 743268, 743299, 743328, 743359, 743389, 743420,
       743450, 743481, 743512, 743542, 743573, 743603,   },
  { 0, 743634, 743665, 743693, 743724, 743754, 743785,
       743815, 743846, 743877, 743907, 743938, 743968,   },
  { 0, 743999, 744030, 744058, 744089, 744119, 744150,
       744180, 744211, 744242, 744272, 744303, 744333,   },
  { 0, 744364, 744395, 744423, 744454, 744484, 744515,
       744545, 744576, 744607, 744637, 744668, 744698,   },
  { 0, 744729, 744760, 744789, 744820, 744850, 744881,
       744911, 744942, 744973, 745003, 745034, 745064,   },
};

const char ProlepticDateImpUtil::s_cachedDaysInMonth[61][13] = {
  { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
  { 0, 31, 29, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 },
};

}  // close package namespace
}  // close enterprise namespace

// ----------------------------------------------------------------------------
// Copyright 2014 Bloomberg Finance L.P.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
// ----------------------------- END-OF-FILE ----------------------------------
